/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import net.sf.dc.actions.ChangeViewAction;
import net.sf.dc.actions.CreateNewAction;
import net.sf.dc.actions.CreateReportAction;
import net.sf.dc.actions.DeleteAction;
import net.sf.dc.actions.ImportAction;
import net.sf.dc.actions.LogAction;
import net.sf.dc.actions.NewItemWizardAction;
import net.sf.dc.actions.OnlineSearchAction;
import net.sf.dc.actions.OpenAction;
import net.sf.dc.actions.OpenFieldsSettingsDialog;
import net.sf.dc.actions.OpenModuleAction;
import net.sf.dc.actions.OpenResourceEditorAction;
import net.sf.dc.actions.SaveAllAction;
import net.sf.dc.actions.SearchAction;
import net.sf.dc.actions.SearchAgainAction;
import net.sf.dc.actions.SettingsAction;
import net.sf.dc.actions.StartProgramAction;
import net.sf.dc.actions.ToggleGroupingPaneAction;
import net.sf.dc.actions.ToggleModuleListAction;
import net.sf.dc.actions.ToggleQuickFilterBar;
import net.sf.dc.actions.ToggleQuickViewAction;
import net.sf.dc.actions.ToggleToolbarLabelsAction;
import net.sf.dc.console.components.DcFrame;
import net.sf.dc.console.components.DcPanel;
import net.sf.dc.console.components.panels.ChartPanel;
import net.sf.dc.console.components.panels.ModuleListPanel;
import net.sf.dc.console.components.renderers.DcTableHeaderRenderer;
import net.sf.dc.console.components.renderers.DcTableHeaderRendererRequired;
import net.sf.dc.console.components.renderers.DcTableRowIndexCellRenderer;
import net.sf.dc.console.menu.DcQuickFilterToolBar;
import net.sf.dc.console.menu.DcToolBar;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.views.View;
import net.sf.dc.console.views.list.ListSearchView;
import net.sf.dc.console.views.table.TableSearchView;
import net.sf.dc.console.windows.AboutDialog;
import net.sf.dc.console.windows.help.HelpDialog;
import net.sf.dc.console.windows.itemforms.DcMinimalisticItemView;
import net.sf.dc.console.windows.messageboxes.QuestionBox;
import net.sf.dc.console.windows.settings.SettingsView;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.data.DataFilters;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.db.DatabaseManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.DcPropertyModule;
import net.sf.dc.core.modules.TemplateModule;
import net.sf.dc.core.objects.DcLookAndFeel;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinitions;

public class MainFrame extends DcFrame {

	private static final long serialVersionUID = 9L;
	public static final int _SEARCHTAB = 0;
    public static final int _INSERTTAB = 1;

    private ModuleListPanel moduleListPanel = new ModuleListPanel();
    private JTabbedPane tabbedPane = ComponentFactory.getTabbedPane();
    private SettingsView settingsView;

    private DcToolBar toolBar;
    private DcQuickFilterToolBar quickFilterTb;

    private DcModule activeModule = null;

    private boolean initialized = false;

    public MainFrame() {
        super(DcRepository.system.product, IconLibrary._icoMain);

        setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        this.getContentPane().setLayout(Layout.getGBL());
        
        toggleModuleList(DcSettings.getBoolean(DcRepository.settings.stShowModuleList));
        tabbedPane.getModel().addChangeListener(new TabIndexListener());

        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        
        try {
            addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    finish(true);
                }
            });

            getContentPane().add(moduleListPanel, Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                                 new Insets(0, 0, 0, 0), 0, 0));
            getContentPane().add(tabbedPane,      Layout.getGBC( 1, 2, 1, 1, 100.0, 100.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                 new Insets(0, 0, 0, 0), 0, 0));

            tabbedPane.setPreferredSize(DcSettings.getDimension(DcRepository.settings.stTabbedPaneSize));
            tabbedPane.setSize(DcSettings.getDimension(DcRepository.settings.stTabbedPaneSize));

        } catch (Error err) {
            DcLog.add(err);
        }

        if (DcModules.getCurrent() == null)
            changeModule(DcModules.get(DcModules._MOVIE).getIndex(), false);
        else
            changeModule(DcModules.getCurrent().getIndex(), false);
        
        applySettings();        
        validate();
        
        setResizable(true);
        setSize(DcSettings.getDimension(DcRepository.settings.stMainViewSize));
        setExtendedState(DcSettings.getInt(DcRepository.settings.stMainViewState));
        registerShortCuts();

        pack();

        setCenteredLocation();

        initialized = true;
        
    }
    
    private void registerShortCuts() {
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_D, InputEvent.CTRL_MASK), new DeleteAction(), "delete");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.CTRL_MASK), new CreateReportAction(), "report");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_F, InputEvent.CTRL_MASK), new SearchAction(), "find");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_G, InputEvent.CTRL_MASK), new SearchAgainAction(), "findagain");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_I, InputEvent.CTRL_MASK), new ImportAction(), "fileimport");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_L, InputEvent.CTRL_MASK), new LogAction(), "showlog");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_M, InputEvent.CTRL_MASK), new SettingsAction(), "settings");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_MASK), new CreateNewAction(), "createnew");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_MASK), new OpenAction(), "open");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_P, InputEvent.CTRL_MASK), new StartProgramAction(), "startProgram");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.CTRL_MASK), new OpenResourceEditorAction(), "resourceeditor");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK), new SaveAllAction(), "saveall");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_T, InputEvent.CTRL_MASK), new OpenFieldsSettingsDialog(), "fieldsettings");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_U, InputEvent.CTRL_MASK), new OnlineSearchAction(), "internetupdateaction");
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.CTRL_MASK), new NewItemWizardAction(), "wizard");
        
        addKeyListener(KeyStroke.getKeyStroke("F2"), new ChangeViewAction(MasterView._TABLE_VIEW), "opentableview");
        addKeyListener(KeyStroke.getKeyStroke("F3"), new ChangeViewAction(MasterView._LIST_VIEW), "openlistview");
        
        addKeyListener(KeyStroke.getKeyStroke("F7"), new ToggleQuickFilterBar(), "togglequickfilterbar");
        addKeyListener(KeyStroke.getKeyStroke("F8"), new ToggleModuleListAction(), "togglemodules");
        addKeyListener(KeyStroke.getKeyStroke("F9"), new ToggleQuickViewAction(), "togglequickview");
        addKeyListener(KeyStroke.getKeyStroke("F10"), new ToggleGroupingPaneAction(), "togglegroupingpane");
        addKeyListener(KeyStroke.getKeyStroke("F11"), new ToggleToolbarLabelsAction(), "toggletoolbarlabels");;

        for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            if (module.isTopModule() && module.getKeyStroke() != null)
                addKeyListener(module.getKeyStroke(), new OpenModuleAction(module.getIndex()), "open" + module.getName() + "module");
        }        
    }
    
    public void setVisible(boolean b) {
        super.setVisible(b);
    }

    public DcToolBar getToolBar() {
        return toolBar;
    }

    public void setSelectedTab(int index) {
    	tabbedPane.setSelectedIndex(index);
    }

    public int getSelectedTab() {
        return tabbedPane.getSelectedIndex();
    }

    public void showSettingsView() {
        settingsView = DcSettings.getView();
    	settingsView.setVisible(true);
        settingsView.dispose();
        settingsView = null;
    }

    public void showAboutDialog() {
        new AboutDialog(this).setVisible(true);
    }

    private void toggleModuleList(boolean enable) {
        moduleListPanel.setVisible(enable);
    }

    public void toggleModuleList() {
    	moduleListPanel.setVisible(!moduleListPanel.isVisible());
        DcSettings.set(DcRepository.settings.stShowModuleList, new Boolean(moduleListPanel.isVisible()));
    }

    public void updateQuickFilterBar() {
        boolean show = DcSettings.getBoolean(DcRepository.settings.stShowQuickFilterBar);
        
        if (quickFilterTb != null)
            getContentPane().remove(quickFilterTb);
        
        if (show) {
            quickFilterTb = new DcQuickFilterToolBar(activeModule);
            getContentPane().add(quickFilterTb, Layout.getGBC( 0, 1, 2, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                     new Insets(0, 0, 0, 0), 0, 0));     
            
            quickFilterTb.revalidate();
        }
        
        repaint();
    }
    
    private void applySettings(DcModule module) {
        Font fontNormal = DcSettings.getFont(DcRepository.settings.stSystemFontNormal);
        Font fontSystem = DcSettings.getFont(DcRepository.settings.stSystemFontBold);
        
        DcLog.setFont(fontNormal);
        tabbedPane.setFont(fontSystem);
        
        remove(moduleListPanel);
        moduleListPanel = new ModuleListPanel();
        getContentPane().add(moduleListPanel, Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                             new Insets(0, 0, 0, 0), 0, 0));
        
        moduleListPanel.setFont(fontSystem);
        moduleListPanel.setSelectedModule(DcModules.getCurrent().getIndex());
        moduleListPanel.revalidate();
        
        moduleListPanel.setVisible(DcSettings.getBoolean(DcRepository.settings.stShowModuleList));

        if (module.getChartPanel(false) != null)
            module.getChartPanel(false).setFont(fontNormal);

        if (settingsView != null)
            settingsView.setFont(fontNormal);
        
        if (    DcModules.getCurrent() != null &&
                module.getIndex() == DcModules.getCurrent().getIndex() && 
                module.getMenuBar() != null)
            module.getMenuBar().setFont(fontSystem);
        
        if (toolBar != null) {
            toolBar.setFloatable(false);
            toolBar.setFont(fontSystem);
        }
        
        if (initialized) {
            if (module.isTopModule() && module.getFilterForm(false) != null) 
                module.getFilterForm(false).setFont(fontNormal);
        }
        
        if (module instanceof DcPropertyModule) { 
            DcMinimalisticItemView form = module.getPropertyModule(DcModules._STATE).getForm();
            form.setFont(fontNormal);
        }

        DcFieldDefinitions definitions = module.getFieldDefinitions();
        module.getDcObject().applySettings(definitions);

        if (module.getSearchView() != null)
            module.getSearchView().applySettings();

        if (module.getInsertView() != null)
            module.getInsertView().applySettings();
    }
    
    public void updateLAF(DcLookAndFeel laf) {
        DcSettings.set(DcRepository.settings.stLookAndFeel, laf);
        
        ComponentFactory.setLookAndFeel();

        tabbedPane.removeAll();
        SwingUtilities.updateComponentTreeUI(moduleListPanel);
        SwingUtilities.updateComponentTreeUI(tabbedPane);
        SwingUtilities.updateComponentTreeUI(toolBar);
        
        getContentPane().setBackground(tabbedPane.getBackground());
        
        if (quickFilterTb != null)
            SwingUtilities.updateComponentTreeUI(quickFilterTb);

        DcLog.reinstallUI();
        
        if (initialized && DcModules.getCurrent().getFilterForm(false) != null)
            SwingUtilities.updateComponentTreeUI(DcModules.getCurrent().getFilterForm(false).getContentPane());
        
        SwingUtilities.updateComponentTreeUI(DcTableHeaderRenderer.button);
        SwingUtilities.updateComponentTreeUI(DcTableHeaderRendererRequired.button);
        SwingUtilities.updateComponentTreeUI(DcTableRowIndexCellRenderer.button);
        
        for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            
            if (module.getMenuBar() != null)
                SwingUtilities.updateComponentTreeUI(module.getMenuBar());
            
            if (module.getChartPanel(false) != null)
                SwingUtilities.updateComponentTreeUI(module.getChartPanel(false));
            
            if (initialized) {
                if (module.isTopModule() && module.getFilterForm(false) != null) 
                    SwingUtilities.updateComponentTreeUI(module.getFilterForm(false).getContentPane());
            }
            
            if (module instanceof DcPropertyModule) { 
                DcMinimalisticItemView view = ((DcPropertyModule) module).getForm();
                if (view != null)  
                    SwingUtilities.updateComponentTreeUI(((DcPropertyModule) module).getForm().getContentPane());
            }
                
            if (module instanceof TemplateModule)
                SwingUtilities.updateComponentTreeUI(((TemplateModule) module).getForm().getContentPane());
            
            MasterView[] views = module.getViews();
            if (views != null) {
                for (int i = 0; i < views.length; i++)  {
                    MasterView view = views[i];
                    if (view.get(MasterView._LIST_VIEW) != null)
                        SwingUtilities.updateComponentTreeUI(view.get(MasterView._LIST_VIEW));
                    if (view.get(MasterView._TABLE_VIEW) != null)
                        SwingUtilities.updateComponentTreeUI(view.get(MasterView._TABLE_VIEW));
                }
            }
        }
            
        changeModule(DcModules.getCurrent().getIndex(), false);
        
        applySettings();        
        validate();

    }    
    
    public void applySettings() {
        for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            
            if (module.isTopModule())
                applySettings(module);    
        }
    }

    public void applyView(int newView) {
        int previousView = DcSettings.getInt(DcRepository.settings.stSelectedView);
        if (previousView != newView) {
            DcSettings.set(DcRepository.settings.stSelectedView, new Integer(newView));
            tabbedPane.removeAll();
            MasterView[] masterViews = activeModule.getViews();
            for (int i = 0; i < masterViews.length; i++) {
                View view = masterViews[i].getCurrent();
                if (view != null) {
                    tabbedPane.addTab(view.getTitle(), view.getIcon(), view);
                    view.applyViewDividerLocation();
                    
                    if (view instanceof TableSearchView || view instanceof ListSearchView)
                    	DataManager.bindData(view, activeModule.getIndex(), 
                    					 DataFilters.getCurrent(activeModule.getIndex()), false);
                }
            }
            ChartPanel chartPanel = DcModules.getCurrent().getChartPanel(true);
            tabbedPane.addTab(chartPanel.getTitle(), chartPanel.getIcon(), chartPanel);
        }
    }
    
    public void updateMenuBar() {
        JMenuBar menubar = getJMenuBar();
        ComponentFactory.clean(menubar);
        setJMenuBar(null);
        
        setJMenuBar(activeModule.getMenuBar());
        repaint();
        validate();
    }

    public void changeModule(int index, boolean loadData) {
        activeModule = DcModules.get(index);
        activeModule = activeModule == null ? DcModules.get(DcModules._SOFTWARE) : activeModule;
        DcSettings.set(DcRepository.settings.stModule, new Integer(activeModule.getIndex()));

        moduleListPanel.setSelectedModule(index);
        
        updateMenuBar();

        if (toolBar != null)
            getContentPane().remove(toolBar);

        toolBar = new DcToolBar(activeModule);
        getContentPane().add(toolBar, Layout.getGBC( 0, 0, 2, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                             new Insets(0, 0, 0, 0), 0, 0));
        
        updateQuickFilterBar();
        tabbedPane.removeAll();
        
        if (loadData)
            bindData();
        
        repaint();
    }
    
    public void bindData() {
        try {
            MasterView[] masterViews = activeModule.getViews();
            for (int i = 0; i < masterViews.length; i++) {
                View view = masterViews[i].getCurrent();
                if (view != null) {
                    tabbedPane.addTab(view.getTitle(), view.getIcon(), view);
                    view.applyViewDividerLocation();
                    if (view instanceof ListSearchView || view instanceof TableSearchView)
                        DataManager.bindData(view, activeModule.getIndex(), DataFilters.getCurrent(activeModule.getIndex()), false);
                }
            }
            
            ChartPanel chartPanel = DcModules.getCurrent().getChartPanel(true);
            tabbedPane.addTab(chartPanel.getTitle(), chartPanel.getIcon(), chartPanel);
            
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }

    public void finish(boolean checkForChanges) {
        
        if (checkForChanges) {
            Collection viewsWithChanges = new ArrayList();
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                
                Collection masterViews = new ArrayList();
                if (module.getInsertView() != null)
                    masterViews.add(module.getInsertView());
                if (module.getSearchView() != null)
                    masterViews.add(module.getSearchView());
                
                for (Iterator iterMasterViews = masterViews.iterator(); iterMasterViews.hasNext(); ) {
                    MasterView masterView = (MasterView) iterMasterViews.next();
                    for (Iterator iterViews = masterView.getViews(); iterViews.hasNext(); ) {
                        View view = (View) iterViews.next();
                        if (!view.isChangesSaved())
                            viewsWithChanges.add(view);
                    }
                } 
            }
            
            if (viewsWithChanges.size() > 0) {
                QuestionBox qb = new QuestionBox(DcResources.getText("msgCancelExitAndSave"));
                if (qb.isAffirmative()) {
                    for (Iterator iterViews = viewsWithChanges.iterator(); iterViews.hasNext(); )
                        ((View) iterViews.next()).save();
                    
                    return;
                }
            } 
        }
        
        DcLog.add(DcResources.getText("msgApplicationStops"), DcLog._STANDARD);

        setVisible(false);

        DatabaseManager.closeDatabases(false);

        DcSettings.set(DcRepository.settings.stTabbedPaneSize, tabbedPane.getSize());
        DcSettings.set(DcRepository.settings.stMainViewSize, getSize());
        DcSettings.set(DcRepository.settings.stMainViewState, new Integer(getExtendedState()));
        
        new DataFilters().save();
        DcSettings.save();
        DcModules.save();
        DcLog.close();
        
        System.exit(0);
    }
    
    public void close() {}

    private class TabIndexListener implements ChangeListener {
        public void stateChanged(ChangeEvent arg0) {
            int index = tabbedPane.getSelectedIndex();
            if (index > -1) {
            	DcPanel panel = (DcPanel) tabbedPane.getComponent(index);
            	HelpDialog.setHelpIndex(panel.getHelpIndex());
            }
	    }
    }
}