/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.objects.DcLookAndFeel;
import net.sf.dc.core.resources.DcResources;

public class DcLookAndFeelSelector extends JComponent implements IComponent {
    
    private JCheckBox checkNoLF = ComponentFactory.getCheckBox(DcResources.getText("lblNoLF"));
    private JCheckBox checkSystemLF = ComponentFactory.getCheckBox(DcResources.getText("lblSystemLF"));
    private JCheckBox checkSkinLF = ComponentFactory.getCheckBox(DcResources.getText("lblSkinLF"));

    private JComboBox comboSkinLF = ComponentFactory.getComboBox();
    private JComboBox comboSystemLF = ComponentFactory.getComboBox();
    
    private DcDialog parent;
    
    private boolean applyModus = true;
   
    public DcLookAndFeelSelector() {
        applyModus = false;
        buildComponent();
        fillSkinLFCombo();
        fillSystemLFCombo();
        applyModus = true;
    }
    
    public void setParent(DcDialog dialog) {
        this.parent = dialog;
    }
    
    public void clear() {
        checkNoLF = null;
        checkSystemLF = null;
        checkSkinLF = null;
        comboSkinLF = null;
        comboSystemLF = null;
        parent = null;
    }    
    
    private void fillSkinLFCombo() {
        String themeDir = DataCrow.baseDir + File.separator + "themes";
        File file = new File(themeDir);
        String [] files = file.exists() && file.isDirectory() ? file.list() : null;
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                String filename = files[i];
                if (filename.toLowerCase().endsWith("zip"))
                    comboSkinLF.addItem(
                            new DcLookAndFeel(filename, null, filename, DcLookAndFeel._SKINLF));
            }
        } else {
            DcLog.add(DcResources.getText("msgNoThemesFound"), DcLog._ERROR);
        }
    }

    private void fillSystemLFCombo() {
        LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();
        for (int i = 0; i < lafs.length; i++) {
            LookAndFeelInfo laf = lafs[i];
            comboSystemLF.addItem(
                    new DcLookAndFeel(laf.getName(), laf.getClassName(), null, DcLookAndFeel._SYSTEMLF));
        }
    }
    
    public Object getValue() {
        return getLookAndFeel();
    }
    
    public DcLookAndFeel getLookAndFeel() {
        return null;
    }
    
    /**
     * Applies a value to this field
     */
    public void setValue(Object o) {
        applyModus = false;
        if (o instanceof DcLookAndFeel) { 
            DcLookAndFeel laf = (DcLookAndFeel) o;
            if (laf.getType() == DcLookAndFeel._NONE) {
                checkNoLF.setSelected(true);
            } else if (laf.getType() == DcLookAndFeel._SKINLF) {
                comboSkinLF.setSelectedItem(laf);
                checkSkinLF.setSelected(true);
            } else { 
                comboSystemLF.setSelectedItem(laf);
                checkSystemLF.setSelected(true);
            }
        }
        applyModus = true;
    }
    
    public void setFont(Font font) {
        super.setFont(font);
        
        if (checkNoLF != null) {
            checkNoLF.setFont(ComponentFactory.getStandardFont());
            checkSkinLF.setFont(ComponentFactory.getStandardFont());
            checkSystemLF.setFont(ComponentFactory.getStandardFont());
            comboSkinLF.setFont(ComponentFactory.getStandardFont());
            comboSystemLF.setFont(ComponentFactory.getStandardFont());
        }
    }     
    
    /**
     * Builds this component
     */
    private void buildComponent() {
        setLayout(Layout.getGBL());
        
        checkNoLF.addActionListener(new TypeActionListener(DcLookAndFeel._NONE));
        checkSkinLF.addActionListener(new TypeActionListener(DcLookAndFeel._SKINLF));
        checkSystemLF.addActionListener(new TypeActionListener(DcLookAndFeel._SYSTEMLF));
        
        comboSystemLF.addActionListener(new LafSelectionListener());
        comboSkinLF.addActionListener(new LafSelectionListener());
        
        add(checkNoLF,       Layout.getGBC( 0, 0, 2, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets( 0, 0, 0, 0), 0, 0));
        add(checkSystemLF,   Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets( 0, 0, 0, 0), 0, 0));
        add(comboSystemLF,   Layout.getGBC( 1, 1, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets( 0, 0, 0, 0), 0, 0));
        add(checkSkinLF,     Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets( 0, 0, 0, 0), 0, 0));
        add(comboSkinLF,     Layout.getGBC( 1, 2, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets( 0, 0, 0, 0), 0, 0));
    }
    
    public class FontStyle {
        final int style;
        final String name;
        
        public FontStyle(int index, String name) {
            this.style = index;
            this.name = name;
        }
        
        public int getIndex() {
            return style;
        }
        
        public String toString() {
            return name;
        }
    }
    
    private void applyLAF() {
        if (applyModus) {
            DcLookAndFeel laf = null;
            
            if (checkNoLF.isSelected()) 
                laf = DcLookAndFeel.getDisabled();
            else if (checkSkinLF.isSelected())
                laf = (DcLookAndFeel) comboSkinLF.getSelectedItem();
            else if (checkSystemLF.isSelected())
                laf = (DcLookAndFeel) comboSystemLF.getSelectedItem();
            
            if (laf != null) {
                DataCrow.mainFrame.updateLAF(laf);
                DataCrow.mainFrame.bindData();
                if (parent != null)
                    SwingUtilities.updateComponentTreeUI(parent.getContentPane());
            }
        }
    }
    
    private class TypeActionListener implements ActionListener {
        
        private int type = DcLookAndFeel._NONE;
        
        public TypeActionListener(int type) {
            this.type = type;
        }
        
        public void actionPerformed(ActionEvent arg0) {
            checkNoLF.setSelected(false);
            checkSkinLF.setSelected(false);
            checkSystemLF.setSelected(false);
            comboSkinLF.setEnabled(false);
            comboSystemLF.setEnabled(false);
            
            if (type == DcLookAndFeel._NONE) {
                checkNoLF.setSelected(true);
            } else if (type == DcLookAndFeel._SKINLF) {
                comboSkinLF.setSelectedIndex(0);
                checkSkinLF.setSelected(true);
                comboSkinLF.setEnabled(true);
            } else if (type == DcLookAndFeel._SYSTEMLF) {
                comboSkinLF.setSelectedIndex(0);
                checkSystemLF.setSelected(true);
                comboSystemLF.setEnabled(true);
            }
            
            applyLAF();
        }
    }
    
    private class LafSelectionListener implements ActionListener {
        public void actionPerformed(ActionEvent a) {
            applyLAF();
        }
    }
}