/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;

import javax.swing.CellRendererPane;
import javax.swing.JComponent;
import javax.swing.JToolTip;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.basic.BasicToolTipUI;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.util.Utilities;

public class DcMultiLineToolTipUI extends BasicToolTipUI {

    private final static DcMultiLineToolTipUI sharedInstance = new DcMultiLineToolTipUI();
    private CellRendererPane rendererPane;
    private static DcTextPane textPane;
    private boolean containsText = true;

    public DcMultiLineToolTipUI() {
        super();
        if (textPane == null) {
            textPane = ComponentFactory.getTextPane();
            textPane.setDocument(new HTMLDocument());
            textPane.setEditorKit(new HTMLEditorKit());
        }
    }

    public static ComponentUI createUI(JComponent c) {
        return sharedInstance;
    }
    
    public void installUI(JComponent c) {
        super.installUI(c);
        rendererPane = new CellRendererPane();
        rendererPane.add(textPane);
        c.add(rendererPane);
    }
    
    public void uninstallUI(JComponent c) {
        super.uninstallUI(c);
        c.remove(rendererPane);
        
        if (rendererPane != null) 
            rendererPane.remove(textPane);
        
        rendererPane = null;
    }
    
    public void paint(Graphics g, JComponent c) {
        if (containsText) {
            Dimension size = c.getSize();
            rendererPane.paintComponent(g, textPane, c, 1, 1, size.width - 1, size.height - 1, true);
        } else {
            g.dispose();
        }
    }
    
    private void setHtmlText(String text) {
        String s = getHtmlText(text);
        
        HTMLDocument document = (HTMLDocument) textPane.getDocument();
        HTMLEditorKit kit = (HTMLEditorKit) textPane.getEditorKit();
        
        try {
            textPane.setText("");
            kit.insertHTML(document, 0, s, 0, 0, null);
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
    
    private String getHtmlText(String s) {
        String text = s;
        
        String html = "";
        int length = text != null ? text.length() : 0;
        if (length > 1) { 
            text = length > 1000 ? text.substring(0, 1000) + "...." : text;
            StringBuffer sb = new StringBuffer();
            sb.append("<html><body><table ");
            sb.append(getHtmlStyle(length));
            sb.append("><tr><td>");
            sb.append(text);
            sb.append("</td></tr></table></body></html>");
            html = sb.toString();
            containsText = true;
        } else {
            containsText = false;
        }
        
        return html;
    }
    
    private String getHtmlStyle(int length) {
      Font font = DcSettings.getFont(DcRepository.settings.stSystemFontNormal);
      
      Color colorEven = DcSettings.getColor(DcRepository.settings.stEvenRowColor);
      Color colorOdd = DcSettings.getColor(DcRepository.settings.stOddRowColor);
      String background = "";
      if (colorEven.getRGB() < colorOdd.getRGB()) 
          background = Utilities.getHexColor(colorOdd);
      else 
          background = Utilities.getHexColor(colorEven);
      
      Color color = ComponentFactory.getCurrentForegroundColor();
      String foreground = Utilities.getHexColor(color);
      
      String width = "";
      if (length > 100)
          width = ";width:400";
      
      String fontstyle = "";
      if (font.isItalic())
          fontstyle = ";font-style:italic";
      else if (font.isBold())
          fontstyle = ";font-weigth:bolder";
      
      String style = "style=" + '"' + 
                     "background:" + background + 
                     ";font-family:" + font.getFamily() + 
                     ";font-size:" + font.getSize() + 
                     fontstyle +
                     width +
                     ";color:" + foreground + ";" + '"';      

      return style;
    }       
    
    public Dimension getPreferredSize(JComponent c) {
        String tipText = ((JToolTip)c).getTipText();
        Dimension dim = new Dimension(0,0);
        setHtmlText(tipText);
        
        if (containsText) {
            int width = ((DcMultiLineToolTip)c).getFixedWidth();
            int columns = ((DcMultiLineToolTip)c).getColumns();
            
            if (columns > 0) {
                textPane.setSize(0,0);
                textPane.setSize( textPane.getPreferredSize() );
            } else if (width > 0) {
                Dimension d = textPane.getPreferredSize();
                d.width = width;
                d.height++;
                textPane.setSize(d);
            } 
            dim = textPane.getPreferredSize();
            dim.height += 1;
            dim.width += 1;
        }
        return dim;
    }
    
    public Dimension getMinimumSize(JComponent c) {
        return getPreferredSize(c);
    }
    
    public Dimension getMaximumSize(JComponent c) {
        return getPreferredSize(c);
    }
}
