/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcPanel;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.data.DataFilter;
import net.sf.dc.core.data.DataFilterEntry;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.data.Operators;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.DcPropertyModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.ext.approximatrix.CoordSystem;
import net.sf.ext.approximatrix.charting.model.ObjectChartDataModel;
import net.sf.ext.approximatrix.charting.render.BarChartRenderer;
import net.sf.ext.approximatrix.charting.render.PieChartRenderer;


public class ChartPanel extends DcPanel {
    
    private JComboBox comboFields;
    private JComboBox comboTypes;
    
    private net.sf.ext.approximatrix.charting.swing.ChartPanel chart;
    
    private final int module;
    
    public ChartPanel(int module) {
        super(DcResources.getText("lblCharts"), IconLibrary._icoChart);
        setHelpIndex("dc.charts");
        this.module = module;
        build();
    }
    
    private void buildChart() {
        DcField field = (DcField) comboFields.getSelectedItem();
        
        if (field == null) return;
        
        if (comboTypes.getSelectedIndex() == 1)
            buildBar(field);
        else
            buildPie(field);
    }
    
    private void deinstall() {
        if (chart != null) {
            chart.removeAll();
            remove(chart);
            chart = null;
            repaint();
        }
    }
    
    private void install() {
        chart.setBorder(ComponentFactory.getTitleBorder(""));
        add(chart, Layout.getGBC( 0, 1, 2, 1, 40.0, 40.0
           ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
            new Insets(5, 5, 5, 5), 0, 0));
        
        revalidate();
    }
    
    private String[] getSortedLabels(Map dataMap) {
        Collection c = new ArrayList();
        c.addAll(dataMap.keySet());
        
        Collections.sort((List) c, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((String) o1).compareTo((String) o2);
            }
        });
        
        String[] s = new String[c.size()];
        return (String[]) c.toArray(s);        
    }
    
    private void buildBar(DcField field) {
        Map dataMap = getDataMap(field);
        
        deinstall();
        
        if (dataMap == null) return;
        
        Double[][] data = new Double[1][dataMap.keySet().size()];
        int maximum = 0;
        String[] labels = getSortedLabels(dataMap);
        for (int i = 0; i < labels.length; i++) {
            String key = labels[i];
            int value = ((Integer) dataMap.get(key)).intValue();
            data[0][i] = new Double(value);
            maximum = maximum < value ? value : maximum;
        }
        
        // create the model
        ObjectChartDataModel model = new ObjectChartDataModel(data, labels, new String[] {field.getLabel()});  
        
        // create the chart
        CoordSystem coord = new CoordSystem(model);

        model.setManualScale(true);
        model.setMinimumValue(new Double(0.0));
        model.setMaximumValue(new Double(maximum * 1.5));
        
        chart = new net.sf.ext.approximatrix.charting.swing.ChartPanel(model, " ");
        chart.setCoordSystem(coord);
        chart.addChartRenderer(new BarChartRenderer(coord, model), 0);
        
        dataMap.clear();
        install();
    }
    
    private void buildPie(DcField field) {
        Map dataMap = getDataMap(field);

        deinstall();
        
        if (dataMap == null) return;
        
        int[][] data = new int[dataMap.keySet().size()][1];
        String[] labels = getSortedLabels(dataMap);
        for (int i = 0; i < labels.length; i++) {
            String key = labels[i];
            data[i][0] = ((Integer) dataMap.get(key)).intValue();
        }
        
        // create the model
        ObjectChartDataModel model = new ObjectChartDataModel(data, new String[] {field.getLabel()}, 
                                                              labels);
        
        // create the chart
        CoordSystem coord = new CoordSystem(model);
        coord.setPaintAxes(false);
        
        chart = new net.sf.ext.approximatrix.charting.swing.ChartPanel(model, field.getLabel());
        chart.setCoordSystem(coord);
        chart.addChartRenderer(new PieChartRenderer(model), 0);
        
        dataMap.clear();
        install();
    }
    
    private Collection getUniqueValues(int module, int field) {
        DcObject[] objects = DataManager.get(module, null);
        Collection values = new ArrayList();
  
        for (int i = 0; i < objects.length; i++) {
            String s = objects[i].getDisplayString(field);    
            if (!values.contains(s) && s.length() > 0)
                values.add(s);
        }
        return values;
    }   
    
    public static int getCount(int module, DcField field, Object value) {
        DcObject[] objects = DataManager.get(module, null);
        int count = 0;
        for (int i = 0; i < objects.length; i++) {
            String s1 = objects[i].getDisplayString(field.getIndex()); 
            String s2 = (String) value;
            
            if ((field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION && (s1.equals(s2) || s1.indexOf(s2) >= 0)) ||
                (s1.equals(s2)))
                count++;
        }
        return count;
    }    
    
    private Map getDataMap(DcField field) {
        Collection categories = new ArrayList();
        int module = field.getModule();
        
        if (module != field.getReferenceIdx()) {
            DcObject[] o = DataManager.get(field.getReferenceIdx(), null);
            for (int i = 0; i < o.length; i++) 
                categories.add(o[i].toString());
        } else {
            categories = getUniqueValues(module, field.getIndex());
        }
        
        if (categories.size() == 0) {
            new MessageBox(DcResources.getText("msgCouldNotCreateChart"), MessageBox._INFORMATION);
            return null;
            
        }
        
        // check for empty values and include these
        DataFilter df = new DataFilter(module);
        df.addEntry(new DataFilterEntry(DataFilterEntry._AND, module, field.getIndex(), Operators.getOperator(Operators._IS_EMPTY), null));
        
        int empty = DataManager.get(field.getModule(), df).length;
        if (field.getFieldType() != ComponentFactory._RATINGCOMBOBOX) {
            if (empty > 0) categories.add(DcResources.getText("lblEmpty"));
        }
        
        // create the data map. exclude zero counts
        Map dataMap = new HashMap();
        for (Iterator iter = categories.iterator(); iter.hasNext(); ) {
            String key = (String) iter.next();
            int count = getCount(module, field, key);
            
            if (key.equals(DcResources.getText("lblEmpty"))) {
                dataMap.put(key, new Integer(empty));
            } else if (count > 0) { 
                dataMap.put(key, new Integer(count));
            }
        }
        
        return dataMap.size() > 0 ? dataMap : null;
    }    

    private void build() {
        setLayout(Layout.getGBL());
        
        JPanel panel = new JPanel();
        
        comboFields = ComponentFactory.getComboBox();
        comboTypes = ComponentFactory.getComboBox();
        JButton button = ComponentFactory.getIconButton(IconLibrary._icoAccept);
        
        panel.add(comboFields);
        panel.add(comboTypes);
        panel.add(button);

        for (Iterator iter = DcModules.get(module).getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            if (((((field.getValueType() == DcRepository.datatypes._DCOBJECTREFERENCE ||
                   (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) &&
                   (DcModules.get(field.getReferenceIdx()) instanceof DcPropertyModule)) ||
                   field.getValueType() == DcRepository.datatypes._LONG ||
                   field.getValueType() == DcRepository.datatypes._BOOLEAN))) && 
                   field.isEnabled() && !field.isTechnicalInfo()) {
                
                comboFields.addItem(field);
            }
        }
        
        comboTypes.addItem(DcResources.getText("lblPie"));
        comboTypes.addItem(DcResources.getText("lblBar"));
        
        button.addActionListener(new SelectionListener());
        add(   panel, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
              ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
               new Insets(5, 5, 5, 5), 0, 0));
    }   
    
    public void setFont(Font font) {
        super.setFont(font);
        if (comboFields != null) {
            comboFields.setFont(font);
            comboTypes.setFont(font);
            
            if (chart != null)
                chart.setFont(font);
        }
    }
    
    private class SelectionListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            buildChart();
        }
    }
}
