/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.StringReader;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;

import net.sf.dc.actions.ToggleQuickViewAction;
import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcPictureField;
import net.sf.dc.console.components.DcPopupMenu;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.windows.itemforms.ItemForm;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.settings.definitions.Definitions;
import net.sf.dc.settings.definitions.QuickViewFieldDefinition;
import net.sf.dc.util.Utilities;

public class QuickViewPanel extends JPanel {
    
    public static final String _DIRECTION_HORIZONTAL = DcResources.getText("lblHorizontal");
    public static final String _DIRECTION_VERTICAL = DcResources.getText("lblVertical");  
    
    protected final DcModule module;
    protected final boolean showImages;
    
    protected DcObject dco;
    private LinkedList pictures = new LinkedList();
    private LinkedList imagePanels = new LinkedList();

    private JTextPane descriptionPane;
    private JScrollPane scroller;
    private final JTabbedPane tabbedPane = ComponentFactory.getTabbedPane();
    
    private final HTMLEditorKit kit = new HTMLEditorKit();
    private final HTMLDocument document = new HTMLDocument();
    
    public QuickViewPanel(DcModule module, boolean showImages) {
        this.module = module;
        this.showImages = showImages;
        setLayout(Layout.getGBL());
        tabbedPane.addChangeListener(new TabListener());
        buildPanel();
    }    
    
    public void setFont(Font font) {
        super.setFont(font);

        if (tabbedPane != null)
            tabbedPane.setFont(ComponentFactory.getSystemFont());
        
        if (dco != null) {
            setObject(dco);            
        }
    }
    
    private synchronized void loadImage() {
        int index = tabbedPane.getSelectedIndex() - 1;
        Picture picture =  (Picture) pictures.get(index);
        picture.loadImage();
        
        JPanel panel =  (JPanel) imagePanels.get(index);
        Component[] components =  panel.getComponents();
        for (int i = 0; i < components.length; i++) {
            if (components[i] instanceof DcPictureField) {
                DcPictureField picField = (DcPictureField) components[i];
                picField.setValue(picture.getValue(Picture._D_IMAGE));
                picField.repaint();
            }
        }
    }
    
    public synchronized void setImages() {
        try {
            clearImages();
    
            if (dco == null)
                return;
            
            DcFieldDefinitions definitions = dco.getModule().getFieldDefinitions(); 
            
            for (Iterator iter =  definitions.iterator(); iter.hasNext(); ) {
                DcFieldDefinition definition = (DcFieldDefinition) iter.next();
                if (    dco.isEnabled(definition.getIndex()) && 
                        dco.getField(definition.getIndex()).getValueType() == DcRepository.datatypes._PICTURE) {
                    
                    Picture picture = (Picture) dco.getValue(definition.getIndex());
                    if (picture != null) {
                        pictures.add(picture);    
                        
                        DcPictureField picField = ComponentFactory.getPictureField(true, false, false, definition.getLabel());
                        picField.setBorder(ComponentFactory.getBorder());
                        
                        JPanel panel = new JPanel();
                        panel.setLayout(Layout.getGBL());
                        panel.addMouseListener(new QuickViewMouseListener());
                        panel.add(picField, Layout.getGBC(0, 0, 1, 1, 10.0, 10.0,
                                GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                new Insets(2, 2, 2, 2), 0, 0));
                            
                        tabbedPane.addTab(dco.getLabel(definition.getIndex()), IconLibrary._icoPicture, panel);
                        imagePanels.add(panel);
                    }
                }
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
    
    public synchronized void setObject(DcObject dco) {
        
        try {
            if (dco == null)
                return;
            
            if (DcModules.getCurrent().getIndex() == DcModules._MEDIA &&
                DcModules.getCurrent().getCurrentSearchView().getIndex() == MasterView._TABLE_VIEW)
                dco.reload();
            
            int tab = tabbedPane.getSelectedIndex();
            clear();
            this.dco = dco;
            
            String html = "<html><body " + getBodyStyle() + ">\n" +
                            getDescriptionTable() +
                          "</body> </html>";
            
            setDescriptionHtml(html);
            
            try {
                descriptionPane.setCaretPosition(0);
            } catch (Exception exp) {}
            
            setImages();
            
            boolean error = true;
            tab += 1;
            
            // prevent endless loop
            int counter = 0;
            while (error && counter < 6) {
                counter ++;
                try {
                    tab -= 1;
                    tabbedPane.setSelectedIndex(tab);
                    error = false;
                } catch (Exception ignore) {}
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
    
    public synchronized void clear() {
        Dimension dim = getSize();
        setPreferredSize(dim);
        if (dco != null) {
            dco = null;
            descriptionPane.setText("");

            try {
                tabbedPane.setSelectedIndex(0);
            } catch (Exception exp) {}
            
            clearImages();
        }
    }
    
    private void removeTabs() {
        tabbedPane.removeAll();
    }
    
    private synchronized void clearImages() {
        removeTabs();
        tabbedPane.addTab("Description", IconLibrary._icoInformation ,scroller);
        
        for (Iterator iter = imagePanels.iterator(); iter.hasNext(); ) {
            JPanel panel = (JPanel) iter.next();
            Component[] components = panel.getComponents();
            for (int i = 0; i < components.length; i++)
                ComponentFactory.clean(components[i]);
        }
        pictures.clear();
        imagePanels.clear();
    }
    
    private String getDescriptionTable() {
        String table = "<table " + getBodyStyle() + ">\n";
        
        Definitions definitions = module.getSettings().getDefinitions(DcRepository.modulesettings.stQuickViewFieldDefinitions);
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            QuickViewFieldDefinition def = (QuickViewFieldDefinition) iter.next();
            
            if (def.isEnabled())
                table = addTableRow(table, def.getField(), def.getDirectrion());    
        }
        table += "</table>";    
        
        if (module.getChild() != null) {
            table += "\n\n";
            table += getChildTable(module.getChild());
        }
        
        return table;
    }
    
    private String getChildTable(DcModule module) {
        String body = "";
        Definitions definitions = module.getSettings().getDefinitions(DcRepository.modulesettings.stQuickViewFieldDefinitions);
        
        dco.removeChildren();
        dco.loadChildren();
        
        int counter = 0;
        for (Iterator iterChildren = dco.getChildren().iterator(); iterChildren.hasNext(); counter++) {
            DcObject child = (DcObject) iterChildren.next();
            
            if (counter > 0)
                body += "<b> / </b>";
            
            boolean first = true;
            for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
                QuickViewFieldDefinition def = (QuickViewFieldDefinition) iter.next();
                String value = child.getDisplayString(def.getField());
                if (def.isEnabled() && value.trim().length() > 0) {
                    if (first)
                        body += "<b>" + value + "</b>";
                    else
                        body += ", " + value;

                    first = false;
                }
            }
        }
        
        String table = "<table " + getBodyStyle() + "><tr><td>\n";
        table += body;
        table += "</tr></td></table>";
        
        return table;
    }
    
    protected synchronized void setDescriptionHtml(String html) {
        try {
            descriptionPane.read(new StringReader(html), "Data Crow");
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
    
    protected String addTableRow(String htmlTable, int index, String direction) {
        String table = htmlTable;
        if (dco.isEnabled(index)) {
            Font font = DcSettings.getFont(DcRepository.settings.stSystemFontNormal);
            boolean horizontal = direction.equals(_DIRECTION_HORIZONTAL);

            String value = dco.getDisplayString(index);
            String realValue = dco.getValue(index) == null ? "" : dco.getValue(index).toString();
            String displayString = value == null ? "" : value.toString();
            if (realValue != null && displayString.length() > 0 ) {
                table += "<tr><td>\n";
                table += "<b>" + dco.getLabel(index) + "</b>";
                
                if (!horizontal) {
                    table += "</td></tr>";
                    table += "<tr><td>\n";
                } else {
                    table += " ";
                }
                
                if (font.getStyle() == Font.BOLD) 
                    value = "<b>" + value + "</b>";
                else
                    value += "<br><br>";
                    
                table += value;
                table += "</td></tr>";
            } else if (dco.getField(index).getValueType() == DcRepository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(index);
                pictures.add(picture);
            }
        }
        return table;
    }    
    
    private synchronized void buildPanel() {
        // description panel
        descriptionPane = ComponentFactory.getTextPane();
        descriptionPane.setEditorKit(kit);
        descriptionPane.setDocument(document);
        descriptionPane.setEditable(false);
        descriptionPane.setBounds(1,1,1,10);
        descriptionPane.addMouseListener(new QuickViewMouseListener());

        scroller = new JScrollPane(descriptionPane);
        scroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        
        // tabbed pane
        tabbedPane.addTab("Description", IconLibrary._icoInformation ,scroller);
        add(tabbedPane, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0,
            GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
            new Insets(4, 0, 3, 0), 0, 0));
        
        String html = "<html><body " + getBodyStyle() + "></body> </html>";
        setDescriptionHtml(html);
    }    
    
    private void showPopupMenu(int x, int y) {
        PopupMenu popupMenu = new PopupMenu();
        popupMenu.show(this, x, y);
    }    
    
    protected String getBodyStyle() {
        Color colorEven = DcSettings.getColor(DcRepository.settings.stEvenRowColor);
        Color colorOdd = DcSettings.getColor(DcRepository.settings.stOddRowColor);
        
        
        Color color = colorEven.getRGB() < colorOdd.getRGB() ? colorOdd : colorEven;
        String background = Utilities.getHexColor(color);
       
        descriptionPane.setBackground(color);
        for (Iterator iter = imagePanels.iterator(); iter.hasNext(); ) {
            ((JPanel) iter.next()).setBackground(color);
        }
        
        String hexColor = Utilities.getHexColor(ComponentFactory.getCurrentForegroundColor());
        Font font = DcSettings.getFont(DcRepository.settings.stSystemFontNormal);
        
        String fontstyle = "";
        if (font.isItalic())
            fontstyle = ";font-style:italic";
        else if (font.isBold())
            fontstyle = ";font-weigth:bolder";
        
        String style = "style=" + '"' + 
                       ";background:" + background + 
                       ";font-family:" + font.getFamily() + 
                       ";font-size:" + font.getSize() + fontstyle +
                       ";color:" + hexColor + ";" + '"';
        return style;
    }
    
    public class PopupMenu extends DcPopupMenu {
        public PopupMenu() {
            JMenuItem menuToggleLabels = ComponentFactory.getMenuItem(DcResources.getText("lblToggleQuickView"));
            menuToggleLabels.addActionListener(new ToggleQuickViewAction());
            this.add(menuToggleLabels);
        }
    }   
    
    private class TabListener implements ChangeListener {
        public void stateChanged(ChangeEvent evt) {
            JTabbedPane pane = (JTabbedPane) evt.getSource();
            
            if (pane.getSelectedIndex() > 0)
                loadImage();
        }
    } 
    
    private class QuickViewMouseListener implements MouseListener {
        public void mouseReleased(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                showPopupMenu(e.getX(), e.getY());
            } 

            if (e.getClickCount() == 2 && dco != null) {
            	dco.markAsUnchanged();
                ItemForm itemform = new ItemForm(null, false, true, dco, false);
                itemform.setVisible(true);
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    }
}
