/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcTree;
import net.sf.dc.console.menu.DcTreePanelMenuBar;
import net.sf.dc.console.views.View;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.data.DataFilter;
import net.sf.dc.core.data.DataFilters;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;

public class TreePanel extends JPanel {

    public static final int  _OBJECT_ADDED = 0;
    public static final int  _OBJECT_UPDATED = 1;
    public static final int  _OBJECT_REMOVED = 2;
    
    private final int module;
    private final View view;
    private int[] fields;
    
    private DcTree tree;
    private JScrollPane scroller;
    private DefaultMutableTreeNode top;
    
    private Object currentUserObject;
    private DcTreePanelMenuBar menu;
    
    private Collection objects = new ArrayList();
    
    private final TreeSelectionListener tsl = new DcTreeSelectionListener();
    
    public TreePanel(int module, View view) {
        this.view = view;
        this.menu = new DcTreePanelMenuBar(module, this);
        this.module = module;
        setLayout(Layout.getGBL());
        initializeMenu();
        fields = (int[]) DcModules.get(module).getSetting(DcRepository.modulesettings.stGroupedBy);
    }
    
    public Object getLastSelectedPathComponent() {
        return tree.getLastSelectedPathComponent();
    }
    
    public synchronized void groupBy(int[] fields) {
        this.fields = fields;
        DcModules.get(module).setSetting(DcRepository.modulesettings.stGroupedBy, fields);
        
        Collection c = getObjects();
        
        DcObject[] objects = new DcObject[c.size()];
        objects = (DcObject[]) c.toArray(objects);
        
        buildTree(objects);
    }

    public void setFont(Font font) {
        if (tree != null) {
            tree.setFont(font);
            menu.setFont(font);
        }
    }
    
    public void add(DcObject[] objects) {
        fields = (int[]) DcModules.get(module).getSetting(DcRepository.modulesettings.stGroupedBy);
        buildTree(objects);
    }

    public synchronized void add(DcObject dco) {
        if (tree == null)  {
            initializeTree();
            buildTree();
        }
        
        if (tree != null) {
            if (!objects.contains(dco)) {
                objects.add(dco);
                revalidateTree(dco, _OBJECT_ADDED);
            }
        }
    }

    public synchronized void remove(DcObject dco) {
        if (tree != null) {
            tree.removeTreeSelectionListener(tsl);
            if (objects.contains(dco)) {
                objects.remove(dco);
                revalidateTree(dco, _OBJECT_REMOVED);
            }
            tree.addTreeSelectionListener(tsl);
        }
    }
        
    public synchronized void remove(String s) {
        DcObject dco = DcModules.get(module).getDcObject();
        dco.setValue(DcObject._ID, s);
        remove(dco);
    }     
    
    public synchronized void remove(String[] ids) {
        for (int i = 0; i < ids.length; i++) 
            remove(ids[i]);
    } 

    
    public synchronized void update(DcObject dco) {
        if (tree != null) {
            int pos = ((List) objects).indexOf(dco);
            if (pos != -1) {
                DcObject o = (DcObject) ((List) objects).get(pos);
                o.reload();
                revalidateTree(o, _OBJECT_UPDATED);
            }
        }
    }    
    
    private void removeElement(DcObject dco) {
        
        Collection nodes = fields != null && fields.length > 0 ? findNodes(dco, top) : new ArrayList();
        
        for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
            DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) iter.next();
            NodeElement parentElement = (NodeElement) parentNode.getUserObject();
            parentElement.removeValue(dco);
            if (fields.length > 1) {
                Collection childNodes = findNodes(dco, parentNode);
                for (Iterator iterChildren = childNodes.iterator(); iterChildren.hasNext(); ) {
                    DefaultMutableTreeNode childNode = (DefaultMutableTreeNode) iterChildren.next();
                    NodeElement childElement = (NodeElement) childNode.getUserObject();
                    childElement.removeValue(dco);
                }
            }
        }
        
        removeEmptyLeafs(top);
    }
    
    private boolean clearSafely = true;

    private void addElement(DcObject dco, boolean select) {
        addElement(dco, top, 0, select);
    }
    
    private synchronized void addElement(DcObject dco, DefaultMutableTreeNode parent, int level, boolean select) {
        try {
            
            boolean grouped = fields != null;
            for (int i = 0; grouped && i < fields.length; i++)
                grouped = fields[i] != -1;
            
            if (!grouped)
                return;
            
            Collection nodes = fields.length > 0 ? findNodes(dco, fields[level], parent) : new ArrayList();
            clearSafely = false;
            
            if (level == 0 && fields != null && fields.length > 0 && nodes.size() == 0) {
//              no nodes available, create all
                DefaultMutableTreeNode parentNode = parent;
                for (int i = 0; i < fields.length; i++) {
                    parentNode = createLeaf(parentNode, dco, i);
                }
                
                if (select && parentNode != null) 
                    setSelected(parentNode);
                
            } else if (nodes.size() > 0) {
//              add and create where needed
                for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
                    DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) iter.next();
                    NodeElement parentElement = (NodeElement) parentNode.getUserObject();
                    parentElement.addValue(dco);
                    if (fields.length > 1 && level < 1) 
                        addElement(dco, parentNode, 1, select);
                    else if (select) 
                        setSelected(parentNode);
                }
            } else if (isActive()) {
                DefaultMutableTreeNode node = createLeaf(parent, dco, level);
                if (select && node != null) setSelected(node);
            }
        } catch (Exception e) {
            DcLog.add(e);
        } finally {
            clearSafely = true;
        }
    }
    
    private void setSelected(DefaultMutableTreeNode node) {
        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
        TreePath treePath = new TreePath(model.getPathToRoot(node));
        tree.setSelectionPath(treePath);
        tree.scrollPathToVisible(treePath);
    }
    
    private synchronized DefaultMutableTreeNode createLeaf(DefaultMutableTreeNode parent, DcObject dco, int level) {
        int fldIdx = fields[level]; 
        
        DefaultMutableTreeNode node = null;
        if (    DcModules.get(module).getField(fldIdx).getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION &&
                dco.getValue(fldIdx) != null && ((Collection) dco.getValue(fldIdx)).size() > 0) {
            
            Collection mappings = (Collection) dco.getValue(fldIdx);
            for (Iterator iter = mappings.iterator(); iter.hasNext(); ) {
                DcObject reference = ((DcMapping) iter.next()).getReferencedObject();
                
                // because references are not garuanteed to exist within Data Crow we have
                // to check whether the reference actually exists
                if (reference != null) {
                	NodeElement ne = new NodeElement();
                	ne.setKey(reference.toString());
                	ne.addValue(dco);
                	node = new DefaultMutableTreeNode(ne);
                	parent.add(node);
                	insertNode(node, parent);
                }
            }
        } else {
            NodeElement element = new NodeElement();
            element.addValue(dco);
            element.setKey(dco.getDisplayString(fields[level]));

            node = new DefaultMutableTreeNode(element);
            parent.add(node);
            insertNode(node, parent);
        }
        return node;
    }
    
    /**
     * Inserts a node. The node needs to have a valid user object defined
     * to insert it at the right position (ordering)
     */
    private synchronized void insertNode(DefaultMutableTreeNode node, DefaultMutableTreeNode parent) {
        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
        
        List elements = new ArrayList();
        for (int i = 0; i < parent.getChildCount(); i++) {
            elements.add(((DefaultMutableTreeNode) parent.getChildAt(i)).getUserObject().toString());
        }
        
        elements.add(node.getUserObject().toString());
        Collections.sort(elements);
        int idx = elements.indexOf(node.getUserObject().toString());
        
        model.insertNodeInto(node, parent, idx);
    }
    
    private synchronized void removeEmptyLeafs(DefaultMutableTreeNode parent) {
        try {
            int count = parent.getChildCount();
            
            tree.removeTreeSelectionListener(tsl);
            for (int i = count; i > 0; i--) {
                DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) parent.getChildAt(i - 1);
                
                if (parentNode.getUserObject() instanceof NodeElement) {
                    NodeElement ne = (NodeElement) parentNode.getUserObject();
                    if (ne.getValues().size() == 0) {
                        tree.setSelectionInterval(0, 0);
                        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
                        model.removeNodeFromParent(parentNode);
                    } else {
                        removeEmptyLeafs(parentNode);
                    }
                }
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        } finally {
            tree.addTreeSelectionListener(tsl);
        }
    }
    
    private synchronized void revalidateTree(DcObject dco, int modus) {
        if (modus == _OBJECT_REMOVED) {
            removeElement(dco);
        } else if (modus == _OBJECT_UPDATED) {
            removeElement(dco);
            addElement(dco, (dco.isBatch() ? dco.isEndOfBatch() : true));
        } else if (modus == _OBJECT_ADDED) {
            removeElement(dco);
            addElement(dco, true);
        }

        repaint();
        revalidate();
    }
    
    private synchronized Collection findNodes(DcObject dco, int field, DefaultMutableTreeNode parentNode) {
        Collection nodes = new ArrayList();
        if (field == -1)
            return nodes;
            
        if (    dco.getField(field).getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION&&
                dco.getValue(field) != null && ((Collection) dco.getValue(field)).size() > 0) {
            
            Collection c = (Collection) dco.getValue(field);
            for (Iterator iter = c.iterator(); iter.hasNext(); ) {
                Object o = iter.next();
                DcMapping mapping = (DcMapping) o;
                String ID = (String) mapping.getValue(DcMapping._B_REFERENCED_ID);
                DcObject so = DataManager.getObject(mapping.getReferencedModuleIdx(), ID);
                
                DefaultMutableTreeNode node = so != null ? findNode(so.toString(), parentNode) : null;
                
                if (node != null) { 
                    nodes.add(node);
                } else if (so != null){
                    // workaround
                    NodeElement ne = new NodeElement();
                    ne.setKey(so.toString());
                    ne.addValue(dco);
                    node = new DefaultMutableTreeNode(ne);
                    parentNode.add(node);
                    insertNode(node, parentNode);
                    nodes.add(node);
                }
            }
        } else {
            DefaultMutableTreeNode node = findNode(dco.getDisplayString(field), parentNode);
            if (node != null) nodes.add(node);
        }
        
        return nodes;
    }

    private synchronized DefaultMutableTreeNode findNode(String key, DefaultMutableTreeNode parentNode) {
        int count = parentNode != null ? parentNode.getChildCount() : 0;
        for (int i = 0; i < count; i++) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode) parentNode.getChildAt(i);
            
            if (node.getUserObject() instanceof NodeElement) {
                NodeElement ne = (NodeElement) node.getUserObject();
                if (ne.getComparableKey().equals(key.toLowerCase()))
                    return node;
            }
        }
        return null;
    }    
    
    private synchronized Collection findNodes(DcObject dco, DefaultMutableTreeNode parentNode) {
        Collection nodes = new ArrayList();
        
        int count = parentNode.getChildCount();
        for (int i = 0; i < count; i++) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode) parentNode.getChildAt(i);
            
            if (node.getUserObject() instanceof NodeElement) {
                NodeElement ne = (NodeElement) node.getUserObject();
                if (ne.getValues().contains(dco))
                    nodes.add(node);
            }
        }
        return nodes;
    }
    
    public void collapseAll() {
        if (top == null) return;
        collapseChildren(top);
        tree.collapsePath(new TreePath(top.getPath()));
        view.clear(false);
    }    
    
    public void collapseChildren(DefaultMutableTreeNode parent) {
        int size = parent.getChildCount();
        for (int i = 0; i < size; i++) {
            try {
                DefaultMutableTreeNode child = (DefaultMutableTreeNode) parent.getChildAt(i);
                tree.collapsePath(new TreePath(child.getPath()));
            } catch (Exception exp) {
                DcLog.add(exp);
            }
        }
    }
    
    public void expandAll() {
        expandAll(top);
    }
    
    private void expandAll(DefaultMutableTreeNode node) {
        try {
            DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
            int count = node.getChildCount();
            tree.expandPath(new TreePath(model.getPathToRoot(node)));
            for (int i = 0; i < count; i++) {
                DefaultMutableTreeNode current = (DefaultMutableTreeNode) node.getChildAt(i);
                tree.expandPath(new TreePath(model.getPathToRoot(current)));
                expandAll(current);
            }
        } catch (Exception exp) {}
    }
    
    public boolean isActive() {
        boolean active = isEnabled() && isVisible() && top != null && fields != null && fields.length > 0;
        
        if (active) {
            for (int i = 0; i < fields.length; i++)
                active &= fields[i] != -1;
        }
            
        return active;
    }
    
    public Collection getObjects() {
        Collection c = new ArrayList();
        c.addAll(objects);
        return c;
    }
    
    private void addLeaf(DefaultMutableTreeNode parentNode, String key) {
        
        DefaultMutableTreeNode existing = findNode(key, parentNode);
        NodeElement currentNode;
        if (existing != null && existing.getUserObject() != null) {
            currentNode = (NodeElement) existing.getUserObject();
        } else {
            currentNode = new NodeElement();
            currentNode.setKey(key);
            parentNode.add(new DefaultMutableTreeNode(currentNode));
        }
        
    }
    
    private synchronized void addLeafs(DefaultMutableTreeNode parentNode, Collection objects, int level) {
        int fldIdx = fields[level];
        
        if (DcModules.get(module).getField(fldIdx).getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
            DcField fld = DcModules.get(module).getField(fldIdx);
            DcObject[] c =  DataManager.get(DcModules.getReferencedModule(fld).getIndex(), null);
            
            for (int i = 0; c != null &&  i < c.length; i++) {
                String key = c[i].toString();
                addLeaf(parentNode, key);
            }
            
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next();
                addElement(dco, false);
            }

            removeEmptyLeafs(parentNode);
            
        } else {
            Collections.sort((List) objects, new DcObjectComparator(fldIdx));
            
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next();
                String key = dco.getDisplayString(fldIdx);
                addLeaf(parentNode, key);
                
                if (view != null) {
                    view.setStatus(DcResources.getText("msgAddingXToView", dco.toString()));
                    view.updateProgressBar();
                }
            
                addElement(dco, false);
            }
        }
    }
    
    private synchronized void buildTree(DcObject[] o) {
        clear();

        objects = new ArrayList();
        for (int i = 0; i < o.length; i++)
            objects.add(o[i]);

        buildTree();
    }
    
    public synchronized void buildTree() {
        
        initializeTree();

        if (!isActive()) {
            NodeElement node = new NodeElement();
            node.setKey("All");
            node.setValues(objects);
            top.add(new DefaultMutableTreeNode(node));
        } else {
            addLeafs(top, objects, 0);

            for (int level = 1; level < fields.length; level++) {
                int size = top.getChildCount();
                for (int pos = 0; pos < size; pos++) {
                    DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) top.getChildAt(pos);
                    NodeElement ne = (NodeElement) parentNode.getUserObject();
                    addLeafs(parentNode, ne.getValues(), level);
                }
            }
        }

        expandAll();
        tree.setSelectionInterval(1, 1);
    }
    
    public void reset() {
        DcModules.get(module).setSetting(DcRepository.modulesettings.stGroupedBy, null);
        fields = null;
        Collection objects = getObjects();
        clear();
        initializeTree();
        
        DcObject[] o = new DcObject[objects.size()];
        add((DcObject[]) objects.toArray(o));
    }
    
    private void clear() {
        if (tree != null) {
            objects.clear();
            
            tree.removeTreeSelectionListener(tsl);
            tree = null;
            
            top.removeAllChildren();
            remove(scroller);
            top = null;
            
            revalidate();
            repaint();
        }
    }
    
    private void initializeTree() {
        if (tree != null) {
            return;
        }
            
        DcModule mod = DcModules.get(module);
        String orderingOn = mod.getObjectNamePlural();
        
        if (isActive()) {
            orderingOn += " by ";
            
            for (int i = 0; i < fields.length; i++)
                orderingOn += (i > 0 ? " & " : "") + mod.getField(fields[i]).getLabel();
        }
        
        top = new DefaultMutableTreeNode(orderingOn);
        tree = new DcTree(new DefaultTreeModel(top));
        tree.setFont(ComponentFactory.getStandardFont());
        tree.addTreeSelectionListener(tsl);
        
        scroller = new JScrollPane(tree);
        
        add(scroller,  Layout.getGBC( 0, 1, 1, 1, 50.0, 50.0
           ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
            new Insets(0, 5, 5, 5), 0, 0));
        
        this.revalidate();
        this.repaint();
    }
    
    private void initializeMenu() {
        
        add(menu,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                  ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                   new Insets(5, 5, 0, 5), 0, 0));
    }
    
    private class DcTreeSelectionListener implements TreeSelectionListener {
        public void valueChanged(TreeSelectionEvent e) {
            try {
                if (e.getNewLeadSelectionPath().equals(e.getOldLeadSelectionPath()))
                    return;
                
                DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
    
                if (node == null) {
                    currentUserObject = null;
                    return;
                }
    
                Object o = node.getUserObject();
                
                if (!o.equals(currentUserObject)) {
                    currentUserObject =  o;
                    if (currentUserObject instanceof NodeElement) {
                        view.clear(clearSafely);
                        NodeElement currentNode = (NodeElement) o;
                        view.add(currentNode.getValues(), false);
                    } else {
                        view.clear(clearSafely);
                    }
                }
            } catch (Exception ignore) {}
        }
    }  
    
    private class NodeElement {
        
        private String key;
        private String keyComparable;
        private Collection values = new ArrayList();
        
        public NodeElement() {
        }

        public void setKey(String key) {
            this.key = key;
            this.keyComparable = key.toLowerCase();
        }

        public void setValues(Collection values) {
            this.values = values;
        }
        
        public void addValue(DcObject dco) {
            if (!values.contains(dco))
                values.add(dco);
        }
        
        public void updateValue(DcObject dco) {
            int pos = ((List) values).indexOf(dco);
            if (pos != -1) {
                DcObject o = (DcObject) ((List) values).get(pos);
                o.reload();
            }
        }
        
        public int hashCode() {
            return key.hashCode();
        }
        
        public void removeValue(DcObject dco) {
            values.remove(dco);
        }
        
        public Collection getValues() {
            if (values != null) {
                DataFilter df = DataFilters.getDefaultDataFilter(module);
                df.sort(values);
            }
            return values;
        }
        
        public String getComparableKey() {
            return keyComparable;
        }
        
        public String getKey() {
            return key;
        }
        
        public String toString() {
            return key + " (" + values.size() + ")";
        }
        
        public void clear() {
            key = null;
            keyComparable = null;
            
            if (values != null)
                values.clear();
            
            values = null;
        }
    }
    
    private class DcObjectComparator implements Comparator {

        private final int field;
        
        public DcObjectComparator(int field) {
            this.field = field;
        }
        
        public int compare(Object o1, Object o2) {
            DcObject dco1 = (DcObject) o1;
            DcObject dco2 = (DcObject) o2;
            
            Object value1 = dco1.getValue(field);
            Object value2 = dco2.getValue(field);
            int valuetype = dco1.getField(field).getValueType(); 
            
            if (valuetype == DcRepository.datatypes._BIGINTEGER ||
                valuetype == DcRepository.datatypes._LONG) {
                
                long long1 = value1 == null ? 0 : ((Number) value1).longValue();
                long long2 = value2 == null ? 0 : ((Number) value2).longValue();
                
                return (int) (long1 - long2);

            } else if (valuetype == DcRepository.datatypes._DATE) {
                
                Date date1 = value1 == null ? new Date(0) : (Date) value1;
                Date date2 = value2 == null ? new Date(0) : (Date) value2;
                
                return date1.compareTo(date2);
            
            } else {
                
                return dco1.getDisplayString(field).toLowerCase().compareTo(
                       dco2.getDisplayString(field).toLowerCase());
            }
        }
    }
}
