/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.menu;

import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JToolBar;
import javax.swing.JToolTip;
import javax.swing.SwingUtilities;

import net.sf.dc.actions.CreateNewAction;
import net.sf.dc.actions.CreateReportAction;
import net.sf.dc.actions.DcAction;
import net.sf.dc.actions.DeleteAction;
import net.sf.dc.actions.HelpAction;
import net.sf.dc.actions.ImportAction;
import net.sf.dc.actions.LogAction;
import net.sf.dc.actions.NewItemWizardAction;
import net.sf.dc.actions.OnlineSearchAction;
import net.sf.dc.actions.OpenAction;
import net.sf.dc.actions.SaveAllAction;
import net.sf.dc.actions.SaveSelectedAction;
import net.sf.dc.actions.SearchAction;
import net.sf.dc.actions.SearchAgainAction;
import net.sf.dc.actions.SettingsAction;
import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.DcMultiLineToolTip;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;

public class DcToolBar extends JToolBar {

	private final PopupMenu popupMenu = new PopupMenu();

    public DcToolBar(DcModule module) {

        add(new ToolBarButton(new NewItemWizardAction()));
        
        addSeparator();
        
		add(new ToolBarButton(new CreateNewAction()));
        add(new ToolBarButton(new OpenAction()));
        add(new ToolBarButton(new SaveSelectedAction()));
        add(new ToolBarButton(new SaveAllAction()));
        add(new ToolBarButton(new DeleteAction()));
        
        addSeparator();
        
        add(new ToolBarButton(new SearchAction()));
        add(new ToolBarButton(new SearchAgainAction()));
        
        if (module.hasOnlineServices())
            add((new ToolBarButton(new OnlineSearchAction())));
        
        addSeparator();
        
        add(new ToolBarButton(new SettingsAction()));
        add(new ToolBarButton(new CreateReportAction()));
        
        if (module.getImporterClass() != null)
            add(new ToolBarButton(new ImportAction()));
        
        addSeparator();
        
        add(new ToolBarButton(new LogAction()));
        add(new ToolBarButton(new HelpAction()));
        
        boolean showLabels = DcSettings.getBoolean(DcRepository.settings.stShowMenuBarLabels);
        showLabels(showLabels);
	}
    
    private void showPopupMenu(int x, int y) {
        popupMenu.show(this, x, y);
    }

    public void toggleLabels() {
        showLabels(!DcSettings.getBoolean(DcRepository.settings.stShowMenuBarLabels));
    }
    
    public void setFont(Font font) {
        Component c;
        int i = 0;
        while ((c = getComponentAtIndex(i++)) != null) {
            if (c instanceof ToolBarButton) {
                ToolBarButton button = (ToolBarButton) c;
                button.setFont(ComponentFactory.getSystemFont());
            }
        }
    }    
    
	public void showLabels(boolean show) {
		Component c;
		int i = 0;
		while ((c = getComponentAtIndex(i++)) != null) {
            if (c instanceof ToolBarButton) {
    			ToolBarButton button = (ToolBarButton) c;
    			if (show) {
    				button.setText(button.getToolTipText());
                } else {
    				button.setText(null);
                }
            }
		}
        DcSettings.set(DcRepository.settings.stShowMenuBarLabels, new Boolean(show));
	}
    
	public class ToolBarButton extends JButton {

		private static final long serialVersionUID = 4L;

		public ToolBarButton(DcAction action) {
			super(action.getIcon());
            setFont(ComponentFactory.getSystemFont());
            setText(action.getLabelShort());
            setVerticalTextPosition(BOTTOM);
			setHorizontalTextPosition(CENTER);
            setToolTipText(action.getLabelShort());
            
            addActionListener(action);
            
            addMouseListener(new ToolBarMouseListener());
		}
        
        public JToolTip createToolTip() {
            return new DcMultiLineToolTip();
        }
	}
    
    public class PopupMenu extends net.sf.dc.console.components.DcPopupMenu {

		private static final long serialVersionUID = 6L;

		public PopupMenu() {
            JMenuItem menuToggleLabels = new JMenuItem(DcResources.getText("lblToggleMenuLabels"));

            menuToggleLabels.addActionListener(new ToggleLabelsAction());
            this.add(menuToggleLabels);
        }
    }
    
    private class ToggleLabelsAction implements ActionListener {
		public void actionPerformed(ActionEvent arg0) {
            boolean showLabels = DcSettings.getBoolean(DcRepository.settings.stShowMenuBarLabels);
			showLabels(!showLabels);
		}
    }
    
    private class ToolBarMouseListener implements MouseListener {
        public void mouseReleased(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                showPopupMenu(e.getX(), e.getY());
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    } 

    
    protected void paintComponent(Graphics g) {
        Graphics2D g2d = (Graphics2D) g;
        
        if (DcSettings.getBoolean(DcRepository.settings.stFontAntiAliasing)) {
           g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
           g2d.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
           g2d.setRenderingHint( RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY );
        }
        
        super.paintComponent( g2d );
    }      
}