/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.views.table;

import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collection;
import java.util.Iterator;
import java.util.StringTokenizer;

import javax.swing.DefaultCellEditor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.SwingUtilities;

import net.sf.dc.actions.AddRowAction;
import net.sf.dc.actions.RemoveRowAction;
import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.panels.QuickViewPanel;
import net.sf.dc.console.components.renderers.DcTableCellRenderer;
import net.sf.dc.console.components.tables.DcTable;
import net.sf.dc.console.menu.DcPopupMenu;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.views.View;
import net.sf.dc.console.windows.messageboxes.QuestionBox;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.util.DataTask;

public abstract class TableBaseView extends View {

    protected DcTable table;

    protected boolean dispatchEvents = true;

    protected QuickViewPanel quickView;
    protected JPanel panelInput  = new JPanel();
    protected JPanel panelStatus = getStatusPanel();
    protected JScrollPane scrollResults1;
    protected JScrollPane scrollResults2;

    protected TableBaseView childView;
    
    protected TableActionsPanel panelTableActions;

    protected final TableMouseListener tableMouseListener = new TableMouseListener();

    private final DcModule moduleChild;

    private final Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

    public TableBaseView(DcModule module, String title, ImageIcon icon) {
        super(module, title, icon, MasterView._TABLE_VIEW);
    	this.moduleChild = module != null ? module.getChild() : null;
    }

    public void reload(String ID) {}    
    
    public JComponent[] getExtraComponents() {
        return new JComponent[0];
    }

    public void markAsUnchanged() {
        table.clearCache();
    }     
    
    public DcModule getChildModule() {
        return moduleChild;
    }
    
    public DcObject getObject(int row) {
    	return table.getDCObject(row);
    }
    
    public DcObject getDcObject(String ID) {
        int row = table.getRowNumberWithID(ID);
        
        DcObject dco = null;
        if (row > -1) 
            dco = table.getDCObject(row);
        
        return dco;
    }  

    public void applyDefinitions() {
        table.applyDefinitions();
    }

    public int getRows() {
        return table.getRowCount();
    }

    public int[] getSelectedRows() {
        return table.getSelectedRows();
    }

    public void applySettings() {
        setFont(DcSettings.getFont(DcRepository.settings.stSystemFontNormal));
        table.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontNormal));
        panelTableActions.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontBold));

        if (quickView != null) {
            quickView.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontBold));
            quickView.setVisible(DcSettings.getBoolean(DcRepository.settings.stShowQuickView));
        }
        
        if (treePanel != null) {
            boolean treeVisibible = treePanel.isVisible();
            boolean treeVisibibleSett = DcSettings.getBoolean(DcRepository.settings.stShowGroupingPanel);
            
            treePanel.setVisible(treeVisibibleSett);
            if (!treeVisibible && treeVisibibleSett)
                treePanel.groupBy(getModule().getSettings().getIntArray(DcRepository.modulesettings.stGroupedBy));
            else if (treeVisibible && !treeVisibibleSett)
                treePanel.reset();
            
            treePanel.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontBold));
            
        }        

        panelInput.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontBold));
        panelStatus.setFont(DcSettings.getFont(DcRepository.settings.stSystemFontBold));
    }

    public void updateRow(int index, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark) {
        table.updateRowAt(index, dco, overwrite, allowDeletes, mark);
    }

    public boolean hasChild() {
        return getChildModule() != null;
    }

    protected boolean isTaskRunning() {
        return task != null && task.isRunning();
    }

    public void add(Collection c) {
        add(c, true);
    }
    
    public void add(DcObject[] objects) {
        add(objects, true);
    }
    
    public void add(Collection c, boolean updateTree) {
        DcObject[] objects = new DcObject[c.size()];
        objects = (DcObject[]) c.toArray(objects);
        add(objects, updateTree);
    }

    public void add(DcObject[] objects, boolean updateTree) {
        if (task != null && task instanceof FillerThread && task.isAlive()) {
            task.stopRunning();
            clear(false);
        }
        
        if (updateTree && treePanel != null) {
            treePanel.add(objects);
        } else {
            task = new FillerThread(objects, this instanceof TableSearchView);
            task.start();
        }
    }
    
    protected void setSelectedRow() {
        try {
            if (this instanceof TableSearchView) {
                table.setSelectedRow(table.getRowCount() - 1);
                ((TableSearchView) this).setDataForSelectedRow();
            } else {
                table.setSelectedRow(0);
            }
        } catch (Exception exp) {}        
    }
    
    public void clear() {
        clear(true);
    }
    
    public void clear(boolean safe) {
        
        if (safe) {
            boolean saved = isChangesSaved();
            if (!saved && getCurrentTask() != null) {
                QuestionBox qb = new QuestionBox(DcResources.getText("msgNotSaved"));
                if (qb.isAffirmative()) {
                    save();
                } else {
                    table.clearCache();
                }
                
                try {
                    getCurrentTask().join();
                } catch (Exception exp) {
                    DcLog.add(exp);
                }
            }
        }
        
        table.clearCache();
        table.clearTable();
        
        if (hasChild()) {
            childView.clear();
            
            if (childView instanceof TableSearchView)
                ((TableSearchView) childView).setParentID(null);
            else 
                ((TableInsertView) childView).setParentID(null);
        }

        if (quickView != null)
            quickView.clear();

        initProgressBar(0);
        setStatus(DcResources.getText("msgTableCleared"));
    }      
    
    public void updateRow(String id, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark) {
        int row = table.getRowNumberWithID(id);
        
        if (row != -1)
            table.updateRowAt(row, dco, overwrite, allowDeletes, mark);
    }

    public void denyActions() {
    	table.setCursor(new Cursor(Cursor.WAIT_CURSOR));
    	
    	removeMouseListeners();

    	panelTableActions.buttonClear.setEnabled(false);
        panelTableActions.buttonSave.setEnabled(false);
    }

    public void allowActions() {
    	table.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    	removeMouseListeners();
        table.addMouseListener(tableMouseListener);
        panelTableActions.buttonClear.setEnabled(true);
        panelTableActions.buttonSave.setEnabled(true);
    }

    private void removeMouseListeners() {
        MouseListener[] listeners = table.getMouseListeners();
        if (listeners != null) {
            for (int i = 0; i < listeners.length; i++) {
                if (listeners[i] instanceof TableMouseListener)
                    table.removeMouseListener(listeners[i]);
            }
        }
    }
    
    public void removeRows(String[] ids) {
        try {
            if (treePanel != null)
                treePanel.remove(ids);        
            
            for (int i = 0; i < ids.length; i++) {
                
                int row = table.getRowNumberWithID(ids[i]);
                
                if (row > -1) {
                    table.removeRow(row);
                    removeFromCache(ids[i]);
                }
            }
    
            if (getRows() == 0 && quickView != null)
                quickView.clear();
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }

    public void removeRows(int[] rows) {
        for (int i = 0; i < rows.length; i++) {

            if (treePanel != null)
                treePanel.remove(getObject(rows[i]));

            table.removeRow(rows[i]);
        }
        
        if (getRows() == 0 && quickView != null)
            quickView.clear();
    }

    public void cancelTask() {
        allowActions();
        if (isTaskRunning()) {
            task.cancel();
        }
        updateProgressBar(0);
    }

    public Collection getObjects() {
    	return table.getObjects();
    }

    protected void addCopyPaste(DcPopupMenu popupMenu) {
        JMenuItem menuItemCopy = new JMenuItem(DcResources.getText("lblCopy"), IconLibrary._icoCopy);
        menuItemCopy.addActionListener(new CopyAction());

        JMenuItem menuItemPaste = new JMenuItem(DcResources.getText("lblPaste"), IconLibrary._icoPaste);
        menuItemPaste.addActionListener(new PasteAction());

        popupMenu.add(new JSeparator());
        popupMenu.add(menuItemCopy);
        popupMenu.add(menuItemPaste);
    }

    public void cancelEdit() {
        int selectedRow = table.getSelectedRow();
        if (selectedRow != -1) {
            for (int i = 0; i < table.getColumnCount(); i++) {
                boolean success = false;
                try {

                    DefaultCellEditor editor = (DefaultCellEditor) table.getCellEditor(selectedRow, i);
                    if (editor.getComponent().isEnabled())
                        editor.cancelCellEditing();
                    
                    success = true;
                } catch (Exception ignore) {}

                if (!success) {
                    table.getCellEditor(selectedRow, i).cancelCellEditing();
                }
            }
        }
    }

    public abstract DcPopupMenu getPopupMenu();
    public abstract void open();
    public abstract void setParentID(String ID);
    public abstract void createReport();
    public abstract boolean changesSaved();


    public void add(DcObject dco) {
    	add(dco, true);
        setSelectedRow();        
    }

    public void add(DcObject dco, boolean select) {
        if (dco.getID() == null || dco.getID().equals("")) 
        	dco.setIDs();
        
    	table.addObject(dco, select);
        
        if (treePanel != null)
            treePanel.add(dco);
    }

    public class TableActionsPanel extends JPanel {
        private JButton buttonClear;
        private JButton buttonSave;
        private JButton buttonAdd;
        private JButton buttonRemove;
        private JButton buttonCancel;
        
        private JPanel panelLeft = new JPanel();
        private JPanel panelRight = new JPanel();

        public TableActionsPanel(boolean insertModus, JComponent[] extraComponents) {
            setLayout(Layout.getGBL());

            panelLeft.setLayout(new FlowLayout(FlowLayout.LEFT));
            panelRight.setLayout(new FlowLayout(FlowLayout.LEFT));
            
            // Create the components
            buttonSave = ComponentFactory.getButton(DcResources.getText("lblSave"));
            buttonClear = ComponentFactory.getButton(DcResources.getText("lblClear"));
            buttonAdd = ComponentFactory.getButton(DcResources.getText("lblAdd"));
            buttonRemove = ComponentFactory.getButton(DcResources.getText("lblRemove"));
            buttonCancel = ComponentFactory.getButton(DcResources.getText("lblCancel"));

            buttonCancel.setMnemonic(KeyEvent.VK_X);
            buttonClear.setMnemonic(KeyEvent.VK_L);
            buttonCancel.setMnemonic(KeyEvent.VK_C);
            buttonSave.setMnemonic(KeyEvent.VK_S);
            buttonAdd.setMnemonic(KeyEvent.VK_A);
            buttonRemove.setMnemonic(KeyEvent.VK_R);
            
            buttonSave.setToolTipText(DcResources.getText("tpSaveChanges"));
            buttonClear.setToolTipText(DcResources.getText("tpClear"));
            buttonAdd.setToolTipText(DcResources.getText("tpAddRow"));
            buttonRemove.setToolTipText(DcResources.getText("tpRemoveRow"));
            buttonCancel.setToolTipText(DcResources.getText("tpCancel"));

            buttonClear.addActionListener(new ClearButtonAction());
            buttonAdd.addActionListener(new AddRowAction(getModule().getIndex()));
            buttonRemove.addActionListener(new RemoveRowAction(getModule().getIndex()));
            buttonCancel.addActionListener(new CancelAction());

            // Build the panel
            if (insertModus) {
                panelLeft.add(buttonAdd);
                panelLeft.add(buttonRemove);
            }

            int counter = 2;
            if (extraComponents != null) {
                for (int i = 0; i < extraComponents.length; i++) {
                    panelLeft.add(extraComponents[i]);
                    counter++;
                }
            }

            panelRight.add(buttonSave);
            
            if (insertModus)
                panelRight.add(buttonClear);
            
            panelRight.add(buttonCancel);
            
            add(panelLeft,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.WEST, GridBagConstraints.NONE,
                            new Insets(0, 5, 0, 5), 0, 0));
            add(panelRight, Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.EAST, GridBagConstraints.NONE,
                            new Insets(0, 5, 0, 5), 0, 0));
        }
        
        public void setFont(Font font) {
            super.setFont(font);
            if (buttonClear != null) {
                buttonClear.setFont(font);
                buttonSave.setFont(font);
                buttonAdd.setFont(font);
                buttonRemove.setFont(font);
                buttonCancel.setFont(font);
            }
        }

        public void addSaveActionListener(ActionListener listener) {
            buttonSave.addActionListener(listener);
        }

        public void removeSaveActionListener(ActionListener listener) {
            buttonSave.removeActionListener(listener);
        }
    }

    public class ClearButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            clear();
        }
    }

    public class CancelAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancelTask();
        }
    }

    protected class CopyAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
          StringBuffer sb = new StringBuffer();
          int[] rows = getSelectedRows();
          DcFieldDefinitions definitions = getModule().getFieldDefinitions();

          for (int i = 0; i < rows.length; i++) {
              DcObject dco = table.getDCObject(rows[i]);
              int fieldCount = 0;
              for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
                  DcFieldDefinition fieldDef = (DcFieldDefinition) iter.next();
                  if (fieldDef.isVisible() && fieldDef.isEnabled()) {
                      if (fieldCount > 0)
                          sb.append("\t");

                      int field = fieldDef.getIndex();
                      Object o = dco.getValue(field);
                      String value = "";
                      value = value == null ? "" : value;
                      if (dco.getField(field).getValueType() == DcRepository.datatypes._PICTURE) {
                          value = "";
                      } else {
                          value = o == null ? "" : o.toString();
                      }

                      sb.append(value);
                      fieldCount++;
                  }
             }
             sb.append("\n");
          }
          StringSelection stSelection = new StringSelection(sb.toString());
          clipboard.setContents(stSelection, stSelection);
       }
    }

    protected class PasteAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
           int row = table.getSelectedRow();
           DcFieldDefinitions definitions = getModule().getFieldDefinitions();

           int[] fields = new int[getModule().getDcObject().getVisibleFields().size()];
           int counter = 0;
           for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
               DcFieldDefinition fieldDef = (DcFieldDefinition) iter.next();
               if (fieldDef.isVisible() && fieldDef.isEnabled()) {
                   fields[counter++] = fieldDef.getIndex();
               }
           }

           try {
              String text = (String) (clipboard.getContents(this).getTransferData(DataFlavor.stringFlavor));
              StringTokenizer rowTokenizer = new StringTokenizer(text, "\n");
              for (int i = 0; rowTokenizer.hasMoreTokens(); i++) {
                 String values = rowTokenizer.nextToken();
                 DcObject dco = getModule().getDcObject();

                 int fldCounter = 0;
                 int index = values.indexOf("\t");
                 while (index > -1) {
                     String value = values.substring(0, index);
                     if (fldCounter <= fields.length) {
                         DcField field = dco.getField(fields[fldCounter]);
                         if (field.getValueType() == DcRepository.datatypes._LONG) {
                             try {
                                 dco.setValue(fields[fldCounter], new Long(value.trim()));
                             } catch (Exception exp) {}
                         } else {
                             dco.setValue(fields[fldCounter], value);
                         }
                     }

                     fldCounter++;
                     values = values.substring(index + 1, values.length());
                     index = values.indexOf("\t");
                 }
                 dco.setValue(fields[fldCounter], values);
                 table.updateRowAt(row++, dco, true, true, true);
             }
          } catch (Exception exp){
              DcLog.add(exp);
          }
       }
    }

    private class TableMouseListener implements MouseListener {
        public void mouseReleased(MouseEvent e) {

            try {
                table.setListeningForChanges(false);
            
                if (SwingUtilities.isRightMouseButton(e)) {
                    int row = table.rowAtPoint(e.getPoint());
    
                    if (table.getSelectedRows() != null && table.getSelectedRows().length < 2) {
    
                        int col = table.getSelectedColumn() != -1 ? table.getSelectedColumn() :
                                                                    table.columnAtPoint(e.getPoint());
                        table.addRowSelectionInterval(row, row);
                        table.addColumnSelectionInterval(col, col);
    
                        ((DcTableCellRenderer) table.getCellRenderer(row, col)).
                            getTableCellRendererComponent(
                                    table,
                                    table.getValueAt(row, col, true),
                                    true,
                                    true,
                                    row,
                                    col);
    
                    }
    
                    DcPopupMenu menu = getPopupMenu();
                    if (menu != null)
                        menu.show(table, e.getX(), e.getY());
                }
    
                if (e.getClickCount() == 2) {
                	cancelEdit();
                    open();
                }
            } finally {
                table.setListeningForChanges(true);
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    }

    public class FillerThread extends DataTask {
        
        private DcObject[] objects;
        private boolean clearViewFirst;
        
        public FillerThread(DcObject[] objects, boolean clearViewFirst) {
            super();
            this.objects = objects;
            this.clearViewFirst = clearViewFirst;
        }
        
        public void run() {
            boolean listening = table.isListeningForChanges();
            
            try {
                table.setListeningForChanges(false);
                
                denyActions();
                
                if (clearViewFirst)
                    clear();
                
                updateProgressBar(0);
                initProgressBar(objects.length);
                dispatchEvents = false;
                
                int i = 0;
                
                if (hasChild()) { 
                    childView.setVisible(false);
                    childView.setOverwriteVisibility(true);
                }

                if (!isVisibilityOverwritten())
                    table.setVisible(false);

                
                for (; i < objects.length && keepOnRunning(); i++) {
                    
                    DcObject dco = objects[i];
                    add(dco, false);
                    
                    try {
                        sleep(2);
                    } catch (Exception exp) {}
                    
                    setStatus(DcResources.getText("msgAddingXToView", dco.toString()));
                    updateProgressBar(i + 1);
                }
    
                if (clearViewFirst)
                    setStatus(DcResources.getText("msgSearchResult", "" + i));
                else 
                    setStatus(DcResources.getText("msgAddedXItemsToView", "" + i));
            } catch (Exception exp) {
                DcLog.add(exp);
            } finally {
                if (!isVisibilityOverwritten()) {
                    table.setVisible(true);
                    table.revalidate();
                }
                
                if (hasChild()) { 
                    childView.setVisible(true);
                    childView.revalidate();
                    childView.setOverwriteVisibility(false);
                }
                
                dispatchEvents = true;
                stopRunning();
                allowActions();
                table.setListeningForChanges(listening);
            }
            
            try {
                sleep(100);
            } catch (Exception exp) {}
            
            table.revalidate();
            setSelectedRow();

            objects = null;
        }
    }     
    
    public class CreateReportAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            createReport();
        }
    }
}

