/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.windows.fileimport;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcFileField;
import net.sf.dc.console.components.DcFrame;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.DcPropertyModule;
import net.sf.dc.core.objects.DcProperty;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.fileimporters.FileImporter;
import net.sf.dc.fileimporters.IFileImportClient;
import net.sf.dc.onlinesearch.DcRegion;
import net.sf.dc.onlinesearch.DcServer;
import net.sf.dc.settings.Settings;
import net.sf.dc.util.FileFilter;

public class FileImportDialog extends DcFrame implements IFileImportClient {

    protected JCheckBox checkVerify;
    protected JCheckBox checkRecursive;
    protected DcFileField fldSource;
    protected JTextField fldLocationName;
    protected JComboBox fldStorageMedium;

    protected JCheckBox checkFilesOnly = ComponentFactory.getCheckBox(DcResources.getText("lblSelectFilesOnly"));
    protected JButton buttonRun = ComponentFactory.getButton(DcResources.getText("lblRun"));
    protected JButton buttonStop = ComponentFactory.getButton(DcResources.getText("lblStop"));
    protected JButton buttonClose = ComponentFactory.getButton(DcResources.getText("lblClose"));

    protected JTextArea textLog = ComponentFactory.getTextArea();
    protected JProgressBar progressBar = new JProgressBar();

    protected Settings settings;
    
    protected boolean cancelled = false; 
    
    protected FileImporter importer;

    public FileImportDialog(FileImporter importer) {
        
        super(DcResources.getText("lblXImport", 
                                  DcModules.get(importer.getModule()).getObjectName()), 
                                  IconLibrary._icoImport);
        
        this.importer = importer;
        this.settings = DcModules.get(importer.getModule()).getSettings();
        
        setHelpIndex("dc.tools.fileinfoimport");

        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                close();
            }
        });

        buildDialog();

        setSize(settings.getDimension(DcRepository.modulesettings.stImportCDDialogSize));
        setCenteredLocation();
    }
    
    public DcProperty getStorageMedium() {
        DcModule module = DcModules.get(importer.getModule());
        
        DcProperty medium = null;
        if (module.getPropertyModule(DcModules._STORAGEMEDIA) != null) {
            try {
                medium = (DcProperty) fldStorageMedium.getSelectedItem();
            } catch (Exception exp) {}
        }
            
        return medium;
    }

    public void initProgressBar(int maxValue) {
        if (progressBar != null) {
            progressBar.setValue(0);
            progressBar.setMaximum(maxValue);
        }
    }

    public void updateProgressBar(int value) {
        if (progressBar != null) progressBar.setValue(value);
    }

    public void addMessage(String message) {
        if (textLog != null) textLog.insert(message + '\n', 0);
    }

    public void addError(String error) {
        new MessageBox(error, MessageBox._ERROR);
    }

    public void finish() {
        if (buttonRun != null) buttonRun.setEnabled(true);
    }

    public boolean cancelled() {
        return cancelled;
    }      
    
    public boolean useOnlineServices() {
        return false;
    }    
    
    public String getLocationName() {
        return fldLocationName.getText();
    }

    public DcRegion getRegion() {
        return null;
    }

    public DcServer getServer() {
        return null;
    }

    public String getSourceDirectory() {
        return fldSource.getFilename();
    }    
    
    protected void startImport() {
        denyActions();
        try {
            cancelled = false;
            importer.parse(this, 
                           checkVerify.isSelected(), 
                           checkRecursive.isSelected(), 
                           fldSource.getFilename());
            
        } catch (Exception e) {
            addError(e.getMessage());
        }
    }
    
    public void denyActions() {
        progressBar.setValue(0);
        textLog.setText("");

        buttonRun.setEnabled(false);
    }

    public void close() {
        cancel();
        settings.set(DcRepository.modulesettings.stImportCDDialogSize, getSize());
        
        checkVerify = null;
        checkRecursive = null;
        fldSource = null;
        fldLocationName = null;
        fldStorageMedium = null;
        checkFilesOnly = null;
        buttonRun = null;
        buttonStop = null;
        buttonClose = null;
        textLog = null;
        progressBar = null;
        settings = null;
        
        super.close();
    }

    protected void cancel() {
        cancelled = true;
    }

    protected void buildDialog() {
        //**********************************************************
        //Input panel
        //**********************************************************
        JPanel panelInput = new JPanel();
        panelInput.setLayout(Layout.getGBL());
        panelInput.setBorder(ComponentFactory.getTitleBorder(""));

        JLabel labelLocation = ComponentFactory.getLabel(DcResources.getText("lblLocation"));
        fldSource = new DcFileField(null);
        checkVerify = ComponentFactory.getCheckBox(DcResources.getText("lblAskPerDir"));

        checkRecursive  = ComponentFactory.getCheckBox(DcResources.getText("lblRecursiveDir"));
        JLabel labelCDName = ComponentFactory.getLabel(DcResources.getText("lblLocationName"));
        fldLocationName = ComponentFactory.getShortTextField(100);

        fldLocationName.setPreferredSize(new Dimension(25, 22));
        fldLocationName.setMinimumSize(new Dimension(25, 22));
        
        JLabel labelMedium = ComponentFactory.getLabel(DcResources.getText("lblStorageMedium"));
        
        DcPropertyModule storageModule = DcModules.get(importer.getModule()).getPropertyModule(DcModules._STORAGEMEDIA);
        fldStorageMedium = ComponentFactory.getMediaCombo(storageModule.getIndex());

        checkRecursive.setSelected(true);
        checkFilesOnly.addChangeListener(new ToggleSearchMode());

        fldSource.setModus(false, true);

        panelInput.add(  labelLocation,  Layout.getGBC( 0, 0, 2, 1, 1.0, 1.0
                                        ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                         new Insets(0, 5, 0, 5), 0, 0));
        panelInput.add(  fldSource,       Layout.getGBC( 1, 0, 3, 1, 1.0, 1.0
                                        ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                         new Insets(0, 5, 0, 5), 0, 0));
        panelInput.add(  labelCDName,    Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                                        ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                         new Insets(0, 5, 0, 5), 0, 0));
        panelInput.add(  fldLocationName,     Layout.getGBC( 1, 1, 2, 1, 20.0, 20.0
                                        ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                         new Insets(0, 5, 0, 5), 0, 0));

        DcModule module = DcModules.get(importer.getModule());
        if (module.getPropertyModule(DcModules._STORAGEMEDIA) != null) {
            panelInput.add(  labelMedium,    Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                     new Insets(0, 5, 0, 5), 0, 0));
            panelInput.add(  fldStorageMedium,    Layout.getGBC( 1, 2, 2, 1, 20.0, 20.0
                    ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                     new Insets(0, 5, 0, 5), 0, 0));
        }

        //**********************************************************
        //Settings Panel
        //**********************************************************
        JPanel panelSettings = new JPanel();
        panelSettings.setLayout(Layout.getGBL());
        panelSettings.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblSettings")));

        panelSettings.add(checkFilesOnly, Layout.getGBC( 0, 0, 3, 1, 1.0, 1.0
                                         ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                          new Insets(5, 5, 0, 5), 0, 0));
        panelSettings.add(checkVerify,    Layout.getGBC( 0, 1, 3, 1, 1.0, 1.0
                                         ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                          new Insets(5, 5, 0, 5), 0, 0));
        panelSettings.add(checkRecursive, Layout.getGBC( 0, 2, 3, 1, 1.0, 1.0
                                         ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                          new Insets(5, 5, 5, 5), 0, 0));

        //**********************************************************
        //Actions Panel
        //**********************************************************
        JPanel panelActions = new JPanel();
        panelActions.setLayout(Layout.getGBL());

        buttonRun.addActionListener(new ExportCDButtonAction());
        buttonStop.addActionListener(new CancelButtonAction());
        buttonClose.addActionListener(new CloseButtonAction());

        buttonRun.setMnemonic(KeyEvent.VK_R);
        buttonStop.setMnemonic(KeyEvent.VK_T);
        buttonClose.setMnemonic(KeyEvent.VK_C);

        panelActions.add(buttonRun,     Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                        new Insets( 5, 5, 5, 5), 0, 0));
        panelActions.add(buttonStop,    Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                        new Insets( 5, 5, 5, 5), 0, 0));
        panelActions.add(buttonClose,  Layout.getGBC( 2, 0, 1, 1, 1.0, 1.0
                                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                        new Insets( 5, 5, 5, 5), 0, 0));

        //**********************************************************
        //Progress panel
        //**********************************************************
        JPanel panelProgress = new JPanel();
        panelProgress.setLayout(Layout.getGBL());
        panelProgress.add(progressBar, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                          new Insets(5, 5, 5, 5), 0, 0));

        //**********************************************************
        //Log Panel
        //**********************************************************
        JPanel panelLog = new JPanel();
        panelLog.setLayout(Layout.getGBL());

        JScrollPane scroller = new JScrollPane(textLog);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        panelLog.add(scroller, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                     new Insets(5, 5, 5, 5), 0, 0));

        panelLog.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblLog")));


        //**********************************************************
        //Main Panel
        //**********************************************************
        this.getContentPane().setLayout(Layout.getGBL());
        this.getContentPane().add(
                 panelInput,      Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(
                panelSettings,    Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(
                 panelActions,    Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE,
                 new Insets( 0, 0, 0, 0), 0, 0));
        this.getContentPane().add(
                 panelProgress,   Layout.getGBC( 0, 3, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets( 0, 0, 0, 0), 0, 0));
        this.getContentPane().add(
                 panelLog,        Layout.getGBC( 0, 4, 1, 1, 40.0, 40.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                 new Insets( 0, 0, 0, 0), 0, 0));

        pack();
        
        setSize(settings.getDimension(DcRepository.modulesettings.stImportCDDialogSize));
        setResizable(true);
        setCenteredLocation();
    }

    protected class ToggleSearchMode implements ChangeListener {
        public void stateChanged(ChangeEvent e) {
            fldSource.setModus(false, !checkFilesOnly.isSelected());
            
            if (!checkFilesOnly.isSelected())
                fldSource.setFileFilter(null);
            else 
                fldSource.setFileFilter(importer.getSupportedExtensions().length > 0 ? 
                                        new FileFilter(importer.getSupportedExtensions()): null);
        }
    }

    protected class CloseButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close();
        }
    }

    protected class CancelButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancel();
        }
    }

    protected class ExportCDButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startImport();
        }
    }
}
