/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.windows.reporting;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcDialog;
import net.sf.dc.console.components.DcFileField;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.reporting.ReportDictionaryGenerator;
import net.sf.dc.reporting.reports.Report;
import net.sf.dc.reporting.reports.Reports;
import net.sf.dc.reporting.templates.ReportTemplate;
import net.sf.dc.reporting.templates.ReportTemplates;
import net.sf.dc.reporting.transformers.XmlTransformer;
import net.sf.dc.reporting.transformers.XmlTransformers;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.util.BrowserLauncher;

public class ReportingDialog extends DcDialog {

    private Report report;
    private XmlTransformer transformer;

    private ReportSettingsPanel panelSettings = new ReportSettingsPanel();
    
    private JButton buttonRun = ComponentFactory.getButton(DcResources.getText("lblRun"));
    private JButton buttonStop = ComponentFactory.getButton(DcResources.getText("lblStop"));
    private JButton buttonClose = ComponentFactory.getButton(DcResources.getText("lblClose"));
    private JButton buttonResults = ComponentFactory.getButton(DcResources.getText("lblOpenReport"));

    private JComboBox cbExportType = ComponentFactory.getComboBox();
    private JComboBox cbTemplates = ComponentFactory.getComboBox();
    private JComboBox cbTransformer = ComponentFactory.getComboBox();

    private JRadioButton rbReport = ComponentFactory.getRadioButton("Create a presentation report", null);
    private JRadioButton rbExport = ComponentFactory.getRadioButton("Create a data report / export", null);

    private JTextArea textLog = ComponentFactory.getTextArea();
    private JProgressBar progressBar = new JProgressBar();
    private DcFileField fileField;
    
    private ReportFileSelectionListener reportFileSelectionListener = new ReportFileSelectionListener();
    
    private Collection objects;

    public ReportingDialog(Collection objects) {
        super(DataCrow.mainFrame);

        new ReportDictionaryGenerator().generate();
        
        try {
            this.objects = objects;
            setHelpIndex("dc.reports");
    
            fileField = ComponentFactory.getFileField(true, false, null);
            buildDialog();
    
            addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    close();
                }
            });
        } catch (Exception exp) {
            new MessageBox(DcResources.getText("msgFailedToOpen", exp.getMessage()) , MessageBox._ERROR);
            DcLog.add(exp);
        }
        
        setModal(true);
    }

    private void saveDefaults() {
        DcSettings.set(DcRepository.settings.stReportingDialogSize, getSize());
        DcSettings.set(DcRepository.settings.stReportFile, fileField.getFilename());
    }

    public void initProgressBar(int maxValue) {
        progressBar.setValue(0);
        progressBar.setMaximum(maxValue);
    }

    public void updateProgressBar(int value) {
        progressBar.setValue(value);
    }

    public void addMessage(String message) {
    	if (textLog != null) { 
    		textLog.insert(message + '\n', 0);
    		textLog.setCaretPosition(0);
    	}
    }

    public void allowActions(boolean b) {
        if (buttonRun != null) {
            buttonRun.setEnabled(b);
            buttonResults.setEnabled(b);
        }
    }

    private void showResults() {
    	File file = fileField.getFile();
        if (file != null && file.exists()) {
            try {
            	BrowserLauncher.openURL(file.toString());
            } catch (Exception exp) {
                new MessageBox(DcResources.getText("msgErrorWhileOpeningX",
                               new String[] {file.toString(), exp.getMessage()}), MessageBox._WARNING);
                DcLog.add(exp);
            }
        } else {
        	new MessageBox(DcResources.getText("msgCouldNotOpenReport"), MessageBox._WARNING);
        }
    }

    private File getTarget(String extension) throws FileNotFoundException {
        File target = fileField.getFile();
        
        if (target == null)
            throw new FileNotFoundException();
        
        String filename = target.toString();
        if (!filename.toLowerCase().endsWith(extension.toLowerCase())) {
            
            if (filename.lastIndexOf(".") > 0) 
                filename = filename.substring(0, filename.lastIndexOf("."));
            
            target = new File(filename + "." + extension);
            fileField.setFile(target);
        }
        
        return target;
    }
    
    private void createReport() {
        try {
            if (rbExport.isSelected()) {
                report = (Report) cbExportType.getSelectedItem();
                report.compile(this, objects, getTarget(report.getFileType()));
            } else {
                transformer = (XmlTransformer) cbTransformer.getSelectedItem();
                ReportTemplate template = (ReportTemplate) cbTemplates.getSelectedItem();
                panelSettings.saveSettings(template.getProperties());
                transformer.transform(this, objects, getTarget(transformer.getFileType()), template);
            }
            
            allowActions(false);
            
        } catch (FileNotFoundException fnfe) {
            new MessageBox(DcResources.getText("msgSelectTargetFile"), MessageBox._WARNING);
        } 
    }

    private void cancel() {
        if (report != null)
            report.cancel();
        
        if (transformer != null)
            transformer.cancel();
        
        allowActions(true);
    }

    public void close() {
        saveDefaults();
        saveReportFileProperties((ReportTemplate) cbTemplates.getSelectedItem());
        
        cancel();

        if (objects != null) { 
            objects.clear();
            objects = null;
        }
        
        report = null;
        transformer = null;
        panelSettings = null;
        buttonRun = null;
        buttonStop = null;
        buttonClose = null;
        buttonResults = null;
        cbExportType = null;
        cbTemplates = null;
        cbTransformer = null;
        rbReport = null;
        rbExport = null;
        textLog = null;
        progressBar = null;
        fileField = null;
        reportFileSelectionListener = null;
        
        super.close();
    }
    
    private void applyReportFileProperties(ReportTemplate reportFile) {
        if (reportFile != null)
            panelSettings.applySettings(reportFile.getProperties());
    }
    
    private void saveReportFileProperties(ReportTemplate reportFile) {
        if (reportFile != null)
            panelSettings.saveSettings(reportFile.getProperties());
    }    
    
    private void applyReportSelection() {
        cbExportType.setEnabled(rbExport.isSelected());
        cbTemplates.setEnabled(rbReport.isSelected());
        cbTransformer.setEnabled(rbReport.isSelected());

        panelSettings.setEnabled(rbReport.isSelected());
        
        if (cbTransformer.isEnabled() && cbTransformer.getSelectedIndex() > -1) {
            
            cbTemplates.removeItemListener(reportFileSelectionListener);
            
            XmlTransformer transformer = (XmlTransformer) cbTransformer.getSelectedItem();
            cbTemplates.removeAllItems();
            Collection templates = new ReportTemplates(true).getReportFiles(transformer.getType());
            for (Iterator iter = templates.iterator(); iter.hasNext(); ) {
                cbTemplates.addItem(iter.next());
            }
            
            if (cbTemplates.getItemCount() == 0) {
                cbTemplates.setEnabled(false);
                cbTransformer.setEnabled(false);
                
                rbExport.setSelected(true);
                rbReport.setEnabled(false);
                cbExportType.setEnabled(true);
            } else {
                cbTemplates.setSelectedIndex(0);
                applyReportFileProperties((ReportTemplate) cbTemplates.getSelectedItem());
            }
        }
        
        cbTemplates.addItemListener(reportFileSelectionListener);
    }
    
    private void buildDialog() {
        //**********************************************************
        //Input Panel
        //**********************************************************
        JPanel panelInput = new JPanel();
        panelInput.setLayout(Layout.getGBL());
        panelInput.add(fileField,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                      ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                       new Insets( 5, 5, 5, 5), 0, 0));
        
        fileField.setValue(DcSettings.getString(DcRepository.settings.stReportFile));
        
        panelInput.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblTargetFile")));        
        
        //**********************************************************
        //Report Type
        //**********************************************************
        JPanel panelReport = new JPanel(false);
        panelReport.setLayout(Layout.getGBL());
        
        ButtonGroup buttonGroup = new ButtonGroup();
        buttonGroup.add(rbReport);
        buttonGroup.add(rbExport);
        
        ReportSelectionListener rl = new ReportSelectionListener();
        rbReport.addActionListener(rl);
        rbExport.addActionListener(rl);
        cbTransformer.addActionListener(rl);

        for (Iterator iter = XmlTransformers.getTransformers().iterator(); iter.hasNext(); ) 
            cbTransformer.addItem(iter.next());
        
        for (Iterator iter = Reports.getReports().iterator(); iter.hasNext(); ) 
            cbExportType.addItem(iter.next());

        JLabel lblTransformers = ComponentFactory.getLabel(DcResources.getText("lblReportFormat"));
        
        panelReport.add(rbReport,       Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                        new Insets( 5, 5, 5, 5), 0, 0));
        panelReport.add(lblTransformers,Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 5, 5, 5), 0, 0));
        panelReport.add(cbTransformer,  Layout.getGBC( 2, 0, 1, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 5, 5, 5), 0, 0));
        panelReport.add(cbTemplates,      Layout.getGBC( 1, 1, 2, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 5, 5, 5), 0, 0));
        panelReport.add(rbExport,       Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                        new Insets( 5, 5, 5, 5), 0, 0));
        panelReport.add(cbExportType,   Layout.getGBC( 1, 2, 2, 1, 1.0, 1.0
                       ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                        new Insets( 5, 5, 5, 5), 0, 0));
        
        panelReport.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblReportSelection")));

    	//**********************************************************
        //Actions Panel
        //**********************************************************
        JPanel panelActions = new JPanel();
        panelActions.setLayout(Layout.getGBL());

        buttonRun.addActionListener(new CreateReportAction());
        buttonStop.addActionListener(new CancelButtonAction());
        buttonResults.addActionListener(new ResultsButtonAction());
        buttonClose.addActionListener(new CloseButtonAction());

        buttonRun.setMnemonic(KeyEvent.VK_R);
        buttonStop.setMnemonic(KeyEvent.VK_T);
        buttonClose.setMnemonic(KeyEvent.VK_C);
        buttonResults.setMnemonic(KeyEvent.VK_O);

        panelActions.add(buttonRun,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 5, 5, 5, 5), 0, 0));
        panelActions.add(buttonStop,    Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 5, 5, 5, 5), 0, 0));
        panelActions.add(buttonResults,   Layout.getGBC( 2, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 5, 5, 5, 5), 0, 0));
        panelActions.add(buttonClose,     Layout.getGBC( 4, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 5, 5, 5, 5), 0, 0));

        //**********************************************************
        //Progress panel
        //**********************************************************
        JPanel panelProgress = new JPanel();
        panelProgress.setLayout(Layout.getGBL());
        panelProgress.add(progressBar, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                          new Insets(5, 5, 5, 5), 0, 0));

        //**********************************************************
        //Log Panel
        //**********************************************************
        JPanel panelLog = new JPanel();
        panelLog.setLayout(Layout.getGBL());

        JScrollPane scroller = new JScrollPane(textLog);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        panelLog.add(scroller, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                     new Insets(5, 5, 5, 5), 0, 0));

        panelLog.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblLog")));

        //**********************************************************
        //Main Panel
        //**********************************************************
        this.getContentPane().setLayout(Layout.getGBL());
        
        
        JPanel panel = new JPanel();
        panel.add(panelSettings);
        panel.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblSettings")));

        this.getContentPane().add(
                panelReport,    Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(
                panel,          Layout.getGBC( 1, 0, 1, 2, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(
                panelInput,     Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(
                panelActions,   Layout.getGBC( 0, 3, 2, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE,
                new Insets( 0, 0, 0, 0), 0, 0));
        this.getContentPane().add(
                panelProgress,  Layout.getGBC( 0, 4, 2, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                new Insets( 0, 0, 0, 0), 0, 0));
        this.getContentPane().add(
                panelLog,       Layout.getGBC( 0, 5, 2, 1, 40.0, 40.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets( 0, 5, 0, 5), 0, 0));
        
        this.setResizable(true);
        this.pack();
        
        Dimension size = DcSettings.getDimension(DcRepository.settings.stReportingDialogSize);
        setSize(size);

        setCenteredLocation();

        rbReport.setSelected(true);
        applyReportSelection();
    }

    private class ResultsButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            showResults();
        }
    }

    private class CloseButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close();
        }
    }

    private class CancelButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancel();
        }
    }

    private class CreateReportAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            createReport();
        }
    }
    
    
    private class ReportFileSelectionListener implements ItemListener {

        public void itemStateChanged(ItemEvent ie) {
            if (ie.getStateChange() == ItemEvent.SELECTED)
                applyReportFileProperties((ReportTemplate) ie.getItem());    
            
            if (ie.getStateChange() == ItemEvent.DESELECTED)
                saveReportFileProperties((ReportTemplate) ie.getItem());
        }
    }
    
    private class ReportSelectionListener implements ActionListener {
        public void actionPerformed(ActionEvent ae) {
            applyReportSelection();
        }
    }
}
