/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.wizards;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeListener;
import java.util.Iterator;
import java.util.List;

import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcFrame;
import net.sf.dc.console.components.DcLongTextField;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.resources.DcResources;

public abstract class Wizard extends DcFrame {

    protected int current = 0;

    private DcLongTextField textHelp;
    
    private List wizardPanels;
    
    private JButton buttonNext;
    private JButton buttonBack;
    private JButton buttonRestart;
    private JButton buttonFinish;
    private JButton buttonClose;
    
    private boolean cancelled = false;
    private boolean restarted = false;

    public Wizard() {
        super("", IconLibrary._icoWizard);
        
        initialize();
        wizardPanels = getWizardPanels();
        
        build();
        
        addKeyListener(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false), new NextOnEnterAction(), "next");
    }
    
    protected abstract String getWizardName();
    protected abstract void saveSettings();
    public abstract void finish() throws WizardException;
    public abstract void next() throws WizardException;
    protected abstract void restart();
    protected abstract void initialize();
    protected abstract List getWizardPanels();    

    public void close() {
        
        if (wizardPanels != null)
            wizardPanels.clear();
        
        wizardPanels = null;
        textHelp = null;
        buttonNext = null;
        buttonBack = null;
        buttonRestart = null;
        buttonFinish = null;
        buttonClose = null;
        
        super.close();
    }
    
    protected void applyPanel() {
        buttonBack.setEnabled(current != 0);
        buttonNext.setVisible(current != wizardPanels.size() - 1);
        buttonFinish.setVisible(current == wizardPanels.size() - 1);
        buttonRestart.setVisible(current == wizardPanels.size() - 1);
        textHelp.setText(getCurrent().getHelpText());

        if (buttonFinish.isVisible())
            buttonFinish.requestFocus();
        else
            buttonNext.requestFocus();

        setTitle(getWizardName());
    }

    protected boolean isRestarted() {
        return restarted;
    }
    
    protected boolean isCancelled() {
        return cancelled;
    }
    
    protected IWizardPanel getCurrent() {
        return (IWizardPanel) wizardPanels.get(current);
    }

    protected int getStepCount() {
        return wizardPanels.size();
    }
    
    protected IWizardPanel getWizardPanel(int step) {
        return (IWizardPanel) wizardPanels.get(step);
    }
    
    protected void back() {
        current -= 1;
        int counter = 0;

        if (current >= 0) {
            for (Iterator iter = wizardPanels.iterator(); iter.hasNext(); counter++) {
                JPanel panel = (JPanel) iter.next();
                panel.setVisible(counter == current);
            }
        } else {
            current += 1;
        }

        applyPanel();
    }
    
    private void build() {
        getContentPane().setLayout(Layout.getGBL());
        getContentPane().add(getInformationPanel(), Layout.getGBC(1, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets(5, 5, 5, 5), 0, 0));
        getContentPane().add(getActionPanel(), Layout.getGBC(1, 3, 1, 2, 1.0, 1.0
                            ,GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                             new Insets(5, 5, 5, 15), 0, 0));

        // the actual content panels
        int counter = 0;
        for (Iterator iter = wizardPanels.iterator(); iter.hasNext(); counter++) {
            JPanel panel = (JPanel) iter.next();
            panel.setMinimumSize(new Dimension(350,350));
            panel.setPreferredSize(new Dimension(350,350));

            panel.setVisible(counter == 0);
            getContentPane().add(panel,   Layout.getGBC(1, 1, 1, 1, 20.0, 20.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                             new Insets(5, 5, 5, 5), 0, 0));
        }

        setResizable(true);
        applyPanel();
        pack();
    }
    
    private JPanel getInformationPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(Layout.getGBL());
        
        textHelp = ComponentFactory.getLongTextField(8000);
        textHelp.setBorder(null);
        
        JScrollPane scroller = new JScrollPane(textHelp);
        scroller.setBorder(null);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        textHelp.setEditable(false);

        panel.add(textHelp, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                            new Insets(5, 5, 5, 5), 0, 0));
        return panel;
    }
    
    private JPanel getActionPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(new FlowLayout(FlowLayout.RIGHT));

        buttonNext = ComponentFactory.getButton(DcResources.getText("lblNext"));
        buttonBack = ComponentFactory.getButton(DcResources.getText("lblBack"));
        buttonRestart = ComponentFactory.getButton(DcResources.getText("lblFinishRestart"));
        buttonFinish = ComponentFactory.getButton(DcResources.getText("lblFinish"));
        buttonClose = ComponentFactory.getButton(DcResources.getText("lblClose"));
        
        buttonNext.addActionListener(new NextAction());
        buttonBack.addActionListener(new BackAction());
        buttonRestart.addActionListener(new RestartAction());
        buttonFinish.addActionListener(new FinishAction());
        buttonClose.addActionListener(new CloseAction());
        
        buttonNext.setMnemonic('N');
        buttonBack.setMnemonic('B');
        buttonFinish.setMnemonic('F');
        buttonRestart.setMnemonic('R');
        buttonClose.setMnemonic('C');
        
        buttonBack.setEnabled(false);
        buttonFinish.setVisible(false);
        buttonRestart.setVisible(false);
        
        panel.add(buttonBack);
        panel.add(buttonNext);
        panel.add(buttonRestart);
        panel.add(buttonFinish);
        panel.add(buttonClose);
        
        return panel;
    }    
    
    private class NextOnEnterAction implements Action {
        public boolean isEnabled() {
            return true;
        }

        public void setEnabled(boolean arg0) {}
        public void actionPerformed(ActionEvent arg0) {
            try {
                if (current < wizardPanels.size() - 1) {
                    next();
                } else {
                    finish();
                }
            } catch (WizardException we) {
                new MessageBox(we.getMessage(), MessageBox._ERROR);
            }
        }
        
        public void putValue(String arg0, Object arg1) {}
        public void addPropertyChangeListener(PropertyChangeListener arg0) {}

        public Object getValue(String arg0) {
            return null;
        }
        public void removePropertyChangeListener(PropertyChangeListener arg0) {}
    }


    public class NextAction implements ActionListener {
        public void actionPerformed(ActionEvent arg0) {
            try {
                next();
            } catch (WizardException we) {
                new MessageBox(we.getMessage(), MessageBox._ERROR);
            }                
        }
    }

    public class RestartAction implements ActionListener {
        public void actionPerformed(ActionEvent arg0) {
            restarted = true;
            restart();
        }
    }

    private class CloseAction implements ActionListener {
        public void actionPerformed(ActionEvent arg0) {
            saveSettings();
            cancelled = true;
            close();
        }
    }

    private class BackAction implements ActionListener {
        public void actionPerformed(ActionEvent arg0) {
            back();
        }
    }

    private class FinishAction implements ActionListener {
        public void actionPerformed(ActionEvent arg0) {
            try {
                finish();
            } catch (Exception exp) {
                new MessageBox(exp.getMessage(), MessageBox._WARNING);
            }
        }
    }
}
