/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core;

import java.awt.Toolkit;
import java.io.File;
import java.io.RandomAccessFile;
import java.net.URL;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.MainFrame;
import net.sf.dc.console.menu.DcEditPopupEventQueue;
import net.sf.dc.console.windows.SplashScreen;
import net.sf.dc.console.windows.TipOfTheDayDialog;
import net.sf.dc.console.windows.messageboxes.NativeMessageBox;
import net.sf.dc.core.data.DataFilters;
import net.sf.dc.core.db.DatabaseManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.enhancers.ValueEnhancers;
import net.sf.dc.settings.DcSettings;

public class DataCrow {

    private static Platform platform = new Platform();
    
    public static String baseDir;
    public static String imageDir;
    public static String moduleDir;
    public static String reportDir;
    
    public static MainFrame mainFrame;
    private static SplashScreen splashScreen;
    
    public static boolean debug = false;

    public static void main(String[] args) {
        String db = null;
        String dir = null;
        for (int i = 0; i < args.length; i++) {
            if (args[i].toLowerCase().startsWith("-dir:")) {
                dir = args[i].substring(5, args[i].length());
            } else if (args[i].toLowerCase().startsWith("-db:")) {
                db = args[i].substring(4, args[i].length());
            } else if (dir != null) {
                dir += " " + args[i];
            } else { 
                System.out.println("The following optional parameters can be used:");
                System.out.println("-dir:<installdir>");
                System.out.println("example: java -jar datacrow.jar -dir:d:/datacrow");
                System.out.println("");
                System.out.println("-db:<databasename>");
                System.out.println("example: java -jar datacrow.jar -db:testdb");
                System.exit(0);
            }
        }
        
        try {
            setBaseDir(dir);
            upgrade();
            createDirectories();
            
            // Initialize the Resources
            new DcResources();
            new DcSettings();
            
            showSplashScreen();
    
            // Initialize all settings
            splashScreen.setStatusMsg("Initializing settings");
    
            // Initialize the Component factory
            splashScreen.setStatusMsg("Loading components");
            new ComponentFactory();
    
            DcLog.add(DcResources.getText("msgApplicationStarts"), DcLog._STANDARD);
    
            // Initialize all modules
            new ModuleUpgrade().upgrade();
            DcModules.load();
    
            DcLog.add(DcResources.getText("msgSettingsLoaded"), DcLog._STANDARD);
    
            ValueEnhancers.initialize();
            
            // set the database name
            DcSettings.set(DcRepository.settings.stConnectionString, "dc");
            if (db != null && db.length() > 0)
                DcSettings.set(DcRepository.settings.stConnectionString, db);
    
            // Establish a connection to the database / server
            splashScreen.setStatusMsg("Initializing database");
            
            DatabaseManager.initialize();
        
            // Start the UI
            if (splashScreen == null)
                showSplashScreen();

            splashScreen.setStatusMsg("Loading items");
            DcModules.loadData();
            
            splashScreen.setStatusMsg("Loading UI");
            Toolkit.getDefaultToolkit().getSystemEventQueue().push( new DcEditPopupEventQueue());

            // load the filters
            new DataFilters().load();
            
            ComponentFactory.setLookAndFeel();
            
            mainFrame = new MainFrame();
            mainFrame.setVisible(true);
            
            Thread splashCloser = new Thread(new SplashScreenCloser());
            splashCloser.start();

            mainFrame.bindData();
            
        } catch (Exception exp) {
            DcLog.add(exp);
            exp.printStackTrace();
            new NativeMessageBox(exp.getMessage());
            System.exit(0);
        }
        
        new FreeResourcesTask();
        
        if (DcSettings.getBoolean(DcRepository.settings.stShowTipsOnStartup)) {
            TipOfTheDayDialog dlg = new TipOfTheDayDialog();
            dlg.setVisible(true);
        }  
    }
    
    public static Platform getPlatform() {
        return platform;
    }
    
    private static void upgrade() throws Exception {
        new SystemUpgrade().upgrade();
    }
    
    private static void createDirectories() {
        // check if the data directory exists. If not, create it
        File file = new File(baseDir + "data");
        if (!file.exists())
            file.mkdir();

        file = new File(imageDir);
        if (!file.exists())
            file.mkdir();
        
        file = new File(reportDir);
        if (!file.exists())
            file.mkdir();
    }    


    private static void showSplashScreen(){
        splashScreen = new SplashScreen();
        splashScreen.splash();
    }

    public static void closeSplashScreen() {
        if (splashScreen != null) {
            splashScreen.dispose();
            splashScreen = null;
        }
    }

    public static void moveSplashScreenToBack() {
        if (splashScreen != null) splashScreen.toBack();
    }

    public static void moveSplashScreenToFront() {
        if (splashScreen != null) splashScreen.toFront();
    }

    private static final class SplashScreenCloser implements Runnable {
        public void run(){
            if (splashScreen != null) 
                splashScreen.dispose();
                
            splashScreen = null;
        }
    }

    private static void setBaseDir(String s) {
        baseDir = s;
        
        if (baseDir == null || baseDir.trim().length() == 0) {
            final URL location;
            final String classLocation = DataCrow.class.getName().replace('.', '/') + ".class";
            final ClassLoader loader = DataCrow.class.getClassLoader();
            if (loader == null) {
              location = ClassLoader.getSystemResource(classLocation);
            } else {
              location = loader.getResource(classLocation);
            }
    
            String dir = location.getFile();
            int index = dir.indexOf("_classes/net/sf/dc");
            if (index > -1)
                dir = dir.substring(0, index);
    
            index = dir.indexOf("_build/net/sf/dc");
            if (index > -1)
                dir = dir.substring(0, index);
    
            index = dir.indexOf("classes/net/sf/dc");
            if (index > -1)
                dir = dir.substring(0, index);
    
            index = dir.indexOf("datacrow.jar");
            if (index > -1)
                dir = dir.substring(0, index);
    
            if (dir.startsWith("file:"))
                dir = dir.substring(5, dir.length());
    
            dir = dir.replaceAll("%20", " ");
            baseDir = dir;
        } else {
            while (baseDir.indexOf('\\') > -1)
                baseDir = baseDir.replace('\\', '/');
            
            baseDir += baseDir.endsWith("/") ? "" : "/";
        }
                
        try {
            File file = new File(DataCrow.baseDir + "data");
            file.mkdirs();
            
            file = new File(DataCrow.baseDir + "data/data_crow.log");
            RandomAccessFile raf = new RandomAccessFile(file, "rw");
            raf.close();
            
            file = new File(DataCrow.baseDir + "icons_system");
            
            if (!file.exists())
                throw new Exception("Directory " + DataCrow.baseDir + " is not the Data Crow installation directory!");
            
        } catch (Exception exp) {
            String message = "Data Crow was unable to get a valid location for its data (" + exp.getMessage() + "). \n" +
                             "Correct the error or specify the installation directory via the parameter '-dir:'\n" +
                             "(java -jar datacrow.jar -dir:d:/datacrow).";
            
            new NativeMessageBox(message);
            System.out.println(message);
            System.exit(0);
        }
        
        imageDir = baseDir + "data/images/";
        reportDir = baseDir + "reports/";
        moduleDir = baseDir + "modules/";
    }
}
