/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JComboBox;

import net.sf.dc.console.MainFrame;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.views.View;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.DcPropertyModule;
import net.sf.dc.core.modules.IChildModule;
import net.sf.dc.core.modules.MappingModule;
import net.sf.dc.core.objects.DcAssociate;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcProperty;
import net.sf.dc.core.objects.Loan;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.objects.helpers.ContactPerson;
import net.sf.dc.core.objects.template.Templates;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.onlinesearch.OnlineSearchHelper;
import net.sf.dc.util.StringUtils;

/**
 * The persistency layer of Data Crow. 
 */ 
public class DataManager {

    private static final Map datasets = new HashMap();
    private static final Map listeners = new HashMap();
    
    public DataManager() {
        initialize();
    }

    public static void bindData(MasterView view, Collection objects) {
        view.add(objects);
    }
    
    public static void bindData(MasterView view, int module, DataFilter df) {
        DcObject[] result = get(module, df);
        
        if (result != null && result.length > 0)
            view.add(result);
        else
            new MessageBox(DcResources.getText("magNoItemsFound"), MessageBox._INFORMATION);
    }
    
    public static void bindData(View view, int module, DataFilter df, boolean refresh) {
        if (view.getRows() == 0 || refresh)
            view.add(get(module, df));
    }
    
    public static void refresh(DcModule module) {
        Collection c = module.loadData();
        datasets.put(new Integer(module.getIndex()), c);
    }
    
    public static void update(DcObject dco, int module) {
        // get the real object
        DcObject o = getObject(dco.getModule().getIndex(), dco.getID());
        
        updatePictures(dco);
        
        o.reload();
        updateRelated(dco, false);
        o.initializeReferences();
        
        updateUiComponents(o.getModule().getIndex());
        updateView(o, 0, module, MainFrame._SEARCHTAB);

        // release if the passed object is just a template
        if (dco != o)
            dco.unload();
    }    
    
    public static void add(DcObject dco, int module) {
    	if (dco.getModule() instanceof DcMediaModule || dco.getModule().getIndex() == DcModules._CONTACTPERSON)
            updatePictures(dco);
        
        Object values = datasets.get(new Integer(dco.getModule().getIndex()));
        if (values instanceof Collection) {
            Collection dataset = (Collection) values;
            dataset.add(dco);
        } else if (values instanceof Map) {
            Map dataset = (Map) values;
            Object key = getKey(dco);
            
            Collection c = (Collection) dataset.get(key);
            c = c == null ? new ArrayList() : c;
            c.add(dco);
            dataset.put(key, c);
        }
        
        if (dco.getModule().canBeLended())
            dco.setValue(DcObject._SYS_AVAILABLE, Boolean.TRUE);

        updateRelated(dco, false);
        updateUiComponents(dco.getModule().getIndex());
        updateView(dco, 1, module, MainFrame._INSERTTAB);
    }  
    
    public static void remove(DcObject dco, int module) {
        Object values = datasets.get(new Integer(dco.getModule().getIndex()));
        if (values instanceof Collection) {
            Collection dataset = (Collection) values;
            dataset.remove(dco);
        } else if (values instanceof Map) {
            Map dataset = (Map) values;
            Object key = getKey(dco);
            Collection c = (Collection) dataset.get(key);
            if (c != null) 
                c.remove(dco);
        }
        
        updateRelated(dco, true);
        updateUiComponents(dco.getModule().getIndex());
        updateView(dco, 2, module, MainFrame._SEARCHTAB);

        // release values
        dco.unload();
    }
    
    public static DcObject createReference(DcObject dco, int fieldIdx, Object value) {
        String name = value != null ? StringUtils.convertHttpString(value.toString()) : null;
        if (name == null || name.trim().length() == 0)
            return null;
        
        int module = DcModules.getReferencedModule(dco.getField(fieldIdx)).getIndex();
        DcObject ref = value instanceof DcObject ? 
                      (DcObject) value : DataManager.getObjectForDisplayValue(module, name);

        if (ref == null) {
            ref = DcModules.get(module).getDcObject();
    
            boolean onlinesearch = false;
            if (ref instanceof DcAssociate) {
                ref.setValue(DcAssociate._A_NAME, name);
                onlinesearch = ref.getModule().getOnlineSearchClass() != null &&
                               dco.getModule().getSettings().getBoolean(DcRepository.modulesettings.stOnlineSearchSubItems);  
            } else if (ref instanceof DcProperty) {
                ref.setValue(DcProperty._A_NAME, name);
            }
            
            if (onlinesearch) {
                OnlineSearchHelper osh = new OnlineSearchHelper(module);
                DcObject queried = osh.query(ref, name, new int[] {DcAssociate._A_NAME});
                ref = queried != null ? queried : ref;
            }
            
            try {
                ref.setIDs();
            } catch (Exception e) {
                DcLog.add(e);
                ref = null;
            }
        }
        
        if (ref != null) {
            if (dco.getField(fieldIdx).getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION)
                DataManager.addMapping(dco, ref, fieldIdx);
            else
                dco.setValue(fieldIdx, ref);
        }
        
        return ref;
    }
    
    public static void addMapping(DcObject parent, DcObject child, int fieldIdx) {
        DcMapping mapping = (DcMapping) DcModules.get(DcModules.getMappingModIdx(parent.getModule().getIndex(), child.getModule().getIndex())).getDcObject();
        mapping.setValue(DcMapping._A_PARENT_ID, parent.getID());
        mapping.setValue(DcMapping._B_REFERENCED_ID, child.getID());
        mapping.setReferencedObject(child);
        
        Collection mappings = (Collection) parent.getValue(fieldIdx);
        mappings = mappings == null ? new ArrayList() : mappings;
        
        // check if a mapping exists already
        for (Iterator iter = mappings.iterator(); iter.hasNext(); ) {
            DcMapping m = (DcMapping) iter.next();
            if ((m.getReferencedObject() != null && m.getReferencedObject().equals(child)) || 
                (m.getReferencedObject() != null && m.getReferencedObject().toString().equals(child.toString()))     )
                return;
        }
        
        mappings.add(mapping);
        parent.setValue(fieldIdx, mappings);
    }    
    
    private static void updateView(DcObject dco, int mode, int module, int tab) {
        if (dco instanceof DcProperty)
            return;
         
        DcModule m = DcModules.get(module);
        if (tab == MainFrame._INSERTTAB) {
        	if (mode == 0) {
        		m.getCurrentInsertView().updateRow(dco.getID(), dco, true, true, false);
        	} else if (mode == 1) {
                
                if (DcModules.get(module).getSearchView() != null) {
                    DcModules.get(module).getSearchView().add(dco);
                    
                    if (module != DcModules._MEDIA && dco instanceof DcMediaObject && dco.getModule().isTopModule())
                        DcModules.get(DcModules._MEDIA).getSearchView().add(dco);
                }
                
                if (DcModules.get(module).getInsertView() != null)
                    DcModules.get(module).getInsertView().removeRows(new String[] {dco.getID()});
        	}
        } else if (tab == MainFrame._SEARCHTAB) {
            
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                try {
                    DcModule mod = (DcModule) iter.next();
                    MasterView masterView = mod.getSearchView();
                    if (masterView != null) {
                        switch (mode) {
                        case 0:
                            masterView.updateRow(dco.getID(), dco, true, true, false);
                            break;
                        case 1:
                            masterView.add(dco);
                            break;
                        case 2:
                            masterView.removeRows(new String[] {dco.getID()});
                            break;
                        }
                    }
                } catch (Exception exp) {
                    DcLog.add(exp);
                }
            }
        }
    }
    
    private static Object getKey(DcObject dco) {
        if (dco instanceof Picture)
            return dco.getValue(Picture._A_OBJECTID);
        else if (dco instanceof Loan)
            return dco.getValue(Loan._D_OBJECTID);
        else if (dco.getParentReferenceFieldIndex() > -1)
            return dco.getValue(dco.getParentReferenceFieldIndex());

        return null;
    }    
    
    private static void updateRelated(DcObject dco, boolean delete) {
        // updated or create the references
        Collection fields =  dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                
                int mappingIdx = DcModules.getMappingModIdx(field.getModule(), field.getReferenceIdx());
                Collection c = getReferences(mappingIdx, dco.getID());
                c = c == null ? c = new ArrayList() : c;

                // avoiding bug where collection in store is same as collection on object
                // in this case; clone the collection, then perform operations
                Collection mappings = (Collection) dco.getValue(field.getIndex());
                if (c.equals(mappings)) {
                    mappings = new ArrayList();
                    mappings.addAll(c);
                    dco.setValue(field.getIndex(), mappings);
                }
                    
                c.clear();
                
                if (!delete && mappings != null) {
                    for (Iterator iterMapping = mappings.iterator(); iterMapping.hasNext(); ) {
                        DcMapping mapping = (DcMapping) iterMapping.next();
                        mapping.setValue(DcMapping._A_PARENT_ID, dco.getID());
                        c.add(mapping);
                    }
                }
                
                if (c.size() > 0) {
                    int mappingModIdx = DcModules.getMappingModIdx(dco.getModule().getIndex(), field.getReferenceIdx());
                    Map map = (Map) datasets.get(new Integer(mappingModIdx));
                    map.put(dco.getID(), c);
                }
            }
        }
        
        if (delete)
            disgardReferences(dco);
        else if (dco instanceof Loan)
            updateLoanedItems((Loan) dco);
    }   
    
    /**
     * Removes references to this object from any other item. This method should only be called
     * when the provided item is about has been deleted from the database as well as the references
     * to this item. 
     * @param dco the deleted item
     */
    private static void disgardReferences(DcObject dco) {
        if (dco.getModule().hasDependingModules() || dco instanceof DcAssociate || dco instanceof DcProperty) {
            // remove collection items holding a reference to this item
            Collection referencingMods = DcModules.getReferencingModules(dco.getModule().getIndex());
            for (Iterator iter = referencingMods.iterator(); iter.hasNext(); ) {
                DcModule referencingMod = (DcModule) iter.next();
                
                DcModule module = referencingMod instanceof MappingModule ?
                                    DcModules.get(((MappingModule) referencingMod).getParentModIdx()) :
                                    referencingMod;
                for (Iterator iterFlds = module.getFields().iterator(); iterFlds.hasNext(); ) {
                    DcField field = (DcField) iterFlds.next();
                    if (field.getReferenceIdx() == dco.getModule().getIndex()) {
                        
                        if (referencingMod instanceof MappingModule) {
                            DataFilter df = new DataFilter(module.getIndex());
                            df.addEntry(new DataFilterEntry(DataFilterEntry._AND, 
                                                            module.getIndex(), 
                                                            field.getIndex(), 
                                                            Operators.getOperator(Operators._IS_FILLED), 
                                                            null));
                            
                            DcObject[] referencingItems = get(module.getIndex(), df);
                            
                            
                            for (int i = 0; i < referencingItems.length; i++) {
                                Collection references = (Collection) referencingItems[i].getValue(field.getIndex());
                                Object val = null;
                                boolean changed = false;
                                for (Iterator iterRefs = references.iterator(); iterRefs.hasNext(); ) {
                                    DcMapping ref = (DcMapping) iterRefs.next();
                                    if (ref.getReferencedId().equals(dco.getID())) {
                                        val = ref;
                                        changed = true;
                                    }
                                }
                                references.remove(val);
                                
                                if (changed) 
                                    updateView(referencingItems[i], 0, referencingItems[i].getModule().getIndex(), MainFrame._SEARCHTAB);
                            }
                        } else {
                            DataFilter df = new DataFilter(module.getIndex());
                            df.addEntry(new DataFilterEntry(DataFilterEntry._AND, 
                                                            module.getIndex(), 
                                                            field.getIndex(), 
                                                            Operators.getOperator(Operators._EQUAL_TO), 
                                                            dco));
                            
                            DcObject[] items = get(module.getIndex(), df);
                            for (int i = 0; i < items.length; i++) {
                                items[i].setValueLowLevel(field.getIndex(), null);
                                updateView(items[i], 0, items[i].getModule().getIndex(), MainFrame._SEARCHTAB);
                            }
                        }
                    }
                }
            }
        }
        
        if (dco.getModule() instanceof DcMediaModule || dco.getModule().getIndex() == DcModules._CONTACTPERSON) {
            Map pictures = (Map) datasets.get(new Integer(DcModules._PICTURE));
            pictures.remove(dco.getID());
        }

        if (dco.getModule().canBeLended()) {
            Map loans = (Map) datasets.get(new Integer(DcModules._LOAN));
            loans.remove(dco.getID());
        }   
    }

    public static void unloadPictures() {
        Map dataset = (Map) datasets.get(new Integer(DcModules._PICTURE));
        if (dataset != null) {
            for (Iterator iter = dataset.values().iterator(); iter.hasNext(); ) {
                Collection c = (Collection) iter.next();
                for (Iterator iterPics = c.iterator(); iterPics.hasNext(); ) {
                    Picture picture = (Picture) iterPics.next();
                    if (!picture.isNew() && !picture.isUpdated())
                        picture.unload();
                }
            }
        }        
    }
    
    private static void updatePictures(DcObject dco) {
        Collection pictures = new ArrayList();
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getValueType() == DcRepository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(field.getIndex());
                if (picture != null && (picture.isUpdated() || picture.isNew() || picture.isDeleted()))
                    pictures.add(picture);
            }
        }

        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            if (picture.isNew()) {
                Map pictureMap = (Map) datasets.get(new Integer(DcModules._PICTURE));
                Collection c = (Collection) pictureMap.get(dco.getID());
                c = c == null ? new ArrayList() : c;
                picture.markAsUnchanged();
                
                if (!c.contains(picture))
                    c.add(picture);
                else 
                    DcLog.add("Trying to add an picture which was already there : " + picture.toString(), 
                                 DcLog._ERROR);
                
                pictureMap.put(dco.getID(), c);
            } else if (picture.isUpdated()) {
                Map pictureMap = (Map) datasets.get(new Integer(DcModules._PICTURE));
                ArrayList c = (ArrayList) pictureMap.get(dco.getID());
                c = c == null ? new ArrayList() : c;
                
                picture.setValue(Picture._D_IMAGE, null);
                if (c.indexOf(picture) > -1) {
                   DcObject o = (DcObject) c.get(c.indexOf(picture));
                   o.copy(picture);
                   o.markAsUnchanged();
                }
            } else if (picture.isDeleted()) {
                Map pictureMap = (Map) datasets.get(new Integer(DcModules._PICTURE));
                ArrayList c = (ArrayList) pictureMap.get(dco.getID());
                c = c == null ? new ArrayList() : c;
                c.remove(picture);
            }
        }
    }
    
    private static void updateLoanedItems(Loan loan) {
        String parentID = (String) loan.getValue(Loan._D_OBJECTID);
        if (parentID != null) {
            DcObject parent = null;
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                
                if (module.canBeLended())
                    parent = getObject(module.getIndex(), parentID, true);
                
                if (parent != null)
                    break;
            }
            
            if (parent != null) {
                parent.reload();
                updateView(parent, 0, parent.getModule().getIndex(), MainFrame._SEARCHTAB);
            }
        }
    }
    
    private static void updateUiComponents(int module) {
    	Collection c = (Collection) listeners.get(new Integer(module));
    	if (c != null) {
    		for (Iterator iter = c.iterator(); iter.hasNext(); ) {
    			JComboBox combo = (JComboBox) iter.next();
    			
                try {
                    Object o = combo.getSelectedItem();
                    combo.removeAllItems();
                    combo.addItem(" ");

                    DcObject[] objects = get(module, null);
                    for (int i = 0; i < objects.length; i++)
                        combo.addItem(objects[i]);

                    if (o != null)
                    	combo.setSelectedItem(o);
                    else
                    	combo.setSelectedIndex(0);

                } catch (Exception exp) {
                    DcLog.add(exp);
                    combo.setSelectedIndex(0);
                }
    		}
    	}
    }
    
    public static void registerUiComponent(JComboBox component, int module) {
    	Collection c = (Collection) listeners.get(new Integer(module));
    	c = c == null ? new ArrayList() : c;
    	c.add(component);
    	listeners.put(new Integer(module), c);
    	
    	updateUiComponents(module);
    }
    
    public static int contains(int module, DataFilter df) {
        DcObject[] objects = get(module, df);
        return objects != null ? objects.length : 0;
    }
    
    public static Collection getLoans(String parentID) {
        Map loans = (Map) datasets.get(new Integer(DcModules._LOAN));
        Collection c = (Collection) loans.get(parentID);
        return c != null ? c : new ArrayList();
    }
    
    public static Loan getCurrentLoan(String parentID) {
        Collection loans = getLoans(parentID);
        
        if (loans != null) {
            for (Iterator iter = loans.iterator(); iter.hasNext(); ) {
                Loan loan = (Loan) iter.next();
                if (loan.getValue(Loan._B_ENDDATE) == null)
                    return loan;
            }
        }
        
        return new Loan();
    }
    
    public static DcObject getObjectForDisplayValue(int module, String s) {
        return getObject(module, s, false);
    }    
    
    public static DcObject getObject(int module, String ID) {
        return getObject(module, ID, true);
    }    
    
    private static DcObject getObject(int module, String s, boolean searchOnId) {
        if (s == null || s.length() == 0)
            return null;
        
        Object values = datasets.get(new Integer(module));
        if (values instanceof ArrayList) {
            ArrayList dataset = (ArrayList) values;
            for (Iterator iter = dataset.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next();
                
                if (searchOnId && dco.getID().equals(s)) 
                    return dco;
                else if (!searchOnId && StringUtils.normalize(dco.toString()).equals(StringUtils.normalize(s)))
                    return dco; 
            }
        } else if (values instanceof Map) {
            Map dataset = (Map) values;
            for (Iterator iterSet = dataset.keySet().iterator(); iterSet.hasNext(); ) {
                Object key = iterSet.next();
                ArrayList c = (ArrayList) dataset.get(key);
                for (Iterator iter = c.iterator(); iter.hasNext(); ) {
                    DcObject dco = (DcObject) iter.next();
                    
                    if (searchOnId && dco.getID().equals(s)) 
                        return dco;
                    else if (!searchOnId && StringUtils.normalize(dco.toString()).equals(StringUtils.normalize(s)))
                        return dco; 
                }
            }
        }
        return null;
    }
    
    public static Collection getReferences(int module, String parentId) {
        Map dataset = (Map) datasets.get(new Integer(module));
        Collection references = (Collection) dataset.get(parentId);
        return references == null ? new ArrayList() : references; 
    }
    
    public static Collection getPictures(String parentID) {
        Map dataset = (Map) datasets.get(new Integer(DcModules._PICTURE));
        Collection pictures = (Collection) dataset.get(parentID);
        return pictures == null ? new ArrayList() : pictures; 
    }
    
    public static Collection getChildren(String parentID, int childModule) {
        Collection children = new ArrayList();
        Map dataset = (Map) datasets.get(new Integer(childModule));
        children = (Collection) dataset.get(parentID);
        
        DcModule childMod =  DcModules.get(childModule);
        DataFilter filter = new DataFilter(childMod.getDcObject());
        filter.setOrder(new DcField[] {childMod.getField(childMod.getDefaultSortFieldIdx())});
        
        children = children == null ? new ArrayList() : children;
        filter.sort(children);
        return children;
    }
    
    public static DcObject[] get(int module, DataFilter df) {
        DataFilter filter = df;
        
    	Object values;
        Collection c = new ArrayList();
        try {
        	if (module == DcModules._MEDIA) {
        		c = new ArrayList();
        		
        		for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
        			DcModule m = (DcModule) iter.next();
        			if (m.getIndex() != module && m.isTopModule() && m instanceof DcMediaModule) {
        				DcObject[] objects = get(m.getIndex(), filter);
        				for (int i = 0; i < objects.length; i++)
        					c.add(objects[i]);
        			}
        		}
        		values = c;
        	} else {
        		values = datasets.get(new Integer(module));
        	}
            
            //Collection c;
            if (values instanceof Map) {
                c = new ArrayList();   //;
            	Collection n = ((Map) values).values();
            	for (Iterator iter = n.iterator(); iter.hasNext(); ) {
                    Collection col = (Collection) iter.next();
                    for (Iterator iterCol = col.iterator(); iterCol.hasNext(); ) {
                        DcObject dco = (DcObject) iterCol.next();
                        if (filter == null || (filter != null && filter.applies(dco)))
                            c.add(dco);
                    }
            	}
            } else {
                Collection dataset = (Collection) values;
                dataset = dataset == null ? new ArrayList() : dataset;
                c = new ArrayList();
                
                DcObject[] objects = new DcObject[dataset.size()];
                objects = (DcObject[]) dataset.toArray(objects);
                for (int i = 0; i < objects.length; i++ ) {
                    DcObject dco = objects[i];
                    if (filter == null || (filter != null && filter.applies(dco)))
                        c.add(dco);
                }
            }
            
            if (filter == null) {
                DcObject dco = DcModules.get(module).getDcObject();
                filter = new DataFilter(dco);

                if (DcModules.get(module) instanceof DcPropertyModule)
                    filter.setOrder(new DcField[] {dco.getField(DcProperty._A_NAME)});
                if (dco instanceof DcAssociate)
                    filter.setOrder(new DcField[] {dco.getField(DcAssociate._A_NAME)});
                if (module == DcModules._CONTACTPERSON)
                    filter.setOrder(new DcField[] {dco.getField(ContactPerson._A_NAME)});
            }
            
            if (filter != null) 
                filter.sort(c);
            
        } catch (Exception exp) {
            DcLog.add(exp);
        }
        
        DcObject[] objects = new DcObject[c.size()];
        return (DcObject[]) c.toArray(objects);
    }    
    
    
    /**********************************
     * Initialization
     **********************************/
    public void setPictures() {
        Map pics = new HashMap();
        Collection pictures = DcModules.get(DcModules._PICTURE).loadData();
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture pic = (Picture) iter.next();
            Object key = pic.getValue(Picture._A_OBJECTID);
            
            Collection c = (Collection) pics.get(key);
            if (c == null)
                c = new ArrayList();
            
            c.add(pic);
            pics.put(key, c);
         }
         
         add(DcModules._PICTURE, pics);
    }
    
    public void setLoans() {
        Map loanMap = new HashMap();
        Collection loans = DcModules.get(DcModules._LOAN).loadData();
        for (Iterator iter = loans.iterator(); iter.hasNext(); ) {
            Loan loan = (Loan) iter.next();
            Object key = loan.getValue(Loan._D_OBJECTID);
            
            Collection c = (Collection) loanMap.get(key);
            if (c == null)
                c = new ArrayList();
            
            c.add(loan);
            loanMap.put(key, c);
         }
         
         add(DcModules._LOAN, loanMap);
    }    
    
    private void setChildren(IChildModule childModule) {
        Map map = new HashMap();

        DcModule module = DcModules.get(childModule.getIndex());
        Collection children = module.loadData();
        for (Iterator iter = children.iterator(); iter.hasNext(); ) {
            DcObject dco = (DcObject) iter.next();
            Object key = dco.getValue(dco.getParentReferenceFieldIndex());
            
            Collection c = (Collection) map.get(key);
            if (c == null)
                c = new ArrayList();
            
            c.add(dco);
            map.put(key, c);
         }
         
         add(childModule.getIndex(), map);
    }    
    
    public void setReferences(MappingModule mappingModule) {
        Map map = new HashMap();
        
        Collection mappings = mappingModule.loadData();
        for (Iterator iter = mappings.iterator(); iter.hasNext(); ) {
            DcMapping mapping = (DcMapping) iter.next();
            Object parentId = mapping.getParentId();
            
            Collection c = (Collection) map.get(parentId);
            c = c == null ? new ArrayList() : c;
            c.add(mapping);

            map.put(parentId, c);
         }
         
         add(mappingModule.getIndex(), map);
    }
    
    private void initialize() {
        datasets.clear();
        
        ThreadGroup group = new ThreadGroup("data-fetchers");
        Collection fetchers = new ArrayList();

        setPictures();
        
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext();) {
            DcModule module = (DcModule) iter.next();
            if (module instanceof MappingModule)
                setReferences((MappingModule) module);
        }         

        setLoans();
        
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext();) {
            DcModule module = (DcModule) iter.next();
            
            if (   !datasets.containsKey(new Integer(module.getIndex())) && 
                   !module.isTopModule() && !module.isChildModule()) {
                
                Thread fetcher = new DataFetcher(module, group);
                fetcher.start();
                fetchers.add(fetcher);
            }
        }        

        for (Iterator iter = DcModules.getAllModules(); iter.hasNext();) {
            DcModule module = (DcModule) iter.next();
            
            if (   !datasets.containsKey(new Integer(module.getIndex())) && 
                    module.hasDependingModules() ) {
                
                Thread fetcher = new DataFetcher(module, group);
                fetcher.start();
                fetchers.add(fetcher);
            }
        }         
        
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext();) {
            DcModule module = (DcModule) iter.next();
            if (module.isChildModule()) 
                setChildren((IChildModule) module);
        }
        
        for (Iterator iter = fetchers.iterator(); iter.hasNext();) {
            Thread thread = (Thread) iter.next();
            try {
                thread.join();
            } catch (InterruptedException e) {}
        }         
        
        fetchers.clear();
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext();) {
            DcModule module = (DcModule) iter.next();
            
            if (!datasets.containsKey(new Integer(module.getIndex())) && module.isTopModule()) {
                Thread fetcher = new DataFetcher(module, group);
                fetcher.start();
                fetchers.add(fetcher);
            }
        }        
        
        for (Iterator iter = fetchers.iterator(); iter.hasNext();) {
            Thread thread = (Thread) iter.next();
            try {
                thread.join();
            } catch (InterruptedException e) {}
        }    
        
        Templates.refresh();
    }
    
    private void add(int module, Collection objects) {
        datasets.put(new Integer(module), objects);
    }

    private void add(int module, Map objects) {
        datasets.put(new Integer(module), objects);
    }
    
    final class DataFetcher extends Thread {
    	
        private final DcModule module;
        
        public DataFetcher(DcModule module, ThreadGroup group) {
            super(group, "datafetcher-" + module.getName());
            this.module = module;
        }
        
        public void run() {
            try {
                if (module.getIndex() == DcModules._SOFTWARE + DcModules._TEMPLATE) 
                    System.out.println("");
                    
                add(module.getIndex(), module.loadData());
            } catch (OutOfMemoryError ome) {
                new MessageBox(DcResources.getText("msgOutOfMemory"), MessageBox._ERROR);
                System.exit(0);
            } catch (Exception exp) {
                exp.printStackTrace();
                new MessageBox(DcResources.getText("msgDataLoadingError"), MessageBox._ERROR);
                System.exit(0);
            }
        }
    }
}
