/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.console.windows.messageboxes.QuestionBox;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.MappingModule;
import net.sf.dc.core.objects.DcAssociate;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.util.StringUtils;

/**
 * Converts the current database before the actual module tables are created / updated.
 * This means that the code here defies workflow logic and is stricktly to be used for
 * table conversions and migration out of the scope of the normal module upgrade code.
 * 
 * @author rj.vanderwaals
 */
public class DatabaseConversion {
    
    public void start() {
        boolean converted = false;
        converted |= convertTablesToVersion3();
        converted |= convertMappingTables();
        converted |= convertReferences();
        converted |= convertReferences();
        converted |= removeGhostReferences();
        
        if (converted) {
            DcLog.show(false);
            new MessageBox("Conversion was successfully completed. Data Crow will now exit.", MessageBox._INFORMATION);
            DatabaseManager.closeDatabases(true);
            System.exit(1);
        }
    }
    
    /**
     * Cleans references to deleted items. These were caused by Data Crow not cleaning up references
     * after a delete of a referenced item. Due to Data Crow's open structure, referential integrity
     * is not managed by the database but is the responsibility of the Query class.
     * @return
     */
    private boolean removeGhostReferences() {
        Connection connection = DatabaseManager.getConnection();
        
        try {
            Statement stmt = connection.createStatement();
            for (Iterator iter = DcModules.getAllModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                
                if (module.hasDependingModules()) {
                    Collection referencingMods = DcModules.getReferencingModules(module.getIndex());
                    for (Iterator iterRefs = referencingMods.iterator(); iterRefs.hasNext(); ) {
                        DcModule referencingMod = (DcModule) iterRefs.next();
                        try {
                            String cond = " NOT IN (SELECT " + module.getField(DcObject._ID).getDatabaseFieldName() + 
                                          "  FROM " + module.getTableName() + ")";
                            
                            if (referencingMod instanceof MappingModule) {
                                String sql = "DELETE FROM " + referencingMod.getTableName() + " WHERE " +
                                             referencingMod.getField(DcMapping._B_REFERENCED_ID).getDatabaseFieldName() + cond;
                                stmt.execute(sql);
                            } else {
                                for (Iterator iterFlds = referencingMod.getFields().iterator(); iterFlds.hasNext(); ) {
                                    DcField field = (DcField) iterFlds.next();
                                    if (!field.isUiOnly() && field.getReferenceIdx() == module.getIndex()) {
                                        String sql = "UPDATE " + referencingMod.getTableName() + " SET " +
                                                     field.getDatabaseFieldName() + " = NULL WHERE " + 
                                                     field.getDatabaseFieldName() + cond;
                                        stmt.execute(sql);
                                    }
                                }
                            }
                        } catch (Exception e) {
                            DcLog.add("Non-critical error: could not clean possible ghost references for " + referencingMod, DcLog._ERROR);
                            DcLog.add(e);
                        }
                    }
                }
            }
        } catch (SQLException se) {
            DcLog.add(se);
        }
        
        try {
            connection.close();
        } catch (Exception exp) {}
        
        // a restart after doing this is not needed in any case!  
        return false;
    }
    
    
    private boolean convertReferences() {
        Connection connection = DatabaseManager.getConnection();
        
        boolean converted = false;
        
        try {
            Statement stmt = connection.createStatement();
          
            try {
                ResultSet rs = stmt.executeQuery("SELECT TOP 1 actors from MOVIE;");
                rs.close();
            } catch (Exception exp) {
                return false;
            }

            QuestionBox qb = new QuestionBox("You are about to coverting text items to real references");
            if (!qb.isAffirmative())
                System.exit(0);
    
            DataCrow.closeSplashScreen();
            DcLog.show(true);

            //**************************************************************************
            //Upgrade movie table <> actors
            createTable(DcModules._ACTOR);
            createTable(DcModules._ACTOR + DcModules._MAPPING + DcModules._MOVIE);
            DcModule actorModule = DcModules.get(DcModules._ACTOR);
            DcModule actorMovieModule = DcModules.get(DcModules._ACTOR + DcModules._MAPPING + DcModules._MOVIE);
            ResultSet rs = stmt.executeQuery("select id, actors from movie;");
            while (rs.next()) {
                createItems(rs.getString("actors"), rs.getString("ID"), 
                            actorMovieModule, actorModule);
            }
            stmt.execute("alter table movie drop column actors");
            rs.close();
            
            //**************************************************************************
            //Upgrade movie table <> directors
            createTable(DcModules._DIRECTOR);
            createTable(DcModules._DIRECTOR + DcModules._MAPPING + DcModules._MOVIE);
            DcModule directorModule = DcModules.get(DcModules._DIRECTOR);
            DcModule directorMovieModule = DcModules.get(DcModules._DIRECTOR + DcModules._MAPPING + DcModules._MOVIE);
            rs = stmt.executeQuery("select id, director from movie;");
            while (rs.next()) {
                createItems(rs.getString("director"), rs.getString("ID"), 
                            directorMovieModule, directorModule);
            }
            stmt.execute("alter table movie drop column director");
            rs.close();

            //**************************************************************************
            //Upgrade book table <> authors
            createTable(DcModules._AUTHOR);
            createTable(DcModules._AUTHOR + DcModules._MAPPING + DcModules._BOOK);
            DcModule authorModule = DcModules.get(DcModules._AUTHOR);
            DcModule authorBookModule = DcModules.get(DcModules._AUTHOR + DcModules._MAPPING + DcModules._BOOK);
            rs = stmt.executeQuery("select id, author from book;");
            while (rs.next()) {
                createItems(rs.getString("author"), rs.getString("ID"), 
                            authorBookModule, authorModule);
            }
            stmt.execute("alter table book drop column author");
            rs.close();

            //**************************************************************************
            //Upgrade book table <> authors
            createTable(DcModules._BOOKPUBLISHER);
            createTable(DcModules._BOOKPUBLISHER + DcModules._MAPPING + DcModules._BOOK);
            DcModule pubModule = DcModules.get(DcModules._BOOKPUBLISHER);
            DcModule pubBookModule = DcModules.get(DcModules._BOOKPUBLISHER + DcModules._MAPPING + DcModules._BOOK);
            rs = stmt.executeQuery("select id, publisher from book;");
            while (rs.next()) {
                createItems(rs.getString("publisher"), rs.getString("ID"), 
                            pubBookModule, pubModule);
            }
            stmt.execute("alter table book drop column publisher");
            rs.close();

            //**************************************************************************
            //Upgrade software table <> publishers
            createTable(DcModules._SOFTWAREPUBLISHER);
            createTable(DcModules._SOFTWAREPUBLISHER + DcModules._MAPPING + DcModules._SOFTWARE);
            DcModule pubSoftModule = DcModules.get(DcModules._SOFTWAREPUBLISHER);
            DcModule pubSoftSoftwareModule = DcModules.get(DcModules._SOFTWAREPUBLISHER + DcModules._MAPPING + DcModules._SOFTWARE);
            rs = stmt.executeQuery("select id, publisher from software;");
            while (rs.next()) {
                String publisher = rs.getString("publisher");
                publisher = publisher != null ? publisher.replaceAll(",", "") : publisher;
                createItems(publisher, rs.getString("ID"), 
                        pubSoftSoftwareModule, pubSoftModule);
            }
            stmt.execute("alter table software drop column publisher");
            rs.close();

            //**************************************************************************
            //Upgrade software table <> developers
            createTable(DcModules._DEVELOPER);
            createTable(DcModules._DEVELOPER + DcModules._MAPPING + DcModules._SOFTWARE);
            DcModule devModule = DcModules.get(DcModules._DEVELOPER);
            DcModule devSoftwareModule = DcModules.get(DcModules._DEVELOPER + DcModules._MAPPING + DcModules._SOFTWARE);
            rs = stmt.executeQuery("select id, developer from software;");
            while (rs.next()) {
                String developer = rs.getString("developer");
                developer = developer != null ? developer.replaceAll(",", "") : developer;
                createItems(developer, rs.getString("ID"), 
                        devSoftwareModule, devModule);
            }
            stmt.execute("alter table software drop column developer");
            rs.close();            

            //**************************************************************************
            //Upgrade audiocd table <> artists
            createTable(DcModules._MUSICARTIST);
            createTable(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._AUDIOCD);
            DcModule artistModule = DcModules.get(DcModules._MUSICARTIST);
            DcModule artistAudioCdModule = DcModules.get(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._AUDIOCD);
            rs = stmt.executeQuery("select id, artist from audiocd;");
            while (rs.next()) {
                String value = rs.getString("artist");
                createItems(value, rs.getString("ID"), 
                        artistAudioCdModule, artistModule);
            }
            stmt.execute("alter table audiocd drop column artist");
            rs.close();            
            
            //**************************************************************************
            //Upgrade musicalbum table <> artists
            createTable(DcModules._MUSICARTIST);
            createTable(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._MUSICALBUM);
            DcModule artistMusicAlbumModule = DcModules.get(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._MUSICALBUM);
            rs = stmt.executeQuery("select id, artist from musicalbum;");
            while (rs.next()) {
                String value = rs.getString("artist");
                createItems(value, rs.getString("ID"), 
                        artistMusicAlbumModule, artistModule);
            }
            stmt.execute("alter table musicalbum drop column artist");
            rs.close();            

            //**************************************************************************
            //Upgrade audiotrack table <> artists
            createTable(DcModules._MUSICARTIST);
            createTable(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._AUDIOTRACK);
            DcModule artistAudioTrackModule = DcModules.get(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._AUDIOTRACK);
            rs = stmt.executeQuery("select id, artist from audiotrack;");
            while (rs.next()) {
                String value = rs.getString("artist");
                createItems(value, rs.getString("ID"), 
                        artistAudioTrackModule, artistModule);
            }
            stmt.execute("alter table audiotrack drop column artist");
            rs.close();                
            
            //**************************************************************************
            //Upgrade musictrack table <> artists
            createTable(DcModules._MUSICARTIST);
            createTable(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._MUSICTRACK);
            DcModule artistMusicTrackModule = DcModules.get(DcModules._MUSICARTIST + DcModules._MAPPING + DcModules._MUSICTRACK);
            rs = stmt.executeQuery("select id, artist from musictrack;");
            while (rs.next()) {
                String value = rs.getString("artist");
                createItems(value, rs.getString("ID"), artistMusicTrackModule, artistModule);
            }
            stmt.execute("alter table musictrack drop column artist");
            rs.close();                
            
            stmt.close();
            converted = true;
            
        } catch (Exception exp) {
            DcLog.add(exp);
            new MessageBox("Conversion was unsuccessful! " + exp.getMessage(), MessageBox._ERROR);
        }

        try {
            connection.close();
        } catch (Exception exp) {}
  
        return converted;
    }
    
    Map existingRefs = new HashMap(); 
    
    private void createItems(String commaString, 
                             String parentID, 
                             DcModule mapping, 
                             DcModule minor) throws SQLException {
        
        String names = commaString;
        if (names == null || names.trim().length() == 0)
            return;
        
        if (names.indexOf(",") < -1)
            names += ", ";
        
        StringTokenizer st = new StringTokenizer(names, ",");
        while (st.hasMoreTokens()) {
            String name = (String) st.nextElement();
            String nameNormalized = StringUtils.normalize(name);
            
            DcObject associate;
            if (existingRefs.containsKey(nameNormalized)) {
                associate = (DcObject) existingRefs.get(nameNormalized);
            } else {
                associate = minor.getDcObject();
                associate.setValue(DcAssociate._A_NAME, name.trim());
                associate.setIDs();
                
                PreparedStatement ps = new Query(Query._INSERT, associate, null, null).getQuery();
                ps.execute();
                existingRefs.put(nameNormalized, associate);
            }
            
            DcObject dcoMapping = mapping.getDcObject();
            dcoMapping.setValue(DcMapping._A_PARENT_ID, parentID);
            dcoMapping.setValue(DcMapping._B_REFERENCED_ID, associate.getID());
            
            PreparedStatement ps = new Query(Query._INSERT, dcoMapping, null, null).getQuery();
            ps.execute();
        }
    }
    
    private void createTable(int module) {
        try {
            DcModule m = DcModules.get(module);
            PreparedStatement ps = new Query(Query._CREATE, m.getDcObject(), null, null).getQuery();
            ps.execute();
        } catch (Exception e) {}
    }
    
    private boolean convertMappingTables() {
        Connection connection = DatabaseManager.getConnection();
        
        boolean converted = false;

        try {
            Statement stmt = connection.createStatement();
          
            try {
                ResultSet rs = stmt.executeQuery("SELECT TOP 1 MEDIAOBJECTID from X_AUDIOCD_MUSICGENRE;");
                rs.close();
            } catch (Exception exp) {
                return false;
            }
        
            QuestionBox qb = new QuestionBox("The mapping tables need to be converted. Continue?");
    
            if (!qb.isAffirmative())
                System.exit(0);
    
            DataCrow.closeSplashScreen();
        
            DcLog.show(true);
        
            stmt.execute("ALTER TABLE X_AUDIOCD_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_AUDIOTRACK_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MOVIE_MOVIE_GENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MUSICALBUM_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MUSICTRACK_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");

            stmt.execute("ALTER TABLE X_AUDIOCD_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_AUDIOTRACK_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MOVIE_MOVIE_GENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MUSICALBUM_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MUSICTRACK_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");

            stmt.close();
    
            converted = true;
            
      } catch (Exception exp) {
          DcLog.add(exp);
          new MessageBox("Conversion was unsuccessful! " + exp.getMessage(), MessageBox._ERROR);
      }
    
      try {
          connection.close();
      } catch (Exception exp) {}
      
      return converted;
    }
    
    private boolean convertTablesToVersion3() {
        Connection connection = DatabaseManager.getConnection();
        
        boolean converted = false;
        
        try {
            Statement stmt = connection.createStatement();
          
            try {
                ResultSet rs = stmt.executeQuery("SELECT TOP 1 * from AUDIOCDS;");
                rs.close();
            } catch (Exception exp) {
                return false;
            }
        
            QuestionBox qb = new QuestionBox("Your database is about to be rebuild. The original data is transferred to the new format. Continue?");
    
            if (!qb.isAffirmative())
                System.exit(0);
    
            DataCrow.closeSplashScreen();
            DcLog.show(true);
        
            stmt.execute("ALTER TABLE AUDIOCDGENREMAPPING RENAME TO X_AUDIOCD_MUSICGENRE");
            stmt.execute("ALTER TABLE AUDIOCDS RENAME TO AUDIOCD");
            stmt.execute("ALTER TABLE AUDIOCDSTATES RENAME TO AUDIOCD_STATE");
            stmt.execute("ALTER TABLE AUDIOCDTEMPLATE RENAME TO AUDIOCD_TEMPLATE");

            stmt.execute("ALTER TABLE AUDIOTRACKS RENAME TO AUDIOTRACK");
            stmt.execute("ALTER TABLE AUDIOTRACKGENREMAPPING RENAME TO X_AUDIOTRACK_MUSICGENRE");
            stmt.execute("ALTER TABLE AUDIOTRACK ALTER COLUMN ALBUM RENAME TO ALBUMID");
            
            stmt.execute("ALTER TABLE BOOKS RENAME TO BOOK");
            stmt.execute("ALTER TABLE BOOKSTATES RENAME TO BOOK_STATE");
            stmt.execute("ALTER TABLE BOOKCATEGORIES RENAME TO BOOK_CATEGORY");

            stmt.execute("ALTER TABLE BOOKTEMPLATE RENAME TO BOOK_TEMPLATE");
            stmt.execute("ALTER TABLE BOOK ALTER COLUMN TYPE RENAME TO CATEGORY");
            stmt.execute("ALTER TABLE BOOK_TEMPLATE ALTER COLUMN TYPE RENAME TO CATEGORY");
            
            stmt.execute("ALTER TABLE IMAGES RENAME TO IMAGE");
            stmt.execute("ALTER TABLE IMAGECATEGORIES RENAME TO IMAGE_CATEGORY");
            stmt.execute("ALTER TABLE IMAGESTATES RENAME TO IMAGE_STATE");
            stmt.execute("ALTER TABLE IMAGESTORAGEMEDIA RENAME TO IMAGE_STORAGEMEDIUM");
            stmt.execute("ALTER TABLE IMAGETEMPLATE RENAME TO IMAGE_TEMPLATE");
            
            stmt.execute("ALTER TABLE MOVIES RENAME TO MOVIE");
            stmt.execute("ALTER TABLE MOVIETEMPLATE RENAME TO MOVIE_TEMPLATE");
            stmt.execute("ALTER TABLE MOVIESTATES RENAME TO MOVIE_STATE");
            stmt.execute("ALTER TABLE MOVIEGENRE RENAME TO MOVIE_GENRE");
            stmt.execute("ALTER TABLE MOVIESTORAGEMEDIA RENAME TO MOVIE_STORAGEMEDIUM");
            stmt.execute("ALTER TABLE MOVIEGENREMAPPINGS RENAME TO X_MOVIE_MOVIE_GENRE");
            
            stmt.execute("ALTER TABLE MUSICALBUMS RENAME TO MUSICALBUM");
            stmt.execute("ALTER TABLE MUSICALBUMTEMPLATE RENAME TO MUSICALBUM_TEMPLATE");
            stmt.execute("ALTER TABLE MUSICALBUMGENREMAPPINGS RENAME TO X_MUSICALBUM_MUSICGENRE");
            stmt.execute("ALTER TABLE MUSICALBUMSTATES RENAME TO MUSICALBUM_STATE");
            stmt.execute("ALTER TABLE MUSICALBUMSTORAGEMEDIA RENAME TO MUSICALBUM_STORAGEMEDIUM");
            
            stmt.execute("ALTER TABLE MUSICTRACKS RENAME TO MUSICTRACK");
            stmt.execute("ALTER TABLE MUSICTRACKSTATES RENAME TO MUSICTRACK_STATE");
            stmt.execute("ALTER TABLE MUSICTRACKGENREMAPPINGS RENAME TO X_MUSICTRACK_MUSICGENRE");
            stmt.execute("ALTER TABLE MUSICTRACK ALTER COLUMN ALBUM RENAME TO ALBUMID");
            stmt.execute("ALTER TABLE MUSICTRACK ALTER COLUMN LENGTH RENAME TO PLAYLENGTH");
            
            stmt.execute("ALTER TABLE MUSICTRACKTEMPLATE RENAME TO MUSICTRACK_TEMPLATE");
            stmt.execute("ALTER TABLE MUSICTRACK_TEMPLATE ALTER COLUMN ALBUM RENAME TO ALBUMID");
            stmt.execute("ALTER TABLE MUSICTRACK_TEMPLATE ALTER COLUMN LENGTH RENAME TO PLAYLENGTH");
            
            stmt.execute("ALTER TABLE PERSONS RENAME TO PERSON");
            stmt.execute("ALTER TABLE PERSONTYPES RENAME TO PERSON_CATEGORY");
            stmt.execute("ALTER TABLE PERSON ALTER COLUMN PERSONTYPE RENAME TO CATEGORY");
            
            stmt.execute("ALTER TABLE PICTURES RENAME TO PICTURE");
            
            stmt.execute("ALTER TABLE SOFTWARECATEGORIES RENAME TO SOFTWARE_CATEGORY");
            stmt.execute("ALTER TABLE SOFTWAREPLATFORMS RENAME TO SOFTWARE_PLATFORM");
            stmt.execute("ALTER TABLE SOFTWARESTATES RENAME TO SOFTWARE_STATE");
            stmt.execute("ALTER TABLE SOFTWARESTORAGEMEDIA RENAME TO SOFTWARE_STORAGEMEDIUM");
            stmt.execute("ALTER TABLE SOFTWARETEMPLATE RENAME TO SOFTWARE_TEMPLATE");
            
            stmt.execute("ALTER TABLE SOFTWARE ALTER COLUMN TYPE RENAME TO CATEGORY");
            stmt.execute("ALTER TABLE SOFTWARE_TEMPLATE ALTER COLUMN TYPE RENAME TO CATEGORY");
            

            stmt.execute("ALTER TABLE X_AUDIOCD_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_AUDIOTRACK_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MOVIE_MOVIE_GENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MUSICALBUM_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");
            stmt.execute("ALTER TABLE X_MUSICTRACK_MUSICGENRE ALTER COLUMN MEDIAOBJECTID RENAME TO OBJECTID");

            stmt.execute("ALTER TABLE X_AUDIOCD_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_AUDIOTRACK_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MOVIE_MOVIE_GENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MUSICALBUM_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");
            stmt.execute("ALTER TABLE X_MUSICTRACK_MUSICGENRE ALTER COLUMN SIMPLEOBJECTID RENAME TO REFERENCEDID");

            stmt.close();
    
            converted = true;
            
      } catch (Exception exp) {
          DcLog.add(exp);
          new MessageBox("Conversion was unsuccessful! " + exp.getMessage(), MessageBox._ERROR);
      }
    
      try {
          connection.close();
      } catch (Exception exp) {}
      
      return converted;
    }
}
