/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;

public class DcDatabase {

    private QueryQueue queue;

    public DcDatabase()  {
    }

    public void initiliaze(Connection connection) throws Exception {
        
        new DatabaseConversion().start();
        
        startQueryQueue();
        initialize(connection);
        setDbProperies(connection);

        try {
            connection.close();
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }

    public int getQueueSize() {
    	return queue.getQueueSize();
    }
    
    private void startQueryQueue() {
        queue = new QueryQueue();
        Thread queryQueue = new Thread(queue, "queryQueue");
        queryQueue.setPriority(Thread.NORM_PRIORITY);
        queryQueue.setDaemon(true);
        queryQueue.start();
    }

    public String getName() {
        return DcSettings.getString(DcRepository.settings.stConnectionString);
    }

    public void addQuery(Query query) {
        queue.addQuery(query);
    }

    public void setDbProperies(Connection connection) {
        try {
            
            Statement stmt = connection.createStatement();
            int cacheScale = DcSettings.getInt(DcRepository.settings.stHsqlCacheScale);
            int cacheSizeScale = DcSettings.getInt(DcRepository.settings.stHsqlCacheSizeScale);

            stmt.execute("SET PROPERTY " +  '"' + "hsqldb.cache_scale" + '"' + " " + cacheScale);
            stmt.execute("SET PROPERTY " +  '"' + "hsqldb.cache_size_scale" + '"' + " " + cacheSizeScale);
            stmt.execute("SET LOGSIZE 200");
            
            stmt.close();
        } catch (Exception exp) {
            DcLog.add(DcResources.getText("msgUnableToChangeDbSettings"), DcLog._ERROR);
        }
    }

    private void initialize(Connection connection) throws Exception {
        Statement stmt = connection.createStatement();

        try {
            stmt.execute("SET DATABASE COLLATION " + '"' + "Latin1_General" + '"');
        } catch (Exception ignore) {}

        for (Iterator iter = DcModules.getAllModules();  iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            DcObject dco = module.getDcObject();

            if (!dco.getTableName().equals("")) {
                String testQuery = "select * from " + dco.getTableName();
                try {
                    stmt.setMaxRows(1);
                    ResultSet result = stmt.executeQuery(testQuery);
                    initializeColumns(connection, result.getMetaData(), dco);
                    DcLog.add(DcResources.getText("msgTableFound", dco.getTableName()), DcLog._STANDARD);
                    result.close();
                } catch (SQLException exp) {
                    DcLog.add(DcResources.getText("msgTableNotFound", dco.getTableName()), DcLog._STANDARD);
                    createTable(dco);
                }
            }
        }
        stmt.close();
    }

    private void initializeColumns(Connection connection, ResultSetMetaData metaData, DcObject dco) throws SQLException {
        String tableName = dco.getTableName();
        Collection columns = new ArrayList();
        for (int i = 1; i < metaData.getColumnCount() + 1; i++)
            columns.add(metaData.getColumnName(i).toUpperCase());
        
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            String column = field.getDatabaseFieldName();
            String type = field.getDataBaseFieldType();
            
            if (!field.isUiOnly() && !columns.contains(column.toUpperCase())) {
                DcLog.add(DcResources.getText("msgTableNeedsUpgrade", tableName), DcLog._STANDARD);
                executeQuery(connection, "alter table " + tableName + " add column " + column + " " + type);
            }
        }
    }

    private void executeQuery(Connection connection, String sql) {
        try {
            executeQuery(connection.prepareStatement(sql));
        } catch (Exception e) {
            DcLog.add(e);
        }
    }
    
    private void executeQuery(PreparedStatement ps) {
        try {
            ps.execute();
            DcLog.add(ps);
            ps.close();
        } catch (Exception e) {
            DcLog.add(e);
        }
    }

    private void createTable(DcObject dco) {
        try {
            Query query = new Query(Query._CREATE, dco, null, null);
            executeQuery(query.getQuery());
            
            DcObject[] objects = dco.getModule().getDefaultData();
            if (objects != null) {
                for (int i = 0; i < objects.length; i++) {
                    objects[i].setIDs();
                    objects[i].keepInSync(false);
                    objects[i].setSilence(true);
                    objects[i].saveNew(false);
                }
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
}
