/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.db;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;

import javax.swing.ImageIcon;

import net.sf.dc.core.DataCrow;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.MappingModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Loan;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.objects.helpers.ContactPerson;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.core.wf.requests.ImageRequest;
import net.sf.dc.core.wf.requests.RequestCollection;

/**
 * The Query class creates SQL statements needed to remove, update, insert and 
 * select items from the database. Queries created by this class ensure the integrity 
 * of the data.
 * @author Robert Jan van der Waals
 */
public class Query {

    public static final int _UNDEFINED = 0;
    public static final int _UPDATE = 1;
    public static final int _CREATE = 2;
    public static final int _INSERT = 3;
    public static final int _SELECT = 4;
    public static final int _DELETE = 5;

    protected static final int _SELECTPRECISE = 6;
    protected static final int _SELECTOR = 7;
    protected static final int _SELECTPRECISEOR = 8;
    
    protected final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd");

    protected final int module;
    protected int queryType;

    protected boolean loadRelations = true;
    protected boolean bPreciseSelect = false;
    protected boolean bComplyToAllConditions = true;

    protected ArrayList queries;
    protected String objectID;

    protected RequestCollection requests;
    
    protected String[] ordering;
    protected boolean bSilence = false;
    protected boolean isBatch = false;
    protected boolean endOfBatch = false;    
    
    public Query(int queryType,
                 DcObject dco,
                 QueryOptions options,
                 RequestCollection requests) throws SQLException {

        long start = DataCrow.debug ? new Date().getTime() : 0;
        
        module = dco.getModule().getIndex();

        setSilence(dco.getSilence());

        if (options != null) {
            bPreciseSelect = options.getPreciseSelect();
            bComplyToAllConditions = options.getComplyAllConditions();
            ordering = options.getOrdering();
        }

        this.requests = requests == null ? new RequestCollection() : requests;
        this.queryType = queryType;
        this.objectID = dco.getID();

        if (this.queryType == _SELECT) {
            if (bPreciseSelect && !bComplyToAllConditions) {
                this.queryType = _SELECTPRECISEOR;
            } else if (bPreciseSelect) {
                this.queryType = _SELECTPRECISE;
            } else if (!bComplyToAllConditions) {
                this.queryType = _SELECTOR;
            }
        }

        queries = getQueries(dco);
        
        if (DataCrow.debug) {
            long end = new Date().getTime();
            String msg = "Query was generated in " + (end - start) + "ms";
            DcLog.add(msg, DcLog._STANDARD);
            System.out.println(msg);
        }
    } 
    
    public PreparedStatement getQuery() {
        return (PreparedStatement) getQueries().get(0);
    }
    
    private ArrayList getQueries(DcObject dco) throws SQLException {
        ArrayList queries = new ArrayList();
        switch (queryType) {
            case _UPDATE :
                queries.addAll(getUpdateQueries(dco));
                break;
            case _CREATE :
                queries.addAll(getCreateTableQuery(dco));
                break;
            case _INSERT :
                queries.addAll(getInsertQueries(dco));
                break;
            case _DELETE :
                queries.addAll(getDeleteQueries(dco));
                break;
            case _UNDEFINED :
            case _SELECT :
            case _SELECTPRECISE :
            case _SELECTOR :
            case _SELECTPRECISEOR :
                queries.addAll(getSelectQueries(dco));
                break;
        }
        return queries;
    }
    
    private boolean isEmpty(Object o) {
        String emptyValue = DcResources.getText("lblEmptySearchValue");
        return o == null || o.equals(emptyValue) || o.equals("0") || o.equals("NULL");
    }

    protected ArrayList getInsertQueries(DcObject dco) throws SQLException {
        Collection values = new ArrayList();
        StringBuffer columns = new StringBuffer();

        createReferences(dco);
        
        Collection references = new ArrayList();
        Collection pictures = new ArrayList();
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (field.getValueType() == DcRepository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(field.getIndex());
                byte[] image = picture != null && picture.isNew() ? (byte[]) picture.getValue(Picture._D_IMAGE) : null;
                if (image != null && image.length > 10) {
                    ImageIcon icon = new ImageIcon(image);
                    picture.setValue(Picture._A_OBJECTID, dco.getID());
                    picture.setValue(Picture._B_FIELD, field.getDatabaseFieldName());
                    picture.setValue(Picture._C_FILENAME, dco.getID() + "_" + field.getDatabaseFieldName() + ".jpg");
                    picture.setValue(Picture._E_HEIGHT, new Integer(icon.getIconHeight()));
                    picture.setValue(Picture._F_WIDTH, new Integer(icon.getIconWidth()));
                    pictures.add(picture);
                }
            } else if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                Collection c = (Collection) dco.getValue(field.getIndex());
                if (c != null) references.addAll(c);                
                
            } else if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (columns.length() > 0)
                    columns.append(", ");

                if (!(dco.getValue(field.getIndex()) instanceof Picture)) {
                    values.add(getQueryValue(dco, field.getIndex()));
                    columns.append(field.getDatabaseFieldName());
                }
            }
        }

        ArrayList queries = new ArrayList();
        if (references.size() > 0) {
            for (Iterator iter = references.iterator(); iter.hasNext(); ) {
                DcMapping mapping = (DcMapping) iter.next();
                queries.add(getPreparedStatement("INSERT INTO " + mapping.getTableName() + 
                                    " (" + mapping.getDatabaseFieldName(DcMapping._A_PARENT_ID) + ", " +
                                           mapping.getDatabaseFieldName(DcMapping._B_REFERENCED_ID) + 
                                     ") \r\n VALUES (" + dco.getID() + ", " + mapping.getReferencedId() + ")"));
            }
        }        
        
        String sqlPart = "";
        for (int i = 0; i < values.size(); i++)
            sqlPart += (sqlPart.length() > 0 ? ", ?" : "?"); 
        
        String sql = "INSERT INTO " + dco.getTableName() + " (" + columns + ") \r\n" + "VALUES (" + sqlPart + ");";
        PreparedStatement ps = getPreparedStatement(sql);
        setValues(ps, values);
        queries.add(ps);
            
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            queries.addAll(getInsertQueries(picture));
            requests.addRequest(
                    new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                     (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));
        }
        
        pictures.clear();
        return queries;
    }

    protected ArrayList getCreateTableQuery(DcObject dco) throws SQLException {
        String columns = "";
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (!field.isUiOnly()) {
                if (columns.length() > 0)
                    columns += ", ";

                columns += field.getDatabaseFieldName() + " " + field.getDataBaseFieldType();
                if (field.getIndex() == DcObject._ID) {
                    columns += " PRIMARY KEY";
                }
            }
        }
        
        String sql = "CREATE MEMORY TABLE " + dco.getTableName() + "\r\n(" + columns + ");";
        ArrayList queries = new ArrayList();
        queries.add(getPreparedStatement(sql));
        return queries;
    }

    protected ArrayList getUpdateQueries(DcObject dco) throws SQLException {
        Collection pictures = new ArrayList();
        Collection references = new ArrayList();
        
        createReferences(dco);
        
        StringBuffer sbValues = new StringBuffer();
        Collection values = new ArrayList();
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (field.getValueType() == DcRepository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(field.getIndex());
                if (picture != null && (picture.isNew() || picture.isUpdated() || picture.isDeleted())) {
                    picture.setValue(Picture._A_OBJECTID, dco.getID());
                    picture.setValue(Picture._B_FIELD, field.getDatabaseFieldName());
                    picture.setValue(Picture._C_FILENAME, dco.getID() + "_" + field.getDatabaseFieldName() + ".jpg");
                    
                    if (picture.getValue(Picture._D_IMAGE) != null) {
                        byte[] image = (byte[]) picture.getValue(Picture._D_IMAGE);
                        ImageIcon icon = new ImageIcon(image);
                        picture.setValue(Picture._E_HEIGHT, new Integer(icon.getIconHeight()));
                        picture.setValue(Picture._F_WIDTH, new Integer(icon.getIconWidth()));
                    }
                    
                    pictures.add(picture);
                }
            } else if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                Collection c = (Collection) dco.getValue(field.getIndex());
                if (c != null) references.addAll(c);
                
            } else if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (sbValues.length() > 0)
                    sbValues.append(", ");

                sbValues.append(field.getDatabaseFieldName());
                sbValues.append(" = ?");
                values.add(getQueryValue(dco, field.getIndex()));
            }
        }

        ArrayList queries = new ArrayList();
        if (!(dco instanceof Picture)) {
            String sql = "UPDATE " + dco.getTableName() + " SET " + sbValues + "\r\n WHERE ID = " + dco.getID();
            PreparedStatement ps = getPreparedStatement(sql);
            setValues(ps, values);
            queries.add(ps);
        } else {
            String sql = "UPDATE " + dco.getTableName() + " SET " + sbValues + "\r\n WHERE " +
                         dco.getDatabaseFieldName(Picture._A_OBJECTID) + " = " + dco.getValue(Picture._A_OBJECTID) + " AND " +
                         dco.getDatabaseFieldName(Picture._B_FIELD) + " = ?";
            
            values.add(dco.getValue(Picture._B_FIELD));
            PreparedStatement ps = getPreparedStatement(sql);
            setValues(ps, values);
            queries.add(ps);
        }

        if (references.size() > 0) {
            boolean first = true;
            for (Iterator iter = references.iterator(); iter.hasNext(); ) {
                DcMapping mapping = (DcMapping) iter.next();
                if (first) {
                    String sql = "DELETE FROM " + mapping.getTableName() + " WHERE " +  
                                 mapping.getDatabaseFieldName(DcMapping._A_PARENT_ID) + " = " + dco.getID();
                    queries.add(getPreparedStatement(sql));  
                }

                String sql = "INSERT INTO " + mapping.getTableName() + 
                             " (" + mapping.getDatabaseFieldName(DcMapping._A_PARENT_ID) + ", " +
                             mapping.getDatabaseFieldName(DcMapping._B_REFERENCED_ID) + 
                             ") \r\n VALUES (" + dco.getID() + ", " + mapping.getReferencedId() + ");";
                queries.add(getPreparedStatement(sql)); 
                first = false;
            }
        }
        
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            
            Collection currentPictures = DataManager.getPictures(dco.getID());
            boolean isReallyNew = picture.isNew() && (currentPictures == null || !currentPictures.contains(picture));
            if (!isReallyNew && picture.isNew())
                picture.isUpdated(!isReallyNew);
            
            if (picture.isUpdated()) {
                queries.addAll(getUpdateQueries(picture));
                requests.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));                
            } else if (picture.isDeleted()) {
                String sql = "DELETE FROM " + picture.getTableName() + " WHERE " +
                             picture.getField(Picture._A_OBJECTID).getDatabaseFieldName() + " = " + dco.getID() + " AND " +
                             picture.getField(Picture._B_FIELD).getDatabaseFieldName() + " = ?";
                
                PreparedStatement ps = getPreparedStatement(sql);
                ps.setString(1, (String) picture.getValue(Picture._B_FIELD));
                
                queries.add(ps);
                requests.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._DELETE));                
            } else if (picture.isNew()) {
                queries.addAll(getInsertQueries(picture));
                requests.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));
            }
        }
        
        pictures.clear();
        return queries;
    }
    
    private void createReferences(DcObject dco) {
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            Object value = dco.getValue(field.getIndex());
            if (field.getValueType() == DcRepository.datatypes._DCOBJECTREFERENCE) {
                
                DcObject obj = value instanceof DcObject ? (DcObject) value : null;
                
                if (obj == null)
                    continue;
                
                try { 
                    DcObject existing = DataManager.getObject(obj.getModule().getIndex(), obj.getID());
                    existing = existing == null ? 
                                   DataManager.getObjectForDisplayValue(obj.getModule().getIndex(), obj.toString()) : 
                                   existing;

                    if (existing == null)
                        obj.saveNew(false);
                    else
                        dco.setValue(field.getIndex(), existing);

                } catch (Exception e) {
                    DcLog.add(e);
                }
                
            } else if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                if (value == null)
                    continue;
                
                for (Iterator iterObj = ((Collection) value).iterator(); iterObj.hasNext(); ) {
                    DcMapping mapping = (DcMapping) iterObj.next();
                    DcObject obj = mapping.getReferencedObject();
                    
                    try { 
                        DcObject existing = DataManager.getObject(obj.getModule().getIndex(), obj.getID());
                        existing = existing == null ? 
                                       DataManager.getObjectForDisplayValue(obj.getModule().getIndex(), obj.toString()) : 
                                       existing;
                        if (existing == null) {
                            obj.saveNew(false);
                        } else {
                            mapping.setReferencedObject(existing);
                            mapping.setValue(DcMapping._B_REFERENCED_ID, existing.getID());
                        }
                    } catch (Exception e) {
                        DcLog.add(e);
                    }
                }
            }
        }
    }
    
    private PreparedStatement getPreparedStatement(String sql) throws SQLException {
        return DatabaseManager.getConnection().prepareStatement(sql);
    }

    private void setValues(PreparedStatement ps, Collection values) throws SQLException {
        int pos = 1;
        for (Iterator iter = values.iterator(); iter.hasNext(); pos++) {
            Object value = iter.next();
            if (value instanceof String)
                ps.setString(pos, (String) value);
            else if (value instanceof Long)
                ps.setLong(pos, ((Long) value).longValue());
            else if (value instanceof Integer)
                ps.setInt(pos, ((Integer) value).intValue());
            else if (value instanceof Boolean)
                ps.setBoolean(pos, ((Boolean) value).booleanValue());
            else if (value instanceof Date)
                ps.setDate(pos, new java.sql.Date(((Date) value).getTime()));
            else
                ps.setNull(pos, Types.NULL);
        }
    }
    
    protected ArrayList getDeleteQueries(DcObject dco) throws SQLException {
        ArrayList queries = new ArrayList();

        Loan loan = new Loan();
        Picture picture = new Picture();
        if (dco.hasPrimaryKey())
            queries.add(getPreparedStatement("DELETE FROM " + dco.getTableName() + " WHERE ID = " + dco.getID()));

        if (dco.getModule().canBeLended())
            queries.add(getPreparedStatement("DELETE FROM " + loan.getTableName() + " WHERE " +
                            loan.getField(Loan._D_OBJECTID).getDatabaseFieldName() + " = " + dco.getID()));

        if (dco.getModule().getChild() != null) {
            DcModule childModule = dco.getModule().getChild(); 
            queries.add(getPreparedStatement("DELETE FROM " + childModule.getTableName() + " WHERE " + 
                        childModule.getField(childModule.getDcObject().getParentReferenceFieldIndex()).getDatabaseFieldName() + " = " + dco.getID()));
        }
        
        // remove any references to the to be deleted item.
        if (dco.getModule().hasDependingModules()) {
            Collection modules = DcModules.getReferencingModules(dco.getModule().getIndex());
            for (Iterator iter = modules.iterator(); iter.hasNext(); ) {
                DcModule m = (DcModule) iter.next();
                
                if (m instanceof MappingModule) {
                    DcObject mapping = m.getDcObject();
                    queries.add(getPreparedStatement(
                            "DELETE FROM " + m.getTableName() + " WHERE " + mapping.getField(DcMapping._B_REFERENCED_ID).getDatabaseFieldName() + " = " + dco.getID()));
                } else {
                    for (Iterator iterFlds = m.getFields().iterator(); iterFlds.hasNext(); ) {
                        DcField field = (DcField) iterFlds.next();
                        if (!field.isUiOnly() && field.getReferenceIdx() == dco.getModule().getIndex()) {
                            queries.add(getPreparedStatement(
                                    "UPDATE " + m.getTableName() + " SET " +  field.getDatabaseFieldName() + " = NULL WHERE " + 
                                                field.getDatabaseFieldName() + " = " + dco.getID()));
                        }
                    }
                }
            }
        }
        
        requests.addRequest(new ImageRequest(dco.getID(), ImageRequest._DELETE));
        
        queries.add(getPreparedStatement("DELETE FROM " + picture.getTableName() + " WHERE " +
                    picture.getField(Picture._A_OBJECTID).getDatabaseFieldName() + " = " + dco.getID()));

        return queries;
    }

    private Object getQueryValue(DcObject object, int index) {
        Object value = object.getValue(index);

        if (value != null && value.toString().trim().length() == 0)
            value = null;
        else if (value instanceof DcObject)
            value = ((DcObject) value).getID();

        return value;
    }    

    protected ArrayList getSelectQueries(DcObject dco) throws SQLException {
        ArrayList queries = new ArrayList();
        Collection objects = new ArrayList();

        Collection tables = new ArrayList();
        if (dco.getModule().getIndex() == DcModules._MEDIA) {
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                
                if (    module instanceof DcMediaModule && module.getIndex() != DcModules._MEDIA &&
                       !module.isChildModule()) {
                    
                    tables.add(module.getTableName());
                }
            }
        } else {
            objects.add(dco);
            objects.addAll(dco.getChildren());
            String table = dco.getTableJoin() != null && dco.getTableJoin().trim().length() > 0 ? dco.getTableJoin() : dco.getTableName();
            tables.add(table);
        }

        int counterTables = 0;
        for (Iterator iterTables = tables.iterator(); iterTables.hasNext(); counterTables++) {
            StringBuffer columns = new StringBuffer();

            String query = "";
            String table = (String) iterTables.next();

            if (dco.getModule().getIndex() == DcModules._MEDIA) {
                objects.clear();
                DcObject o = DcModules.getObjectForTable(table);
                o.copy(dco);
                objects.add(o);
            }

            int counter = 0;
            StringBuffer conditions = new StringBuffer();
            Collection values = new ArrayList();

            for (Iterator iter = objects.iterator(); iter.hasNext(); counter++) {
                DcObject object = (DcObject) iter.next();
                Collection fields = object.getFields();

                for (Iterator iterFields = fields.iterator(); iterFields.hasNext(); ) {
                    DcField field = (DcField) iterFields.next();

                    if (!field.isUiOnly() && field.getValueType() != DcRepository.datatypes._PICTURE)  {
                        String column = field.getDatabaseFieldName();
                        if (objects.size() > 1)
                            column = object.getTableShortName() + "." + column;

                        if (counter == 0) {
                            if (columns.length() > 0)
                                columns.append(", ");

                            // only select fields for the current module, child fields are added to the condition only
                            columns.append(column);
                        }

                        if (object.isChanged(field.getIndex())) {
                            Object value = getQueryValue(object, field.getIndex());

                            String appender = queryType == _SELECTOR || queryType == _SELECTPRECISEOR ? " OR " : " AND ";
                            if (conditions.length() > 0)
                                conditions.append(appender);
                            
                            if (isEmpty(value)) {
                                conditions.append(getEmptyCondition(column, field));
                            } else {
                                String condition;
                                
                                if (queryType == _SELECT || queryType == _SELECTOR) {
                                    if (value instanceof String)
                                        condition = "UPPER (" + column + ") LIKE '%" + value.toString().substring(1, value.toString().length() -1).toUpperCase() + "%'";
                                    else 
                                        condition = column + " = ?";
                                } else {
                                    if (value instanceof String) 
                                        condition = "UPPER (" + column + ") = " + value.toString().toUpperCase();
                                    else
                                        condition = column + " = " + value;
                                }
                                values.add(value);
                                conditions.append(condition);
                            }
                        }
                    }
                }
            }

            query += "SELECT DISTINCT " + columns + " FROM " + table;

            DcObject objectClone = DcModules.getObjectForTable(table);
            if (objectClone  != null)
                objectClone.copy(dco);

            addAvailabilityCondition(objectClone == null ? dco : objectClone, conditions);

            if (conditions.length() > 0)
                query +=  "\r\n WHERE " + conditions;

            if (ordering != null) {
                for (int i = 0; i < ordering.length; i++) {
                    if (i == 0)
                        query += " ORDER BY ";
                    else
                        query += ", ";

                    query += ordering[i];
                }
            } else if (dco.hasPrimaryKey()) {
                query += " ORDER BY " + dco.getField(DcObject._ID).getDatabaseFieldName();
            }

            PreparedStatement ps = getPreparedStatement(query);
            setValues(ps, values);
            queries.add(ps);
        }

        return queries;
    }
    
    public ArrayList getQueries() {
        return queries;
    }
    
    protected boolean isBoolean(DcField field) {
        return  field.getValueType() == DcRepository.datatypes._BOOLEAN;
    }
    
    protected boolean isInteger(DcField field) {
        return  field.getValueType() == DcRepository.datatypes._LONG ||
                field.getValueType() == DcRepository.datatypes._BIGINTEGER || 
                field.getValueType() == DcRepository.datatypes._DCOBJECTREFERENCE ||
                field.getValueType() == DcRepository.datatypes._DCPARENTREFERENCE;
    }
    
    public void unload() {
        if (requests != null)
            requests.clear();
        
        requests = null;
        queries = null;
        objectID = null;
        ordering = null;
    }
    
    public int getModule() {
        return module;
    }

    public String[] getOrdering() {
        return ordering;
    }

    public void setLoadRelations(boolean b) {
        loadRelations = b;
    }

    public boolean getLoadRelations() {
        return loadRelations;
    }

    public void setBatch(boolean endOfBatch) {
        this.isBatch = true;
        this.endOfBatch = endOfBatch;
    }

    public boolean isBatch() {
        return isBatch;
    }

    public boolean isEndOfBatch() {
        return isBatch() && endOfBatch;
    }

    public String getObjectID() {
        return objectID;
    }

    public int getType() {
        return queryType;
    }

    public boolean getSilence() {
        return bSilence;
    }

    public void setSilence(boolean silence) {
        this.bSilence = silence;
    }

    public RequestCollection getRequestors() {
        return requests;
    }
    
    public void clear() {
        if (queries != null)
            queries.clear();
        
        queries = null;
        objectID = null;
        requests = null;
        ordering = null;
    }    
    
    protected String getEmptyCondition(String column, DcField field) {
        String emptyCondition;
        if (isInteger(field)) 
            emptyCondition = "(" + column + " = 0 OR " + column + " IS NULL) ";
        else
            emptyCondition = "(" + column + " IS NULL OR " + column + " = '') ";

        return emptyCondition;
    } 
    
    protected void addAvailabilityCondition(DcObject dco, StringBuffer conditions) {
        if (dco.getModule().canBeLended()) {
            ContactPerson loanedBy = (ContactPerson) dco.getValue(DcObject._SYS_LOANEDBY);
            Integer duration = (Integer) dco.getValue(DcObject._SYS_LOANDURATION);
            String s = (String) dco.getValue(DcObject._SYS_AVAILABLE);

            if (s != null || loanedBy != null || duration != null) {
                boolean available = s == null || duration != null ? false : new Boolean(s).booleanValue();
                loanedBy = available ? null : loanedBy;

                if (conditions.length() > 0)
                    conditions.append(" AND");

                String column = dco.getChildren().size() > 0 ? " " + dco.getTableShortName() + ".ID" : " ID";
                String tablename = dco.getChildren().size() > 0 ? dco.getTableShortName() : dco.getTableName();

                String current = formatter.format(new Date());
                String daysCondition = duration != null ? " AND DATEDIFF('dd', startDate , '" + current + "') >= " + duration.intValue() : "";
                String personCondition = loanedBy != null ? " AND PersonID = " + loanedBy.getID() : "";

                if (available)
                    conditions.append(column + " NOT in (select objectID from Loans where objectID = " +
                                      tablename + ".ID AND enddate IS NULL AND startDate <= '" + current +  "')");
                else
                    conditions.append(column + " in (select objectID from Loans where objectID = " +
                                      tablename + ".ID " + daysCondition + " AND enddate IS NULL AND startDate <= '" + current +  "'" +
                                      personCondition + ")");
            }
        }
    }   
    
    public String toString() {
        String sql = "";
        for (Iterator iter = queries.iterator(); iter.hasNext(); ) {
            sql += (sql.length() > 0 ? "\r\n" : "");
            sql += iter.next().toString();
        }
        return sql;
    }
}
