/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.db;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.core.wf.WorkFlow;
import net.sf.dc.core.wf.requests.IRequest;
import net.sf.dc.core.wf.requests.IUpdateUIRequest;
import net.sf.dc.core.wf.requests.ImageRequest;
import net.sf.dc.core.wf.requests.RequestCollection;

public class QueryQueue extends Thread {

    private final LinkedList lQueryQueue = new LinkedList();

    private boolean isLazy = true;
    private boolean success = true;
    
    public QueryQueue() {}

    public int getQueueSize() {
    	return lQueryQueue.size();
    }
    
    public void addQuery(Query query) {
        lQueryQueue.addLast(query);
    }

    public void run() {
        synchronized(this) {
            while (true) {
                try {
                    if (lQueryQueue.size() > 0 && isLazy) {
                        success = true;
                        isLazy = false;

                        Query query = (Query) lQueryQueue.removeFirst();
                        Collection result = executeQuery(query);
                    	handleRequests(query, result, success);
                        query.unload();
                    } else {
                        isLazy = true;
                        sleep(100);
                    }
                } catch (Exception exp) {
                    DcLog.add(exp);
                }
            }
        }
    }

    private Collection executeQuery(Query query) {
        if (DataCrow.debug)
            DcLog.add(query.getQuery());

        switch(query.getType()) {
            case Query._UPDATE:
                success = executeUpdateQuery(query);
                break;
            case Query._DELETE:
                success = executeDeleteQuery(query);
                break;
            case Query._INSERT:
                success = executeInsertQuery(query);
                break;
            case Query._UNDEFINED:
                if (!query.getSilence()) {
                    DcLog.add(query.getQuery());
                }
                return executeUndefinedQuery(query);
            default:
                if (!query.getSilence())
                    DcLog.add(query.getQuery());

            return executeSelectQuery(query);
        }

        return null;
    }

    private boolean executeUpdateQuery(Query query) {
        Collection queries = query.getQueries();

        boolean success = true;
        boolean messageShown = false;
        for (Iterator iter = queries.iterator(); iter.hasNext(); ) {
            PreparedStatement ps = (PreparedStatement) iter.next();
            try {
                ps.execute();
                if ((query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                    (!query.getSilence() && !messageShown)) {
                    messageShown = true;
                    new MessageBox(DcResources.getText("msgDataSaved"), MessageBox._INFORMATION, false);
                }
            } catch (SQLException sqlExp) {
                success = false;
                DcLog.add(ps);
                DcLog.add(sqlExp);
                new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
            } finally {
                close(ps);
            }
        }
        return success;
    }

    private boolean executeInsertQuery(Query query) {
        Collection queries = query.getQueries();

        boolean success = true;
        boolean messageShown = false;
        for (Iterator iter = queries.iterator(); iter.hasNext(); ) {
            PreparedStatement ps = (PreparedStatement) iter.next();
            
            try {
                ps.execute();
                if ((query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                        (!query.getSilence() && !messageShown)) {
                    messageShown = true;
                    new MessageBox(DcResources.getText("msgDataSaved"), MessageBox._INFORMATION, false);
                }
            } catch (SQLException sqlExp) {
                success = false;
                DcLog.add(ps);
                DcLog.add(sqlExp);
            } finally {
                close(ps);
            }
        }
        return success;
    }

    private boolean executeDeleteQuery(Query query) {
        Collection queries = query.getQueries();

        boolean messageShown = false;
        boolean success = true;
        for (Iterator iter = queries.iterator(); iter.hasNext(); ) {
            PreparedStatement ps = (PreparedStatement) iter.next();
            
            try {
                ps.execute();
                if (    (query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                        (!query.getSilence() && !messageShown)) {
                    messageShown = true;
                    new MessageBox(DcResources.getText("msgDeleteSuccessfull"), MessageBox._INFORMATION, false);
                }
            } catch (SQLException sqlExp) {
                success = false;
                DcLog.add(ps);
                DcLog.add(sqlExp);
                new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
            } finally {
                close(ps);
            }
        }
        
        return success;
    }

    private Collection executeSelectQuery(Query query) {
        Collection objects = new ArrayList();
        for (Iterator iter = query.getQueries().iterator(); iter.hasNext(); ) {
            PreparedStatement ps = (PreparedStatement) iter.next();
            
            ResultSet result = null;
            try {
                result = ps.executeQuery();
            } catch (SQLException sqlExp) {
                DcLog.add(ps);
                DcLog.add(sqlExp);
                if (sqlExp.getMessage().toUpperCase().indexOf("MEMORY") > -1) {
                    new MessageBox(DcResources.getText("msqMemoryError"), MessageBox._ERROR, false);
                    System.gc();
                } else {
                    new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
                }
            }

            objects.addAll(new WorkFlow().convertToDCObjects(result));

            if (query.getLoadRelations()) {
                for (Iterator iterObj = objects.iterator(); iterObj.hasNext(); )
                    ((DcObject) iterObj.next()).loadChildren();
            }

            try {
                if (result != null)
                    result.close();
                
                close(ps);
            } catch (SQLException exp) {
                DcLog.add(exp);
            }
        }

        return objects;
    }
    
    private Collection executeUndefinedQuery(Query query) {
        PreparedStatement ps = query.getQuery();
        ResultSet result = null;
        try {
            result = ps.executeQuery();
        } catch (SQLException sqlExp) {
            DcLog.add(ps);
            DcLog.add(sqlExp);

            if (sqlExp.getMessage().toUpperCase().indexOf("MEMORY") > -1) {
                new MessageBox(DcResources.getText("msqMemoryError"),
                               MessageBox._ERROR, false);
                System.gc();
            } else {
                new MessageBox(DcResources.getText("msgQueryError", "" + sqlExp),
                               MessageBox._ERROR, false);
            }
        }

        Collection objects = new WorkFlow().convertToDCObjects(result);
        try {
            
            if (result != null)
                result.close();
            
            close(ps);
            
        } catch (SQLException exp) {
            DcLog.add(exp);
        }
        
        return objects;
    }

    private void handleRequests(Query qry,
                                Collection objects,
                                boolean qryWasSuccess) {
        
        RequestCollection requestors = qry.getRequestors();
        if (requestors != null)  {
            RequestCollection uiRequests = new RequestCollection();
            RequestCollection imageRequests = new RequestCollection();
            
            IRequest[] requests = requestors.getRequests();
            for (int i = 0; i < requests.length; i++) {
                requestors.removeRequestor(requests[i]);
                
                if (requests[i] instanceof ImageRequest)
                    imageRequests.addRequest(requests[i]);
                else if (requests[i] instanceof IUpdateUIRequest)
                    uiRequests.addRequest(requests[i]);
                else
                    requests[i].execute(objects);
            }
            
            boolean hasImageRequest = imageRequests.size() > 0;
            IRequest[] imgRequests = imageRequests.getRequests();
            for (int i = 0; i < imgRequests.length; i++) {
                imageRequests.removeRequestor(imgRequests[i]);
                imgRequests[i].execute(objects);
            }
            
            if (hasImageRequest && qry.getObjectID() != null) {
                MasterView[] views = DcModules.get(qry.getModule()).getViews();
                for (int i = 0; i < views.length; i++)
                    views[i].reload(qry.getObjectID());
            }
            
            WorkFlow.handleRequests(objects, uiRequests, qryWasSuccess);
        }
    }

    private void close(PreparedStatement ps) {
        try {
            ps.getConnection().commit();
            ps.getConnection().close();
            ps.close();
        } catch (SQLException exp) {
            DcLog.add(exp);
        }
    }
}
