/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.log;

import java.awt.Font;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.sql.PreparedStatement;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;

import javax.swing.SwingUtilities;

import net.sf.dc.console.windows.LogForm;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.resources.DcResources;

/**
 * A very simple implementation of a log.
 * @author rj.vanderwaals
 */
public class DcLog extends Thread {

    public static final int _STANDARD = 0;
    public static final int _SIMPLE = 1;
    public static final int _ERROR = 3;
    public static final int _QUERY = 4;

    private LogForm logForm;
    private RandomAccessFile raf;
    private LinkedList messages = new LinkedList();
    
    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy/MM/dd hh:mm:ss");
    private static DcLog log;
    
    private static boolean show = false;
    private static boolean toFile = true;
    private static boolean toScreen = true;
    
    private boolean cancelled = false;
    
    private DcLog() {
        logForm = new LogForm();
        initializeFile();
    }
    
    private static void initialize() {
        if (log == null || !log.isAlive()) {
            
            if (log != null) log.closeLog();
            
            log = new DcLog();
            log.setPriority(Thread.NORM_PRIORITY);
            log.start();
        }
    }
    
    public static void close() {
        log.closeLog();
    }
    
    public static void setFont(Font font) {
        log.getForm().setFont(font);
    }

    public static void reinstallUI() {
        SwingUtilities.updateComponentTreeUI(log.getForm().getContentPane());
    }    
    
    public static void show(boolean b) {
        show = b;
    }

    public static void toFile(boolean b) {
        toFile = b;
    }
    
    public static void add(Throwable throwable) {
        StringBuffer sb = new StringBuffer();
        sb.append(throwable.getMessage());
        StackTraceElement[] elements = throwable.getStackTrace();
        for (int i = 0; i < elements.length; i++) {
            sb.append(DcRepository.operators.opTab);
            sb.append(elements[i].toString());
            sb.append("\r\n");
        }
        add(sb.toString(), _ERROR);
    }    
    
    public static void add(PreparedStatement ps) {
        String sql = ps.toString();
        
        int idx = sql.toLowerCase().indexOf("sql=[");
        if (idx > -1)
            sql = sql.substring(idx + 5, (idx + 5 < sql.length() - 2 ? sql.length() - 2 : sql.length()));
        
        add(sql, _QUERY);
    }      
    
    public static void add(String message, int type) {
        
        initialize();
        
        StringBuffer sb = new StringBuffer();
        switch (type) {
            case _STANDARD :
                sb.append(DcRepository.operators.opTab);
                sb.append(sdf.format(new Date()));
                sb.append(DcRepository.operators.opSeparator);
                sb.append(message);
                break;
            case _SIMPLE :
                sb.append(DcRepository.operators.opTab);
                sb.append(message);
            case _QUERY :
                sb.append("\r\n");
                sb.append(DcRepository.operators.opTab);
                sb.append(message);
                sb.append("\r\n");
                break;
            case _ERROR :
                sb.append(DcRepository.operators.opTab);
                sb.append(DcRepository.operators.opTab);
                sb.append(DcRepository.operators.opTab);
                sb.append(DcResources.getText("msgError"));
                sb.append(message);
                break;
        }
        
        log.add(sb.toString());
    }
    
    protected LogForm getForm() {
        return logForm;
    }    
    
    public void run() {
        while (!cancelled) {
            if (show) {
                logForm.setVisible(true);
                logForm.toFront();
                show = false;
            }
            
            if (messages.size() > 0) {
                try {
                    String message = (String) messages.removeFirst();
                    if (message != null) {
                        write("\r\n");
                        write(message);
                    }
                } catch (Exception e) {}
            } else {
                try {
                    sleep(500);
                } catch (InterruptedException ignore) {}
            }
        }
    }

    protected void add(String message) {
        messages.add(message);
    }    
    
    protected synchronized void write(String message) {
        try {
            if (toFile) {
                raf.seek(raf.length());
                raf.writeBytes(message);
            }
            
            if (toScreen)
                logForm.writeToScreen(message);
            
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    protected void closeLog() {
        cancelled = true;

        try {
            raf.close();
        } catch (Exception e) {}
    }
    
    private void initializeFile() {
        File file = new File(DataCrow.baseDir + "data/data_crow.log");
        
        try {
            
            long fileLenght = file.length();
            if (fileLenght > 1014220) {
                String add = new SimpleDateFormat("yyyyMMdd").format(new Date());
                
                try {
                    file.renameTo(new File(DataCrow.baseDir + "data/data_crow" + add + ".log"));
                } catch (Exception e) {
                    file.delete();
                }
                
                file = new File(DataCrow.baseDir + "data/data_crow.log");
                raf = new RandomAccessFile(file, "rw");
                DcLog.add(DcResources.getText("msgNewLogUsed"), DcLog._STANDARD);
            } else {
                raf = new RandomAccessFile(file, "rw");
            }
            
        } catch (IOException exp) {
            DcLog.add(exp);
        }
    }      
}