/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.modules;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.ImageIcon;
import javax.swing.KeyStroke;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.panels.ChartPanel;
import net.sf.dc.console.components.panels.QuickViewPanel;
import net.sf.dc.console.menu.DcMenuBar;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.console.views.View;
import net.sf.dc.console.views.list.ListInsertView;
import net.sf.dc.console.views.list.ListSearchView;
import net.sf.dc.console.views.table.TableInsertView;
import net.sf.dc.console.views.table.TableSearchView;
import net.sf.dc.console.windows.filtering.FilterDialog;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.db.DatabaseManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.xml.XmlField;
import net.sf.dc.core.modules.xml.XmlModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.enhancers.IValueEnhancer;
import net.sf.dc.fileimporters.FileImporter;
import net.sf.dc.onlinesearch.OnlineSearch;
import net.sf.dc.settings.DcModuleSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.settings.definitions.QuickViewFieldDefinition;
import net.sf.dc.settings.definitions.QuickViewFieldDefinitions;
import net.sf.dc.synchronizers.Synchronizer;
import net.sf.dc.util.CSVReader;
import net.sf.dc.util.Utilities;

public class DcModule {

    private final int index;
    private int indexDiff = 0;
    
    private int displayIndex;
    private int defaultSortFieldIdx;
    private int nameFieldIdx;
    
    private String label;
    
    private final String name;
    private final String tableName;
    private final String tableShortName;
    private final String tableJoin;
    private final String description;
    
    private final String objectName;
    private final String objectNamePlural;
    
    private OnlineSearch onlineSearch;
    
    private net.sf.dc.settings.Settings settings;
    
    private Class synchronizerClass;
    private Class objectClass;
    private Class importerClass;
    private Class onlineSearchClass;
    
    private int childIdx = -1;
    private int parentIdx = -1;
    
    private boolean hasSearchView = true;
    private boolean hasInsertView  = true;
    
    private ImageIcon icon16;
    private ImageIcon icon32;
    
    private String icon16filename;
    private String icon32filename;
    
    protected MasterView insertView;
    protected MasterView searchView;
    private FilterDialog filterForm;
    private ChartPanel chartPanel;
    
    private Map fields = new HashMap();
    private Map systemFields = new HashMap();
    private Collection sortedFields;

    private KeyStroke keyStroke;
    
    private boolean canBeLended = false;
    private boolean topModule = false;
    private boolean hasDependingModules = true;
    
    private XmlModule xmlModule;
    
    protected DcModule(int index, 
                       String name,
                       String description,
                       String objectName,
                       String objectNamePlural,
                       String tableName, 
                       String tableShortName, 
                       String tableJoin,
                       boolean topModule) { 

        this.index = index;
        this.tableName = (tableName == null ? "" : tableName).toLowerCase();
        this.tableShortName  = (tableShortName == null ? "" : tableShortName).toLowerCase();
        this.tableJoin = tableJoin;
        this.name = name;
        this.label = name;
        this.description = description;
        
        this.topModule = topModule;
        this.objectName = objectName;
        this.objectNamePlural = objectNamePlural;
    }
    
    public DcModule(int index, 
                    boolean topModule, 
                    String name,
                    String description,
                    String objectName,
                    String objectNamePlural,
                    String tableName, 
                    String tableShortName, 
                    String tableJoin) {

        this(index, name, description, objectName, objectNamePlural, tableName, 
             tableShortName, tableJoin, topModule);
        
        initializeSystemFields();
        initializeFields();
        initializeSettings();
    }
    
    public DcModule(XmlModule module) {
        this(module.getIndex(), module.getName(), module.getDescription(), module.getObjectName(), 
             module.getObjectNamePlural(), module.getTableName(), module.getTableNameShort(),
             module.getTableName(), true);

        this.xmlModule = module;
        
        childIdx = module.getChildIndex();
        parentIdx = module.getParentIndex();
        label = module.getLabel();
        
        hasInsertView = module.hasInsertView();
        hasSearchView = module.hasSearchView();
        keyStroke = module.getKeyStroke();
        objectClass = module.getObject();
        synchronizerClass = module.getSynchronizer();
        importerClass = module.getImporter();
        onlineSearchClass = module.getOnlineSearch();
        hasDependingModules = module.hasDependingModules();
        
        icon16filename = module.getIcon16Filename();
        icon32filename = module.getIcon32Filename();
        icon16 = new ImageIcon(module.getIcon16());
        icon32 = new ImageIcon(module.getIcon32());
        
        nameFieldIdx = module.getNameFieldIdx();
        canBeLended = module.canBeLended();
        
        displayIndex = module.getDisplayIndex();
        defaultSortFieldIdx = module.getDefaultSortFieldIdx();
        
        initializeSystemFields();
        initializeFields();
        
        if (module.getChildIndex() > -1)
            setChild(module.getChildIndex());
        
        for (Iterator iter = module.getFields().iterator(); iter.hasNext(); )
            addField(new DcField((XmlField) iter.next(), getIndex()));
        
        initializeSettings();
        canBeLended(module.canBeLended());
        
        // Set it to disabled only if the XML module is defined as disabled.
        // There is no use case for this (yet).
        if (!module.isEnabled()) isEnabled(false);
    }
    
    public boolean isChildModule() {
        return false;
    }
    
    public boolean isParentModule() {
        return false;
    }    
    
    public void setIndexDiff(int diff) {
        this.indexDiff = diff;
    }

    public int getIndexDiff() {
        return indexDiff;
    }    
    
    public int getIndex() {
        return index + indexDiff;
    }    
    
    public ImageIcon getIcon16() {
        return icon16;
    }
    
    public ImageIcon getIcon32() {
        return icon32;
    }

    public String getObjectName() {
        return objectName;
    }
    
    public String getObjectNamePlural() {
        return objectNamePlural;
    }
    
    public DcObject getDcObject() {
        try {
            try {
                return (DcObject) objectClass.getConstructors()[0].newInstance(new Object[] {});    
            } catch (Exception exp) {
                return (DcObject) objectClass.getConstructors()[0].newInstance(new Object[] {new Integer(getIndex())});
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }

        return null;
    }    
    
    public KeyStroke getKeyStroke() {
        return keyStroke;
    }    
    
    public boolean isEnabled() {
        return settings.getBoolean(DcRepository.modulesettings.stEnabled);
    }

    public void isEnabled(boolean b) {
        settings.set(DcRepository.modulesettings.stEnabled, new Boolean(b));
    }    
    
    public String getName() {
        return name;
    }

    public  String getDescription() {
        return description;
    }    

    public String getTableName() {
        return tableName;
    }

    public String getTableShortName() {
        return tableShortName;
    }

    public String getTableJoin() {
        return tableJoin;
    }

    public boolean hasOnlineServices() {
        return onlineSearchClass != null;
    }
    
    public Synchronizer getSynchronizer() {
        if (synchronizerClass != null) {
            try {
                return (Synchronizer) synchronizerClass.getConstructors()[0].newInstance(new Object[] {});    
            } catch (Exception exp) {
                DcLog.add(exp);
            }
        }
        return null;
    }
    
    public MasterView getInsertView() {
        initializeUI();
        return insertView;
    }

    public MasterView getSearchView() {
        initializeUI();
        return searchView;
    }

    public View getCurrentSearchView() {
        initializeUI();
        return getSearchView() != null ? getSearchView().getCurrent() : null;
    }

    public View getCurrentInsertView() {
        initializeUI();
        return getInsertView() != null ? getInsertView().getCurrent() : null;
    }    
    
    public DcPropertyModule getPropertyModule(int modIdx) {
        return (DcPropertyModule) DcModules.get(getIndex() + modIdx);
    }

    public TemplateModule getTemplateModule() {
        if ((isTopModule() || isChildModule()) && getIndex() != DcModules._MEDIA)
            return (TemplateModule) DcModules.get(getIndex() + DcModules._TEMPLATE);
        else 
            return null;
    }
    
    public Collection getOnlineServices() {
        return getOnlineSearch().getServers();
    }
    
    public void addField(DcField field) {
        fields.put(new Integer(field.getIndex()), field);
    }
    
    public Collection loadData() {
        try {
            return DatabaseManager.executeQuery(getDcObject(), true);
        } catch (Exception e) {
            DcLog.add(e);
        }
        return null;
    }    

    public MasterView[] getViews() {
        if (getSearchView() != null && getInsertView() != null)
            return new MasterView[] {getSearchView(), getInsertView()};
        else if (getSearchView() != null)
            return new MasterView[] {getSearchView()};
        else if (getInsertView() != null)
            return new MasterView[] {getInsertView()};
        else
        	return new MasterView[0];
    }

    public DcField getField(int index) {
        DcField field = (DcField) fields.get(new Integer(index));
        return field == null ? getSystemField(index) : field;
    }

    public DcField getField(String dbName) {
        for (Iterator iter = getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getDatabaseFieldName().equals(dbName))
                return field;
        }
        return null;
    }
    
    public void addValueEnhancer(IValueEnhancer enhancer, int field) {
        getField(field).addValueEnhancer(enhancer);
    }
    
    public DcField getSystemField(int index) {
        return (DcField) systemFields.get(new Integer(index));
    }

    public void setChild(int module) {
    	this.childIdx = module;
    }
    
    public DcMenuBar getMenuBar() {
        return isTopModule() ? new DcMenuBar(this) : null;
    }

    public DcModule getParent() {
        return parentIdx > 0 ? DcModules.get(parentIdx) : null;
    }
    
    public DcModule getChild() {
    	return childIdx > 0 ? DcModules.get(childIdx) : null;
    }

    public FilterDialog getFilterForm(boolean create) {
        if (create)
            filterForm = filterForm == null && isTopModule() ? new FilterDialog(this, getSearchView()) : filterForm;
            
        return filterForm;
    }

    public ChartPanel getChartPanel(boolean create) {
        if (create)
            chartPanel = chartPanel == null && isTopModule() ? new ChartPanel(getIndex()) : chartPanel;
            
        return chartPanel;
    }
    
    public boolean isTopModule() {
        return topModule;
    }

    public int getFieldCount() {
        return fields.size();
    }

    public boolean canBeLended() {
        return canBeLended && DcModules.get(DcModules._CONTACTPERSON).isEnabled();
    }

    public void canBeLended(boolean b) {
        this.canBeLended = b;
    }
    
    public boolean hasDependingModules() {
        return hasDependingModules;    
    }
    
    public Object getSetting(String key) {
        return settings.get(key);
    }
    
    public void setSetting(String key, Object value) {
        settings.set(key, value);
    }

    public QuickViewPanel getQuickView() {
        return new QuickViewPanel(this, true);
    }

    public boolean hasReferenceTo(int module) {
        for (Iterator iter = getFields().iterator(); iter.hasNext(); ) {
            if (((DcField) iter.next()).getSourceModuleIdx() == module)
                return true;
        }
        return false;
    }

    public Collection getFields() {
        if (sortedFields == null || sortedFields.size() < fields.size()) {
            sortedFields = new ArrayList();
            sortedFields.addAll(fields.values());
            Collections.sort((List) sortedFields, new Comparator() {
                public int compare(Object o1, Object o2) {
                    return ((DcField) o1).getLabel().compareTo(((DcField) o2).getLabel());
                }
            });
        }
        return sortedFields;
    }

    public DcFieldDefinitions getFieldDefinitions() {
        return (DcFieldDefinitions) getSetting(DcRepository.modulesettings.stFieldDefinitions);
    }
    
    public int[] getFieldIndices() {
        Set keys = fields.keySet();
        int[] indices = new int[keys.size()];
        int counter = 0;
        for (Iterator iter = keys.iterator(); iter.hasNext(); counter++) {
            indices[counter] = ((Integer) iter.next()).intValue();
        }
        return indices;
    }
    
    public void removeEnhancers() {
        for (Iterator iter = fields.values().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            field.removeEnhancers();
        }
    }    

    public Class getImporterClass() {
        return importerClass;
    }
    
    public FileImporter getImporter() {
        if (importerClass != null) {
            try {
                return (FileImporter) importerClass.newInstance();
            } catch (Exception exp) {
                DcLog.add(exp);
            }
        }
        return null;
    }
    
    public OnlineSearch getOnlineSearch()  {
        if (onlineSearch == null) {
            try {
                onlineSearch = (OnlineSearch) onlineSearchClass.newInstance();
            } catch (Exception exp) {
                try {
                    onlineSearch = (OnlineSearch) onlineSearchClass.getConstructors()[0].newInstance(new Object[] {new Integer(getIndex())});  
                } catch (Exception exp2) {
                    DcLog.add(exp);    
                }
            }
        }
        return onlineSearch;
    }  
    
    public net.sf.dc.settings.Settings getSettings() {
        return settings;
    }
    
    protected void initializeFields() {
        addField(new DcField(DcObject._ID, getIndex(), "ID",
                             false, true, true, false, false,
                             15, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._BIGINTEGER,
                             "ID"));
        addField(new DcField(DcObject._SYS_CREATED, getIndex(), "Created",
                             false, true, true, true, true,
                             10, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                             "Created"));
        addField(new DcField(DcObject._SYS_MODIFIED, getIndex(), "Modified",
                             false, true, true, true, true,
                             10, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                             "Modified"));
        
        if ((isTopModule() || isChildModule()) && !(this instanceof DcPropertyModule)) {
            addField(new DcField(DcMediaObject._U1_USER_LONGTEXT, getIndex(), "User Long Text Field", 
                    false, false, false, true, false, 
                    4000, ComponentFactory._LONGTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                    "UserLongText1"));         
            addField(new DcField(DcMediaObject._U2_USER_SHORTTEXT1, getIndex(), "User Short Text Field 1",  
                    false, false, false, true, false, 
                    255, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                    "UserShortText1")); 
            addField(new DcField(DcMediaObject._U3_USER_SHORTTEXT2, getIndex(), "User Short Text Field 2",  
                    false, false, false, true, false, 
                    255, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                    "UserShortText2")); 
            addField(new DcField(DcMediaObject._U4_USER_NUMERIC1, getIndex(), "User Numeric Field 1",  
                    false, false, false, true, false, 
                    255, ComponentFactory._NUMBERFIELD, getIndex(), DcRepository.datatypes._LONG,
                    "UserInteger1"));
            addField(new DcField(DcMediaObject._U5_USER_NUMERIC2, getIndex(), "User Numeric Field 2",  
                    false, false, false, true, false, 
                    255, ComponentFactory._NUMBERFIELD, getIndex(), DcRepository.datatypes._LONG,
                    "UserInteger2"));            
        }

        if (isTopModule() && getOnlineSearchClass() != null) {
            addField(getField(DcObject._SYS_SERVICE));
            addField(getField(DcObject._SYS_SERVICEURL));
        }
        
        if (canBeLended()) {
            addField(getField(DcObject._SYS_AVAILABLE));
            addField(getField(DcObject._SYS_LOANEDBY));
            addField(getField(DcObject._SYS_LOANDURATION));
        }

        addField(getField(DcObject._SYS_MODULE));
    }    

    /**
     * Initializes and corrects the module settings (if necessary)
     */
    protected void initializeSettings() {
        settings = new DcModuleSettings(this);
    
        // check whether the definitions are still correct; 
        // - there can be more definitions for a field; the actual field can be removed by a user
        QuickViewFieldDefinitions qvDefinitions = 
            (QuickViewFieldDefinitions) settings.get(DcRepository.modulesettings.stQuickViewFieldDefinitions);
        
        QuickViewFieldDefinitions newQvDefinitions = new QuickViewFieldDefinitions();  
        for (Iterator iter = qvDefinitions.iterator(); iter.hasNext(); ) {
            QuickViewFieldDefinition definition = (QuickViewFieldDefinition) iter.next();
            if (getField(definition.getField()) != null)
                newQvDefinitions.addDefinition(definition);
        }  
        
        DcFieldDefinitions definitions = 
            (DcFieldDefinitions) settings.get(DcRepository.modulesettings.stFieldDefinitions);
        
        DcFieldDefinitions newDefinitions = new DcFieldDefinitions();  
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            if (getField(definition.getIndex()) != null)
                newDefinitions.addDefinition(definition);
        }
        
        settings.set(DcRepository.modulesettings.stQuickViewFieldDefinitions, newQvDefinitions);
        settings.set(DcRepository.modulesettings.stFieldDefinitions, newDefinitions);
    }
    
    protected void initializeSystemFields() {
        systemFields.put(new Integer(DcObject._SYS_MODULE),
                         new DcField(DcObject._SYS_MODULE, getIndex(), "Item",
                                     true, true, true, true, false,
                                     255, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                                     ""));
        systemFields.put(new Integer(DcObject._SYS_AVAILABLE),
                         new DcField(DcObject._SYS_AVAILABLE, getIndex(), "Available",
                                     true, true, true, true, false,
                                     4, ComponentFactory._AVAILABILITYCOMBO, getIndex(), DcRepository.datatypes._BOOLEAN,
                                     ""));
        systemFields.put(new Integer(DcObject._SYS_LOANEDBY),
                         new DcField(DcObject._SYS_LOANEDBY, getIndex(), "Loaned By",
                                     true, true, true, true, false,
                                     255, ComponentFactory._REFERENCEFIELD, DcModules._CONTACTPERSON, DcRepository.datatypes._DCOBJECTREFERENCE,
                                     ""));
        systemFields.put(new Integer(DcObject._SYS_LOANDURATION),
                         new DcField(DcObject._SYS_LOANDURATION, getIndex(), "Days Loaned",
                                     true, true, true, true, false,
                                     10, ComponentFactory._NUMBERFIELD, getIndex(), DcRepository.datatypes._LONG,
                                     ""));
        
        if (isTopModule() && getOnlineSearchClass() != null) {
            systemFields.put(new Integer(DcObject._SYS_SERVICE),
                             new DcField(DcObject._SYS_SERVICE, getIndex(), "Service",
                                         false, true, true, false, true,
                                         255, ComponentFactory._SHORTTEXTFIELD, getIndex(), DcRepository.datatypes._STRING,
                                         "service"));
            systemFields.put(new Integer(DcObject._SYS_SERVICEURL),
                             new DcField(DcObject._SYS_SERVICEURL, getIndex(), "Service URL",
                                         false, true, true, false, true,
                                         255, ComponentFactory._URLFIELD, getIndex(), DcRepository.datatypes._STRING,
                                         "serviceurl"));        
        }
    }    
    
    public boolean hasInsertView() {
        return hasInsertView;
    }

    public boolean hasSearchView() {
        return hasSearchView;
    }
    
    protected void initializeUI()  {
        if (searchView == null && insertView == null) {
            if (hasInsertView()) {
                insertView = new MasterView();
                insertView.addView(MasterView._TABLE_VIEW, new TableInsertView(this, DcResources.getText("lblNewItem", getObjectNamePlural()), IconLibrary._icoItemsNew));
                insertView.addView(MasterView._LIST_VIEW, new ListInsertView(this, DcResources.getText("lblNewItem", getObjectNamePlural()), IconLibrary._icoItemsNew));
            }

            if (hasSearchView()) {
                searchView = new MasterView();
                searchView.addView(MasterView._TABLE_VIEW, new TableSearchView(this, getObjectNamePlural()));
                searchView.addView(MasterView._LIST_VIEW, new ListSearchView(this, getObjectNamePlural(), getIcon16()));
            }
        }
    }

    public String toString() {
        return getName();
    }

    public int getDefaultSortFieldIdx() {
        return defaultSortFieldIdx;
    }

    public int getDisplayIndex() {
        return displayIndex;
    }

    public int getNameFieldIdx() {
        return nameFieldIdx;
    }

    public String getIcon16Filename() {
        return icon16filename;
    }

    public String getIcon32Filename() {
        return icon32filename;
    }

    public int hashCode() {
        return index;
    }
    
    public boolean equals(Object o) {
        return (o instanceof DcModule ? ((DcModule) o).getIndex() == getIndex() : false);
    }

    public Class getOnlineSearchClass() {
        return onlineSearchClass;
    }

    public XmlModule getXmlModule() {
        return xmlModule;
    }

    public void setXmlModule(XmlModule xmlModule) {
        this.xmlModule = xmlModule;
    }    
    
    public String getLabel() {
        return label;
    }
    
    public DcObject[] getDefaultData() throws Exception  {
        String filename = getTableName() + ".data";
        File file = new File(DataCrow.moduleDir + "data/" + filename);
        
        DcObject[] data = null;
        if (file.exists()) {
            InputStreamReader reader = new InputStreamReader(new FileInputStream(file), "UTF-8");
            CSVReader csvReader = new CSVReader(reader, "\t", null);
            List lines = csvReader.readAll();  
            
            data = new DcObject[lines.size() - 1];
            
            Object[] headers = (Object[]) lines.get(0);
            for (int i = 1; i < lines.size(); i++) {
                Object[] line = (Object[]) lines.get(i);
                DcObject dco = getDcObject();
                for (int j = 0; j < headers.length; j++) {
                    Object value = line[j];
                    if (value != null && !value.equals("")) {
                        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
                            DcField field = (DcField) iter.next();
                            if (field.getSystemName().equalsIgnoreCase((String) headers[j])) {
                                if (field.getValueType() == DcRepository.datatypes._PICTURE) {
                                    Picture picture = new Picture();
                                    picture.isNew(true);
                                    picture.setValue(Picture._C_FILENAME, DataCrow.baseDir + value);
                                    picture.setValue(Picture._D_IMAGE, Utilities.readFile(new File(DataCrow.baseDir + value)));
                                    value = picture;
                                } else if (field.getValueType() == DcRepository.datatypes._ICON) {
                                    String s = DataCrow.baseDir + value;
                                    value = Utilities.fileToBase64String(new URL("file://" + s));
                                }
                                
                                dco.setValue(field.getIndex(), value);
                                break;
                            }
                        }                        
                    }
                }
                data[i - 1] = dco;
            }
        }
        
        return data;
    }

    public void setIcon16(ImageIcon icon16) {
        this.icon16 = icon16;
    }

    public void setIcon32(ImageIcon icon32) {
        this.icon32 = icon32;
    }    
}
