/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.modules;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.sf.dc.core.DataCrow;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.ModuleUpgradeException;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.xml.XmlModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.util.FileNameFilter;

/**
 * Things to know:
 * - Property modules are stored in "propertyBaseModules".
 *   These modules are used as templates for the actual references.
 *   (DcPropertyModule.getIndex() + DcModule.getIndex())
 * - Multi relation references are stored in the same way except they also get a
 *   mapping module (<ParentModule>.getIndex() + <ReferencedModule>.getIndex() + DcModules._MAPPING)
 * - If defined, a base property module serves multiple modules (such as the music genres).
 * 
 * @author rj.vanderwaals
 */
public class DcModules {

    public static final int _SOFTWARE = 50;
    public static final int _MOVIE = 51;
    public static final int _MUSICALBUM = 52;
    public static final int _AUDIOCD = 53;
    public static final int _BOOK = 54;
    public static final int _IMAGE = 55;
    public static final int _CONTACTPERSON = 56;
    public static final int _MEDIA = 57;
    public static final int _MUSICTRACK = 58;
    public static final int _AUDIOTRACK = 59;
    public static final int _PICTURE = 60;
    public static final int _LOAN = 61;
    
    public static final int _CATEGORY = 10000;
    public static final int _STORAGEMEDIA = 11000;
    public static final int _PLATFORM = 12000;
    public static final int _TEMPLATE = 13000;
    public static final int _MUSICGENRE = 14000;
    public static final int _STATE = 15000;
    public static final int _GENRE = 16000;
    
    public static final int _ACTOR = 30000;
    public static final int _DIRECTOR = 31000;
    public static final int _AUTHOR = 32000;
    public static final int _BOOKPUBLISHER = 33000;
    public static final int _SOFTWAREPUBLISHER = 34000;
    public static final int _DEVELOPER = 35000;
    public static final int _MUSICARTIST = 36000;

    public static final int _MAPPING = 50000;

    private static final Map propertyBaseModules = new HashMap();
    private static final HashMap modules = new LinkedHashMap();
    private static final Map indexMapping = new HashMap();
    
    public static void load() throws ModuleUpgradeException, InvalidModuleXmlException, ModuleJarException {
        initReferenceBaseModules();
        loadSystemModules();
        loadModuleJars();
        initIndexMapping();
    }
    
    public static void save() {
        for (Iterator iter = getAllModules(); iter.hasNext(); ) {
            try {
                DcModule module = (DcModule) iter.next();
                XmlModule xmlModule = module.getXmlModule();
                if (xmlModule != null && xmlModule.isChanged()) 
                    new ModuleJar(module.getXmlModule()).save();
            } catch (Exception e) {
                DcLog.add(e);
            }
        }
    }
    
    private static void initIndexMapping() {
        for (Iterator iter = getAllModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            indexMapping.put(new Integer(module.getIndex()), new Integer(module.getIndex() - module.getIndexDiff()));
        }
    }
    
    private static void initReferenceBaseModules() {
        propertyBaseModules.put(new Integer(DcModules._CATEGORY), 
                       new DcPropertyModule(DcModules._CATEGORY, "Category", 
                                            "category", "cat",
                                            DcResources.getText("sysCategory"),
                                            DcResources.getText("sysCategoryPlural")));
        propertyBaseModules.put(new Integer(DcModules._STATE), 
                       new DcPropertyModule(DcModules._STATE, "State", 
                                            "state", "st",
                                            DcResources.getText("sysState"),
                                            DcResources.getText("sysStatePlural")));
        propertyBaseModules.put(new Integer(DcModules._PLATFORM), 
                       new DcPropertyModule(DcModules._PLATFORM, "Platform", 
                                            "platform", "pf",
                                            DcResources.getText("sysPlatform"),
                                            DcResources.getText("sysPlatformPlural")));    
        propertyBaseModules.put(new Integer(DcModules._STORAGEMEDIA), 
                       new DcPropertyModule(DcModules._STORAGEMEDIA, "Platform", 
                                            "storagemedium", "stme",
                                            DcResources.getText("sysStorageMedium"),
                                            DcResources.getText("sysStorageMediumPlural")));            
        propertyBaseModules.put(new Integer(DcModules._GENRE), 
                       new DcPropertyModule(DcModules._GENRE, "Genre", 
                                            "genre", "gr",
                                            DcResources.getText("sysGenre"),
                                            DcResources.getText("sysGenrePlural")));
        
        DcPropertyModule musicGenreMod =
                        new DcPropertyModule(DcModules._MUSICGENRE, "Music Genre", "musicgenre", "musgr",
                                           DcResources.getText("sysMusicGenre"), 
                                           DcResources.getText("sysMusicGenrePlural"));
        
        musicGenreMod.setServesMultipleModules(true);
        propertyBaseModules.put(new Integer(DcModules._MUSICGENRE), musicGenreMod);
    }
    
    public static int getMappingModIdx(int module, int referenceModIdx) {
        int baseModIdx = DcModules.get(module) instanceof TemplateModule ? 
                        ((TemplateModule) DcModules.get(module)).getTemplatedModule().getIndex() : module;
        return DcModules.get(referenceModIdx) instanceof DcPropertyModule ? 
               DcModules._MAPPING + referenceModIdx : baseModIdx + referenceModIdx + DcModules._MAPPING;
    }
    
    public static int getAvailableIdx(XmlModule module) {
        int add = module.getModuleClass().equals(DcPropertyModule.class) ? 1000 : 1;
        int index = module.getModuleClass().equals(DcPropertyModule.class) ? 20000 : 1;
        while (modules.containsKey(new Integer(index)) || propertyBaseModules.containsKey(new Integer(index)))
            index += add;
        
        return index;
    }    
    
    private static void loadSystemModules() {
        register(new PictureModule());
        register(new LoanModule());
    }
    
    private static void loadModuleJars() throws ModuleUpgradeException, InvalidModuleXmlException, ModuleJarException  {
        File file = new File(DataCrow.moduleDir);
        String[] files = file.list(new FileNameFilter("jar"));
        
        Collection dependingModules = new ArrayList();
        Collection mods = new ArrayList();
        for (int i = 0; i < files.length; i++) {
            ModuleJar mj = new ModuleJar(DataCrow.moduleDir + files[i]);
            
            mj.load();
            XmlModule xmlModule = mj.getModule();
            if (xmlModule.getModuleClass().equals(DcPropertyModule.class))
                registerBasePropertyModule((DcPropertyModule) convert(xmlModule));
            else if (xmlModule.hasDependingModules()) 
                mods.add(xmlModule);
            else 
                dependingModules.add(xmlModule);
        }

        for (Iterator iter = mods.iterator(); iter.hasNext(); ) {
            XmlModule xmlModule = (XmlModule) iter.next();
            register(convert(xmlModule));
        }
        
        for (Iterator iter = dependingModules.iterator(); iter.hasNext(); ) {
            XmlModule xmlModule = (XmlModule) iter.next();
            register(convert(xmlModule));
        }
    }
    
    public static DcModule convert(XmlModule xmlModule) {
        Class parent = xmlModule.getModuleClass();
        try {
            return (DcModule) parent.getConstructor(new Class[] {XmlModule.class}).newInstance(new Object[] {xmlModule});
        } catch (Exception exp) {
            DcLog.add(exp);
        }
        return null;
    }
    
    private static int getIndexDiff(DcModule module) {
        int index = module.getIndex();
        while (modules.get(new Integer(index)) != null)
            ++index;
        
        return index - module.getIndex();
    }

    /**
     * Holds the global property module definitions. These are used as a base to create
     * sepecific referenced modules from.
     * @param module
     */
    public static void registerBasePropertyModule(DcPropertyModule module) {
        propertyBaseModules.put(new Integer(module.getIndex()), module);
    }
    
    public static void register(DcModule module) {
        registerModule(module);
        registerPropertyModules(module);
    }
    
    private static void registerModule(DcModule module) {
        module.setIndexDiff(getIndexDiff(module));
        modules.put(new Integer(module.getIndex()), module);
    }

    private static void registerPropertyModules(DcModule module) {
        if (module.getIndex() == DcModules._MEDIA) // special case
            return;
        
        Collection fields = module.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            DcModule mod = module instanceof TemplateModule ? 
                               ((TemplateModule) module).getTemplatedModule() : module; 
   
            int sourceIdx = field.getSourceModuleIdx();
            int realIdx = sourceIdx;
            if (propertyBaseModules.containsKey(new Integer(sourceIdx))) {
                DcPropertyModule propMod = (DcPropertyModule) propertyBaseModules.get(new Integer(sourceIdx));
                
                if (propMod.isServingMultipleModules()) {
                    DcPropertyModule pm = new DcPropertyModule(realIdx, propMod.getName(),  
                                                                  propMod.getTableName(), propMod.getTableShortName(),
                                                                  propMod.getObjectName(), propMod.getObjectNamePlural());
                    pm.setServesMultipleModules(true);
                    addMissingPropertyFields(propMod, pm);
                    registerModule(pm);
                } else {
                    realIdx = sourceIdx + mod.getIndex();
                    
                    DcPropertyModule pm = new DcPropertyModule(realIdx, propMod.getName(),  
                                                               mod.getTableName() + "_" + propMod.getTableName(), 
                                                               mod.getTableShortName() + propMod.getTableShortName(),
                                                               propMod.getObjectName(), propMod.getObjectNamePlural());
                    addMissingPropertyFields(propMod, pm);
                    registerModule(pm);
                }
                
                if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION)
                    registerModule(new MappingModule(mod, DcModules.get(realIdx)));
            
            } else {
                if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION)
                    registerModule(new MappingModule(mod, DcModules.get(sourceIdx)));                
            }
        }
        
        if (module.isTopModule() || module.isChildModule())
            registerModule(new TemplateModule(module));
    }
    
    /**
     * As users can define extra fields for base property modules these fields need to be
     * added to the property module when assigned (by default a property module has only a name 
     * and an icon field).
     * @param base
     * @param pm
     */
    private static void addMissingPropertyFields(DcPropertyModule base, DcPropertyModule pm) {
        for (Iterator iter1 = base.getFields().iterator(); iter1.hasNext(); ) {
            DcField fld1 = (DcField) iter1.next();
            boolean exists = false;
            for (Iterator iter2 = pm.getFields().iterator(); iter2.hasNext(); ) {
                DcField fld2 = (DcField) iter2.next();
                if (fld2.getIndex() == fld1.getIndex())
                    exists = true;
            }
            
            if (!exists) {
                DcField field = new DcField(fld1.getIndex(), pm.getIndex(), fld1.getLabel(), fld1.isUiOnly(), 
                                            fld1.isEnabled(), fld1.isReadOnly(), fld1.isSearchable(), 
                                            fld1.isTechnicalInfo(), fld1.getMaximumLength(), 
                                            fld1.getFieldType(), pm.getIndex(), fld1.getValueType(), 
                                            fld1.getDatabaseFieldName());
                pm.addField(field);
            }
        }
        
        pm.setIcon16(base.getIcon16());
        pm.setIcon32(base.getIcon32());
        pm.setXmlModule(base.getXmlModule());
        pm.initializeSettings();
    }
    
    public static Collection getPropertyModules(DcModule module) {
        Collection references = new ArrayList();
        for (Iterator iter = module.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            DcPropertyModule pm = getPropertyModule(field);
            if (pm != null && !references.contains(pm))
                references.add(pm);
        }
        return references;
    }
    
    public static DcModule getReferencedModule(DcField field) {
        int parentModule = DcModules.get(field.getModule()) instanceof TemplateModule ?
                           ((TemplateModule) DcModules.get(field.getModule())).getTemplatedModule().getIndex() : 
                           field.getModule();
        
        DcModule module = DcModules.get(field.getSourceModuleIdx());
        return module == null ? DcModules.get(field.getSourceModuleIdx() + parentModule) : module;
    }
    
    public static DcPropertyModule getPropertyModule(DcField field) {
        DcModule module = getReferencedModule(field);
        return module instanceof DcPropertyModule ? (DcPropertyModule) module : null;
    }

    public static Collection getPropertyBaseModules() {
        return propertyBaseModules.values();
    }
    
    public static DcModule get(String name) {
        for (Iterator iter = modules.values().iterator(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            if (module.getName().equalsIgnoreCase(name)) {
                return module;
            }
        }
        return null; 
    }
    
    public static DcModule get(int index) {
        Integer key = new Integer(index); 
        if (modules.containsKey(key)) { 
            return (DcModule) modules.get(key);
        } else {
            return (DcModule) modules.get(key);
        }
    }
    
    public static DcModule getCurrent() {
        return DcModules.get(DcSettings.getInt(DcRepository.settings.stModule));
    }    

    public static DcObject getObjectForTable(String tableName) {
        for (Iterator iter = modules.values().iterator(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            if (module.getTableName().equalsIgnoreCase(tableName)) {
                return module.getDcObject();
            }
        }
        return null;
    }

    public static void loadData() {
        new DataManager();
    }

    
    public static Collection getReferencingModules(int modIdx) {
        Collection refs = new ArrayList();
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            
            if (    !module.getName().equals(DcModules.get(modIdx).getName()) && 
                     module.getIndex() != modIdx && !(module instanceof TemplateModule)) {
                
                if (module instanceof MappingModule) {
                    MappingModule mappingMod = (MappingModule) module;
                    if (mappingMod.getReferencedModIdx() == modIdx) 
                        refs.add(mappingMod);
                } else {
                    // check modules, field based
                    for (Iterator iterFlds = module.getFields().iterator(); iterFlds.hasNext(); ) {
                        DcField field = (DcField) iterFlds.next();
                        if (field.getReferenceIdx() == modIdx)
                            refs.add(module);
                    }
                }
            }
        }
        
        return refs;
    }    
    
    public static Iterator getAllModules() {
        List c = new ArrayList();
        for (Iterator iter = modules.values().iterator(); iter.hasNext(); )
            c.add(iter.next());
        
        Collections.sort(c, new ModuleComparator()); 
        
        return c.iterator();
    }
    
    public static Iterator getModules() {
        Collection c = new ArrayList();
        for (Iterator iter = getAllModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            if (module.isEnabled())
                c.add(module);
        }
        return c.iterator();
    }
    
    private static class ModuleComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            return ((DcModule) o1).getLabel().toLowerCase().compareTo(((DcModule) o2).getLabel().toLowerCase());
        }
    }
}

