/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.objects;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.ImageIcon;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.db.DatabaseManager;
import net.sf.dc.core.db.Query;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.template.Templates;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.core.wf.WorkFlow;
import net.sf.dc.core.wf.requests.IRequest;
import net.sf.dc.core.wf.requests.RequestCollection;
import net.sf.dc.core.wf.requests.SynchronizeWithManagerRequest;
import net.sf.dc.enhancers.IValueEnhancer;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.settings.definitions.Definitions;
import net.sf.dc.util.Utilities;

public class DcObject {

    protected Map values = new HashMap();
    private RequestCollection requests = new RequestCollection();
    protected Collection children = new ArrayList();

    protected int uiRowNumber = -1;
    private final int module;

    protected boolean bValidation = true;
    private boolean bSilence = false;
    protected static boolean alwaysGenerateID = false;

    private boolean isBatch = false;
    private boolean endOfBatch = false;
    protected boolean keepInSync = true;
    
    public static final int _ID = 0;
    public static final int _SYS_MODULE = 201;
    public static final int _SYS_AVAILABLE = 202;
    public static final int _SYS_LOANEDBY = 203;
    public static final int _SYS_LOANDURATION = 204;
    public static final int _SYS_CREATED = 205;
    public static final int _SYS_MODIFIED = 206;
    public static final int _SYS_SERVICE = 207;
    public static final int _SYS_SERVICEURL = 208;
    
    public DcObject(int module) {
        this.module = module;
        
        initializeValues();
        initializeSystemValues();
        applySettings(DcModules.get(module).getFieldDefinitions());
    }    
    
    /**
     * Educated guess..
     */
    public int getDisplayFieldIdx() {
        Definitions definitions = DcModules.get(module).getFieldDefinitions();
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            if (definition.isDescriptive())
                return definition.getIndex();
        }
        return DcObject._ID;
    }
    
    public String getName() {
        Definitions definitions = DcModules.get(module).getFieldDefinitions();
        
        String name = "";
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            if (definition.isDescriptive()) {
                int idx = definition.getIndex();
                String disp = getDisplayString(idx);
                if (disp.length() > 0)
                    name += (name.length() > 0 ? ", " + disp : disp);
            }
        }
        
        if (name.length() == 0)
            name = getDisplayString(getModule().getNameFieldIdx());    
        
        return name;
    }
    
    public boolean hasPrimaryKey() {
        return getField(_ID) != null;
    }
    
    public void removeRequestors() {
        requests.clear();
    }

    public void setValueForColumn(String column, Object value) {
        Collection fields = getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getDatabaseFieldName().equalsIgnoreCase(column)) {
                setValue(field.getIndex(), value);
                break;
            }
        }
    }
    
    public void setValueForName(String sysName, Object value) {
        Collection fields = getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getSystemName().equalsIgnoreCase(sysName)) {
                setValue(field.getIndex(), value);
                break;
            }
        }
    }    

    public void keepInSync(boolean b) {
    	keepInSync = b;
    }
    
    public RequestCollection getRequests() {
    	return requests;
    }

    public String getTableName() {
    	return getModule().getTableName();
    }

    public String getTableShortName() {
        return getModule().getTableShortName();
    }

    public String getTableJoin() {
        return getModule().getTableJoin();
    }

    public int getDatabaseFieldCount() {
        Collection fields = getFields();
        int count = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            count += field.isUiOnly() ? 0 : 1;
        }
        return count;
    }

    public String getLabel(int index) {
        return getField(index).getLabel();
    }

    public void initializeImages() {
        Collection pictures = DataManager.getPictures(getID());
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            picture.setValue(Picture._D_IMAGE, null);
            setValueForColumn((String) picture.getValue(Picture._B_FIELD), picture);
        }
    }
    
    public void initializeReferences() {
        int[] fields = getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            DcField field = getField(fields[i]);
            if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                int mappingModIdx = DcModules.getMappingModIdx(getModule().getIndex(), field.getReferenceIdx());
                Collection mo = DataManager.getReferences(mappingModIdx, getID());
                setValue(fields[i], mo);
            }
        }
    }
    
    public void addRequest(IRequest request) {
        requests.addRequest(request);
    }

    public void setEndOfBatch(boolean end) {
    	endOfBatch = end;
    }

    public void setBatch(boolean batch) {
    	isBatch = batch;
    }

    public boolean isBatch() {
        return isBatch;
    }

    public boolean isEndOfBatch() {
    	return endOfBatch;
    }

    public boolean isFilled(int index) {
        Object value = getValue(index);
        return value != null && value.toString().trim().length() > 0 && !value.toString().equals("-1");
    }

    public void removeChildren() {
    	children.clear();
        children = new ArrayList();
    }

    public void loadChildren() {
        if (getModule().getChild() == null)
            return;
        
        if (children == null || children.size() == 0) {
            children = new ArrayList();
            children.addAll(DataManager.getChildren(getID(), getModule().getChild().getIndex()));
        }
    }  

    public void setChildren(Collection children) {
        this.children = new ArrayList();
        this.children.addAll(children);
        for (Iterator iter = children.iterator(); iter.hasNext(); ) {
            DcObject child = (DcObject) iter.next();
            child.setValue(child.getParentReferenceFieldIndex(), getID());
        }
    }

    public void addChild(DcObject child) {
        child.setBatch(isBatch());

        if (endOfBatch) {
        	setEndOfBatch(false);

            for (Iterator iter = children.iterator(); iter.hasNext(); ) {
                ((DcObject) iter.next()).setEndOfBatch(false);
            }
            child.setEndOfBatch(true);
        }

        children.add(child);
    }

    public Collection getChildren() {
        return children;
    }

    public int getParentReferenceFieldIndex() {
        Collection fields = getModule().getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
        	if (field.getValueType() == DcRepository.datatypes._DCPARENTREFERENCE) {
        		return field.getIndex();
            }
        }
        return -1;
    }

    public ImageIcon getIcon() {
        return null;
    }

    public DcField getFileField() {
        Collection fields = getModule().getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getFieldType() == ComponentFactory._FILELAUNCHFIELD)
                return field;
        }
        return null;
    }
    
    public String getFilename() {
        DcField field = getFileField();
        String filename = field != null ? (String) getValue(field.getIndex()) : null;
        return filename != null ? filename.replaceAll("`", "'") : filename;
    }

    public DcModule getModule() {
        return DcModules.get(module);
    }

    public Collection getFields() {
        return getModule().getFields();
    }

    public void markAsUnchanged() {
        for (Iterator iter = values.values().iterator(); iter.hasNext(); ) {
            DcValue value = (DcValue) iter.next();
            value.setChanged(false);
        }
    }
    
    public void reload() {
        resetImages();
        
        String query = "SELECT * FROM " + getTableName() + " WHERE ID = " + getID();
        Collection objects = DatabaseManager.executeQuery(query, Query._SELECT, false);
        for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
            DcObject dco = (DcObject) iter.next();
            int[] fields = getModule().getFieldIndices();
            for (int i = 0; i < fields.length; i++) {
                if (!dco.getField(fields[i]).isUiOnly())
                    setValue(fields[i], dco.getValue(fields[i]));
            }
            break;
        }
        
        if (getModule().canBeLended()) {
            Loan loan = DataManager.getCurrentLoan(getID());
            setValue(DcObject._SYS_AVAILABLE, new Boolean(loan.isAvailable(getID())));
            setValue(DcObject._SYS_LOANEDBY, loan.getPersonDescription());
            setValue(DcObject._SYS_LOANDURATION, loan.getDaysLoaned());
        }
        
        initializeImages();
        markAsUnchanged();
    }
    
    protected void beforeSave() {}
    
    private void resetImages() {
        for (Iterator iter = getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.getValueType() == DcRepository.datatypes._PICTURE)
                setValueLowLevel(field.getIndex(), null);
        }
    }
    
    public void setValue(int index, Object o) {
        DcValue value = getValueDef(index);
        if (value != null)
            value.setValue(o, getModule().getField(index));
    }    
    
    public void setValueLowLevel(int index, Object o) {
        DcValue value = getValueDef(index);
        if (value != null)
            value.setValueLowLevel(o, getModule().getField(index));
    }

    public void markAsChanged() {
        for (Iterator iter = values.values().iterator(); iter.hasNext(); ) {
            DcValue value = (DcValue) iter.next();
            value.setChanged(true);
        }

        for (Iterator iter = children.iterator(); iter.hasNext(); ) {
        	DcObject child =  (DcObject) iter.next();
        	child.markAsChanged();
        }
    }

    public boolean isChanged() {
        for (Iterator iter = getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            boolean changed = getValueDef(field.getIndex()).isChanged();
            if ((!field.isUiOnly() || field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) && changed)
                return true;
        }
        
        return false;
    }

    public boolean isChanged(int index) {
        return getValueDef(index).isChanged();
    }

    public void setChanged(int index, boolean b) {
    	getValueDef(index).setChanged(b);
    }

    public boolean isEnabled(int index) {
        return getModule().getField(index).isEnabled();
    }

    public boolean isRequired(int index) {
        return getModule().getField(index).isRequired();
    }

    public boolean isSearchable(int index) {
        return getModule().getField(index).isSearchable();
    }

    public int[] getDisabledFields() {
        Collection fields = getModule().getFields();

        int counter = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isEnabled()) {
                counter++;
            }
        }


        int[] disabled = new int[counter];
        counter = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isEnabled()) {
                disabled[counter++] = field.getIndex();
            }
        }

        return disabled;
    }

    public Collection getHiddenFields() {
        Collection fields = getModule().getFields();
        Collection hidden = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.isHidden())
                hidden.add(field);
        }
        return hidden;
    }

    public String getID() {
        return hasPrimaryKey() ? getValueDef(_ID).getValueAsString() : null;
    }

    public Collection getNotSearchableFields() {
        Collection fields = getModule().getFields();
        Collection notSearchable = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isSearchable())
                notSearchable.add(field);
        }
        return notSearchable;
    }

    public Collection getVisibleFields() {
        Collection fields = getModule().getFields();
        Collection visible = new ArrayList();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (!field.isHidden())
                visible.add(field);
        }
        return visible;
    }

    public void unload() {
        clearValues();
        values.clear();
        values = null;
        
        requests.clear();
        requests = null;
        
        children.clear();
        children = null;
    }
    
    public void clearValues() {
        Set keys = values.keySet();
        for (Iterator iter = keys.iterator(); iter.hasNext(); ) {
            Integer key = (Integer) iter.next();

            if (key.intValue() != _ID) {
                DcValue value = (DcValue) values.get(key);
                value.clear();;
            }
        }
    }

    public int getMaxFieldLength(int index) {
        return getField(index).getMaximumLength();
    }

    public Object getValue(int index) {
        Object value = null;
        if (getValueDef(index) != null)
            value = getValueDef(index).getValue();
        return value;
    }

    public String getDisplayString(int index) {
    	String s = "";
        try {
            if (getFieldType(index) != ComponentFactory._PICTUREFIELD)
                s = getValueDef(index).getDisplayString(getFieldType(index));
        } catch (Exception exp) {}

        return s;
    }

    public void setRow(int row) {
        uiRowNumber = row;
    }

    public int getRow() {
    	return uiRowNumber;
    }

    public int getFieldType(int index) {
        return getField(index).getFieldType();
    }

    public String getDatabaseFieldName(int index) {
        return getField(index).getDatabaseFieldName();
    }

    public void applyEnhancers(boolean update) {
        for (Iterator iter = getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            Object value = getValue(field.getIndex());
            
            IValueEnhancer[] enhancers = field.getValueEnhancers();
            for (int i = 0; i < enhancers.length; i++) {
                if (    enhancers[i].isEnabled() && 
                        (update && enhancers[i].isRunOnUpdating() || !update && enhancers[i].isRunOnInsert())) {
                    Object newVal = enhancers[i].apply(field, value);
                    Object oldVal = getValue(field.getIndex());
                    
                    if (newVal != null && ((newVal != null && oldVal == null) || (!newVal.equals(oldVal)))) {
                        setValue(field.getIndex(), newVal);
                    } 
                }
            }
        }
    }    
    
    private String getCurrentDate() {
        SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd");
        return df.format(new Date());
    }
    
    public void saveNew(boolean queued) throws ValidationException {
        try {
            markAsChanged();
            applyEnhancers(false);
            checkIntegrity(false);
            
            beforeSave();
            setValue(_SYS_CREATED, getCurrentDate());
            
            if (queued) {
            	if (keepInSync)
            		addRequest(new SynchronizeWithManagerRequest(SynchronizeWithManagerRequest._ADD, this));
            	
            	WorkFlow.insertValues(this);
            } else {
                Query query = new Query(Query._INSERT, this, null, null);
            	DatabaseManager.executeQuery(query, false);

            	if (keepInSync)
            		new SynchronizeWithManagerRequest(SynchronizeWithManagerRequest._ADD, this).execute(null);
            }
        } catch (ValidationException exp) {
            executeRequests(false);
            throw exp;
        } catch (Exception e) {
            DcLog.add(e);
        }
    }    
    
    public void saveUpdate(boolean queued) throws ValidationException {
        try {
            applyEnhancers(true);
            this.checkIntegrity(true);

            beforeSave();
            setValue(_SYS_MODIFIED, getCurrentDate());
            
            if (queued) {
            	if (keepInSync)
            		addRequest(new SynchronizeWithManagerRequest(SynchronizeWithManagerRequest._UPDATE, this));
            	
            	WorkFlow.updateValues(this);
            } else {
                Query query = new Query(Query._UPDATE, this, null, null);
                DatabaseManager.executeQuery(query, false);
            	if (keepInSync)
            		new SynchronizeWithManagerRequest(SynchronizeWithManagerRequest._UPDATE, this).execute(null);
            }
        } catch (ValidationException exp) {
            executeRequests(false);
            throw exp;
        } catch (Exception e) {
            DcLog.add(e);
        }
    }

    public void delete() {
    	if (keepInSync)
    		addRequest(new SynchronizeWithManagerRequest(SynchronizeWithManagerRequest._DELETE, this));
    	
        WorkFlow.deleteValues(this);
    }

    public void setSilence(boolean silence) {
        this.bSilence = silence;
        
        if (getChildren() != null) {
            for (Iterator iter = getChildren().iterator(); iter.hasNext(); ) {
                ((DcObject) iter.next()).setSilence(silence);
            }
        }
    }

    public boolean getSilence() {
        return this.bSilence;
    }

    public void setValidation(boolean validate) {
        bValidation = validate;
    }

    public DcField getField(int index) {
        return getModule().getField(index);
    }

    public void checkIntegrity(boolean update) throws ValidationException {
        if (DcSettings.getBoolean(DcRepository.settings.stCheckRequiredFields)) {
            validateRequiredFields();
        }
        if (DcSettings.getBoolean(DcRepository.settings.stCheckUniqueness)) {
            isUnique(this, update);
        }
    }

    public void isUnique(DcObject o, boolean bUpdateQuery) throws ValidationException {
        boolean bUnique = WorkFlow.checkUniqueness(o, bUpdateQuery);
        if (!bUnique && bValidation) {
            String message;
            if (getRow() != -1) {
                // shown by the table panels (ui row ID is set)
                message = DcResources.getText("msgItemNotUniqueRow",
                          new String[] {toString(), String.valueOf(uiRowNumber + 1)});
            } else {
                // shown by item forms and the likes
                message = DcResources.getText("msgItemNotUnique");
            }

        	throw new ValidationException(message);
        }
    }

    private String getUnfilledRequiredFields() {
        String required = "";
        Collection fields = getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            if (field.isRequired()) {
                Object value = getValue(field.getIndex());

                if (value == null || value.toString().equals("")) {
                    if (required.length() > 0)
                        required += ",";

                    required += getLabel(field.getIndex());
                }
            }
        }
        return required;
    }

    public int[] getFieldIndices() {
        return getModule().getFieldIndices();
    }

    public void applySettings(DcFieldDefinitions definitions) {
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            DcFieldDefinition definition = (DcFieldDefinition) iter.next();
            DcField field = getField(definition.getIndex());

            String label = definition.getLabel();
            
            if (label != null && label.trim().length() > 0)
                field.setLabel(label);
            else
            	field.setLabel(field.getSystemName());

            field.setEnabled(definition.isEnabled());
            field.setRequired(definition.isRequired());
            field.setHidden(!definition.isVisible());
        }
    }

    public void setIDs() {
        if (hasPrimaryKey()) {
            String sID = Utilities.getUniqueID();
            
            while (sID == null || sID.length() < 15)
                sID = Utilities.getUniqueID();
            
            sID = sID.substring(0, 15);
            setValue(DcObject._ID, sID);
            for (Iterator iter = children.iterator(); iter.hasNext(); ) {
                DcObject child = (DcObject) iter.next();
                child.setIDs();
                child.setValue(child.getParentReferenceFieldIndex(), sID);
            }
        }
    }

    protected DcValue getValueDef(int index) {
        return (DcValue) values.get(new Integer(index));
    }

    private void initializeSystemValues() {
        values.put(new Integer(DcObject._SYS_MODULE), new DcValue());
        setValue(DcObject._SYS_MODULE, getModule());
    }

    private void initializeValues() {
        int[] fields = getModule().getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            values.put(new Integer(fields[i]), new DcValue());
        }
    }

    protected void executeRequests(boolean saveSuccessful) {
        if (requests != null) {
            IRequest[] requestArray = requests.getRequests();
            for (int i = 0; i < requestArray.length; i++) {
                IRequest request = requestArray[i];
                if (saveSuccessful || request.getExecuteOnFail())
                    request.execute(new ArrayList());
            }
        }
    }
    
    protected void validateRequiredFields() throws ValidationException {
        String s = this.getUnfilledRequiredFields();

        if (s.trim().length() > 1) {
            String message;
            if (getRow() != -1) {
            	//shown by the table panels (ui row ID is set)
                message = DcResources.getText("msgIncompleteRecRow",
                		  new String[] {s, toString(), String.valueOf(getRow() + 1)});
            } else {
                //shown by item forms and the likes
                message = DcResources.getText("msgIncompleteRec", s);
            }
            throw new ValidationException(message);
        }
    }

    /**
     * Copies all values from the specified Data Crow object
     * @param dco source object
     */
    public void copy(DcObject dco) {
        int[] fields = dco.getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            int field = dco.getFieldIndices()[i];

            if (field != _ID && dco.getValue(field) != null) {
                
                if (getField(field).getValueType() == DcRepository.datatypes._PICTURE) {
                    Picture curPic = (Picture) dco.getValue(field);
                    
                    Picture newPic = new Picture();
                    newPic.copy(curPic);
                    
                    newPic.isNew(curPic.isNew());
                    newPic.isDeleted(curPic.isDeleted());
                    newPic.isUpdated(curPic.isUpdated());
                    
                    setValue(field, newPic);
                } else {
                    setValue(field, dco.getValue(field));    
                }
            }
        }
    }

    public void applyTemplate() {
        if (getModule().getTemplateModule() != null) {
            DcTemplate template = Templates.getDefault(getModule().getTemplateModule().getIndex());
            if (template != null)
                applyTemplate(template);
        }
    }
    
    public void applyTemplate(DcTemplate template) {
        if (template == null) {
            applyTemplate();
            return;
        } 
        
        int[] fields = getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            int idx = fields[i];

            DcField field = getField(idx);
            Object templateVal = template.getValue(idx); 
            
            if (    idx != _ID && 
                    field.getValueType() != DcRepository.datatypes._PICTURE && 
                    templateVal != null) {
                
                setValue(idx, template.getValue(idx));    
            } else if ( field.getValueType() == DcRepository.datatypes._PICTURE && 
                        template.getValue(idx) != null) {
                
                Picture templatePic = (Picture) template.getValue(idx);
                
                Picture pic = new Picture();
                templatePic.loadImage();
                pic.setValue(Picture._D_IMAGE, templatePic.getValue(Picture._D_IMAGE));
                pic.setValue(Picture._E_HEIGHT, templatePic.getValue(Picture._E_HEIGHT));
                pic.setValue(Picture._F_WIDTH, templatePic.getValue(Picture._F_WIDTH));
                pic.isNew(true);
                
                setValue(idx, pic);                
            }
        }
    }    
    
    public void merge(DcObject dco) {
        int[] fields = dco.getFieldIndices();
        for (int i = 0; i < fields.length; i++) {
            int field = fields[i];
            Object value = getValue(field);
            String s = value == null || value.toString().equals("-1") ? "" : value.toString().trim();
        
            if ((field != _ID && dco.getValue(field) != null) && s.length() > 0) { 
                setValue(field, dco.getValue(field));
            }
        }
    }    
    
    public Object clone() {
        DcObject dco = getModule().getDcObject();
        
        dco.copy(this);
        dco.setValue(DcObject._ID, getID());
        dco.markAsUnchanged();
        
        Collection children = getChildren();
        if (children != null) {
            for (Iterator iter = children.iterator(); iter.hasNext(); ) {
                DcObject child = (DcObject) iter.next();
                DcObject childCopy = child.getModule().getDcObject();
                childCopy.copy(child);
                dco.addChild(child);
            }
        }        
        
        int[] indices = getFieldIndices();
        for (int i = 0; i < indices.length; i++) {
            dco.setChanged(indices[i], isChanged(indices[i]));
        }
        
        return dco;
    }
    
    public int hashCode() {
        return getID() != null ? getID().hashCode() : super.hashCode();
    }

    public String toString() {
        return getName();
    }
    
    public boolean equals(Object o) {
        boolean equals = false;
        
        if (o instanceof String) {
            equals = getID() != null ? getID().equals(o) : false;
        } else if (o instanceof DcObject) {
            String id1 = ((DcObject) o).getDisplayString(DcObject._ID);
            String id2 = getDisplayString(DcObject._ID);
            if (id1.length() == 0 && id2.length() == 0) 
                return o == this;
            else 
                return id1.equals(id2);
        } else {
            equals = super.equals(o);
        }
        
        return equals;
    }
}
