/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.settings;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JComponent;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.DcColorSelector;
import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.objects.DcLookAndFeel;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.dc.settings.definitions.Definition;
import net.sf.dc.settings.definitions.Definitions;
import net.sf.dc.settings.definitions.ProgramDefinitions;
import net.sf.dc.settings.definitions.QuickViewFieldDefinitions;
import net.sf.dc.util.StringUtils;

/**
 * A Settings definition. The setting is always added to a file
 * (the settings file) and if specified, the setting is also
 * editable thru the SettingsView.
 * 
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class Setting {

    private int iDataType;
    private String sKey;
    private Object oValue;

    private int settingsGroup = 0; 
    private int parentGroup = -1; // no parent as default    
    private int componentType = ComponentFactory._SHORTTEXTFIELD;
    private String helpText = "";
    private boolean displayLabel = true; 
    private boolean showToUser = true;
    private String labelText;
    
    private boolean temporary = false;
    
    /**
     * Creates a setting
     * 
     * @param iDataType the data type of the value
     * @param sKey a unique identifier, also used for the settings file
     * @param oValue the default value for thsi setting
     * @param componentType the UI component
     * @param helptext helptext for this setting, used as tooltip
     * @param labeltext the display text for the label
     * @param displaylabel show the label on screen ?
     * @param showToUser is the setting meant to be display in the panel ?
     */
    public Setting(int iDataType,
                   String sKey,
                   Object oValue,
                   int componentType,
                   String helpText,
                   String labelText,
                   boolean displayLabel,
                   boolean showToUser) {

        this.iDataType = iDataType;
        this.sKey = sKey;
        this.oValue = oValue;
        this.componentType = componentType;
        this.displayLabel = displayLabel;
        this.showToUser = showToUser;
        this.labelText = labelText;
        this.helpText = helpText;
    }
    
    public void isTemporary(boolean b) {
        this.temporary = b;
    }
    
    public boolean isTemporary() {
        return temporary;
    }
    
    public String getLabelText() {
        return labelText;   
    }
    
    public boolean displayLabel() {
        return displayLabel;   
    }
    
    public boolean showToUser() {
        return showToUser;   
    }
    
    public int getSettingsGroup() {
        return settingsGroup;   
    }

    public int getSettingsGroupParent() {
        return parentGroup;   
    }
    
    public JComponent getUIComponent() {
        if (componentType == ComponentFactory._COLORSELECTOR)
            return new DcColorSelector(sKey);
        
        return ComponentFactory.getComponent(-1, -1, componentType, labelText, 400);
    }

    public int getDataType() {
        return iDataType;
    }

    public String getKey() {
        return sKey;
    }

    public Object getValue() {
        return oValue;
    }
    
    public String getHelpText() {
        return helpText;   
    }

    public String getValueAsString() {
        String toString = "";
        switch (iDataType) {
	        case DcRepository.datatypes._DEFINITIONGROUP:
	        	Definitions definitions = (Definitions) oValue;
	            for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
	            	Definition definition = (Definition) iter.next();
	                toString += "{" + definition.toSettingValue() + "}";
	            }
	            break;
            case DcRepository.datatypes._DIMENSION:
                int x = new Double(((Dimension) oValue).getWidth()).intValue();
                int y = new Double(((Dimension) oValue).getHeight()).intValue();

                toString = "" + x + '/' + y;
                break;
            case DcRepository.datatypes._FONT:
                toString =  toString +
                            ((Font) oValue).getName() + '/' +
                            ((Font) oValue).getSize() + '/' +
                            ((Font) oValue).getStyle();
                break;
            case DcRepository.datatypes._LOOKANDFEEL:
                toString =  toString +
                            ((DcLookAndFeel) oValue).getClassName() + "%%" +
                            ((DcLookAndFeel) oValue).getFileName() + "%%" +
                            ((DcLookAndFeel) oValue).getName() + "%%" +
                            ((DcLookAndFeel) oValue).getType();
                break;                
            case DcRepository.datatypes._COLOR:
                toString =  toString +
                            ((Color) oValue).getRed() + '/' +
                            ((Color) oValue).getGreen() + '/' +
                            ((Color) oValue).getBlue();
                break;
            case DcRepository.datatypes._LONG:
                toString =  toString + oValue;
                break;
            case DcRepository.datatypes._STRINGARRAY:
                String[] sArray = (String[]) oValue;
                for (int i = 0; i < sArray.length; i++) {
                    toString = toString + sArray[i] + '#';
                }
                break;
            case DcRepository.datatypes._INTEGERARRAY:
                int[] iArray = (int[]) oValue;
                if (iArray != null) {
                    for (int i = 0; i < iArray.length; i++) {
                        toString = toString + iArray[i] + '#';
                    }
                }
                break;                
            default:
                toString =  toString + (oValue != null ? oValue.toString() : "");
        }
        return toString;
    }

    public void setValue(Object oValue) {
        this.oValue = oValue;
    }

    public void setStringAsValue(String s) {
        String sValue = s;
        try {
            switch (iDataType) {

            case DcRepository.datatypes._DEFINITIONGROUP:
                    int group = sValue.indexOf("}");
                    Definitions definitions = (Definitions) getValue();
                    
                    if (definitions == null) {
                        if (sKey.equals(DcRepository.settings.stProgramDefinitions))
                            definitions = new ProgramDefinitions();
                        else if (sKey.equals(DcRepository.modulesettings.stQuickViewFieldDefinitions))
                            definitions = new QuickViewFieldDefinitions();
                        else 
                            definitions = new DcFieldDefinitions();
                    }
                    
                    while (group > -1) {
                        String value = sValue.substring(1, group);
                        definitions.addDefinition(value);
                        if (sValue.length() > group + 1) {
                            sValue = sValue.substring(group + 1, sValue.length());
                            group = sValue.indexOf("}");
                        } else {
                            group = -1;
                        }
                    }
                    
                    if (definitions.size() > 0) {
                        oValue = definitions;
                    }
                    break;
                case DcRepository.datatypes._DIMENSION:
                    int index1 = sValue.indexOf("/");
                    String testX = sValue.substring(0, index1);
                    String testY = sValue.substring(index1 + 1, sValue.length());

                    int x = Integer.valueOf(testX).intValue();
                    int y = Integer.valueOf(testY).intValue();

                    oValue = new Dimension(x, y);
                    break;
                case DcRepository.datatypes._COLOR:
                    String[] colorValue = StringUtils.getListElements(sValue, "/");
                    int red = Integer.valueOf(colorValue[0]).intValue();
                    int green = Integer.valueOf(colorValue[1]).intValue();
                    int blue = Integer.valueOf(colorValue[2]).intValue();
                    oValue = new Color(red, green, blue);
                    break;
                case DcRepository.datatypes._FONT:
                    String[] fontValues = StringUtils.getListElements(sValue, "/");
                    String fontName = fontValues[0];
                    int size = Integer.valueOf(fontValues[1]).intValue();
                    int style = Integer.valueOf(fontValues[2]).intValue();
                	oValue = new Font(fontName, style, size);
                    break;
                case DcRepository.datatypes._LOOKANDFEEL:  
                    String[] lafValues =  StringUtils.getListElements(sValue, "%%");
                    String className = lafValues[0];
                    String fileName = lafValues[1];
                    String name = lafValues[2];
                    int lafType = Integer.valueOf(lafValues[3]).intValue();
                    oValue = new DcLookAndFeel(name, className, fileName, lafType);
                    break;                    
                case DcRepository.datatypes._LONG:
                    oValue = new Integer(Integer.parseInt(sValue));
                    break;
                case DcRepository.datatypes._BOOLEAN:
                    oValue = Boolean.valueOf(sValue);
                    break;
                case DcRepository.datatypes._INTEGERARRAY:
                    Vector vInt = new Vector(0);

                    int index = sValue.indexOf("#");
                    while (index > -1) {
                        vInt.add(new Integer(sValue.substring(0, index)));
                        sValue = sValue.substring(index + 1, sValue.length());
                        index = sValue.indexOf("#");
                    }
                    Integer[] iFields = new Integer[vInt.size()];
                    iFields = (Integer[]) vInt.toArray(iFields);
                    
                    int[] intArray = new int[iFields.length];
                    for (int i = 0; i < iFields.length; i++) {
                        intArray[i] = iFields[i].intValue();
                    }
                    oValue = intArray;
                    break;                    
                case DcRepository.datatypes._STRINGARRAY:
                    Vector fields = new Vector(0);
                    index = sValue.indexOf("#");
                    while (index > -1) {
                        fields.add(sValue.substring(0, index));
                        sValue = sValue.substring(index + 1, sValue.length());
                        index = sValue.indexOf("#");
                    }

                    String[] sFields = new String[fields.size()];
                    oValue = fields.toArray(sFields);
                    break;
                default:
                    oValue = sValue;
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
}
