/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.settings;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.io.File;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Vector;

import net.sf.dc.core.objects.DcLookAndFeel;
import net.sf.dc.settings.definitions.Definitions;

/**
 * Contains all Settings Groups and there settings. 
 * The Settings class is the only class which needs to be referenced
 * to add groups, add settings, get settings and get values of settings.
 * 
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class Settings {

    private File settingsFile = new File("unnamed.ini");
    private LinkedHashMap htGroups = new LinkedHashMap();
    
    /**
     * Adds a group to the hashtable. A group may contain sub-groups.
     * Only the parent, to which a sub-group belongs, should be added.
     * 
     * @param sKey unique identifier for this group
     * @param group a top-level group (parent)
     */ 
    public void addGroup(String sKey, SettingsGroup group) {
        htGroups.put(sKey, group);
    }

    /**
     * Specifies the location and name of the settings file
     * 
     * @param file the settings file
     */
    public void setSettingsFile(File file) {
        settingsFile = file;
    }
    
    public File getSettingsFile() {
        return settingsFile;   
    }

    /**
     * Retrieves a settings group with the key
     */
    private SettingsGroup getSettingGroup(String sGroupKey) {
        Collection settings = getFlatSettingsGroups();
        for (Iterator iter = settings.iterator(); iter.hasNext(); ) {
            SettingsGroup settingsGroup = (SettingsGroup) iter.next();
            if (settingsGroup.getKey().equals(sGroupKey)) {
                return settingsGroup; 
            }
        }
        return null;
    }

    /**
     * Adds a setting to a specified group. A group can be either
     * a parent or a child.
     * 
     * @param sGroupKey the key of the group
     * @param setting the setting to be added to the specified group
     */
    public void addSetting(String sGroupKey, Setting setting) {
        SettingsGroup stGroup = getSettingGroup(sGroupKey);
        stGroup.add(setting);
    }

    public boolean isSettingKeyValid(String sKey) {
        boolean bValid = getSetting(sKey) == null ? false : true;
        return bValid;
    }

    /**
     * Retrieves a setting
     * 
     * @param sSettingsKey the settings key
     */
    public Setting getSetting(String sSettingsKey) {
        for (Iterator iter = htGroups.values().iterator(); iter.hasNext(); ) {
            SettingsGroup group = (SettingsGroup) iter.next();
            Setting setting = group.getSetting(sSettingsKey);
            
            if (setting != null) return setting;
        }
        return null;
    }

    /**
     * Retrieves a value of a settings
     * 
     * @param sKey the settings key
     */
    public Object getValue(String sKey) {
        Setting setting = getSetting(sKey);
        Object oValue = setting != null ? setting.getValue() : null;
        return oValue;
    }

    /**
     * Returns the value of the setting as a String array
     */
    public String[] getValueAsStringArray(String key) {
        return (String[]) getValue(key);
    }
    
    public int[] getValueAsIntegerArray(String key) {
        return (int[]) getValue(key);
    }

    public Definitions getValueAsDefinitionGroup(String key) {
        return (Definitions) getValue(key);
    }      
    
    public Font getValueAsFont(String key) {
        return (Font) getValue(key);
    }    

    public DcLookAndFeel getValueAsLookAndFeel(String key) {
        return (DcLookAndFeel) getValue(key);
    } 
    
    /**
     * Returns the value of the setting as an integer
     * 
     * @param sKey the settings key
     */
    public int getValueAsInt(String sKey) {
        Object o = getValue(sKey);
        return o != null ? ((Integer) o).intValue() : -1;
    }

    /**
     * Returns the value of the setting as a boolean
     */
    public boolean getValueAsBoolean(String key) {
        Object o = getValue(key);
        return o == null ? false : ((Boolean) o).booleanValue();
    }

    /**
     * Returns the value of the setting as a color object
     */
    public Color getValueAsColor(String key) {
        return (Color) getValue(key);
    }
    
    /**
     * Returns the value of the setting as a String
     */
    public Dimension getValueAsDimension(String key) {
        return (Dimension) getValue(key);
    }

    /**
     * Returns the value of the setting as a String
     * 
     * @param sKey the settings key
     */
    public String getValueAsString(String sKey) {
        Object o = getValue(sKey);
        return o != null ? o.toString() : null;
    }

    /**
     * Sets the value of a setting
     * 
     * @param sKey the settings key
     * @param oValue the value 
     */
    public void setValue(String sKey, Object oValue) {
        Setting setting = getSetting(sKey);
        if (setting != null) {
            setting.setValue(oValue);
        }
    }

    /**
     * Sets a string as a value for the setting (by parsing the 
     * string)
     * 
     * @param sKey the settings key
     * @param sValue the string value to be parsed
     */
    public void setStringAsValue(String sKey, String sValue) {
        Setting setting = getSetting(sKey);
        if (setting != null) {
            setting.setStringAsValue(sValue);
        }
    }

    /**
     * Retrieves all the top-level groups
     */    
    public LinkedHashMap getSettingsGroups() {
        return htGroups;   
    }

    /**
     * Retrieves all settings groups without an hierarchy
     */
    public Vector getFlatSettingsGroups() {
        Vector vSettings = new Vector();
        vSettings.addAll(htGroups.values());
        for (Iterator iter = htGroups.values().iterator(); iter.hasNext(); ) {
            SettingsGroup settingsGroup = (SettingsGroup) iter.next();
            vSettings.addAll(settingsGroup.getChildren().values());
        }
        return vSettings;
    }
    
    /**
     * Retrieves all the settings
     */
    public Vector getSettings() {
        Collection groups = getFlatSettingsGroups();
        Vector settings = new Vector(0);
        for (Iterator iter = groups.iterator(); iter.hasNext(); ) {
            SettingsGroup settingsGroup = (SettingsGroup) iter.next();
            settings.addAll(settingsGroup.getSettings().values());
        }        
        return settings;
    }
}
