/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.fileimporters;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.console.windows.fileimport.FileImportDialog;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.util.Directory;

/**
 * Base for all file importers
 * @author rj.vanderwaals
 *
 */
public abstract class FileImporter {

    private final int module;
    
    public FileImporter(int module) {
        this.module = module;
    }
    
    public int getModule() {
        return module;
    }
    
    public void showUI() {
        FileImportDialog dlg = new FileImportDialog(this);
        dlg.setVisible(true);
    }
    
    public boolean allowReparsing() {
        return false;
    }
    
    public abstract DcObject parse(String filename) throws ParseException;
    public abstract String[] getSupportedExtensions();
    
    public void parse(IFileImportClient listener, 
                      boolean isVerficationNeeded,
                      boolean isRecursiveSearch,
                      String sourceDir) throws Exception {
        
        Batch batch = new Batch(listener, this, isVerficationNeeded, isRecursiveSearch, sourceDir);
        batch.start();
    }

    protected int getFileSize(String path) {
        File file = new File(path);
        long fileSize = 0;
        if (file.isDirectory()) {
            Collection files = Directory.read(path, false, true, false, null, false);
            for (Iterator iter = files.iterator(); iter.hasNext(); ) {
                File folderFile = new File(iter.next().toString());
                fileSize += folderFile.length();
            }
        } else {
            fileSize = file.length();
        }
        return Math.round(fileSize / 1024f);
    }  
    
    protected String getName(String file) {
        String name = file;
        int index = name.lastIndexOf(File.separator);
        if (index >=0) {
            name = name.substring(index + 1, name.length());

            index = name.lastIndexOf(".");
            if (index > 0 && index > name.length() - 5) {
                name = name.substring(0, index);
            }

            name = name.replaceAll("\\.", " ");
            name = name.replaceAll("_", " ");
            name = name.replaceAll("-", " ");
        }
        return name;
    }    
    
    private class Batch extends Thread {
        
        private IFileImportClient listener;
        private FileImporter importer;
        
        private boolean isVerficationNeeded;
        private boolean isRecursiveSearch;
        private String sourceDir;
        
        public Batch(IFileImportClient listener, 
                     FileImporter importer,
                     boolean isVerficationNeeded,
                     boolean isRecursiveSearch,
                     String sourceDir) throws Exception {
            
            this.listener = listener;
            this.importer = importer;
            this.isVerficationNeeded = isVerficationNeeded;
            this.isRecursiveSearch = isRecursiveSearch;
            this.sourceDir = sourceDir;
            
            if (sourceDir == null || sourceDir.trim().length() == 0)
                throw new Exception(DcResources.getText("msgSelectDirectory"));
        }
        
        public void run() {
            try {
                listener.addMessage(DcResources.getText("msgImportStarts"));
                listener.addMessage(DcResources.getText("msgImportReadingDirs"));
                
                File file = new File(sourceDir);
                Collection files = new ArrayList();
                if (file.isFile()) {
                    files.add(sourceDir);
                } else {
                    files.addAll(Directory.read( 
                            sourceDir, isVerficationNeeded, isRecursiveSearch, true, getSupportedExtensions(), false));
                }

                DataCrow.mainFrame.setSelectedTab(net.sf.dc.console.MainFrame._INSERTTAB);
                parse(files);
            } catch (Exception exp) {
                listener.addMessage("msgProcessCrash");
                DcLog.add(exp);
                new MessageBox(DcResources.getText("msgProcessCrash"), MessageBox._ERROR);
            }
        }    
        
        public void parse(Collection files) {
            try {
                listener.addMessage(DcResources.getText("msgImportFoundXResults", "" + files.size()));
                listener.initProgressBar(files.size());
                int counter = 1;
                for (Iterator iter = files.iterator(); iter.hasNext() && !listener.cancelled(); counter++) {

                    String filename = iter.next().toString();
                    listener.addMessage(DcResources.getText("msgProcessingFileX", filename));
                    
                    DcObject dco = null; 
                    try {
                        dco = importer.parse(filename);
                    } catch (ParseException pe) {
                        listener.addError(DcResources.getText("msgCouldNotReadInfoFrom", filename));
                    }
                        
                    if (dco != null) {
                        if (listener.getStorageMedium() != null) { 
                            for (Iterator iterFields = dco.getFields().iterator(); iterFields.hasNext(); ) {
                                DcField field = (DcField) iterFields.next();
                                if (field.getSourceModuleIdx() == DcModules._STORAGEMEDIA)
                                    dco.setValue(field.getIndex(), listener.getStorageMedium());
                            }
                        }

                        if (    dco instanceof DcMediaObject && 
                                listener.getLocationName() != null &&
                                listener.getLocationName().trim().length() > 0) {
                            
                            dco.setValue(DcMediaObject._D_LOCATION, listener.getLocationName());
                        }
                        
                        dco.applyTemplate();
                        DcModule module = DcModules.getCurrent();
                        module.getCurrentInsertView().add(dco, false);

                        try {
                            sleep(200);
                        } catch (Exception ignore) {}
                    }
                    
                    listener.updateProgressBar(counter);
                }
            } finally {
                files.clear();
                listener.addMessage(DcResources.getText("msgImportStops"));
                listener.finish();
            }
        }        
    }
}