/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.fileimporters;

import java.io.File;
import java.util.Date;

import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.objects.helpers.Image;
import net.sf.dc.util.Utilities;
import net.sf.ext.ij.ImagePlus;

import com.drew.imaging.jpeg.JpegMetadataReader;
import com.drew.imaging.jpeg.JpegProcessingException;
import com.drew.metadata.Directory;
import com.drew.metadata.Metadata;
import com.drew.metadata.exif.ExifDirectory;
import com.drew.metadata.iptc.IptcDirectory;

public class ImageImporter extends FileImporter {

    public ImageImporter() {
        super(DcModules._IMAGE);
    }
    
    public String[] getSupportedExtensions() {
        return new String[] {"jpg", "gif", "jpeg", "png"};
    }
    
    public boolean allowReparsing() {
        return true;
    }    
    
    public DcObject parse(String filename) throws ParseException {
        Image image = new Image();
        
        try {
            ImagePlus imagePlus = new ImagePlus(filename);
            
            image.setIDs();
            
            String title = filename.substring(filename.lastIndexOf(File.separator) + 1, filename.lastIndexOf("."));
            image.setValue(Image._A_TITLE, title);
            
            int width = imagePlus.getWidth();
            int height = imagePlus.getHeight();
            
            image.setValue(Image._F_WIDTH, width != -1 ? new Long(width) : null);
            image.setValue(Image._G_HEIGHT, height != -1 ? new Long(height) : null);
            image.setValue(Image._H_FILENAME, filename);
            image.setValue(Image._M_FILESIZE, new Long(getFileSize(filename)));
            
            Picture pic = new Picture(); 
            byte[] img = Utilities.readFile(new File(filename));
            pic.setValue(Picture._A_OBJECTID, image.getID());
            pic.setValue(Picture._B_FIELD, image.getField(Image._I_IMAGE).getDatabaseFieldName());
            pic.setValue(Picture._C_FILENAME, filename);
            pic.setValue(Picture._D_IMAGE, img);
            pic.isNew(true);
            
            image.setValue(Image._I_IMAGE, pic);
            imagePlus.getOriginalFileInfo().getBytesPerPixel();
            
            //int bitDepath = imagePlus.getBitDepth();
            //int nrOfChannels = imagePlus.getNChannels();

            File jpegFile = new File(filename); 
            
            try {
                Metadata metadata = JpegMetadataReader.readMetadata(jpegFile);
                if (metadata.containsDirectory(ExifDirectory.class)) {
                    Directory exifDirectory = metadata.getDirectory(ExifDirectory.class);
                    
                    String camera = exifDirectory.getString(ExifDirectory.TAG_MODEL);
                    int compression = exifDirectory.getInt(ExifDirectory.TAG_COMPRESSION);
                    Date date = exifDirectory.getDate(ExifDirectory.TAG_DATETIME);
                    String description = exifDirectory.getString(ExifDirectory.TAG_IMAGE_DESCRIPTION);
    
                    image.setValue(Image._N_DATE, date);
                    image.setValue(Image._B_DESCRIPTION, description);
                    image.setValue(Image._Q_CAMERA, camera);
                    image.setValue(Image._O_COMPRESSION, new Long(compression));
                }
                
                if (metadata.containsDirectory(IptcDirectory.class)) {
                    Directory iptcDirectory = metadata.getDirectory(IptcDirectory.class);
                    
                    String city = iptcDirectory.getString(IptcDirectory.TAG_CITY);
                    String country = iptcDirectory.getString(IptcDirectory.TAG_COUNTRY_OR_PRIMARY_LOCATION);
                    String state = iptcDirectory.getString(IptcDirectory.TAG_PROVINCE_OR_STATE);
                    
                    String location = "";
                    location += country != null ? country + ", " : "";
                    location += state != null ? state  + ", " : "";
                    location += city != null ? city : "";
                    
                    image.setValue(Image._P_PLACE, location);
                }
                
                imagePlus.flush();
                
            } catch (JpegProcessingException jpe) {}
            
        } catch (Exception exp) {
            throw new ParseException(exp);
        }
        
        return image;
    }
}
