/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.fileimporters;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import net.sf.dc.console.windows.fileimport.MusicFileImportDialog;
import net.sf.dc.console.windows.messageboxes.MessageBox;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.helpers.MusicAlbum;
import net.sf.dc.core.objects.helpers.MusicTrack;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.util.Converter;
import net.sf.dc.util.Directory;
import net.sf.dc.util.StringUtils;

public class MusicAlbumImporter extends FileImporter {

    private static final int _DONOTUSE = 0;
    private static final int _DIRISALBUM = 1;
    private static final int _DIRISARTISTS = 2;
    private static final int _DIRISARTIST_SUBDIRISALBUM = 3;
    
    protected final Collection albums = new ArrayList();
    
    public MusicAlbumImporter() {
        super(DcModules._MUSICALBUM);
    }

    public String[] getSupportedExtensions() {
        return new String[] {"mp3", "ogg", "mp4", "flac", "flc", "ape", "asf", "wav", "mpc"};
    }

    public boolean allowReparsing() {
        return true;
    }    
    
    public void showUI() {
        MusicFileImportDialog dlg = new MusicFileImportDialog(this);
        dlg.setVisible(true);
    }
    
    public DcObject parse(String filename) throws ParseException {
        DcObject dco = parse(filename, _DONOTUSE, filename);
        albums.clear();
        return dco;
    }
    
    public DcObject parse(String filename, int directoryUsage, String sourceDir) throws ParseException {
        MusicAlbum ma = new MusicAlbum();

        try {
            MusicFile musicFile = new MusicFile(filename);
            
            String artist = musicFile.getArtist();
            String album = musicFile.getAlbum();
            
            String s = null; 
            switch (directoryUsage) {
            case _DONOTUSE:
                break;
            case _DIRISALBUM:
                s = getDirectoryName(filename, sourceDir, 0);
                album = s != null ? s : album;
                break;
            case _DIRISARTISTS:
                s = getDirectoryName(filename, sourceDir, 0);
                artist = s != null ? s : artist;
                break;
            case _DIRISARTIST_SUBDIRISALBUM:
                s = getDirectoryName(filename, sourceDir, 0);
                artist = s != null ? s : artist;

                s = getDirectoryName(filename, sourceDir, 1);
                album = s != null ? s : album;

                break;
            }

            if (album != null) {
                album = Converter.databaseValueConverter(album);
                ma = getAlbum(album, albums);
            }
            
            if (ma == null) {
                ma = new MusicAlbum();
                ma.setValue(MusicAlbum._A_TITLE, album);
                DataManager.createReference(ma, MusicAlbum._F_ARTISTS, artist);
            } 
            
            DcObject genre = DataManager.createReference(ma, MusicAlbum._G_GENRES, musicFile.getGenre());
            if (ma != null) {
                MusicTrack mt = new MusicTrack();
                
                mt.setValue(MusicTrack._K_QUALITY, new Long(musicFile.getBitrate()));
                mt.setValue(MusicTrack._J_PLAYLENGTH, new Long(musicFile.getLength()));
                mt.setValue(MusicTrack._L_ENCODING, musicFile.getEncodingType());
                mt.setValue(MusicTrack._A_TITLE, musicFile.getTitle());
                mt.setValue(MusicTrack._C_YEAR, musicFile.getYear());
                mt.setValue(MusicTrack._F_TRACKNUMBER, new Long(musicFile.getTrack()));
                mt.setValue(MusicTrack._I_FILENAME, filename);
                mt.setValue(MusicTrack._N_FILESIZE, new Long(getFileSize(filename)));

                if (genre != null)
                    DataManager.createReference(mt, MusicTrack._H_GENRES, genre);
                
                DataManager.createReference(mt, MusicTrack._G_ARTIST, artist);                

                ma.addChild(mt);
            }
            
        } catch (Exception exp) {
            throw new ParseException(exp);
        }

        return ma;
    }     
    
    private MusicAlbum getAlbum(String title, Collection albums) {
        if (title == null)
            return null;
        
        for (Iterator iter = albums.iterator(); iter.hasNext(); ) {
            MusicAlbum ma = (MusicAlbum) iter.next();
            
            String albumName = (String) ma.getValue(MusicAlbum._A_TITLE);
            albumName = albumName == null ? "" : StringUtils.normalize(albumName);
            if (albumName.toLowerCase().equals(StringUtils.normalize(title.toLowerCase())))
                return ma;
        }
        
        return null;
    }    
    
    private String getDirectoryName(String filename, String sourceDir, int pos) {
        String dir = filename.indexOf(sourceDir) > -1 ?
                     filename.substring(filename.indexOf(sourceDir) + sourceDir.length(), filename.length()) : filename;
        
        StringTokenizer st = new StringTokenizer(dir, File.separator);
        
        LinkedList ll = new LinkedList();
        while (st.hasMoreElements()) {
            ll.add(st.nextToken());
        }
        
        String s = null;
        if (pos + 1 < ll.size()) {
            try {
                s = (String) ll.get(pos);   
            } catch (Exception ignore) {}
        }
        
        return s;
    }    
    
    public void parse(IFileImportClient listener, 
            boolean isVerficationNeeded,
            boolean isRecursiveSearch,
            String sourceDir) throws Exception {

        Batch batch = new Batch(listener, this, isVerficationNeeded, isRecursiveSearch, sourceDir);
        batch.start();
    }
    
    private class Batch extends Thread {
        
        private IFileImportClient listener;
        private FileImporter importer;
        
        private boolean isVerficationNeeded;
        private boolean isRecursiveSearch;
        private String sourceDir;
        
        public Batch(IFileImportClient listener, 
                     FileImporter importer,
                     boolean isVerficationNeeded,
                     boolean isRecursiveSearch,
                     String sourceDir) throws Exception {
            
            this.listener = listener;
            this.importer = importer;
            this.isVerficationNeeded = isVerficationNeeded;
            this.isRecursiveSearch = isRecursiveSearch;
            this.sourceDir = sourceDir;
            
            if (sourceDir == null || sourceDir.trim().length() == 0)
                throw new Exception(DcResources.getText("msgSelectDirectory"));
        }
        
        public void run() {
            try {
                listener.addMessage(DcResources.getText("msgImportStarts"));
                listener.addMessage(DcResources.getText("msgImportReadingDirs"));
                
                File file = new File(sourceDir);
                Collection files = new ArrayList();
                if (file.isFile()) {
                    files.add(sourceDir);
                } else {
                    files.addAll(Directory.read( 
                            sourceDir, isVerficationNeeded, isRecursiveSearch, true, getSupportedExtensions(), false));
                }

                DataCrow.mainFrame.setSelectedTab(net.sf.dc.console.MainFrame._INSERTTAB);
                parse(files);
            } catch (Exception exp) {
                listener.addMessage("msgProcessCrash");
                DcLog.add(exp);
                new MessageBox(DcResources.getText("msgProcessCrash"), MessageBox._ERROR);
            }
        }    
        
        public void parse(Collection filenames) {
            try {
                listener.addMessage(DcResources.getText("msgImportFoundXResults", "" + filenames.size()));
                listener.initProgressBar(filenames.size());
                int counter = 1;
                
                DcModule module = DcModules.getCurrent();
                for (   Iterator iter = filenames.iterator(); iter.hasNext() && !listener.cancelled(); counter++) {

                    String filename = iter.next().toString();
                    listener.addMessage(DcResources.getText("msgProcessingFileX", filename));
                    
                    DcObject dco = null;
                    try {
                        dco = ((MusicAlbumImporter) importer).parse(
                                                    filename, 
                                                    ((MusicFileImportDialog) listener).getDirectoryUsage(), 
                                                    sourceDir);
                    } catch (ParseException pe) {
                        listener.addError(DcResources.getText("msgCouldNotReadInfoFrom", filename));
                    }
                    
                    if (dco != null) {
                        dco.setValue(MusicAlbum._I_STORAGEMEDIUM, listener.getStorageMedium());
                        dco.setValue(MusicAlbum._D_LOCATION, listener.getLocationName());

                        dco.applyTemplate();
                        
                        if (!albums.contains(dco))
                            albums.add(dco);
                        
                        try {
                            sleep(200);
                        } catch (Exception ignore) {}
                    }
                    
                    listener.updateProgressBar(counter);
                }
                
                if (listener.useOnlineServices())
                    onlineUpdate();

                module.getCurrentInsertView().add(albums);
                
            } finally {
                filenames.clear();
                listener.addMessage(DcResources.getText("msgImportStops"));
                listener.finish();
                albums.clear();
            }
        }
        
        private void onlineUpdate() {
            if (listener.useOnlineServices()) {
                int counter = 1;
                listener.initProgressBar(albums.size());
                
                for (Iterator iter = albums.iterator(); iter.hasNext() && !listener.cancelled(); counter++) {
                    MusicAlbum ma = (MusicAlbum) iter.next();
                    
                    listener.addMessage(DcResources.getText("msgSearchingOnlineFor", "" + ma));
                    
                    MusicFile musicFile = new MusicFile();
                    musicFile.onlineUpdate(ma, listener.getServer(), listener.getRegion());
                    listener.updateProgressBar(counter);
                }
            }
        }        
    }    
}
