/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.onlinesearch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.DcRepository;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.util.StringUtils;

public class OnlineSearchHelper implements IOnlineSearchClient {
    
    private final int module;
    
    private int maximum = 2;
    private DcServer server;
    private DcRegion region;
    private DcSearchMode mode;
    
    private Collection result = new ArrayList();
    
    public OnlineSearchHelper(int module) {
        this.module = module;
    }
    
    public void setServer(DcServer server) {
        this.server = server;
    }

    public void setRegion(DcRegion region) {
        this.region = region;
    }

    public void setMode(DcSearchMode mode) {
        this.mode = mode;
    }

    public void setMaximum(int maximum) {
        this.maximum = maximum;
    }
    
    public DcObject query(DcObject base, String query, int[] matcherFieldIdx) {
        DcServer server = getServer();
        DcRegion region = getRegion(server);
        
        OnlineSearchTask task = server.getTask(this, query, getSearchMode(server), region);
        task.setMaximum(maximum);
        task.run();
        
        return getMatchingItem(base, matcherFieldIdx);
    }
    
    public Collection query(String query) {
        DcServer server = getServer();
        DcRegion region = getRegion(server);
        
        OnlineSearchTask task = server.getTask(this, query, getSearchMode(server), region);
        task.setMaximum(maximum);
        task.run();
        
        Collection c = new ArrayList();
        c.addAll(result);
        return c;
    }    

    private DcServer getServer() {
        OnlineSearch os = DcModules.get(module).getOnlineSearch();
        DcServer defaultSrv = 
            os.getServer(DcModules.get(module).getSettings().getString(DcRepository.modulesettings.stOnlineSearchDefaultServer));
        
        DcServer server = this.server != null ? this.server : defaultSrv;
        return server == null ? (DcServer) os.getServers().toArray()[0] : server;
    }

    private DcRegion getRegion(DcServer server) {
        OnlineSearch os = DcModules.get(module).getOnlineSearch();
        DcRegion region = this.region != null ? this.region : os.getDefaultRegion();
        if (region != null) {
            boolean partOfServer = false;
            for (Iterator iter = server.getRegions().iterator(); iter.hasNext(); ) {
                DcRegion serverRegion = (DcRegion) iter.next();
                if (serverRegion.getCode().equals(region.getCode())) {
                    region = serverRegion;
                    partOfServer = true;
                }
            }
            if (!partOfServer) region = null;
        } 
            
        return region == null && server.getRegions().size() > 0 ? (DcRegion) server.getRegions().toArray()[0] : region;
    }
    
    private DcSearchMode getSearchMode(DcServer server) {
        OnlineSearch os = DcModules.get(module).getOnlineSearch();
        DcSearchMode mode = this.mode == null ? os.getDefaultSearchMode() : this.mode;
        
        if (mode != null) {
            boolean partOfServer = false;
            for (Iterator iter = server.getModes().iterator(); iter.hasNext(); ) {
                DcSearchMode serverMode = (DcSearchMode) iter.next();
                if (serverMode.getDisplayName().equals(mode.getDisplayName())) {
                    mode = serverMode;
                    partOfServer = true;
                }
            }
            if (!partOfServer) mode = null;
        } 
        
        if (mode == null) {
            for (Iterator iter = server.getModes().iterator(); iter.hasNext(); ) {
                DcSearchMode sm = ((DcSearchMode) iter.next());
                mode = sm.keywordSearch() ? sm : mode;
            }
        }
            
        return mode;
    }
    
    private DcObject getMatchingItem(DcObject base, int[] matcherFieldIdx) {
        for (Iterator iter = result.iterator(); iter.hasNext(); ) {
            DcObject dco = (DcObject) iter.next();
            boolean match = true;
            for (int i = 0; i < matcherFieldIdx.length; i++) {
                Object o1 = base.getValue(matcherFieldIdx[i]);
                Object o2 = dco.getValue(matcherFieldIdx[i]);
                
                String value1 = o1 == null || o1.toString().equals("-1") ? "" : StringUtils.normalize(o1.toString().trim());
                String value2 = o2 == null || o2.toString().equals("-1") ? "" : StringUtils.normalize(o2.toString().trim());
                
                match &= value1.equals(value2);
            }
            
            if (match)
                return dco;
        }
        
        return null;
    }
    
    public void clear() {
        server = null;
        region = null;
        result.clear();
        result = null;
    }
    
    public void addError(Throwable t) {
        DcLog.add(t);
    }

    public void addError(String message) {
        DcLog.add(message, DcLog._ERROR);
    }

    public void addMessage(String message) {}

    public void addObject(DcObject dco) {
        if (result != null)
            result.add(dco);
    }

    public void addWarning(String warning) {}

    public DcObject getDcObject() {
        return null;
    }

    public DcModule getModule() {
        return DcModules.get(module);
    }

    public void processed(int i) {}

    public void processing() {}

    public void processingTotal(int i) {}

    public int resultCount() {
        return result.size();
    }

    public void stopped() {}
}
