/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.onlinesearch.amazon;

import java.io.PrintWriter;
import java.net.URL;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.objects.DcAssociate;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.helpers.AudioCD;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.onlinesearch.HttpConnection;
import net.sf.dc.onlinesearch.IOnlineSearchClient;

public class AmazonCoverSource {

    private final IOnlineSearchClient listener;
    private final AudioCD audioCD;

    private boolean keepOnRunning = true;

	private static final String SOAP_ACTION = "SearchArtistRequest";
	private static final String SOAP_URL = "http://soap.amazon.com/onca/soap3";
	private static final Pattern SOAP_PATTERN = Pattern.compile(".+?ProductName.+?>(.+?)</ProductName.+?Artist .+?>(.+?)</Artist.+?ImageUrlSmall.+?>(.+?)</ImageUrlSmall.+?ImageUrlMedium.+?>(.+?)</ImageUrlMedium.+?ImageUrlLarge.+?>(.+?)</ImageUrlLarge.*");

    private static final Pattern[] stripPatterns = {
            Pattern.compile(" ?\\(dis[ck] \\d+?\\)"),
            Pattern.compile("^(the|a|an|el|le|les|la|los|die|der|das|den) "),
            Pattern.compile(" (the|a|an|el|le|les|la|los|die|der|das|den) "),
            Pattern.compile("('|~|\\!|@|#|\\$|%|\\^|\\*|_|\\[|\\{|\\]|\\}|\\||\\\\|;|:|`|\"|<|,|>|\\.|\\?|/|&)")};

	public AmazonCoverSource(IOnlineSearchClient listener, AudioCD audioCD) {
        this.listener = listener;
        this.audioCD = audioCD;
    }

    public void stopSearch() {
        keepOnRunning = false;
    }

	private String getSoapRequest(String artist, String album) {
		StringBuffer sb = new StringBuffer();
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		sb.append("<SOAP-ENV:Envelope SOAP-ENV:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:SOAP-ENC=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:xsi=\"http://www.w3.org/1999/XMLSchema-instance\" xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xsd=\"http://www.w3.org/1999/XMLSchema\">");
		sb.append("<SOAP-ENV:Body>");
		sb.append("<ns1:ArtistSearchRequest xmlns:ns1=\"http://soap.amazon.com\" SOAP-ENC:root=\"1\">");
		sb.append("<v1>" + "<sort xsi:type=\"xsd:string\">+pmrank</sort>");
		sb.append("<artist xsi:type=\"xsd:string\">").append(artist).append("</artist>");
		sb.append("<devtag xsi:type=\"xsd:string\">D1R5WXCELBX4NF</devtag>");
		sb.append("<tag xsi:type=\"xsd:string\">accessonline</tag>");
		sb.append("<mode xsi:type=\"xsd:string\">music</mode>");
		sb.append("<keywords xsi:type=\"xsd:string\">").append(album).append("</keywords>");
		sb.append("<type xsi:type=\"xsd:string\">lite</type>");
		sb.append("<page xsi:type=\"xsd:int\">1</page>" + "</v1>");
		sb.append("</ns1:ArtistSearchRequest>" + "</SOAP-ENV:Body>");
		sb.append("</SOAP-ENV:Envelope>");

		return sb.toString();
	}

	private String sendSoapRequest(String request) throws Exception {
		//Set up the request
        URL url = new URL(SOAP_URL);
		HttpConnection connection = new HttpConnection(url);
		connection.setContentLength(request.length());
		connection.setToSoapRequest(SOAP_ACTION);

		//Send our xml request
        //ui.addMessage(DcResources.getText("msgSendingSearchAmazon"));
		PrintWriter out = new PrintWriter(connection.getOutputStream());
		out.println(request);
		out.close();

		//Receive the answer
        String answer = connection.getString();
		return answer;
	}

	public void search() {
        String album = audioCD.getValue(AudioCD._A_TITLE).toString();

        // TODO: this can never work!
        Collection artists = (Collection) audioCD.getValue(AudioCD._F_ARTIST);
        String artist = "";
        if (artists != null) {
            for (Iterator iter = artists.iterator(); iter.hasNext(); ) {
                artist += (artist.length() > 0 ? "," : "");
                artist += (((DcObject) iter.next()).getValue(DcAssociate._A_NAME));
            }
        }

		//Transform the artist and album strings to better search
		artist = normalize(artist);
		album = normalize(album);

        String answer;
        try {
        	answer = sendSoapRequest(getSoapRequest(artist, album));
        	answer = answer.replaceAll("\n", "");
        	answer = answer.replaceAll("&amp;", "&");
        } catch (Exception exp) {
        	//ui.addMessage(DcResources.getText("msgCouldNotConnectAmazon", exp.getMessage()));
            DcLog.add(exp);
            return;
        }

		int nextSubIndex = 0;
        Vector vURLs = new Vector(0);
		while (answer.length() > nextSubIndex && keepOnRunning) {
			Matcher matcher = SOAP_PATTERN.matcher(answer.substring(nextSubIndex));
			if (matcher.matches()) {
				//Retrieve the best url (non-empty image)
				String imageURL = "";
				for (int i = 5; i >= 3; i--) {
					imageURL = matcher.group(i);
					if (!imageURL.equals("")) {
                        vURLs.add(imageURL);
                        break;
                    }
				}
				nextSubIndex += matcher.end(matcher.groupCount());
			} else {
				break;
            }
		}
        retrieveImages(vURLs);
	}

    private void retrieveImages(Collection urls) {
        int results = urls.size();
        if (results == 0)
            return;

        int counter = 0;
        for (Iterator iter = urls.iterator(); iter.hasNext(); counter++) {
            try {
                URL url = new URL(iter.next().toString());
                HttpConnection connection = new HttpConnection(url);
                // check if this image is empty or not
                if (connection.getContentLength() != 807) {
                    // In case more then one image is found, add the item
                    // again with the new image
                    audioCD.setValue(AudioCD._H_PICTUREFRONT, connection.getBytes());
                    break;
                } 
            } catch (Exception exp) {
                listener.addMessage(DcResources.getText("msgErrorWithImage", exp.getMessage()));
                DcLog.add(exp);
            }
        }
    }

    public static String normalize(String string) {
        String s = string.toLowerCase();
        for (int i = 0; i < stripPatterns.length; i++) {
            Matcher ma = stripPatterns[i].matcher(s);
            s = ma.replaceAll("");
        }
        return s;
    }
}
