/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.onlinesearch.imdb;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;

import net.sf.dc.core.DcRepository;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcAssociate;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.helpers.Movie;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.onlinesearch.DcRegion;
import net.sf.dc.onlinesearch.DcSearchMode;
import net.sf.dc.onlinesearch.DcServer;
import net.sf.dc.onlinesearch.HttpConnection;
import net.sf.dc.onlinesearch.IOnlineSearchClient;
import net.sf.dc.util.StringUtils;

public class ImdbMovieSearch extends ImdbSearch {
    
    public ImdbMovieSearch(IOnlineSearchClient listener, 
                           String searchString, 
                           DcServer server, 
                           DcRegion region,
                           DcSearchMode mode) {

        super(listener, searchString, server, region, mode);
    }
    
    public DcObject getDcObject(String id) throws Exception {
        URL url = new URL("http://www.imdb.com/title/tt" + id);
        return getDcObject(url);
    }      
    
    public DcObject getDcObject(URL url) throws Exception {
        Movie movie = new Movie();
        
        HttpConnection connection = new HttpConnection(url);
        String webpage = connection.getString();
        connection.close();

        String s = url.toString();
        String id = s.substring(s.indexOf("/title/tt") + 9, s.length());
        
        movie.setValue(Movie._G_WEBPAGE, getAddress() + "/title/tt" + id);
        movie.setValue(DcObject._SYS_SERVICEURL, s);
        
        setTitleAndYear(webpage, movie);
        setDescription(id, movie);
        setDirector(webpage, movie);
        setGenres(webpage, movie);
        setRating(webpage, movie);
        setCountry(webpage, movie);
        setRuntime(webpage, movie);
        setCertification(webpage, movie);
        setActors(id, movie);
        setPoster(webpage, id, movie);

        return movie;
    }
    
    public Collection getIDs() {
        Collection keys = new ArrayList();

        try {
            listener.addMessage(DcResources.getText("msgConnectingToServer", getAddress()));
            URL url = new URL(getAddress() + "/Find?" + getQuery());

            HttpConnection connection = new HttpConnection(url);
            String sWebPage = connection.getString();
            connection.close();

            String sTemp = sWebPage;
            if (sTemp.indexOf("/title/tt") > -1) {
                int temp = 0;
                int counter = 0;
                while (temp != -1 && counter < getMaximum()) {
                    counter++;
                    temp = sTemp.indexOf("/title/tt");
                    if (temp != -1) {
                        sTemp = sTemp.substring(temp + 9, sTemp.length());
                        String key  = sTemp.substring(0, 7);
                        if (!keys.contains(key))
                            keys.add(key);
                    }
                }
            }
        } catch (Exception e) {
            listener.addError(DcResources.getText("msgCouldNotConnectToImdb"));
        }
        return keys;
    }
    
    /**
     * Queries Imdb for the specific person (using the imdb id). Then it checks if the
     * person exists already in the database. If not, it is created. The new or existing
     * person is then returned.
     * @throws Exception
     */
    private DcObject getConcretePerson(String imdbId, String name, int module) throws Exception {
        DcObject person = DataManager.getObjectForDisplayValue(module, name);
        
        if (person == null) {
            if (DcModules.get(DcModules._MOVIE).getSettings().getBoolean(DcRepository.modulesettings.stOnlineSearchSubItems)) {
                ImdbPerson imdbPerson = new ImdbPerson(module);
                person = imdbPerson.get(imdbId);
                imdbPerson.clear();
            } else {
                person = DcModules.get(module).getDcObject();
                person.setValue(DcAssociate._A_NAME, name);
            }
            
            person.setIDs();
        }
        return person;
    }    
    
    private void setTitleAndYear(String webpage, Movie movie) {
        String title = StringUtils.getValueBetween("<title>", "</title>", webpage);
        movie.setValue(Movie._A_TITLE, StringUtils.convertHttpString(title));
        movie.setValue(Movie._C_YEAR,  StringUtils.getValueBetween("(", ")", title));
    }
    
    private void setPoster(String webpage, String id, Movie movie) {
        byte[] poster = getPoster(id);
        if (poster == null)
            poster =  getThumbnailImage(webpage);

        movie.setValue(Movie._X_PICTUREFRONT, poster);
    }
    
    private void setCertification(String webpage, Movie movie) {
        String certification = getPart(_CERTIFICATION, "<br>", webpage);
        if (certification.indexOf("</div>") > -1)
            certification = certification.substring(0, certification.indexOf("</div>"));
        
        int certificationIdx = certification.indexOf("\">");
        String s = "";
        while (certificationIdx != -1) {
            certification = certification.substring(certificationIdx + 2);
            s += certification.substring(0, certification.indexOf("<")) + (",");
            certificationIdx = certification.indexOf("\">");
        }
        certification = s.length() > 1 ? s.substring(0, s.length() -1) : s;
        movie.setValue(Movie._3_CERTIFICATION, StringUtils.convertHttpString(certification));
    }
    
    private void setActors(String movieId, Movie movie) throws Exception {
        URL url = new URL("http://www.imdb.com/title/tt" + movieId + "/fullcredits#cast");
        HttpConnection connection = new HttpConnection(url);
        String webpage = connection.getString();
        connection.close();
        
        int idx = webpage.indexOf("<table class=\"cast\">");

        if (idx == -1) return;

        String s = webpage.substring(idx, webpage.length());
        if (webpage.indexOf("</table>") > -1)
            s = s.substring(0, s.indexOf("</table>"));
        
        while (s.indexOf("/name/nm") > -1) {
            String name = StringUtils.getValueBetween("/name/nm", "</a>", s);
            name = name.substring(name.indexOf(">") + 1, name.length());
            
            if (name.indexOf("img src") > -1) {
                s = s.substring(s.indexOf("/name/nm") + 8, s.length());
                continue;
            }
            
            String id = StringUtils.getValueBetween("/name/nm", "/", s);
            if (id != null && id.trim().length() > 0) {
                try {
                    DcObject person = getConcretePerson(id, name, DcModules._ACTOR);
                    DataManager.createReference(movie, Movie._I_ACTORS, person);
                } catch (Exception e) {
                    DcLog.add(e);
                }
            }
            s = s.substring(s.indexOf("/name/nm") + 8, s.length());
        }
    }
    
    private void setCountry(String webpage, Movie movie) {
        String country = getPart(_COUNTRY, "/a>", webpage);
        country = StringUtils.getValueBetween("/" + '"' + '>', "<", country);
        movie.setValue(Movie._K_COUNTRY, StringUtils.convertHttpString(country));        
    }
    
    private void setRating(String webpage, Movie movie) {
        int ratingIdx = indexOf(_USERRATING, webpage);
        if (ratingIdx != -1) {
            String sTemp = webpage.substring(ratingIdx + 20);
            if (sTemp.indexOf("</b>") > -1) {
                try {
                    String vote = sTemp.substring(1, sTemp.indexOf("</b>"));
                    vote = vote.substring(0, vote.indexOf("/"));
                    int value = Math.round(Float.valueOf(vote).floatValue() / 2);
                    movie.setValue(Movie._E_RATING, new Long(value));
                } catch (Exception exp) {}
            }
        }        
    }

    private void setDirector(String webpage, Movie movie) {
        String director = getPart(_DIRECTEDBY, "</table>", webpage);
        
        String name = StringUtils.getValueBetween("/name/nm", "</a>", director);
        name = name.substring(name.indexOf(">") + 1, name.length());
        String id = StringUtils.getValueBetween("name/nm", "/", director);
        
        if (id == null || id.trim().length() == 0) return;

        try {
            DcObject person = getConcretePerson(id, name, DcModules._DIRECTOR);
            DataManager.createReference(movie, Movie._J_DIRECTOR, person);
        } catch (Exception exp) {
            DcLog.add(exp);
        }
    }
    
    private void setGenres(String webpage, Movie movie) {
        String genre = getPart(_GENRE, "</div>", webpage);
        int genreIdx = genre.indexOf("/\">");
        while (genreIdx != -1) {
            genre = genre.substring(genreIdx + 3);
            DataManager.createReference(movie, Movie._H_GENRES, genre.substring(0, genre.indexOf("<")).trim());
            genreIdx = genre.indexOf("/\">");
        }
    }
    
    private void setRuntime(String webpage, Movie movie) {
        int pos = indexOf(_RUNTIME, webpage);
        
        if (pos == -1) return;

        String runtime = "";
        String tmp = webpage.substring(pos + 13);
        if (tmp.indexOf("<") < tmp.indexOf(" ")) {
            runtime = tmp.substring(0, tmp.indexOf("<"));
        } else {
            pos = tmp.indexOf("\t");
            if (pos != -1)
                runtime = tmp.indexOf(" ") < pos ? 
                            tmp.substring(0, tmp.indexOf(" ")) : tmp.substring(0, tmp.indexOf("\t"));
            else
                runtime = tmp.substring(0, tmp.indexOf(" "));
        }

        try {
            pos = new Integer(runtime.trim()).intValue() * 60;
            runtime = "" + pos;
            movie.setValue(Movie._L_PLAYLENGTH, runtime);
        } catch (NumberFormatException e) {}
    }

    private void setDescription(String id, Movie movie) {
        try {
            URL url = new URL(getAddress() + "/Plot?" + id);
            HttpConnection connection = new HttpConnection(url);
            String plot = connection.getString();
            connection.close();

            plot = StringUtils.getValueBetween("<p class=\"plotpar\">", "<i>", plot);
            if (plot != null && plot.trim().length() > 0) {
                plot = StringUtils.convertHttpString(plot);
                movie.setValue(Movie._B_DESCRIPTION, plot);
            }
            
        } catch (Exception ignore) {}
    }

    private byte[] getPoster(String id) {
        try {
            URL url = new URL(getAddress() + "/title/tt" + id + "/posters");
            HttpConnection connection = new HttpConnection(url);
            String picture = connection.getString();
            connection.close();

            int idx = picture.indexOf("Poster images");
            if (idx > -1) {
                picture = picture.substring(idx);
                picture = StringUtils.getValueBetween("background=" + '"', '"' +">", picture);

                String picUrl = "posters.imdb.com";
                int idx2 = picture.indexOf(picUrl);
                if (idx2 != -1) {
                    String picUrl2 = picture.substring(idx2 + picUrl.length());
                    url = new URL("http://posters.imdb.com" + picUrl2);
                    connection = new HttpConnection(url);
                    byte[] image = connection.getBytes();

                    return image.length < 100 ? null : image;
                }
            }
        } catch (Exception exp) {
            listener.addMessage(DcResources.getText("msgPictureCouldNotBeRetrieved"));
        }

        return null;
    }
    
    private byte[] getThumbnailImage(String webpage) {
        byte[] poster = null;

        int pos = webpage.indexOf("<a name=\"poster\"");
        if (pos !=-1) {
            String tmp = webpage.substring(pos);
            try {
                String picture = StringUtils.getValueBetween("src=\"" , "\"", tmp);
                URL url = new URL(picture);
                HttpConnection connection = new HttpConnection(url);
                poster = connection.getBytes();
            } catch (Exception exp) {
                DcLog.add(exp);
            }
        }
        return poster;
    }
}
