/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.onlinesearch.wikipedia;

import java.net.URL;
import java.util.StringTokenizer;

import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.helpers.Movie;
import net.sf.dc.onlinesearch.DcRegion;
import net.sf.dc.onlinesearch.DcServer;
import net.sf.dc.onlinesearch.HttpConnection;
import net.sf.dc.onlinesearch.IOnlineSearchClient;
import net.sf.dc.util.StringUtils;

public abstract class WikipediaMovieSearch extends WikipediaSearch {

    protected static final int _DIRECTOR = 0;
    protected static final int _PRODUCER = 1;
    protected static final int _RELEASEDATE = 2;
    protected static final int _STARRING = 3;
    protected static final int _PLAYLENTGH = 4;
    protected static final int _COUNTRY = 5;
    protected static final int _LANGUAGE = 6;
    
    protected static final int _DESC_START = 0;
    protected static final int _DESC_END = 1;
    
    public WikipediaMovieSearch(IOnlineSearchClient listener, DcServer server, DcRegion region, String query) {
        super(listener, server, region, query);
    }
    
    public abstract String[] getTags();
    public abstract String[] getDescriptionTags();
    public abstract RemovalTag[] getRemovalTags();
    
    private boolean isMovie(String[] tags, String webpage) {
        String page = StringUtils.getValueBetween("<body", "</body>", webpage);
        page = page.replaceAll("\r", "\n");
        page = page.replaceAll("\n\n", "\n");
        
        StringTokenizer st = new StringTokenizer(page, "\n");
        String header = "";
        int count = 0;
        while (st.hasMoreElements() && count < 80) {
            header += "\n" + st.nextElement();
            count++;
        }
        
        int matched = 0;
        for (int i = 0; i < tags.length; i++)
            matched += header.toLowerCase().indexOf(tags[i]) > -1 ? 1 : 0;
        
        return matched >= getTagConstraint(); 
    }

    public int getTagConstraint() {
        return 4;
    }
    
    public void setAdditionInfo(Movie movie, String[] headers, String text, String webpage) {}
    
    public DcObject getDcObject(String link) throws Exception {
    	return getDcObject(new URL(link));
    }
    
    public DcObject getDcObject(URL url) throws Exception  {
        Movie movie = null;
        
        try {
            HttpConnection connection = new HttpConnection(url);
            String webpage = connection.getString();
            connection.close();

            String[] tags = getTags();
            String[] descTags = getDescriptionTags();
            if (isMovie(tags, webpage)) {
                movie = new Movie();

                String text = getCleanText(webpage);
                
                int idx = text.indexOf(descTags[_DESC_START]);

                StringTokenizer st = new StringTokenizer(text.substring(0, idx > 0 ? idx : text.length()), "\n");
                String[] header = new String[st.countTokens()];
                int counter = 0;
                while (st.hasMoreElements()) 
                    header[counter++] = (String) st.nextElement();
                
                movie.setValue(Movie._A_TITLE, getTitle(webpage));
                movie.setValue(Movie._B_DESCRIPTION, getDescription(text));
                movie.setValue(Movie._0_LANGUAGE, getValueForTag(header, tags[_LANGUAGE]));
                movie.setValue(Movie._G_WEBPAGE, url.toString());
                
                setDirector(movie, header, tags[_DIRECTOR]);
                setActors(movie, header, tags[_STARRING]);
                
                movie.setValue(Movie._K_COUNTRY, getValueForTag(header, tags[_COUNTRY]));
                movie.setValue(Movie._C_YEAR, getYear(header));
                movie.setValue(Movie._L_PLAYLENGTH, getPlaylength(header));
                
                setServiceInfo(movie);
                movie.setValue(Movie._SYS_SERVICEURL, url.toString());
                
                setAdditionInfo(movie, header, text, webpage);
                setImage(movie, webpage);
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
        
        return movie;
    }
    
    private String getTitle(String webpage) {
        String title = StringUtils.getValueBetween("<title>", "</title>", webpage);
        if (title.indexOf(" - Wikipedia") > -1)
            title = title.substring(0, title.indexOf(" - Wikipedia"));
        
        return title;
    }
    
    private String getDescription(String s) {
        String[] tags = getDescriptionTags();
        String tag1 = tags[_DESC_START];
        String tag2 = tags[_DESC_END];
        
        int idx = s.toLowerCase().indexOf(tag1);
        
        if (idx == -1)
            return "";
            
        String description = idx == -1 ? s : s.substring(idx + tag1.length());
        
        idx = description.toLowerCase().indexOf(tag2) > -1 ? description.toLowerCase().indexOf(tag2) :
              description.toLowerCase().indexOf(tag1) > -1 ? description.toLowerCase().indexOf(tag1) : description.length();
        description = description.substring(0, idx);

        return cleanup(getRemovalTags(), description);
    }
    
    protected String getValueForTag(String[] header, String tag) {
        for (int i = 0; i < header.length; i++) {
            String line = header[i].toLowerCase().trim();
            if (line.startsWith(tag) || line.equals(tag))
                return header[i+1];
        }
        return null;
    }

    protected void setDirector(Movie movie, String[] header, String tag) {
        String director = getValueForTag(header, tag);
        DataManager.createReference(movie, Movie._J_DIRECTOR, director);
    }
    
    protected void setActors(Movie movie, String[] header, String tag) {
        String s = getValueForTag(header, tag);
        StringTokenizer st = new StringTokenizer(s,  s.indexOf(",") > -1 ? "," : s.indexOf("  ") > -1 ? "  " : ">");
        while (st.hasMoreElements())
            DataManager.createReference(movie, Movie._I_ACTORS, st.nextElement());
    }

    private Long getYear(String[] headers) {
        String[] tags = getTags();
        String s = getValueForTag(headers, tags[_RELEASEDATE]);
        
        String nr = "";
        for (int i = 0; i < s.length(); i++) {
            nr = Character.isDigit(s.charAt(i)) ? nr + s.charAt(i) : "";
            if (nr.length() == 4)
                return new Long(nr);
        }
        return null;
    }
    
    private Long getPlaylength(String[] headers) {
        String[] tags = getTags();
        String s = getValueForTag(headers, tags[_PLAYLENTGH]);
        
        if (s == null)
            return null;
            
        String nr = "";
        for (int i = 0; i < s.length(); i++) {
            nr += Character.isDigit(s.charAt(i)) ? "" + s.charAt(i) : "";
            if (!Character.isDigit(s.charAt(i)) && nr.length() > 1)
                break;
        }
        return new Long(Integer.valueOf(nr).intValue() * 60);
    }   
    
    protected void setImage(Movie movie, String webpage) {}
}
