/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.reporting.writer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.DcRepository;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.TemplateModule;
import net.sf.dc.core.objects.DcAssociate;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcProperty;
import net.sf.dc.reporting.reports.XmlReference;

public class XmlSchemaWriter extends XmlBaseWriter {
    
    private Collection references = new ArrayList();
    
    public XmlSchemaWriter(String filename) throws IOException {
        super(filename);
    }
    
    public void create(DcObject dco) throws IOException {
        startDocument(dco);
        
        Collection handled = new ArrayList();
        if (dco.getModule().getIndex() == DcModules._MEDIA) {
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                if (module instanceof DcMediaModule && 
                    module.isTopModule() &&
                    module.getIndex() != DcModules._MEDIA &&
                  !(module instanceof TemplateModule)) {
                    
                    handle(module.getDcObject(), handled);
                }
            }
        } else {
            handle(dco, handled);
        }
        
        endDocument();
    }
    
    private void handle(DcObject dco, Collection handled) throws IOException {
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
                DcModule sm = DcModules.get(field.getReferenceIdx());
                DcObject so = sm.getDcObject();

                if (!handled.contains(so.getModule().getObjectName())) {
                    writeDco(so);
                    newLine();
                }
                
                handled.add(so.getModule().getObjectName());
            }
        }
        
        if (dco.getModule().getChild() != null) {
            writeDco(dco.getModule().getChild().getDcObject());
            newLine();
            
            handled.add(dco.getModule().getChild().getObjectName());
        }
        
        writeDco(dco);
        handled.add(dco.getModule().getObjectName());
    }
    
    private void addReference(String name, String reference) {
        XmlReference xmlReference = new XmlReference(name, reference);
        if (!references.contains(xmlReference))
            references.add(xmlReference);
    }
    
    private void writeField(DcField field) throws IOException {
        String label = getValidTag(field.getSystemName());

        if (field.getValueType() == DcRepository.datatypes._DCOBJECTCOLLECTION) {
            DcModule sm = DcModules.get(field.getReferenceIdx());
            String name = getValidTag(field.getSystemName());
            String reference = getValidTag(sm.getObjectName());

            writeLine("<xsd:element name=\"" + name + "\"/>", 3);
            addReference(name, reference);
        } else {
            String type;
            switch (field.getValueType()) {
            case DcRepository.datatypes._BIGINTEGER :
                type = "long";
                break;
            case DcRepository.datatypes._BOOLEAN :
                type = "boolean";
                break;
            case DcRepository.datatypes._DATE :
                type = "date";
                break;
            case DcRepository.datatypes._LONG :
                type = "integer";
                break;
            default:
                type = "string";
            }
            
            writeLine("<xsd:element name=\"" + label + "\" type=\"xsd:" + type + "\"/>", 3);
        }        
    }
    
    private void writeDco(DcObject dco) throws IOException {
        String baseName = getValidTag(dco.getModule().getObjectName());
        
        newLine();
        writeLine("<xsd:element name=\"" + baseName + "\" type=\"type-" + baseName + "\"/>", 1);
        writeLine("<xsd:complexType name=\"type-" + baseName + "\">", 1);
        writeLine("<xsd:sequence>", 2);
        
        if (dco instanceof DcProperty || dco instanceof DcAssociate) {
            int field = dco instanceof DcProperty ? DcProperty._A_NAME : DcAssociate._A_NAME;
            String label = getValidTag(dco.getField(field).getSystemName());
            writeLine("<xsd:element name=\"" + label + "\" type=\"xsd:string\"/>", 3);
        } else if (dco instanceof DcMediaObject) {
            
            writeField(dco.getField(DcObject._SYS_MODULE));
            for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
                DcField field = (DcField) iter.next();
                writeField(field);
            }
        }
        
        if (dco.getModule().getChild() != null) {
            String name = getValidTag(dco.getModule().getChild().getObjectNamePlural());
            String reference = getValidTag(dco.getModule().getChild().getObjectName());
            
            writeLine("<xsd:element name=\"" + name + "\"/>", 3);
            addReference(name, reference);
        }
        
        writeLine("</xsd:sequence>", 2);
        writeLine("</xsd:complexType>", 1);
    }
    
    private void startDocument(DcObject dco) throws IOException{
        writeLine("<?xml version=\"1.0\"?>", 0);
        writeLine("<xsd:schema xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" elementFormDefault=\"qualified\">", 0);
        writeLine("<xsd:element name=\"data-crow-objects\">", 1);
        writeLine("<xsd:complexType>", 1);
        writeLine("<xsd:sequence>", 2);
        writeLine("<xsd:element maxOccurs=\"unbounded\" ref=\"" + 
                    getValidTag(dco.getModule().getObjectName())  + 
                  "\"/>", 3);
        writeLine("</xsd:sequence>", 2);
        writeLine("</xsd:complexType>", 1);
        writeLine("</xsd:element>", 0);
    }
    
    private void endDocument() throws IOException {
        writeReferences();
        writeLine("</xsd:schema>", 0);
        bos.flush();
        bos.close();
    }
    
    private void writeReferences() throws IOException {
        for (Iterator iter = references.iterator(); iter.hasNext(); ) {
            newLine();

            XmlReference reference = (XmlReference) iter.next();
            
            writeLine("<xsd:element name=\"" + reference.getName() + "\">", 1);
            writeLine("<xsd:complexType>", 2);
            writeLine("<xsd:sequence>", 3);
            writeLine("<xsd:element maxOccurs=\"unbounded\" ref=\"" + reference.getReference() + "\"/>", 4);
            writeLine("</xsd:sequence>", 3);
            writeLine("</xsd:complexType>", 2);
            writeLine("</xsd:element>", 1);
        }        
    }
}