/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.synchronizers;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.console.views.View;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.db.DatabaseManager;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.ValidationException;
import net.sf.dc.core.objects.helpers.MusicAlbum;
import net.sf.dc.core.objects.helpers.MusicTrack;
import net.sf.dc.core.resources.DcResources;
import net.sf.dc.fileimporters.MusicFile;
import entagged.audioformats.exceptions.CannotReadException;

public class MusicAlbumSynchronizer extends Synchronizer {

    private DcObject dco;
    
    public MusicAlbumSynchronizer() {
        super(DcResources.getText("lblMassItemUpdate", DcModules.get(DcModules._MUSICALBUM).getObjectName()),
              DcModules._MUSICALBUM);
    }
    
    public String getHelpText() {
        return DcResources.getText("msgMusicFileMassUpdateHelp");
    }
    
    public boolean canParseFiles() {
        return true;
    }

    public boolean canUseOnlineServices() {
        return true;
    }
    
    public DcObject getDcObject() {
        return dco;
    }    
    
    public Thread getTask() {
        return new Task();
    }    
    
    private class Task extends Thread {
        public void run() {
            try {
                initialize();
                
                View view = DcModules.get(module).getCurrentSearchView();
                
                int items = view.getRows();
                Collection objects = new ArrayList();
                for (int i = 0; i < items && !isCancelled(); i++)
                    objects.add(view.getObject(i));
                
                initProgressBar(items);
                for (Iterator iterItems = objects.iterator(); iterItems.hasNext() && !isCancelled(); ) {
                    dco = (DcObject) iterItems.next();
                    
                    dco.loadChildren();
                    Collection children = dco.getChildren();

                    boolean updated = false;
                    
                    if (isReparseFiles()) {
                        for (Iterator iter = children.iterator(); iter.hasNext() && !isCancelled(); ) {
                            DcObject child = (DcObject) iter.next();
                            
                            String filename = child.getFilename();
                            
                            if (filename == null || filename.trim().length() == 0)
                                continue;
                            
                            File tst = new File(filename);
                            if (!tst.exists())
                                filename = filename.replaceAll("`", "'");
                            
                            try {
                                addMessage(DcResources.getText("msgParsing", filename));
                                
                                MusicFile musicFile = new MusicFile(filename);
        
                                dco.setValue(MusicAlbum._A_TITLE, musicFile.getAlbum());
                                dco.setValue(MusicAlbum._F_ARTISTS, musicFile.getArtist());
        
                                setValue(child, MusicTrack._K_QUALITY, new Long(musicFile.getBitrate()));
                                setValue(child, MusicTrack._J_PLAYLENGTH, new Long(musicFile.getLength()));
                                setValue(child, MusicTrack._L_ENCODING, musicFile.getEncodingType());
                                setValue(child, MusicTrack._A_TITLE, musicFile.getTitle());
                                setValue(child, MusicTrack._G_ARTIST, musicFile.getArtist());
                                setValue(child, MusicTrack._C_YEAR, musicFile.getYear());
                                setValue(child, MusicTrack._F_TRACKNUMBER, new Long(musicFile.getTrack()));
                                
                                DataManager.createReference(child, MusicTrack._H_GENRES, musicFile.getGenre());
                                
                                child.setSilence(true);
                                updated = true;
                            } catch (CannotReadException exp) {
                                child.markAsUnchanged();
                                addMessage(exp.getMessage());
                            }
                        }
                    }
                    
                    if (useOnlineService()) {
                        try {
                            MusicFile musicFile = new MusicFile();
                        
                            addMessage(DcResources.getText("msgSearchingOnlineFor", "" + dco));
                            DcObject dcoNew = dco.getModule().getOnlineSearch().query(dco);
                            
                            if (dcoNew != null) {
                                dco.setValue(MusicAlbum._J_PICTUREFRONT, dcoNew.getValue(MusicAlbum._J_PICTUREFRONT));
                                dco.setValue(MusicAlbum._K_PICTUREBACK, dcoNew.getValue(MusicAlbum._K_PICTUREBACK));
                                dco.setValue(MusicAlbum._L_PICTURECD, dcoNew.getValue(MusicAlbum._L_PICTURECD));
                                dco.setValue(MusicAlbum._SYS_SERVICEURL, dcoNew.getValue(MusicAlbum._SYS_SERVICEURL));
                                musicFile.merge(dco, dcoNew);
                                updated = true;
                            } else {
                                updated = musicFile.onlineUpdate((MusicAlbum) dco, getServer(), getRegion());    
                            }
                            
                        } catch (Exception e) {
                            DcLog.add(e);
                        }
                    }

                    updateProgressBar();
                    
                    try {
                        if (updated) {
                            dco.setSilence(true);
                            dco.saveUpdate(true);
                            
                            while (DatabaseManager.getQueueSize() > 0) {
                                try {
                                    sleep(100);
                                } catch (Exception exp) {}
                            }
                        }
                    } catch (ValidationException ve) {
                        addMessage(ve.getMessage());
                    }
                }
            } finally {
                addMessage(DcResources.getText("msgSynchronizerEnded"));
                enableAction(true);
            }  
        }
    }
}
