/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                       This file is part of Data Crow.                      *
 *       Data Crow is free software; you can redistribute it and/or           *
 *        modify it under the terms of the GNU General Public                 *
 *       License as published by the Free Software Foundation; either         *
 *              version 3 of the License, or any later version.               *
 *                                                                            *
 *        Data Crow is distributed in the hope that it will be useful,        *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             *
 *           See the GNU General Public License for more details.             *
 *                                                                            *
 *        You should have received a copy of the GNU General Public           *
 *  License along with this program. If not, see http://www.gnu.org/licenses  *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Vector;

import net.sf.dc.console.windows.messageboxes.QuestionBox;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.resources.DcResources;

/**
 * Reads all files from a specific location
 * @author Robert Jan van der Waals
 */
public class Directory {

    /**
     * Retrieves all files and directories from the given locations
     *
     * @param baseDir starting location for reading data
     * @param bConfirmation ask for confirmation for eah directory
     * @param bLeveledSearch read subdirectories
     * @param bReturnDirectories allow directory names in the result
     * @param vExtensions extensions to filter on
     * @param logInformation write messages to the screen and the log
     */
    public static Collection read(String dir,
                              boolean bConfirmation,
                              boolean bLeveledSearch,
                              boolean bReturnDirectories,
                              String[] extensions,
                              boolean logInformation) {

        String baseDir = dir;
        Collection result = new ArrayList();
        File directory = new File(baseDir);
        baseDir = directory.toString();

        if (!directory.exists()) {
            return null;
        }

        boolean useFilter = false;
        if (extensions != null && extensions.length > 0)
            useFilter = true;
        

        Vector unhandled = new Vector();
        unhandled.add(directory);
        
        File tempFile = null;

        while (unhandled.size() > 0) {
            String s = unhandled.firstElement().toString();
            String current = s.startsWith(baseDir) ? s : baseDir + (baseDir.endsWith(File.separator) ? "" : File.separator) + s;
            tempFile = new File(current);
            if (tempFile.isDirectory()) {
                boolean affirmative = true;
                if (bConfirmation) {
                    QuestionBox qb = new QuestionBox(DcResources.getText("msgImport", current),
                                        SwingUtil.getRootFrame());
                    affirmative = qb.isAffirmative();
                    
                }
                
                if ((bConfirmation && affirmative) || !bConfirmation) {
                        
                    if (logInformation) {
                        DcLog.add(
                                DcResources.getText("msgReadingFrom", current), 
                                DcLog._STANDARD);
                    }

                    String[] list = tempFile.list();
                    if (list != null) {
                        for (int j = 0; j < list.length; j++) {
                            String sTempCurrent = current + (current.endsWith(File.separator) ? "" : File.separator) + list[j];
                            if (!unhandled.contains(sTempCurrent)) {
                                tempFile = new File(sTempCurrent);
                                if (tempFile.isDirectory()) {
                                    String part = sTempCurrent.substring(baseDir.length());
                                    part = part.startsWith(File.separator) ? part.substring(1) : part;
                                    unhandled.add(part);
                                } else {
                                    if (useFilter && isValid(sTempCurrent, extensions)) {
                                        result.add(sTempCurrent);
                                    } else if (!useFilter) {
                                        result.add(sTempCurrent);
                                    }
                                }
                            }
                        }
                    }
                }  else {
                    if (bReturnDirectories) {
                        if (logInformation) {
                            DcLog.add(
                                DcResources.getText("msgRegDirName", current), 
                                DcLog._STANDARD);
                        }
                        result.add(current);
                    }
                }
            } else {
                if (useFilter && isValid(current, extensions)) {
                    result.add(current);
                } else if (!useFilter) {
                    result.add(current);
                }
            }

            unhandled.removeElementAt(0);

            if (!bLeveledSearch)
                return result;
        }

        return result;
    }

    private static boolean isValid(String file, String[] extensions) {
        String filename = file;
        int last = filename.lastIndexOf(".");
        boolean valid = false;
        if (last + 1 < filename.length()) {
            filename = filename.substring(last + 1, filename.length());
            for (int i = 0; i < extensions.length; i++) {
                if (extensions[i].equals(filename.toLowerCase())) {
                    valid = true;
                    break;
                }
            }
        }
        return valid;
    }
}