/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.util;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.HeadlessException;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.net.URL;
import java.util.Calendar;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.ImageStore;
import net.sf.dc.core.log.DcLog;
import net.sf.dc.core.resources.DcResources;

public class Utilities {

    private static Random random = new Random();
    private static final Toolkit tk = Toolkit.getDefaultToolkit();
    private static final GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    private static final GraphicsDevice gs = ge.getDefaultScreenDevice();
    private static final GraphicsConfiguration gc = gs.getDefaultConfiguration();
    
    /**
     * Returns a centered location for a window / form / dialog 
     * @param windowSize size of the window
     * @return centered location
     */
    public static Point getCenteredWindowLocation(Dimension windowSize) {
        Dimension dim = tk.getScreenSize();
        dim.height = (dim.height-windowSize.height)/2;
        dim.width = (dim.width-windowSize.width) /2;
        return new Point(dim.width, dim.height);
    }
    
    public static boolean sameImage(byte[] img1, byte[] img2) {
        boolean same = img1.length == img2.length;
        if (same) {
            for (int i = 0; i < img1.length; i++) {
                same = img1[i] == img2[i];
                if (!same)
                    break;
            }
        }
        return same;
    }    
    
    public static String fillSubtitles(String filePath, String subs) {
        String s = subs;
        try {
            File f = new File(filePath);
            File dir = f.getParentFile();
            String movieName = f.getName().substring(0,f.getName().lastIndexOf("."));
            String [] allFiles = dir.list();
            for (int i = 0; i < allFiles.length; i++){
                String name = allFiles[i];
                if (    name.startsWith(movieName) && 
                        ( name.toUpperCase().endsWith("SRT") ||
                          name.toUpperCase().endsWith("SUB") ||
                          name.toUpperCase().endsWith("SSA"))) {
                    
                    if (!s.equals("")) {
                        s =s + " ";
                    }
                    s = s + name.substring(movieName.length());
                }
            }
        } catch (Exception exp) {
            DcLog.add(exp);
        }
        
        return s;
    }
    
    private static final Calendar calendar = Calendar.getInstance();

    /**
     * Creates a unique ID. Can be used for custom IDs in the database.
     * Based on date / time + random number
     * @return unique ID as String
     */
    public static String getUniqueID() {
        long temp = random.nextLong();
        if (temp < 0l) {
            temp = temp * -1;
        }
        
        StringBuffer sb = new StringBuffer();
        sb.append(calendar.get(Calendar.MONTH) + 1);
        sb.append(calendar.get(Calendar.DAY_OF_MONTH));
        sb.append(temp);
        
        return sb.toString();
    }
    
    /**
     * Retrieved the file extension of a file
     * @param f file to get the extension from
     * @return extension or empty string
     */
    public static String getExtension(File f) {
        String name = f.getName().toLowerCase();
        int i = name.lastIndexOf( "." );
        if (i == -1) {
            return "";
        }
        return name.substring( i + 1 );
    }    
    
    public static int getIntegerValue(String s) {
        char[] characters = s.toCharArray();
        String test = "";
        for (int i = 0; i < characters.length; i++) {
            if (Character.isDigit(characters[i])) test += "" + characters[i];
        }
        
        int number = 0;
        try {
            number = Integer.valueOf(test).intValue();
        } catch (Exception ignore) {}
        
        return number;
    }
   
    /**
     * Reads the content of a file (fully)
     * @param file file to retrieve the content from
     * @return content of the file as a byte array
     * @throws Exception
     */
    public static byte[] readFile(File file) throws Exception {
        RandomAccessFile access = new RandomAccessFile(file, "r");
        byte[] byties = new byte[new Long(access.length()).intValue()];
        access.readFully(byties);
        access.close();
        return byties;        
    }

    public static ImageIcon base64ToImage(String base64) {
        byte[] bytes = Base64.decode(base64.toCharArray());
        return new ImageIcon(bytes);
    }
    
    public static byte[] getBytes(ImageIcon icon) throws Exception {
        return getBytes(icon.getImage());
    }
    
    public static byte[] getBytes(Image image) throws Exception {
    	BufferedImage bi;
    	if (image instanceof BufferedImage)
    		bi = (BufferedImage) image;
    	else 
    		bi = Utilities.toBufferedImage(new ImageIcon(image), true);
    	
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        byte[] bytes = null;
        try {
            ImageIO.write(bi, "JPG", baos);
            baos.flush();
            bytes = baos.toByteArray();
            bi.flush();
        } finally {
            baos.close();
        }
        return bytes;
    }
    
    public static void writeToFile(byte[] b, String filename) throws Exception {
        FileOutputStream fos = new FileOutputStream(filename);
        BufferedOutputStream bos = new BufferedOutputStream(fos);

        bos.write(b);
        bos.flush();
        bos.close();
    }   
    
    public static Image getScaledImage(byte[] bytes) {
        return getScaledImage(bytes, 190, 145);
    }
    
    public static Image getScaledImage(byte[] bytes, int width, int height) {
        int h = height;
        int w = width;
        
        ImageIcon icon = new ImageIcon(bytes);
        int imageWidth = icon.getIconWidth();
        int imageHeight = icon.getIconHeight();
        
        Image image;
        if (imageWidth <= w && imageHeight <= h) {
            image = icon.getImage();
        } else {
            double scaledRatio = (double) w / (double) h;
            double imageRatio = (double) imageWidth / (double) imageHeight;
            if (scaledRatio < imageRatio)
                h = (int) (w / imageRatio);
            else
                w = (int) (h * imageRatio);
            
            image = icon.getImage().getScaledInstance(w, h, BufferedImage.SCALE_SMOOTH);
        }

        ImageStore.addImage(image);
        return image;
    }    
    
    public static void writeScaledImageToFile(byte[] bytes, String filename) throws Exception {
    	writeScaledImageToFile(bytes, filename, 190, 145);
    }

    public static void writeScaledImageToFile(byte[] bytes, String filename, int w, int h) throws Exception {
    	writeScaledImageToFile(bytes, filename, true, w, h);
    }     
    
    public static void writeScaledImageToFile(byte[] bytes, String filename, boolean alpha, int w, int h) throws Exception {
        Image image = getScaledImage(bytes, w, h);
        BufferedImage bufferedImage = toBufferedImage(new ImageIcon(image), alpha, true);
        ImageIO.write(bufferedImage, "JPG", new File(filename));
        bufferedImage.flush();
    }       
    
    public static String getHexColor(Color color) {
        String hexColor = "#" + Integer.toHexString(color.getRed());
        hexColor += Integer.toHexString(color.getGreen());
        hexColor += Integer.toHexString(color.getBlue()); 
        return hexColor.toUpperCase();
    }    
    
    public static boolean isEmpty(Object o) {
        return o == null || o.toString().equals("-1") || o.toString().equals("0") || o.toString().trim().equals("") || o.toString().equals("[]");     
    }

    public static String getComparableString(Object o) {
        return Utilities.isEmpty(o) ? "" : o.toString().trim();
    }
    
    public static ImageIcon getRatingIcon(Object value) {
        int rating = value != null ? ((Number) value).intValue() : -1;
        if (rating == 0)
        	return IconLibrary._icoRating0;
        if (rating == 1)
        	return IconLibrary._icoRating1;
        if (rating == 2)
        	return IconLibrary._icoRating2;
        if (rating == 3)
        	return IconLibrary._icoRating3;
        if (rating == 4)
        	return IconLibrary._icoRating4;
        if (rating == 5)
        	return IconLibrary._icoRating5;

        return null;
    }    

    public static String getCurrentDirectory() throws Exception {
    	File fl = new File(".");
    	fl = fl.getCanonicalFile();
        return fl.toString();
    }
    
    /**
     * Gets the content of a file and converts it to a base64 string
     * @param url URL of file
     * @return base64 content of the file
     */
    public static String fileToBase64String(URL url) {
        try {
            if (url != null) {
                File file = new File(url.getFile());
                byte[] b = Utilities.readFile(file);
                file = null;
                return String.valueOf(Base64.encode(b));
            }
        } catch (Exception exp) {
            DcLog.add(DcResources.getText("msgCouldNotLoadPicture"), DcLog._ERROR);
            DcLog.add(exp);
        }
        return "";
    }
    
    public static BufferedImage toBufferedImage(ImageIcon icon, boolean flush) {
    	return toBufferedImage(icon, true, flush);
    }
    
    public static BufferedImage toBufferedImage(ImageIcon icon, boolean alpha, boolean flush) {
        Image image = icon.getImage();
    
        // Create a buffered image with a format that's compatible with the screen
        BufferedImage bi = null;
        try {
            // Determine the type of transparency of the new buffered image and create the buffered image
            int transparency = alpha ? Transparency.BITMASK : Transparency.OPAQUE;
            bi = gc.createCompatibleImage(image.getWidth(null), image.getHeight(null), transparency);
        } catch (HeadlessException e) {
        	DcLog.add(e);
        }
    
        if (bi == null) {
            int type = alpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;
            bi = new BufferedImage(image.getWidth(null), image.getHeight(null), type);
        }
    
        Graphics g = bi.createGraphics();
        g.drawImage(image, 0, 0, null);
        g.dispose();
        
        ImageStore.addImage(bi);
        ImageStore.addImage(image);
        
        if (flush) image.flush();
    
        return bi;
    }       
}
