package net.sf.ext.ij;
import java.applet.Applet;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.CharArrayWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.net.URL;
import java.util.Enumeration;
import java.util.Properties;

import net.sf.dc.core.log.DcLog;
import net.sf.ext.ij.gui.ImageCanvas;
import net.sf.ext.ij.gui.NewImage;
import net.sf.ext.ij.gui.PlotWindow;
import net.sf.ext.ij.gui.Roi;
import net.sf.ext.ij.gui.Toolbar;
import net.sf.ext.ij.io.ImportDialog;
import net.sf.ext.ij.io.OpenDialog;
import net.sf.ext.ij.plugin.Animator;
import net.sf.ext.ij.plugin.GelAnalyzer;
import net.sf.ext.ij.plugin.JpegWriter;
import net.sf.ext.ij.plugin.filter.Analyzer;
import net.sf.ext.ij.plugin.filter.Filters;
import net.sf.ext.ij.plugin.filter.ParticleAnalyzer;
import net.sf.ext.ij.process.ColorProcessor;
import net.sf.ext.ij.process.FloatBlitter;
import net.sf.ext.ij.util.Tools;

/**
This class contains the ImageJ preferences, which are 
loaded from the "IJ_Props.txt" and "IJ_Prefs.txt" files.
@see net.sf.ext.ij.ImageJ
*/
public class Prefs {

	public static final String PROPS_NAME = "imagej.properties";
	public static final String PREFS_NAME = "imagej.preferences";
	public static final String DIR_IMAGE = "dir.image";
	public static final String FCOLOR = "fcolor";
	public static final String BCOLOR = "bcolor";
	public static final String ROICOLOR = "roicolor";
	public static final String SHOW_ALL_COLOR = "showcolor";
	public static final String JPEG = "jpeg";
	public static final String FPS = "fps";
    public static final String DIV_BY_ZERO_VALUE = "div-by-zero";
    public static final String NOISE_SD = "noise.sd";
    public static final String MENU_SIZE = "menu.size";
    public static final String THREADS = "threads";
	public static final String KEY_PREFIX = ".";
 
	private static final int USE_POINTER=1, ANTIALIASING=2, INTERPOLATE=4, ONE_HUNDRED_PERCENT=8,
		BLACK_BACKGROUND=16, JFILE_CHOOSER=32, BLACK_CANVAS=128, WEIGHTED=256, 
		AUTO_MEASURE=512, REQUIRE_CONTROL=1024, USE_INVERTING_LUT=2048, ANTIALIASED_TOOLS=4096,
		INTEL_BYTE_ORDER=8192, DOUBLE_BUFFER=16384, NO_POINT_LABELS=32768, NO_BORDER=65536,
		SHOW_ALL_SLICE_ONLY=131072; 
    public static final String OPTIONS = "prefs.options";

	/** file.separator system property */
	public static String separator = System.getProperty("file.separator");
	/** Use pointer cursor instead of cross */
	public static boolean usePointerCursor;
	/** No longer used */
	public static boolean antialiasedText;
	/** Display images scaled <100% using bilinear interpolation */
	public static boolean interpolateScaledImages;
	/** Open images at 100% magnification*/
	public static boolean open100Percent;
	/** Backgound is black in binary images*/
	public static boolean blackBackground;
	/** Use JFileChooser instead of FileDialog to open and save files. */
	public static boolean useJFileChooser;
	/** Color to grayscale conversion is weighted (0.299, 0.587, 0.114) if the variable is true. */
	public static boolean weightedColor;
	/** Use black image border. */
	public static boolean blackCanvas;
	/** Point tool auto-measure mode. */
	public static boolean pointAutoMeasure;
	/** Point tool auto-next slice mode (not saved in IJ_Prefs). */
	public static boolean pointAutoNextSlice;
	/** Require control or command key for keybaord shortcuts. */
	public static boolean requireControlKey;
	/** Open 8-bit images with inverting LUT so 0 is white and 255 is black. */
	public static boolean useInvertingLut;
	/** Draw tool icons using antialiasing. */
	public static boolean antialiasedTools;
	/** Export Raw using little-endian byte order. */
	public static boolean intelByteOrder;
	/** Double buffer display of selections. */
	public static boolean doubleBuffer;
	/** Do not label multiple points created using point tool. */
	public static boolean noPointLabels = true;
	/** Disable Edit/Undo command. */
	public static boolean disableUndo;
	/** Do not draw black border around image. */
	public static boolean noBorder;
	/** Only show ROIs associated with current slice in Roi Manager "Show All" mode. */
	public static boolean showAllSliceOnly;

	static Properties ijPrefs = new Properties();
	static Properties props = new Properties(Prefs.ijPrefs);
	static String prefsDir;
	static String imagesURL;
	static String homeDir; // ImageJ folder
	static int threads;

	/** Finds and loads the ImageJ configuration file, "IJ_Props.txt".
		@return	an error message if "IJ_Props.txt" not found.
	*/
	public static String load(Object ij, Applet applet) {
		InputStream f = ij.getClass().getResourceAsStream("/"+Prefs.PROPS_NAME);
		if (applet!=null)
			return Prefs.loadAppletProps(f,applet);
		if (Prefs.homeDir==null)
			Prefs.homeDir = System.getProperty("user.dir");
		String userHome = System.getProperty("user.home");
		String osName = System.getProperty("os.name");
		if (osName.indexOf("Windows",0)>-1) {
			Prefs.prefsDir = Prefs.homeDir; //ImageJ folder on Windows
			if (Prefs.prefsDir.endsWith("Desktop"))
				Prefs.prefsDir = userHome;
		} else {
			Prefs.prefsDir = userHome; // Mac Preferences folder or Unix home dir
			if (IJ.isMacOSX())
				Prefs.prefsDir += "/Library/Preferences";
		} 
		if (f==null) {
			try {f = new FileInputStream(Prefs.homeDir+"/"+Prefs.PROPS_NAME);}
			catch (FileNotFoundException e) {f=null;}
		}
		if (f==null)
			return Prefs.PROPS_NAME+" not found in ij.jar or in "+Prefs.homeDir;
		f = new BufferedInputStream(f);
		try {Prefs.props.load(f); f.close();}
		catch (IOException e) {return("Error loading "+Prefs.PROPS_NAME);}
		Prefs.imagesURL = Prefs.props.getProperty("images.location");
		Prefs.loadPreferences();
		Prefs.loadOptions();
		return null;
	}

	/*
	static void dumpPrefs(String title) {
		IJ.log("");
		IJ.log(title);
		Enumeration e = ijPrefs.keys();
		while (e.hasMoreElements()) {
			String key = (String) e.nextElement();
			IJ.log(key+": "+ijPrefs.getProperty(key));
		}
	}
	*/

	static String loadAppletProps(InputStream f, Applet applet) {
		if (f==null)
			return Prefs.PROPS_NAME+" not found in ij.jar";
		try {
			Prefs.props.load(f);
			f.close();
		}
		catch (IOException e) {return("Error loading "+Prefs.PROPS_NAME);}
		try {
			URL url = new URL(applet.getDocumentBase(), "images/");
			Prefs.imagesURL = url.toString();
		}
		catch (Exception e) {}
		return null;
	}

	/** Returns the URL of the directory that contains the ImageJ sample images. */
	public static String getImagesURL() {
		return Prefs.imagesURL;
	}

	/** Sets the URL of the directory that contains the ImageJ sample images. */
	public static void setImagesURL(String url) {
		Prefs.imagesURL = url;
	}

	/** Returns the path to the ImageJ directory. */
	public static String getHomeDir() {
		return Prefs.homeDir;
	}

	/** Gets the path to the directory where the 
		preferences file (IJPrefs.txt) is saved. */
	public static String getPrefsDir() {
		return Prefs.prefsDir;
	}

	/** Sets the path to the ImageJ directory. */
	static void setHomeDir(String path) {
		if (path.endsWith(File.separator))
			path = path.substring(0, path.length()-1);
		Prefs.homeDir = path;
	}

	/** Finds an string in IJ_Props or IJ_Prefs.txt. */
	public static String getString(String key) {
		return Prefs.props.getProperty(key);
	}

	/** Finds an string in IJ_Props or IJ_Prefs.txt. */
	public static String getString(String key, String defaultString) {
		if (Prefs.props==null)
			return defaultString;
		String s = Prefs.props.getProperty(key);
		if (s==null)
			return defaultString;
		else
			return s;
	}

	/** Finds a boolean in IJ_Props or IJ_Prefs.txt. */
	public static boolean getBoolean(String key, boolean defaultValue) {
		if (Prefs.props==null) return defaultValue;
		String s = Prefs.props.getProperty(key);
		if (s==null)
			return defaultValue;
		else
			return s.equals("true");
	}

	/** Finds an int in IJ_Props or IJ_Prefs.txt. */
	public static int getInt(String key, int defaultValue) {
		if (Prefs.props==null) //workaround for Netscape JIT bug
			return defaultValue;
		String s = Prefs.props.getProperty(key);
		if (s!=null) {
			try {
				return Integer.decode(s).intValue();
			} catch (NumberFormatException e) {IJ.write(""+e);}
		}
		return defaultValue;
	}

	/** Looks up a real number in IJ_Props or IJ_Prefs.txt. */
	public static double getDouble(String key, double defaultValue) {
		if (Prefs.props==null)
			return defaultValue;
		String s = Prefs.props.getProperty(key);
		Double d = null;
		if (s!=null) {
			try {d = new Double(s);}
			catch (NumberFormatException e){d = null;}
			if (d!=null)
				return(d.doubleValue());
		}
		return defaultValue;
	}

	/** Finds a color in IJ_Props or IJ_Prefs.txt. */
	public static Color getColor(String key, Color defaultColor) {
		int i = Prefs.getInt(key, 0xaaa);
		if (i == 0xaaa)
			return defaultColor;
		return new Color((i >> 16) & 0xFF, (i >> 8) & 0xFF, i & 0xFF);
	}

	/** Returns the file.separator system property. */
	public static String getFileSeparator() {
		return Prefs.separator;
	}

	/** Opens the IJ_Prefs.txt file. */
	static void loadPreferences() {
		String path = Prefs.prefsDir+Prefs.separator+Prefs.PREFS_NAME;
		boolean ok =  Prefs.loadPrefs(path);
		if (!ok && IJ.isMacOSX()) {
			path = System.getProperty("user.home")+Prefs.separator+Prefs.PREFS_NAME;
			ok = Prefs.loadPrefs(path); // look in home dir
			if (ok)
				new File(path).delete();
		}

	}

	static boolean loadPrefs(String path) {
		try {
			InputStream is = new BufferedInputStream(new FileInputStream(path));
			Prefs.ijPrefs.load(is);
			is.close();
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	/** Saves user preferences in the IJ_Prefs.txt properties file. */
	public static void savePreferences() {
		try {
			Properties prefs = new Properties();
			String dir = OpenDialog.getDefaultDirectory();
			if (dir!=null)
				prefs.put(Prefs.DIR_IMAGE, dir);
			prefs.put(Prefs.ROICOLOR, Tools.c2hex(Roi.getColor()));
			prefs.put(Prefs.SHOW_ALL_COLOR, Tools.c2hex(ImageCanvas.getShowAllColor()));
			prefs.put(Prefs.FCOLOR, Tools.c2hex(Toolbar.getForegroundColor()));
			prefs.put(Prefs.BCOLOR, Tools.c2hex(Toolbar.getBackgroundColor()));
			prefs.put(Prefs.JPEG, Integer.toString(JpegWriter.getQuality()));
			prefs.put(Prefs.FPS, Double.toString(Animator.getFrameRate()));
			prefs.put(Prefs.DIV_BY_ZERO_VALUE, Double.toString(FloatBlitter.divideByZeroValue));
			prefs.put(Prefs.NOISE_SD, Double.toString(Filters.getSD()));
			if (Prefs.threads>1) prefs.put(Prefs.THREADS, Integer.toString(Prefs.threads));
			if (IJ.isMacOSX()) Prefs.useJFileChooser = false;
			Prefs.saveOptions(prefs);
			Prefs.savePluginPrefs(prefs);
			IJ.getInstance().savePreferences(prefs);
			Menus.savePreferences(prefs);
			ParticleAnalyzer.savePreferences(prefs);
			Analyzer.savePreferences(prefs);
			ImportDialog.savePreferences(prefs);
			PlotWindow.savePreferences(prefs);
			GelAnalyzer.savePreferences(prefs);
			NewImage.savePreferences(prefs);
			String path = Prefs.prefsDir+Prefs.separator+Prefs.PREFS_NAME;
			Prefs.savePrefs(prefs, path);
		} catch (Throwable t) {
			CharArrayWriter caw = new CharArrayWriter();
            DcLog.add(t);
			IJ.log(caw.toString());
			IJ.log("<Unable to save preferences>");
			IJ.wait(3000);
		}
	}

	static void loadOptions() {
		int options = Prefs.getInt(Prefs.OPTIONS, Prefs.ANTIALIASING);
		Prefs.usePointerCursor = (options&Prefs.USE_POINTER)!=0;
		//antialiasedText = (options&ANTIALIASING)!=0;
		Prefs.antialiasedText = false;
		Prefs.interpolateScaledImages = (options&Prefs.INTERPOLATE)!=0;
		Prefs.open100Percent = (options&Prefs.ONE_HUNDRED_PERCENT)!=0;
		Prefs.open100Percent = (options&Prefs.ONE_HUNDRED_PERCENT)!=0;
		Prefs.blackBackground = (options&Prefs.BLACK_BACKGROUND)!=0;
		Prefs.useJFileChooser = (options&Prefs.JFILE_CHOOSER)!=0;
		Prefs.weightedColor = (options&Prefs.WEIGHTED)!=0;
		if (Prefs.weightedColor)
			ColorProcessor.setWeightingFactors(0.299, 0.587, 0.114);
		Prefs.blackCanvas = (options&Prefs.BLACK_CANVAS)!=0;
		Prefs.pointAutoMeasure = (options&Prefs.AUTO_MEASURE)!=0;
		Prefs.requireControlKey = (options&Prefs.REQUIRE_CONTROL)!=0;
		Prefs.useInvertingLut = (options&Prefs.USE_INVERTING_LUT)!=0;
		Prefs.antialiasedTools = (options&Prefs.ANTIALIASED_TOOLS)!=0;
		Prefs.intelByteOrder = (options&Prefs.INTEL_BYTE_ORDER)!=0;
		Prefs.doubleBuffer = (options&Prefs.DOUBLE_BUFFER)!=0;
		Prefs.noPointLabels = (options&Prefs.NO_POINT_LABELS)!=0;
		Prefs.noBorder = (options&Prefs.NO_BORDER)!=0;
		Prefs.showAllSliceOnly = (options&Prefs.SHOW_ALL_SLICE_ONLY)!=0;
	}

	static void saveOptions(Properties prefs) {
		int options = (Prefs.usePointerCursor?Prefs.USE_POINTER:0) + (Prefs.antialiasedText?Prefs.ANTIALIASING:0)
			+ (Prefs.interpolateScaledImages?Prefs.INTERPOLATE:0) + (Prefs.open100Percent?Prefs.ONE_HUNDRED_PERCENT:0)
			+ (Prefs.blackBackground?Prefs.BLACK_BACKGROUND:0) + (Prefs.useJFileChooser?Prefs.JFILE_CHOOSER:0)
			+ (Prefs.blackCanvas?Prefs.BLACK_CANVAS:0) + (Prefs.weightedColor?Prefs.WEIGHTED:0) 
			+ (Prefs.pointAutoMeasure?Prefs.AUTO_MEASURE:0) + (Prefs.requireControlKey?Prefs.REQUIRE_CONTROL:0)
			+ (Prefs.useInvertingLut?Prefs.USE_INVERTING_LUT:0) + (Prefs.antialiasedTools?Prefs.ANTIALIASED_TOOLS:0)
			+ (Prefs.intelByteOrder?Prefs.INTEL_BYTE_ORDER:0) + (Prefs.doubleBuffer?Prefs.DOUBLE_BUFFER:0)
			+ (Prefs.noPointLabels?Prefs.NO_POINT_LABELS:0) + (Prefs.noBorder?Prefs.NO_BORDER:0)
			+ (Prefs.showAllSliceOnly?Prefs.SHOW_ALL_SLICE_ONLY:0);
		prefs.put(Prefs.OPTIONS, Integer.toString(options));
	}

	/** Saves the value of the string <code>text</code> in the preferences
		file using the keyword <code>key</code>. This string can be 
		retrieved using the appropriate <code>get()</code> method. */
	public static void set(String key, String text) {
		if (key.indexOf('.')<1)
			throw new IllegalArgumentException("Key must have a prefix");
		Prefs.ijPrefs.put(Prefs.KEY_PREFIX+key, text);
	}

	/** Saves <code>value</code> in the preferences file using 
		the keyword <code>key</code>. This value can be retrieved 
		using the appropriate <code>getPref()</code> method. */
	public static void set(String key, int value) {
		Prefs.set(key, Integer.toString(value));
	}

	/** Saves <code>value</code> in the preferences file using 
		the keyword <code>key</code>. This value can be retrieved 
		using the appropriate <code>getPref()</code> method. */
	public static void set(String key, double value) {
		Prefs.set(key, ""+value);
	}

	/** Saves the boolean variable <code>value</code> in the preferences
		 file using the keyword <code>key</code>. This value can be retrieved 
		using the appropriate <code>getPref()</code> method. */
	public static void set(String key, boolean value) {
		Prefs.set(key, ""+value);
	}

	/** Uses the keyword <code>key</code> to retrieve a string from the
		preferences file. Returns <code>defaultValue</code> if the key
		is not found. */
	public static String get(String key, String defaultValue) {
		String value = Prefs.ijPrefs.getProperty(Prefs.KEY_PREFIX+key);
		if (value == null)
			return defaultValue;
		else
			return value;
	}

	/** Uses the keyword <code>key</code> to retrieve a number from the
		preferences file. Returns <code>defaultValue</code> if the key
		is not found. */
	public static double get(String key, double defaultValue) {
		String s = Prefs.ijPrefs.getProperty(Prefs.KEY_PREFIX+key);
		Double d = null;
		if (s!=null) {
			try {d = new Double(s);}
			catch (NumberFormatException e) {d = null;}
			if (d!=null)
				return(d.doubleValue());
		}
		return defaultValue;
	}

	/** Uses the keyword <code>key</code> to retrieve a boolean from
		the preferences file. Returns <code>defaultValue</code> if
		the key is not found. */
	public static boolean get(String key, boolean defaultValue) {
		String value = Prefs.ijPrefs.getProperty(Prefs.KEY_PREFIX+key);
		if (value==null)
			return defaultValue;
		else
			return value.equals("true");
	}

	/** Saves the Point <code>loc</code> in the preferences
		 file as a string using the keyword <code>key</code>. */
	public static void saveLocation(String key, Point loc) {
		Prefs.set(key, loc.x+","+loc.y);
	}

	/** Uses the keyword <code>key</code> to retrieve a location
		from the preferences file. Returns null if the
		key is not found or the location is not valid (e.g., offscreen). */
	public static Point getLocation(String key) {
		String value = Prefs.ijPrefs.getProperty(Prefs.KEY_PREFIX+key);
		if (value==null) return null;
		int index = value.indexOf(",");
		if (index==-1) return null;
		double xloc = Tools.parseDouble(value.substring(0, index));
		if (Double.isNaN(xloc) || index==value.length()-1) return null;
		double yloc = Tools.parseDouble(value.substring(index+1));
		if (Double.isNaN(yloc)) return null;
		Point p = new Point((int)xloc, (int)yloc);
		Dimension screen = IJ.getScreenSize();
		if (p.x>screen.width-100 || p.y>screen.height-40)
			return null;
		else
			return p;
	}

	/** Save plugin preferences. */
	static void savePluginPrefs(Properties prefs) {
		Enumeration e = Prefs.ijPrefs.keys();
		while (e.hasMoreElements()) {
			String key = (String) e.nextElement();
			if (key.indexOf(Prefs.KEY_PREFIX) == 0)
				prefs.put(key, Prefs.ijPrefs.getProperty(key));
		}
	}

	public static void savePrefs(Properties prefs, String path) throws IOException{
		FileOutputStream fos = new FileOutputStream(path);
		BufferedOutputStream bos = new BufferedOutputStream(fos);
		prefs.store(bos, "ImageJ "+ImageJ.VERSION+" Preferences");
		bos.close();
	}
	
	/** Returns the number of threads used by PlugInFilters to process stacks. */
	public static int getThreads() {
		if (Prefs.threads==0) {
			Prefs.threads = Prefs.getInt(Prefs.THREADS, 0);
			int processors = Runtime.getRuntime().availableProcessors();
			if (Prefs.threads<1 || Prefs.threads>processors) Prefs.threads = processors;
		}
		return Prefs.threads;
	}
	
	/** Sets the number of threads (1-32) used by PlugInFilters to process stacks. */
	public static void setThreads(int n) {
		if (n<1) n = 1;
		if (n>32) n = 32;
		Prefs.threads = n;
	}

}

