package net.sf.ext.ij.io;

import java.io.File;
import java.util.Properties;

import net.sf.ext.ij.IJ;
import net.sf.ext.ij.ImagePlus;
import net.sf.ext.ij.ImageStack;
import net.sf.ext.ij.Prefs;
import net.sf.ext.ij.gui.GenericDialog;
import net.sf.ext.ij.measure.Calibration;
import net.sf.ext.ij.plugin.FolderOpener;
import net.sf.ext.ij.process.ImageProcessor;


/** This is a dialog box used to imports raw 8, 16, 24 and 32-bit images. */
public class ImportDialog {
	private String fileName;
    private String directory;
	static final String TYPE = "raw.type";
	static final String WIDTH = "raw.width";
	static final String HEIGHT = "raw.height";
	static final String OFFSET = "raw.offset";
	static final String N = "raw.n";
	static final String GAP = "raw.gap";
	static final String OPTIONS = "raw.options";
	static final int WHITE_IS_ZERO = 1;
	static final int INTEL_BYTE_ORDER = 2;
	static final int OPEN_ALL = 4;
	
    // default settings
    private static int choiceSelection = Prefs.getInt(ImportDialog.TYPE,0);
    private static int width = Prefs.getInt(ImportDialog.WIDTH,512);
    private static int height = Prefs.getInt(ImportDialog.HEIGHT,512);
    private static long offset = Prefs.getInt(ImportDialog.OFFSET,0);
    private static int nImages = Prefs.getInt(ImportDialog.N,1);
    private static int gapBetweenImages = Prefs.getInt(ImportDialog.GAP,0);
	private static int options;
    private static boolean whiteIsZero,intelByteOrder;
    private boolean openAll;
    private static String[] types = {"8-bit", "16-bit Signed", "16-bit Unsigned",
		"32-bit Signed", "32-bit Unsigned", "32-bit Real", "64-bit Real", "24-bit RGB", 
		"24-bit RGB Planar", "24-bit BGR", "24-bit Integer", "32-bit ARGB", "1-bit Bitmap"};
    	
    static {
    	ImportDialog.options = Prefs.getInt(ImportDialog.OPTIONS,0);
    	ImportDialog.whiteIsZero = (ImportDialog.options&ImportDialog.WHITE_IS_ZERO)!=0;
    	ImportDialog.intelByteOrder = (ImportDialog.options&ImportDialog.INTEL_BYTE_ORDER)!=0;
    	//openAll = (options&OPEN_ALL)!=0;
    }
	
    public ImportDialog(String fileName, String directory) {
        this.fileName = fileName;
        this.directory = directory;
		IJ.showStatus("Importing: " + fileName);
	}

    public ImportDialog() {
	}

	boolean showDialog() {
		if (ImportDialog.choiceSelection>=ImportDialog.types.length)
			ImportDialog.choiceSelection = 0;
		GenericDialog gd = new GenericDialog("Import...", IJ.getInstance());
		gd.addChoice("Image Type:", ImportDialog.types, ImportDialog.types[ImportDialog.choiceSelection]);
		gd.addNumericField("Width:", ImportDialog.width, 0, 6, "pixels");
		gd.addNumericField("Height:", ImportDialog.height, 0, 6, "pixels");
		gd.addNumericField("Offset to First Image:", ImportDialog.offset, 0, 6, "bytes");
		gd.addNumericField("Number of Images:", ImportDialog.nImages, 0, 6, null);
		gd.addNumericField("Gap Between Images:", ImportDialog.gapBetweenImages, 0, 6, "bytes");
		gd.addCheckbox("White is Zero", ImportDialog.whiteIsZero);
		gd.addCheckbox("Little-Endian Byte Order", ImportDialog.intelByteOrder);
		gd.addCheckbox("Open All Files in Folder", openAll);
		gd.showDialog();
		if (gd.wasCanceled())
			return false;
		ImportDialog.choiceSelection = gd.getNextChoiceIndex();
		ImportDialog.width = (int)gd.getNextNumber();
		ImportDialog.height = (int)gd.getNextNumber();
		ImportDialog.offset = (long)gd.getNextNumber();
		ImportDialog.nImages = (int)gd.getNextNumber();
		ImportDialog.gapBetweenImages = (int)gd.getNextNumber();
		ImportDialog.whiteIsZero = gd.getNextBoolean();
		ImportDialog.intelByteOrder = gd.getNextBoolean();
		openAll = gd.getNextBoolean();
		IJ.register(ImportDialog.class);
		return true;
	}
	
	/** Opens all the images in the directory. */
	void openAll(String[] list, FileInfo fi) {
		//StringSorter.sort(list);
		FolderOpener fo = new FolderOpener();
		list = fo.trimFileList(list);
		list = fo.sortFileList(list);
		if (list==null) return;
		ImageStack stack=null;
		ImagePlus imp=null;
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for (int i=0; i<list.length; i++) {
			if (list[i].startsWith("."))
				continue;
			fi.fileName = list[i];
			imp = new FileOpener(fi).open(false);
			if (imp==null)
				IJ.log(list[i] + ": unable to open");
			else {
				if (stack==null)
					stack = imp.createEmptyStack();	
				try {
					ImageProcessor ip = imp.getProcessor();
					if (ip.getMin()<min) min = ip.getMin();
					if (ip.getMax()>max) max = ip.getMax();
					stack.addSlice(list[i], ip);
				}
				catch(OutOfMemoryError e) {
					IJ.outOfMemory("OpenAll");
					stack.trim();
					break;
				}
				IJ.showStatus((stack.getSize()+1) + ": " + list[i]);
			}
		}
		if (stack!=null) {
			imp = new ImagePlus("Imported Stack", stack);
			if (imp.getBitDepth()==16 || imp.getBitDepth()==32)
				imp.getProcessor().setMinAndMax(min, max);
                Calibration cal = imp.getCalibration();
                if (fi.fileType==FileInfo.GRAY16_SIGNED) {
                    double[] coeff = new double[2];
                    coeff[0] = -32768.0;
                    coeff[1] = 1.0;
                    cal.setFunction(Calibration.STRAIGHT_LINE, coeff, "gray value");
                }
			imp.show();
		}
	}
	
	/** Displays the dialog and opens the specified image or images.
		Does nothing if the dialog is canceled. */
	public void openImage() {
		FileInfo fi = getFileInfo();
		if (fi==null)
			return;
		if (openAll) {
			String[] list = new File(directory).list();
			if (list==null)
				return;
			openAll(list, fi);
		} else {
			FileOpener fo = new FileOpener(fi);
			fo.open();
		}
	}

	/** Displays the dialog and returns a FileInfo object that can be used to
		open the image. Returns null if the dialog is canceled. The fileName 
		and directory fields are null if the no argument constructor was used. */
	public FileInfo getFileInfo() {
		if (!showDialog())
			return null;
		String imageType = ImportDialog.types[ImportDialog.choiceSelection];
		FileInfo fi = new FileInfo();
		fi.fileFormat = FileInfo.RAW;
		fi.fileName = fileName;
		fi.directory = directory;
		fi.width = ImportDialog.width;
		fi.height = ImportDialog.height;
		if (ImportDialog.offset>2147483647)
			fi.longOffset = ImportDialog.offset;
		else
			fi.offset = (int)ImportDialog.offset;
		fi.nImages = ImportDialog.nImages;
		fi.gapBetweenImages = ImportDialog.gapBetweenImages;
		fi.intelByteOrder = ImportDialog.intelByteOrder;
		fi.whiteIsZero = ImportDialog.whiteIsZero;
		if (imageType.equals("8-bit"))
			fi.fileType = FileInfo.GRAY8;
		else if (imageType.equals("16-bit Signed"))
			fi.fileType = FileInfo.GRAY16_SIGNED;
		else if (imageType.equals("16-bit Unsigned"))
			fi.fileType = FileInfo.GRAY16_UNSIGNED;
		else if (imageType.equals("32-bit Signed"))
			fi.fileType = FileInfo.GRAY32_INT;
		else if (imageType.equals("32-bit Unsigned"))
			fi.fileType = FileInfo.GRAY32_UNSIGNED;
		else if (imageType.equals("32-bit Real"))
			fi.fileType = FileInfo.GRAY32_FLOAT;
		else if (imageType.equals("64-bit Real"))
			fi.fileType = FileInfo.GRAY64_FLOAT;
		else if (imageType.equals("24-bit RGB"))
			fi.fileType = FileInfo.RGB;
		else if (imageType.equals("24-bit RGB Planar"))
			fi.fileType = FileInfo.RGB_PLANAR;
		else if (imageType.equals("24-bit BGR"))
			fi.fileType = FileInfo.BGR;
		else if (imageType.equals("24-bit Integer"))
			fi.fileType = FileInfo.GRAY24_UNSIGNED;
		else if (imageType.equals("32-bit ARGB"))
			fi.fileType = FileInfo.ARGB;
		else if (imageType.equals("1-bit Bitmap"))
			fi.fileType = FileInfo.BITMAP;
		else
			fi.fileType = FileInfo.GRAY8;
		if (IJ.debugMode) IJ.log("ImportDialog: "+fi);
		return fi;
	}

	/** Called once when ImageJ quits. */
	public static void savePreferences(Properties prefs) {
		prefs.put(ImportDialog.TYPE, Integer.toString(ImportDialog.choiceSelection));
		prefs.put(ImportDialog.WIDTH, Integer.toString(ImportDialog.width));
		prefs.put(ImportDialog.HEIGHT, Integer.toString(ImportDialog.height));
		prefs.put(ImportDialog.OFFSET, Integer.toString(ImportDialog.offset>2147483647?0:(int)ImportDialog.offset));
		prefs.put(ImportDialog.N, Integer.toString(ImportDialog.nImages));
		prefs.put(ImportDialog.GAP, Integer.toString(ImportDialog.gapBetweenImages));
		int options = 0;
		if (ImportDialog.whiteIsZero)
			options |= ImportDialog.WHITE_IS_ZERO;
		if (ImportDialog.intelByteOrder)
			options |= ImportDialog.INTEL_BYTE_ORDER;
		//if (openAll)
		//	options |= OPEN_ALL;
		prefs.put(ImportDialog.OPTIONS, Integer.toString(options));
	}

}