package net.sf.ext.ij.plugin;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.image.IndexColorModel;
import java.util.Properties;
import java.util.Vector;

import net.sf.ext.ij.IJ;
import net.sf.ext.ij.ImagePlus;
import net.sf.ext.ij.Prefs;
import net.sf.ext.ij.Undo;
import net.sf.ext.ij.WindowManager;
import net.sf.ext.ij.gui.GenericDialog;
import net.sf.ext.ij.gui.ImageCanvas;
import net.sf.ext.ij.gui.ImageWindow;
import net.sf.ext.ij.gui.Line;
import net.sf.ext.ij.gui.ProfilePlot;
import net.sf.ext.ij.gui.Roi;
import net.sf.ext.ij.gui.Toolbar;
import net.sf.ext.ij.measure.Calibration;
import net.sf.ext.ij.process.ByteProcessor;
import net.sf.ext.ij.process.ImageConverter;
import net.sf.ext.ij.process.ImageProcessor;
import net.sf.ext.ij.process.ImageStatistics;

/** This plugin generates gel profile plots that can be analyzed using
the wand tool. It is similar to the "Gel Plotting Macros" in NIH Image. */
public class GelAnalyzer implements PlugIn {

    static final String GEL = "gel.options"; 
    static final int OD=1, PERCENT=2, OUTLINE=4, INVERT=8;
	static int saveID;
	static int nLanes, saveNLanes;
	static Rectangle firstRect;
	static final int MAX_LANES = 100;
	static int[] x = new int[GelAnalyzer.MAX_LANES+1];
	static PlotsCanvas plotsCanvas;
	static ImageProcessor ipLanes;
	static ImagePlus  gel;
	static int plotHeight;
	static int options = Prefs.getInt(GelAnalyzer.GEL, GelAnalyzer.PERCENT+GelAnalyzer.INVERT);
	static boolean uncalibratedOD = (GelAnalyzer.options&GelAnalyzer.OD)!=0;
	static boolean labelWithPercentages = (GelAnalyzer.options&GelAnalyzer.PERCENT)!=0;;
	static boolean outlineLanes;
	static boolean invertPeaks = (GelAnalyzer.options&GelAnalyzer.INVERT)!=0;
	static Vector roiList;
	boolean invertedLut;
	
	ImagePlus imp;
	Font f;
	double odMin=Double.MAX_VALUE, odMax=-Double.MAX_VALUE;
	static boolean isVertical;
	
	public void run(String arg) {
		if (arg.equals("options")) {
			showDialog();
			return;
		}
		
		imp = WindowManager.getCurrentImage();
		if (imp==null) {
			IJ.noImage();
			return;
		}

		if (arg.equals("reset")) {
			GelAnalyzer.nLanes = 0;
			GelAnalyzer.saveNLanes = 0;
			GelAnalyzer.saveID = 0;
			if (GelAnalyzer.plotsCanvas!=null)
				GelAnalyzer.plotsCanvas.reset();
			GelAnalyzer.ipLanes = null;
			GelAnalyzer.roiList = null;
			if (GelAnalyzer.gel!=null) {
				ImageCanvas ic = GelAnalyzer.gel.getCanvas();
				if (ic!=null) ic.setDisplayList(null);
				GelAnalyzer.gel.draw();
			}
			return;
		}

		if (arg.equals("percent") && GelAnalyzer.plotsCanvas!=null) {
			GelAnalyzer.plotsCanvas.displayPercentages();
			return;
		}

		if (arg.equals("label") && GelAnalyzer.plotsCanvas!=null) {
			if (GelAnalyzer.plotsCanvas.counter==0)
				show("There are no peak area measurements.");
			else
				GelAnalyzer.plotsCanvas.labelPeaks();
			return;
		}

		if (imp.getID()!=GelAnalyzer.saveID) {
			GelAnalyzer.nLanes=0;
			GelAnalyzer.ipLanes = null;
			GelAnalyzer.saveID = 0;
		}

		if (arg.equals("replot")) {
			if (GelAnalyzer.saveNLanes==0) {
				show("The data needed to re-plot the lanes is not available");
				return;
			}
			GelAnalyzer.nLanes = GelAnalyzer.saveNLanes;
			plotLanes(GelAnalyzer.gel, true);
			return;
		}
		
		if (arg.equals("draw")) {
			outlineLanes();
			return;
		}

		Roi roi = imp.getRoi();
		if (roi==null || roi.getType()!=Roi.RECTANGLE) {
			show("Rectangular selection required.");
			return;
		}
		Rectangle rect = roi.getBounds();
		if (GelAnalyzer.nLanes==0) {
			invertedLut = imp.isInvertedLut();
			IJ.register(GelAnalyzer.class);  // keeps this class from being GC'd
		}

		if (arg.equals("first")) {
			selectFirstLane(rect);
			return;
		}

		if (GelAnalyzer.nLanes==0) {
			show("You must first use the \"Outline First Lane\" command.");
			return;
		}
		if (arg.equals("next")) {
			selectNextLane(rect);
			return;
		}
		
		if (arg.equals("plot")) {
			if (( GelAnalyzer.isVertical && (rect.x!=GelAnalyzer.x[GelAnalyzer.nLanes]) ) || ( !(GelAnalyzer.isVertical) && (rect.y!=GelAnalyzer.x[GelAnalyzer.nLanes]) )) {
				selectNextLane(rect);
			}
			plotLanes(GelAnalyzer.gel, false);
			return;
		}

	}

	void showDialog() {
		GenericDialog gd = new GenericDialog("Gel Analyzer");
		gd.addCheckbox("Uncalibrated OD", GelAnalyzer.uncalibratedOD);
		gd.addCheckbox("Label with Percentages", GelAnalyzer.labelWithPercentages);
		gd.addCheckbox("Invert Peaks", GelAnalyzer.invertPeaks);
		gd.showDialog();
		if (gd.wasCanceled())
			return;
		GelAnalyzer.uncalibratedOD = gd.getNextBoolean();
		GelAnalyzer.labelWithPercentages = gd.getNextBoolean();
		GelAnalyzer.invertPeaks = gd.getNextBoolean();
		GelAnalyzer.options = 0;
		if (GelAnalyzer.uncalibratedOD) GelAnalyzer.options |= GelAnalyzer.OD;
		if (GelAnalyzer.labelWithPercentages) GelAnalyzer.options |= GelAnalyzer.PERCENT;
		if (GelAnalyzer.invertPeaks) GelAnalyzer.options |= GelAnalyzer.INVERT;
		return;
	}


	void selectFirstLane(Rectangle rect) {
		if (rect.width/rect.height>=2 || IJ.altKeyDown())
			GelAnalyzer.isVertical = false;
		else
			GelAnalyzer.isVertical = true;
			
		/*
		if ( (isVertical && (rect.height/rect.width)<2 ) || (!isVertical && (rect.width/rect.height)<2 ) ) {
			GenericDialog gd = new GenericDialog("Lane Orientation");
			String[] orientations = {"Vertical","Horizontal"};
			int defaultOrientation = isVertical?0:1;
			gd.addChoice("Lane Orientation:", orientations, orientations[defaultOrientation]);
			gd.showDialog();
			if (gd.wasCanceled())
				return;
			String orientation = gd.getNextChoice();
			if(orientation.equals(orientations[0]))
				isVertical=true;
			else
				isVertical=false;
		}
		*/

		IJ.showStatus("Lane 1 selected ("+(GelAnalyzer.isVertical?"vertical":"horizontal")+" lanes)");
		GelAnalyzer.firstRect = rect;
		GelAnalyzer.nLanes = 1;
		GelAnalyzer.saveNLanes = 0;
		if(GelAnalyzer.isVertical)
			GelAnalyzer.x[1] = rect.x;
		else
			GelAnalyzer.x[1] = rect.y;
		GelAnalyzer.gel = imp;
		GelAnalyzer.saveID = imp.getID();
		GelAnalyzer.roiList = null;
		updateRoiList(rect);
	}

	void selectNextLane(Rectangle rect) {
		if (rect.width!=GelAnalyzer.firstRect.width || rect.height!=GelAnalyzer.firstRect.height) {
			show("Selections must all be the same size.");
			return;
		}
		if (GelAnalyzer.nLanes<GelAnalyzer.MAX_LANES)
			GelAnalyzer.nLanes += 1;
		IJ.showStatus("Lane " + GelAnalyzer.nLanes + " selected");

		if(GelAnalyzer.isVertical)
			GelAnalyzer.x[GelAnalyzer.nLanes] = rect.x;
		else
			GelAnalyzer.x[GelAnalyzer.nLanes] = rect.y;
		if (GelAnalyzer.isVertical && rect.y!=GelAnalyzer.firstRect.y) {
			rect.y = GelAnalyzer.firstRect.y;
			GelAnalyzer.gel.setRoi(rect);
		} else if (!GelAnalyzer.isVertical && rect.x!=GelAnalyzer.firstRect.x) {
			rect.x = GelAnalyzer.firstRect.x;
			GelAnalyzer.gel.setRoi(rect);
		}
		updateRoiList(rect);
	}
	
	void updateRoiList(Rectangle rect) {
			if (GelAnalyzer.gel==null) return;
			if (GelAnalyzer.roiList==null) {
				GelAnalyzer.roiList = new Vector();
				ImageCanvas ic = GelAnalyzer.gel.getCanvas();
				if (ic!=null) ic.setDisplayList(GelAnalyzer.roiList);
			}
			GelAnalyzer.roiList.addElement(new Roi(rect.x, rect.y, rect.width, rect.height, null));
			GelAnalyzer.gel.draw();
	}

	void plotLanes(ImagePlus imp, boolean replot) {
		int topMargin = 16;
		int bottomMargin = 2;
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		int plotWidth;
		double[][] profiles;
		profiles = new double[GelAnalyzer.MAX_LANES+1][];
		IJ.showStatus("Plotting " + GelAnalyzer.nLanes + " lanes");
		ImageProcessor ipRotated = imp.getProcessor();
		if(GelAnalyzer.isVertical) {
			ipRotated = ipRotated.rotateLeft();
		}
		ImagePlus imp2 = new ImagePlus("", ipRotated);
		imp2.setCalibration(imp.getCalibration());
		if (GelAnalyzer.uncalibratedOD && (imp2.getType()==ImagePlus.GRAY16 || imp2.getType()==ImagePlus.GRAY32))
			new ImageConverter(imp2).convertToGray8();
		if (GelAnalyzer.invertPeaks) {
				ImageProcessor ip2 = imp2.getProcessor().duplicate();
				ip2.invert();
				imp2.setProcessor(null, ip2);
		}
		//imp2.show();

		for (int i=1; i<=GelAnalyzer.nLanes; i++) {
			if(GelAnalyzer.isVertical)
				imp2.setRoi(GelAnalyzer.firstRect.y,
							ipRotated.getHeight() - GelAnalyzer.x[i] - GelAnalyzer.firstRect.width,
							GelAnalyzer.firstRect.height, GelAnalyzer.firstRect.width);
			else
				imp2.setRoi(GelAnalyzer.firstRect.x, GelAnalyzer.x[i], GelAnalyzer.firstRect.width, GelAnalyzer.firstRect.height);
			ProfilePlot pp = new ProfilePlot(imp2);
			profiles[i] = pp.getProfile();
			if (pp.getMin()<min)
				min = pp.getMin();
			if (pp.getMax()>max)
				max = pp.getMax();
			if (GelAnalyzer.uncalibratedOD)
				profiles[i] = od(profiles[i]);
		}
		if (GelAnalyzer.uncalibratedOD) {
			min = odMin;
			max = odMax;
		}

		if(GelAnalyzer.isVertical)
			plotWidth = GelAnalyzer.firstRect.height;
		else
			plotWidth = GelAnalyzer.firstRect.width;
		if (plotWidth<500)
			plotWidth = 500;
		if(GelAnalyzer.isVertical) {
			if (plotWidth>2*GelAnalyzer.firstRect.height)
				plotWidth = 2*GelAnalyzer.firstRect.height;
		} else {
			if (plotWidth>2*GelAnalyzer.firstRect.width)
				plotWidth = 2*GelAnalyzer.firstRect.width;
		}

		Dimension screen = IJ.getScreenSize();
		if (plotWidth>screen.width-40)
			plotWidth = screen.width - 40;
		GelAnalyzer.plotHeight = plotWidth/2;
		if (GelAnalyzer.plotHeight<200)
			GelAnalyzer.plotHeight = 200;
		if (GelAnalyzer.plotHeight>400)
			GelAnalyzer.plotHeight = 400;
		ImageProcessor ip = new ByteProcessor(plotWidth, topMargin+GelAnalyzer.nLanes*GelAnalyzer.plotHeight+bottomMargin);
		ip.setColor(Color.white);
		ip.fill();
		ip.setColor(Color.black);
		//draw border
		int h= ip.getHeight();
		ip.moveTo(0,0);
		ip.lineTo(plotWidth-1,0);
		ip.lineTo(plotWidth-1, h-1);
		ip.lineTo(0, h-1);
		ip.lineTo(0, 0);
		ip.moveTo(0, h-2);
		ip.lineTo(plotWidth-1, h-2);
		String s = imp.getTitle()+"; ";
		Calibration cal = imp.getCalibration();
		if (cal.calibrated())
			s += cal.getValueUnit();
		else if (GelAnalyzer.uncalibratedOD)
			s += "Uncalibrated OD";
		else
			s += "Uncalibrated";
		ip.moveTo(5,topMargin);
		ip.drawString(s);
		double xScale = (double)plotWidth/profiles[1].length;
		double yScale;
		if ((max-min)==0.0)
			yScale = 1.0;
		else
			yScale = GelAnalyzer.plotHeight/(max-min);
		for (int i=1; i<=GelAnalyzer.nLanes; i++) {
			double[] profile = profiles[i];
			int top = (i-1)*GelAnalyzer.plotHeight + topMargin;
			int base = top+GelAnalyzer.plotHeight;
			ip.moveTo(0, base);
			ip.lineTo((int)(profile.length*xScale), base);
			ip.moveTo(0, base-(int)((profile[0]-min)*yScale));
			for (int j = 1; j<profile.length; j++)
				ip.lineTo((int)(j*xScale+0.5), base-(int)((profile[j]-min)*yScale+0.5));
		}
		Line.setWidth(1);
		ImagePlus plots = new Plots();
		plots.setProcessor("Plots of "+imp.getShortTitle(), ip);
		plots.changes = true;
		ip.setThreshold(0,0,ImageProcessor.NO_LUT_UPDATE); // Wand tool works better with threshold set
		if (cal.calibrated()) {
			double pixelsAveraged = GelAnalyzer.isVertical?GelAnalyzer.firstRect.width:GelAnalyzer.firstRect.height;
			double scale = Math.sqrt((xScale*yScale)/pixelsAveraged);
			Calibration plotsCal = plots.getCalibration();
			plotsCal.setUnit("unit");
			plotsCal.pixelWidth = 1.0/scale;
			plotsCal.pixelHeight = 1.0/scale;
		}
		plots.show();
		GelAnalyzer.saveNLanes = GelAnalyzer.nLanes;
		GelAnalyzer.nLanes = 0;
		GelAnalyzer.saveID = 0;
		//gel = null;
		GelAnalyzer.ipLanes = null;
		Toolbar toolbar = Toolbar.getInstance();
		toolbar.setColor(Color.black);
		toolbar.setTool(Toolbar.LINE);
		ImageWindow win = WindowManager.getCurrentWindow();
		ImageCanvas canvas = win.getCanvas();
		if (canvas instanceof PlotsCanvas)
			GelAnalyzer.plotsCanvas = (PlotsCanvas)canvas;
		else
			GelAnalyzer.plotsCanvas = null;
	}

	double[] od(double[] profile) {
		double v;
		for (int i=0; i<profile.length; i++) {
			v = 0.434294481*Math.log(255.0/(255.0-profile[i]));
			//v = 0.434294481*Math.log(255.0/v);
			if (v<odMin) odMin = v;
			if (v>odMax) odMax = v;
			profile[i] = v;
		}
		return profile;
	}

	void outlineLanes() {
		if (GelAnalyzer.gel==null || GelAnalyzer.roiList==null) {
			show("Data needed to outline lanes is no longer available.");
			return;
		}
		//IJ.write("outlining lane "+x);
		int lineWidth = (int)(1.0/GelAnalyzer.gel.getCanvas().getMagnification());
		if (lineWidth<1)
			lineWidth = 1;
		Font f = new Font("Helvetica", Font.PLAIN, 12*lineWidth);
		ImageProcessor ip = GelAnalyzer.gel.getProcessor();
		ImageProcessor ipLanes = ip.duplicate();
		if (!(ipLanes instanceof ByteProcessor))
			ipLanes = ipLanes.convertToByte(true);
		ipLanes.setFont(f);
		ipLanes.setLineWidth(lineWidth);
		setCustomLut(ipLanes);
		ImagePlus lanes = new ImagePlus("Lanes of "+GelAnalyzer.gel.getShortTitle(), ipLanes);
		lanes.changes = true;
		lanes.setRoi(GelAnalyzer.gel.getRoi());
		GelAnalyzer.gel.killRoi();
		for (int i=0; i<GelAnalyzer.roiList.size(); i++) {
			Roi roi = (Roi)GelAnalyzer.roiList.elementAt(i);
			Rectangle r = roi.getBounds();
			ipLanes.drawRect(r.x, r.y, r.width, r.height);
			String s = ""+(i+1);
			if(GelAnalyzer.isVertical) {
				int yloc = r.y;
				if (yloc<lineWidth*12) yloc += lineWidth*14;
				ipLanes.drawString(s, r.x+r.width/2-ipLanes.getStringWidth(s)/2, yloc);
			} else {
				int xloc = r.x-ipLanes.getStringWidth(s)-2;
				if (xloc<lineWidth*10) xloc = r.x + 2;
				ipLanes.drawString(s, xloc, r.y+r.height/2+6);
			}
		}
		lanes.killRoi();
		lanes.show();
	}
	
	void setCustomLut(ImageProcessor ip) {
		IndexColorModel cm = (IndexColorModel)ip.getColorModel();
		byte[] reds = new byte[256];
		byte[] greens = new byte[256];
		byte[] blues = new byte[256];
		cm.getReds(reds);
		cm.getGreens(greens);
		cm.getBlues(blues);
		reds[1] =(byte) 255;
		greens[1] = (byte)0;
		blues[1] = (byte)0;
		ip.setColorModel(new IndexColorModel(8, 256, reds, greens, blues));
		byte[] pixels = (byte[])ip.getPixels();
		for (int i=0; i<pixels.length; i++)
			if ((pixels[i]&255)==1)
			   pixels[i] = 0;
		ip.setColor(1);
	}

	void show(String msg) {
		IJ.showMessage("Gel Analyzer", msg);
	}

	/** Called once when ImageJ quits. */
	public static void savePreferences(Properties prefs) {
		prefs.put(GelAnalyzer.GEL, Integer.toString(GelAnalyzer.options));
	}

}


class Plots extends ImagePlus {

	/** Overrides ImagePlus.show(). */
	public void show() {
		img = ip.createImage();
		ImageCanvas ic = new PlotsCanvas(this);
		win = new ImageWindow(this, ic);
		IJ.showStatus("");
		if (ic.getMagnification()==1.0)
			return;
		while(ic.getMagnification()<1.0)
			ic.zoomIn(0,0);
		Point loc = win.getLocation();
		int w = getWidth()+20;
		int h = getHeight()+30;
		Dimension screen = IJ.getScreenSize();
		if (loc.x+w>screen.width)
			w = screen.width-loc.x-20;
		if (loc.y+h>screen.height)
			h = screen.height-loc.y-30;
		win.setSize(w, h);
		win.validate();
		repaintWindow();
	}

}


class PlotsCanvas extends ImageCanvas {

	public static final int MAX_PEAKS = 200;

	double[] actual = {428566.00,351368.00,233977.00,99413.00,60057.00,31382.00,
					   14531.00,7843.00,2146.00,752.00,367.00};
	double[] measured = new double[PlotsCanvas.MAX_PEAKS];
	Rectangle[] rect = new Rectangle[PlotsCanvas.MAX_PEAKS];
	int counter;

	public PlotsCanvas(ImagePlus imp) {
		super(imp);
	}

	public void mousePressed(MouseEvent e) {
		super.mousePressed(e);
		Roi roi = imp.getRoi();
		if (roi==null)
			return;
		if (roi.getType()==Roi.LINE)
			Roi.setColor(Color.blue);
		else
			Roi.setColor(Color.yellow);
		if (Toolbar.getToolId()!=Toolbar.WAND || IJ.spaceBarDown())
			return;
		ImageStatistics s = imp.getStatistics();
		if (counter==0)
			IJ.setColumnHeadings(" \tArea");
		double perimeter = roi.getLength();
		String error = "";
		double circularity = 4.0*Math.PI*(s.pixelCount/(perimeter*perimeter));
		if (circularity<0.025)
			error = " (error?)";
		double area = s.pixelCount+perimeter/2.0; // add perimeter/2 to account area under border
		Calibration cal = imp.getCalibration();
		area = area*cal.pixelWidth*cal.pixelHeight;
		rect[counter] = roi.getBounds();

		//area += (rect[counter].width/rect[counter].height)*1.5;
		// adjustment for small peaks from NIH Image gel macros

		int places = cal.scaled()?3:0;
		IJ.write((counter+1)+"\t"+IJ.d2s(area, places)+error);
		measured[counter] = area;
		if (counter<PlotsCanvas.MAX_PEAKS)
			counter++;
	}

	public void mouseReleased(MouseEvent e) {
		super.mouseReleased(e);
		Roi roi = imp.getRoi();
		if (roi!=null && roi.getType()==Roi.LINE) {
			Undo.setup(Undo.FILTER, imp);
			imp.getProcessor().snapshot();
			roi.drawPixels();
			imp.updateAndDraw();
			imp.killRoi();
		}
	}

	void reset() {
		counter = 0;
	}

	void labelPeaks() {
		imp.killRoi();
		double total = 0.0;
		for (int i=0; i<counter; i++)
			total += measured[i];
		ImageProcessor ip = imp.getProcessor();
		ip.setFont(new Font("SansSerif", Font.PLAIN, 9));
		for (int i=0; i<counter; i++) {
			Rectangle r = rect[i];
			String s;
			if (GelAnalyzer.labelWithPercentages)
				s = IJ.d2s((measured[i]/total)*100, 2);
			else
				s = IJ.d2s(measured[i], 0);
			int swidth = ip.getStringWidth(s);
			int x = r.x + r.width/2 - swidth/2;
			int y = r.y + r.height*3/4 + 9;
			int[] data = new int[swidth];
			ip.getRow(x, y, data, swidth);
			boolean fits = true;
			for (int j=0; j<swidth; j++)
				if (data[j]!=255) {
					fits = false;
					break;
				}
			fits = fits && measured[i]>500;
			if (r.height>=(GelAnalyzer.plotHeight-11))
				fits = true;
			if (!fits)
				y = r.y - 2;
			ip.drawString(s, x, y);
			//IJ.write(i+": "+x+" "+y+" "+s+" "+ip.StringWidth(s)/2);
		}
		imp.updateAndDraw();
		displayPercentages();
		//Toolbar.getInstance().setTool(Toolbar.RECTANGLE);
		reset();
	}

	void displayPercentages() {
		IJ.setColumnHeadings(" \tarea\tpercent");
		double total = 0.0;
		for (int i=0; i<counter; i++)
			total += measured[i];
		if (IJ.debugMode && counter==actual.length) {
			debug();
			return;
		}
		for (int i=0; i<counter; i++) {
			double percent = (measured[i]/total)*100;
			IJ.write((i+1)+"\t"+IJ.d2s(measured[i],3)+"\t"+IJ.d2s(percent,3));
		}
	}

	void debug() {
		for (int i=0; i<counter; i++) {
			double a = (actual[i]/actual[0])*100;
			double m = (measured[i]/measured[0])*100;
			IJ.write(IJ.d2s(a, 4)+" "
					 +IJ.d2s(m, 4)+" "
					 +IJ.d2s(((m-a)/m)*100, 4));
		}
	}

}

