package net.sf.ext.ij.plugin.filter;
import java.awt.Color;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.util.Properties;

import net.sf.ext.ij.IJ;
import net.sf.ext.ij.ImageJ;
import net.sf.ext.ij.ImagePlus;
import net.sf.ext.ij.ImageStack;
import net.sf.ext.ij.Macro;
import net.sf.ext.ij.Prefs;
import net.sf.ext.ij.WindowManager;
import net.sf.ext.ij.gui.GenericDialog;
import net.sf.ext.ij.gui.Line;
import net.sf.ext.ij.gui.PolygonRoi;
import net.sf.ext.ij.gui.ProfilePlot;
import net.sf.ext.ij.gui.Roi;
import net.sf.ext.ij.gui.Toolbar;
import net.sf.ext.ij.gui.YesNoCancelDialog;
import net.sf.ext.ij.macro.Interpreter;
import net.sf.ext.ij.measure.Calibration;
import net.sf.ext.ij.measure.Measurements;
import net.sf.ext.ij.measure.ResultsTable;
import net.sf.ext.ij.plugin.MeasurementsWriter;
import net.sf.ext.ij.process.FloatProcessor;
import net.sf.ext.ij.process.FloatStatistics;
import net.sf.ext.ij.process.ImageProcessor;
import net.sf.ext.ij.process.ImageStatistics;
import net.sf.ext.ij.text.TextPanel;

/** This plugin implements ImageJ's Analyze/Measure and Analyze/Set Measurements commands. */
public class Analyzer implements PlugInFilter, Measurements {
	
	private ImagePlus imp;
	private ResultsTable rt;
	private int measurements;
	private StringBuffer min,max,mean,sd;
	
	// Order must agree with order of checkboxes in Set Measurements dialog box
	private static final int[] list = {Measurements.AREA,Measurements.MEAN,Measurements.STD_DEV,Measurements.MODE,Measurements.MIN_MAX,
		Measurements.CENTROID,Measurements.CENTER_OF_MASS,Measurements.PERIMETER,Measurements.RECT,Measurements.ELLIPSE,Measurements.CIRCULARITY, Measurements.FERET,
		Measurements.INTEGRATED_DENSITY,Measurements.MEDIAN,Measurements.SKEWNESS,Measurements.KURTOSIS,Measurements.AREA_FRACTION,Measurements.SLICE,
		Measurements.LIMIT,Measurements.LABELS,Measurements.INVERT_Y};

	private static final int UNDEFINED=0,AREAS=1,LENGTHS=2,ANGLES=3,POINTS=4;
	private static int mode = Analyzer.AREAS;
	private static final String MEASUREMENTS = "measurements";
	private static final String MARK_WIDTH = "mark.width";
	private static final String PRECISION = "precision";
	//private static int counter;
	private static boolean unsavedMeasurements;
	public static Color darkBlue = new Color(0,0,160);
	private static int systemMeasurements = Prefs.getInt(Analyzer.MEASUREMENTS,Measurements.AREA+Measurements.MEAN+Measurements.MIN_MAX);
	public static int markWidth = Prefs.getInt(Analyzer.MARK_WIDTH,0);
	public static int precision = Prefs.getInt(Analyzer.PRECISION,3);
	private static float[] umeans = new float[Measurements.MAX_STANDARDS];
	private static ResultsTable systemRT = new ResultsTable();
	private static int redirectTarget;
	private static String redirectTitle = "";
	static int firstParticle, lastParticle;
	private static boolean summarized;
	private static boolean switchingModes;
	
	public Analyzer() {
		rt = Analyzer.systemRT;
		rt.setPrecision(Analyzer.precision);
		measurements = Analyzer.systemMeasurements;
	}
	
	/** Constructs a new Analyzer using the specified ImagePlus object
		and the system-wide measurement options and results table. */
	public Analyzer(ImagePlus imp) {
		this();
		this.imp = imp;
	}
	
	/** Construct a new Analyzer using an ImagePlus object and private
		measurement options and results table. */
	public Analyzer(ImagePlus imp, int measurements, ResultsTable rt) {
		this.imp = imp;
		this.measurements = measurements;
		this.rt = rt;
	}
	
	public int setup(String arg, ImagePlus imp) {
		this.imp = imp;
		IJ.register(Analyzer.class);
		if (arg.equals("set"))
			{doSetDialog(); return PlugInFilter.DONE;}
		else if (arg.equals("sum"))
			{summarize(); return PlugInFilter.DONE;}
		else if (arg.equals("clear")) {
			if (IJ.macroRunning()) Analyzer.unsavedMeasurements = false;
			Analyzer.resetCounter();
			return PlugInFilter.DONE;
		} else
			return PlugInFilter.DOES_ALL+PlugInFilter.NO_CHANGES;
	}

	public void run(ImageProcessor ip) {
		measure();
	}

	void doSetDialog() {
		String NONE = "None";
		String[] titles;
        int[] wList = WindowManager.getIDList();
        if (wList==null) {
        	titles = new String[1];
            titles[0] = NONE;
        } else {
			titles = new String[wList.length+1];
			titles[0] = NONE;
			for (int i=0; i<wList.length; i++) {
				ImagePlus imp = WindowManager.getImage(wList[i]);
				titles[i+1] = imp!=null?imp.getTitle():"";
			}
		}
		ImagePlus tImp = WindowManager.getImage(Analyzer.redirectTarget);
		String target = tImp!=null?tImp.getTitle():NONE;
		
 		GenericDialog gd = new GenericDialog("Set Measurements", IJ.getInstance());
		String[] labels = new String[18];
		boolean[] states = new boolean[18];
		labels[0]="Area"; states[0]=(Analyzer.systemMeasurements&Measurements.AREA)!=0;
		labels[1]="Mean Gray Value"; states[1]=(Analyzer.systemMeasurements&Measurements.MEAN)!=0;
		labels[2]="Standard Deviation"; states[2]=(Analyzer.systemMeasurements&Measurements.STD_DEV)!=0;
		labels[3]="Modal Gray Value"; states[3]=(Analyzer.systemMeasurements&Measurements.MODE)!=0;
		labels[4]="Min & Max Gray Value"; states[4]=(Analyzer.systemMeasurements&Measurements.MIN_MAX)!=0;
		labels[5]="Centroid"; states[5]=(Analyzer.systemMeasurements&Measurements.CENTROID)!=0;
		labels[6]="Center of Mass"; states[6]=(Analyzer.systemMeasurements&Measurements.CENTER_OF_MASS)!=0;
		labels[7]="Perimeter"; states[7]=(Analyzer.systemMeasurements&Measurements.PERIMETER)!=0;
		labels[8]="Bounding Rectangle"; states[8]=(Analyzer.systemMeasurements&Measurements.RECT)!=0;
		labels[9]="Fit Ellipse"; states[9]=(Analyzer.systemMeasurements&Measurements.ELLIPSE)!=0;
		labels[10]="Circularity"; states[10]=(Analyzer.systemMeasurements&Measurements.CIRCULARITY)!=0;
		labels[11]="Feret's Diameter"; states[11]=(Analyzer.systemMeasurements&Measurements.FERET)!=0;
		labels[12]="Integrated Density"; states[12]=(Analyzer.systemMeasurements&Measurements.INTEGRATED_DENSITY)!=0;
		labels[13]="Median"; states[13]=(Analyzer.systemMeasurements&Measurements.MEDIAN)!=0;
		labels[14]="Skewness"; states[14]=(Analyzer.systemMeasurements&Measurements.SKEWNESS)!=0;
		labels[15]="Kurtosis"; states[15]=(Analyzer.systemMeasurements&Measurements.KURTOSIS)!=0;
		labels[16]="Area_Fraction"; states[16]=(Analyzer.systemMeasurements&Measurements.AREA_FRACTION)!=0;
		labels[17]="Slice Number"; states[17]=(Analyzer.systemMeasurements&Measurements.SLICE)!=0;
		gd.setInsets(0, 0, 0);
		gd.addCheckboxGroup(10, 2, labels, states);
		labels = new String[3];
		states = new boolean[3];
		labels[0]="Limit to Threshold"; states[0]=(Analyzer.systemMeasurements&Measurements.LIMIT)!=0;
		labels[1]="Display Label"; states[1]=(Analyzer.systemMeasurements&Measurements.LABELS)!=0;
		labels[2]="Invert Y Coordinates"; states[2]=(Analyzer.systemMeasurements&Measurements.INVERT_Y)!=0;
		gd.setInsets(0, 0, 0);
		gd.addCheckboxGroup(2, 2, labels, states);
		gd.setInsets(15, 0, 0);
        gd.addChoice("Redirect To:", titles, target);
		gd.setInsets(5, 0, 0);
		gd.addNumericField("Decimal Places (0-9):", Analyzer.precision, 0, 2, "");
		gd.showDialog();
		if (gd.wasCanceled())
			return;
		setOptions(gd);
		int index = gd.getNextChoiceIndex();
		Analyzer.redirectTarget = index==0?0:wList[index-1];
		Analyzer.redirectTitle = titles[index];

		int prec = (int)gd.getNextNumber();
		if (prec<0) prec = 0;
		if (prec>9) prec = 9;
		if (prec!=Analyzer.precision) {
			Analyzer.precision = prec;
			rt.setPrecision(Analyzer.precision);
			if (Analyzer.mode==Analyzer.AREAS && IJ.isResultsWindow()) {
				IJ.setColumnHeadings("");
				updateHeadings();
			}
		}
	}
	
	void setOptions(GenericDialog gd) {
		int oldMeasurements = Analyzer.systemMeasurements;
		int previous = 0;
		boolean b = false;
		for (int i=0; i<Analyzer.list.length; i++) {
			//if (list[i]!=previous)
			b = gd.getNextBoolean();
			previous = Analyzer.list[i];
			if (b)
				Analyzer.systemMeasurements |= Analyzer.list[i];
			else
				Analyzer.systemMeasurements &= ~Analyzer.list[i];
		}
		if ((oldMeasurements&(~Measurements.LIMIT))!=(Analyzer.systemMeasurements&(~Measurements.LIMIT))) {
			if (IJ.macroRunning()) {
				Analyzer.unsavedMeasurements = false;
				reset();
				Analyzer.mode = Analyzer.AREAS;
			} else
				Analyzer.mode = Analyzer.UNDEFINED;
		}
		if ((Analyzer.systemMeasurements&Measurements.LABELS)==0)
			Analyzer.systemRT.disableRowLabels();
	}
	
	void measure() {
		String sliceHdr = rt.getColumnHeading(ResultsTable.SLICE);
		if (sliceHdr==null || sliceHdr.charAt(0)!='S') {
			if (!reset()) return;
		}
		Analyzer.firstParticle = Analyzer.lastParticle = 0;
		Roi roi = imp.getRoi();
		if (roi!=null && roi.getType()==Roi.POINT) {
			measurePoint(roi);
			return;
		}
		if (roi!=null && roi.isLine()) {
			measureLength(roi);
			return;
		}
		if (roi!=null && roi.getType()==Roi.ANGLE) {
			measureAngle(roi);
			return;
		}
		if (Analyzer.mode!=Analyzer.AREAS) {
			Analyzer.switchingModes = true;
			if (!reset())
				return;
			Analyzer.mode = Analyzer.AREAS;
		}
		ImageStatistics stats;
		if (Analyzer.isRedirectImage()) {
			stats = getRedirectStats(measurements, roi);
			if (stats==null) return;
		} else
			stats = imp.getStatistics(measurements);
		if (!IJ.isResultsWindow())
			reset();
		saveResults(stats, roi);
		displayResults();
	}
	
	boolean reset() {
		boolean ok = true;
		if (rt.getCounter()>0)
			ok = Analyzer.resetCounter();
		if (ok && rt.getColumnHeading(ResultsTable.SLICE)==null)
			rt.setDefaultHeadings();
		return ok;
	}

	/** Returns <code>true</code> if an image is selected in the "Redirect To:"
		popup menu of the Analyze/Set Measurements dialog box. */
	public static boolean isRedirectImage() {
		return Analyzer.redirectTarget!=0;
	}
	
	/** Returns the image selected in the "Redirect To:" popup
		menu of the Analyze/Set Measurements dialog or null
		if "None" is selected, the image was not found or the 
		image is not the same size as <code>currentImage</code>. */
	public static ImagePlus getRedirectImage(ImagePlus currentImage) {
		ImagePlus rImp = WindowManager.getImage(Analyzer.redirectTarget);
		if (rImp==null) {
			IJ.error("Analyzer", "Redirect image (\""+Analyzer.redirectTitle+"\")\n"
				+ "not found.");
			Analyzer.redirectTarget = 0;
			Macro.abort();
			return null;
		}
		if (rImp.getWidth()!=currentImage.getWidth() || rImp.getHeight()!=currentImage.getHeight()) {
			IJ.error("Analyzer", "Redirect image (\""+Analyzer.redirectTitle+"\") \n"
				+ "is not the same size as the current image.");
			Macro.abort();
			return null;
		}
		return rImp;
	}

	ImageStatistics getRedirectStats(int measurements, Roi roi) {
		ImagePlus redirectImp = Analyzer.getRedirectImage(imp);
		if (redirectImp==null)
			return null;
		int depth = redirectImp.getStackSize();
		if (depth>1 && depth==imp.getStackSize())
			redirectImp.setSlice(imp.getCurrentSlice());
		ImageProcessor ip = redirectImp.getProcessor();
		if (imp.getTitle().equals("mask") && imp.getBitDepth()==8) {
			ip.setMask(imp.getProcessor());
			ip.setRoi(0, 0, imp.getWidth(), imp.getHeight());
		} else
			ip.setRoi(roi);
		return ImageStatistics.getStatistics(ip, measurements, redirectImp.getCalibration());
	}
	
	void measurePoint(Roi roi) {
		if (Analyzer.mode!=Analyzer.POINTS) {
			Analyzer.switchingModes = true;
			if (!reset())
				return;
			//IJ.setColumnHeadings(" \tX\tY\tValue");		
			Analyzer.mode = Analyzer.POINTS;
		}
		Polygon p = roi.getPolygon();
		ImageProcessor ip = imp.getProcessor();
		Calibration cal = imp.getCalibration();
		ip.setCalibrationTable(cal.getCTable());
		for (int i=0; i<p.npoints; i++) {
			incrementCounter();
			int x = p.xpoints[i];
			int y = p.ypoints[i];
			double value = ip.getPixelValue(x,y);
			if (Analyzer.markWidth>0) {
				ip.setColor(Toolbar.getForegroundColor());
				ip.setLineWidth(Analyzer.markWidth);
				ip.moveTo(x,y);
				ip.lineTo(x,y);
				imp.updateAndDraw();
				ip.setLineWidth(Line.getWidth());
			}
			if ((measurements&Measurements.LABELS)!=0)
				rt.addLabel("Label", getFileName());
			rt.addValue("X", cal.getX(x));
			rt.addValue("Y", cal.getY(y, imp.getHeight()));
			rt.addValue("Z", cal.getZ(imp.getCurrentSlice()-1));
			if (imp.getProperty("FHT")!=null) {
				double center = imp.getWidth()/2.0;
				y = imp.getHeight()-y-1;
				double r = Math.sqrt((x-center)*(x-center) + (y-center)*(y-center));
				if (r<1.0) r = 1.0;
				double theta = Math.atan2(y-center, x-center);
				theta = theta*180.0/Math.PI;
				if (theta<0) theta = 360.0+theta;
				rt.addValue("R", (imp.getWidth()/r)*cal.pixelWidth);
				rt.addValue("Theta", theta);
			}
			rt.addValue("Value", value);
			displayResults();
		}
		//IJ.write(rt.getCounter()+"\t"+n(cal.getX(x))+n(cal.getY(y))+n(value));
	}
	
	void measureAngle(Roi roi) {
		if (Analyzer.mode!=Analyzer.ANGLES) {
			Analyzer.switchingModes = true;
			if (!reset())
				return;
			if ((measurements&Measurements.LABELS)!=0)
				IJ.setColumnHeadings(" \tName\tangle");
			else		
				IJ.setColumnHeadings(" \tangle");
			Analyzer.mode = Analyzer.ANGLES;
		}
		incrementCounter();
		if ((measurements&Measurements.LABELS)!=0)
			rt.addLabel("Label", getFileName());
		rt.addValue("Angle", ((PolygonRoi)roi).getAngle());
		displayResults();
		//IJ.write(rt.getCounter()+"\t"+n(((PolygonRoi)roi).getAngle()));
	}
	
	void measureLength(Roi roi) {
		if (Analyzer.mode!=Analyzer.LENGTHS) {
			Analyzer.switchingModes = true;
			if (!reset())
				return;
			if ((measurements&Measurements.LABELS)!=0)
				IJ.setColumnHeadings(" \tName\tlength");
			else		
				IJ.setColumnHeadings(" \tlength");
			Analyzer.mode = Analyzer.LENGTHS;
		}
		incrementCounter();
		if ((measurements&Measurements.LABELS)!=0)
			rt.addLabel("Label", getFileName());
		rt.addValue("Length", roi.getLength());
		double angle = 0.0;
		if (roi.getType()==Roi.LINE) {
			Line l = (Line)roi;
			angle = roi.getAngle(l.x1, l.y1, l.x2, l.y2);
		}
		rt.addValue("Angle", angle);
		boolean moreParams = (measurements&Measurements.MEAN)!=0||(measurements&Measurements.STD_DEV)!=0||(measurements&Measurements.MODE)!=0||(measurements&Measurements.MIN_MAX)!=0;
		if (moreParams) {
			ProfilePlot profile = new ProfilePlot(imp);
			double[] values = profile.getProfile();
			ImageProcessor ip2 = new FloatProcessor(values.length, 1, values);
			ImageStatistics stats = ImageStatistics.getStatistics(ip2, Measurements.MEAN+Measurements.STD_DEV+Measurements.MODE+Measurements.MIN_MAX, null);
			if ((measurements&Measurements.MEAN)!=0) rt.addValue(ResultsTable.MEAN,stats.mean);
			if ((measurements&Measurements.STD_DEV)!=0) rt.addValue(ResultsTable.STD_DEV,stats.stdDev);
			if ((measurements&Measurements.MODE)!=0) rt.addValue(ResultsTable.MODE, stats.dmode);
			if ((measurements&Measurements.MIN_MAX)!=0) {
				rt.addValue(ResultsTable.MIN,stats.min);
				rt.addValue(ResultsTable.MAX,stats.max);
			}
		}
		if ((measurements&Measurements.RECT)!=0) {
			Rectangle r = roi.getBounds();
			Calibration cal = imp.getCalibration();
			rt.addValue(ResultsTable.ROI_X, cal.getX(r.x));
			rt.addValue(ResultsTable.ROI_Y, cal.getY(r.y, imp.getHeight()));
			rt.addValue(ResultsTable.ROI_WIDTH, r.width*cal.pixelWidth);
			rt.addValue(ResultsTable.ROI_HEIGHT, r.height*cal.pixelHeight);
		}
		displayResults();
	}
	
	/** Saves the measurements specified in the "Set Measurements" dialog,
		or by calling setMeasurments(), in the system results table.
	*/
	public void saveResults(ImageStatistics stats, Roi roi) {
		if (rt.getColumnHeading(ResultsTable.SLICE)==null)
			reset();
		incrementCounter();
		int counter = rt.getCounter();
		if (counter<=Measurements.MAX_STANDARDS) {
			if (Analyzer.umeans==null) Analyzer.umeans = new float[Measurements.MAX_STANDARDS];
			Analyzer.umeans[counter-1] = (float)stats.umean;
		}
		if ((measurements&Measurements.LABELS)!=0)
			rt.addLabel("Label", getFileName());
		if ((measurements&Measurements.AREA)!=0) rt.addValue(ResultsTable.AREA,stats.area);
		if ((measurements&Measurements.MEAN)!=0) rt.addValue(ResultsTable.MEAN,stats.mean);
		if ((measurements&Measurements.STD_DEV)!=0) rt.addValue(ResultsTable.STD_DEV,stats.stdDev);
		if ((measurements&Measurements.MODE)!=0) rt.addValue(ResultsTable.MODE, stats.dmode);
		if ((measurements&Measurements.MIN_MAX)!=0) {
			rt.addValue(ResultsTable.MIN,stats.min);
			rt.addValue(ResultsTable.MAX,stats.max);
		}
		if ((measurements&Measurements.CENTROID)!=0) {
			rt.addValue(ResultsTable.X_CENTROID,stats.xCentroid);
			rt.addValue(ResultsTable.Y_CENTROID,stats.yCentroid);
		}
		if ((measurements&Measurements.CENTER_OF_MASS)!=0) {
			rt.addValue(ResultsTable.X_CENTER_OF_MASS,stats.xCenterOfMass);
			rt.addValue(ResultsTable.Y_CENTER_OF_MASS,stats.yCenterOfMass);
		}
		if ((measurements&Measurements.PERIMETER)!=0 || (measurements&Measurements.CIRCULARITY)!=0) {
			double perimeter;
			if (roi!=null)
				perimeter = roi.getLength();
			else
				perimeter = 0.0;
			if ((measurements&Measurements.PERIMETER)!=0) 
				rt.addValue(ResultsTable.PERIMETER,perimeter);
			if ((measurements&Measurements.CIRCULARITY)!=0) {
				double circularity = perimeter==0.0?0.0:4.0*Math.PI*(stats.area/(perimeter*perimeter));
				if (circularity>1.0) circularity = 1.0;
				rt.addValue(ResultsTable.CIRCULARITY, circularity);
			}
		}
		if ((measurements&Measurements.RECT)!=0) {
			rt.addValue(ResultsTable.ROI_X,stats.roiX);
			rt.addValue(ResultsTable.ROI_Y,stats.roiY);
			rt.addValue(ResultsTable.ROI_WIDTH,stats.roiWidth);
			rt.addValue(ResultsTable.ROI_HEIGHT,stats.roiHeight);
		}
		if ((measurements&Measurements.ELLIPSE)!=0) {
			rt.addValue(ResultsTable.MAJOR,stats.major);
			rt.addValue(ResultsTable.MINOR,stats.minor);
			rt.addValue(ResultsTable.ANGLE,stats.angle);
		}
		if ((measurements&Measurements.FERET)!=0)
			rt.addValue(ResultsTable.FERET, roi!=null?roi.getFeretsDiameter():0.0);
		if ((measurements&Measurements.INTEGRATED_DENSITY)!=0)
			rt.addValue(ResultsTable.INTEGRATED_DENSITY,stats.area*stats.mean);
		if ((measurements&Measurements.MEDIAN)!=0) rt.addValue(ResultsTable.MEDIAN, stats.median);
		if ((measurements&Measurements.SKEWNESS)!=0) rt.addValue(ResultsTable.SKEWNESS, stats.skewness);
		if ((measurements&Measurements.KURTOSIS)!=0) rt.addValue(ResultsTable.KURTOSIS, stats.kurtosis);
		if ((measurements&Measurements.AREA_FRACTION)!=0) rt.addValue(ResultsTable.AREA_FRACTION, stats.areaFraction);
		if ((measurements&Measurements.SLICE)!=0) rt.addValue(ResultsTable.SLICE, imp!=null?imp.getCurrentSlice():1.0);
	}
	
	String getFileName() {
		String s = "";
		if (imp!=null) {
			if (Analyzer.mode==Analyzer.AREAS && Analyzer.redirectTarget!=0) {
				ImagePlus rImp = WindowManager.getImage(Analyzer.redirectTarget);
				if (rImp!=null) s = rImp.getTitle();				
			} else
				s = imp.getTitle();
			int len = s.length();
			if (len>4 && s.charAt(len-4)=='.' && !Character.isDigit(s.charAt(len-1)))
				s = s.substring(0,len-4); 
			Roi roi = imp.getRoi();
			String roiName = roi!=null?roi.getName():null;
			if (roiName!=null)
				s += ":"+roiName;
			if (imp.getStackSize()>1) {
				ImageStack stack = imp.getStack();
				int currentSlice = imp.getCurrentSlice();
				String label = stack.getShortSliceLabel(currentSlice);
				String colon = s.equals("")?"":":";
				if (label!=null && !label.equals(""))
					s += colon+label;
				else
					s += colon+currentSlice;
			}
		}
		return s;
	}

	/** Writes the last row in the results table to the ImageJ window. */
	public void displayResults() {
		int counter = rt.getCounter();
		if (counter==1)
			IJ.setColumnHeadings(rt.getColumnHeadings());		
		IJ.write(rt.getRowAsString(counter-1));
	}

	/** Updates the displayed column headings. Does nothing if
	    the results table headings and the displayed headings are
        the same. Redisplays the results if the headings are
        different and the results table is not empty. */
	public void updateHeadings() {
		TextPanel tp = IJ.getTextPanel();
		if (tp==null)
			return;
		String worksheetHeadings = tp.getColumnHeadings();		
		String tableHeadings = rt.getColumnHeadings();		
		if (worksheetHeadings.equals(tableHeadings))
			return;
		IJ.setColumnHeadings(tableHeadings);
		int n = rt.getCounter();
		if (n>0) {
			StringBuffer sb = new StringBuffer(n*tableHeadings.length());
			for (int i=0; i<n; i++)
				sb.append(rt.getRowAsString(i)+"\n");
			tp.append(new String(sb));
		}
	}

	/** Converts a number to a formatted string with a tab at the end. */
	public String n(double n) {
		String s;
		if (Math.round(n)==n)
			s = IJ.d2s(n,0);
		else
			s = IJ.d2s(n,Analyzer.precision);
		return s+"\t";
	}
		
	void incrementCounter() {
		//counter++;
		if (rt==null) rt = Analyzer.systemRT;
		rt.incrementCounter();
		Analyzer.unsavedMeasurements = true;
	}
	
	public void summarize() {
		rt = Analyzer.systemRT;
		if (rt.getCounter()==0)
			return;
		if (Analyzer.summarized)
			rt.show("Results");
		measurements = Analyzer.systemMeasurements;
		min = new StringBuffer(100);
		max = new StringBuffer(100);
		mean = new StringBuffer(100);
		sd = new StringBuffer(100);
		min.append("Min\t");
		max.append("Max\t");
		mean.append("Mean\t");
		sd.append("SD\t");
		if ((measurements&Measurements.LABELS)!=0) {
			min.append("\t");
			max.append("\t");
			mean.append("\t");
			sd.append("\t");
		}
		if (Analyzer.mode==Analyzer.POINTS) 
			summarizePoints(rt);
		else if (Analyzer.mode==Analyzer.LENGTHS) 
			summarizeLengths(rt);
		else if (Analyzer.mode==Analyzer.ANGLES) 
			add2(rt.getColumnIndex("Angle"));
		else
			summarizeAreas();
		TextPanel tp = IJ.getTextPanel();
		if (tp!=null) {
			String worksheetHeadings = tp.getColumnHeadings();		
			if (worksheetHeadings.equals(""))
				IJ.setColumnHeadings(rt.getColumnHeadings());
		}		
		IJ.write("");		
		IJ.write(new String(mean));		
		IJ.write(new String(sd));		
		IJ.write(new String(min));		
		IJ.write(new String(max));
		IJ.write("");		
		mean = null;		
		sd = null;		
		min = null;		
		max = null;
		Analyzer.summarized = true;		
	}
	
	void summarizePoints(ResultsTable rt) {
		add2(rt.getColumnIndex("X"));
		add2(rt.getColumnIndex("Y"));
		add2(rt.getColumnIndex("Z"));
		add2(rt.getColumnIndex("Value"));
	}

	void summarizeLengths(ResultsTable rt) {
		int index = rt.getColumnIndex("Mean");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("StdDev");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("Mode");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("Min");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("Max");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("Angle");
		if (rt.columnExists(index)) add2(index);
		index = rt.getColumnIndex("Length");
		if (rt.columnExists(index)) add2(index);
	}
	
	void summarizeAreas() {
		if ((measurements&Measurements.AREA)!=0) add2(ResultsTable.AREA);
		if ((measurements&Measurements.MEAN)!=0) add2(ResultsTable.MEAN);
		if ((measurements&Measurements.STD_DEV)!=0) add2(ResultsTable.STD_DEV);
		if ((measurements&Measurements.MODE)!=0) add2(ResultsTable.MODE);
		if ((measurements&Measurements.MIN_MAX)!=0) {
			add2(ResultsTable.MIN);
			add2(ResultsTable.MAX);
		}
		if ((measurements&Measurements.CENTROID)!=0) {
			add2(ResultsTable.X_CENTROID);
			add2(ResultsTable.Y_CENTROID);
		}
		if ((measurements&Measurements.CENTER_OF_MASS)!=0) {
			add2(ResultsTable.X_CENTER_OF_MASS);
			add2(ResultsTable.Y_CENTER_OF_MASS);
		}
		if ((measurements&Measurements.PERIMETER)!=0)
			add2(ResultsTable.PERIMETER);
		if ((measurements&Measurements.RECT)!=0) {
			add2(ResultsTable.ROI_X);
			add2(ResultsTable.ROI_Y);
			add2(ResultsTable.ROI_WIDTH);
			add2(ResultsTable.ROI_HEIGHT);
		}
		if ((measurements&Measurements.ELLIPSE)!=0) {
			add2(ResultsTable.MAJOR);
			add2(ResultsTable.MINOR);
			add2(ResultsTable.ANGLE);
		}
		if ((measurements&Measurements.CIRCULARITY)!=0)
			add2(ResultsTable.CIRCULARITY);
		if ((measurements&Measurements.FERET)!=0)
			add2(ResultsTable.FERET);
		if ((measurements&Measurements.INTEGRATED_DENSITY)!=0)
			add2(ResultsTable.FERET);
		if ((measurements&Measurements.MEDIAN)!=0)
			add2(ResultsTable.MEDIAN);
		if ((measurements&Measurements.SKEWNESS)!=0)
			add2(ResultsTable.SKEWNESS);
		if ((measurements&Measurements.KURTOSIS)!=0)
			add2(ResultsTable.KURTOSIS);
		if ((measurements&Measurements.AREA_FRACTION)!=0)
			add2(ResultsTable.AREA_FRACTION);
	}

	private void add2(int column) {
		float[] c = column>=0?rt.getColumn(column):null;
		if (c!=null) {
			ImageProcessor ip = new FloatProcessor(c.length, 1, c, null);
			if (ip==null)
				return;
			ImageStatistics stats = new FloatStatistics(ip);
			if (stats==null)
				return;
			mean.append(n(stats.mean));
			min.append(n(stats.min));
			max.append(n(stats.max));
			sd.append(n(stats.stdDev));
		} else {
			mean.append("-\t");
			min.append("-\t");
			max.append("-\t");
			sd.append("-\t");
		}
	}

	/** Returns the current measurement count. */
	public static int getCounter() {
		return Analyzer.systemRT.getCounter();
	}

	/** Sets the measurement counter to zero. Displays a dialog that
	    allows the user to save any existing measurements. Returns
	    false if the user cancels the dialog.
	*/
	public synchronized static boolean resetCounter() {
		TextPanel tp = IJ.isResultsWindow()?IJ.getTextPanel():null;
		int counter = Analyzer.systemRT.getCounter();
		int lineCount = tp!=null?IJ.getTextPanel().getLineCount():0;
		ImageJ ij = IJ.getInstance();
		boolean macro = (IJ.macroRunning()&&!Analyzer.switchingModes) || Interpreter.isBatchMode();
		Analyzer.switchingModes = false;
		if (counter>0 && lineCount>0 && Analyzer.unsavedMeasurements && !macro && ij!=null && !ij.quitting()) {
			YesNoCancelDialog d = new YesNoCancelDialog(ij, "ImageJ", "Save "+counter+" measurements?");
			if (d.cancelPressed())
				return false;
			else if (d.yesPressed()) {
				if (!(new MeasurementsWriter()).save(""))
					return false;
			}
		}
		Analyzer.umeans = null;
		Analyzer.systemRT.reset();
		Analyzer.unsavedMeasurements = false;
		if (tp!=null) {
			tp.selectAll();
			tp.clearSelection();
		}
		Analyzer.summarized = false;
		return true;
	}
	
	public static void setUnsavedMeasurements(boolean b) {
		Analyzer.unsavedMeasurements = b;
	}
	
	// Returns the measurements defined in the Set Measurements dialog. */
	public static int getMeasurements() {
		return Analyzer.systemMeasurements;
	}

	// Sets the system-wide measurements. */
	public static void setMeasurements(int measurements) {
		Analyzer.systemMeasurements = measurements;
	}

	/** Called once when ImageJ quits. */
	public static void savePreferences(Properties prefs) {
		prefs.put(Analyzer.MEASUREMENTS, Integer.toString(Analyzer.systemMeasurements));
		prefs.put(Analyzer.MARK_WIDTH, Integer.toString(Analyzer.markWidth));
		prefs.put(Analyzer.PRECISION, Integer.toString(Analyzer.precision));	}

	/** Returns an array containing the first 20 uncalibrated means. */
	public static float[] getUMeans() {
		return Analyzer.umeans;
	}

	/** Returns the ImageJ results table. This table should only
		be displayed in a the "Results" window. */
	public static ResultsTable getResultsTable() {
		return Analyzer.systemRT;
	}

	/** Returns the number of digits displayed to the right of decimal point. */
	public static int getPrecision() {
		return Analyzer.precision;
	}

	/** Sets the number of digits displayed to the right of decimal point. */
	public static void setPrecision(int decimalPlaces) {
		if (decimalPlaces<0) decimalPlaces = 0;
		if (decimalPlaces>9) decimalPlaces = 9;
		Analyzer.precision = decimalPlaces;
	}

	/** Returns an updated Y coordinate based on
		the current "Invert Y Coordinates" flag. */
	public static int updateY(int y, int imageHeight) {
		if ((Analyzer.systemMeasurements&Measurements.INVERT_Y)!=0)
			y = imageHeight-y-1;
		return y;
	}
	
	/** Returns an updated Y coordinate based on
		the current "Invert Y Coordinates" flag. */
	public static double updateY(double y, int imageHeight) {
		if ((Analyzer.systemMeasurements&Measurements.INVERT_Y)!=0)
			y = imageHeight-y-1;
		return y;
	}
	
	/** Sets the default headings ("Area", "Mean", etc.). */
	public static void setDefaultHeadings() {
		Analyzer.systemRT.setDefaultHeadings();
	}


}
	
